/*
 * @(#)file      Introspector.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.32
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package com.sun.management.jmx;


// Java import
import java.lang.reflect.*;
import java.util.Iterator;

// RI Import
import javax.management.MBeanInfo;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanConstructorInfo;
import javax.management.NotCompliantMBeanException;
import javax.management.MBeanParameterInfo;


/**
 * This class contains the methods for performing all the tests needed to verify
 * that a class represents a JMX compliant MBean.
 *
 */
public class Introspector {
    

    /*
     * ------------------------------------------
     *  PRIVATE VARIABLES
     * ------------------------------------------
     */
    
    private static java.util.Vector attributes = null;
    private static java.util.Vector operations = null;
    
    private static String className = null;
    private static Class baseClass = null;
    private static final String attributeDescription = "Attribute exposed for management";
    private static final String operationDescription = "Operation exposed for management";
    private static final String constructorDescription = "Public constructor of the MBean";
    private static final String mbeanInfoDescription = "Information on the management interface of the MBean";
    

     /*
     * ------------------------------------------
     *  PRIVATE CONSTRUCTORS
     * ------------------------------------------
     */

    // private constructor defined to "hide" the default public constructor
    private Introspector() {
	
    }
    
    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */
    
    
    /**
     * Basic method for testing if a given class is a JMX compliant MBean
     *
     * @param c The class to be tested
     *
     * @exception NotCompliantMBeanException The specified class is not a JMX compliant MBean
     *
     */    
    public static synchronized MBeanInfo testCompliance(Class c) throws NotCompliantMBeanException {
	
	baseClass = c;
	className = c.getName();
	
	//Check if the class is a concrete class
	int mods = c.getModifiers(); 
	if (Modifier.isAbstract(mods)) {
	    throw new NotCompliantMBeanException("The MBean must be have a concrete class");
	}
	
	//Check if the MBean has a public constructor 
	Constructor[] consList = c.getConstructors();     
	if (consList.length == 0) throw new NotCompliantMBeanException("The MBean does not have a public constructor");
	
	//Check if the MBean implements the MBean or the Dynamic MBean interface
	Class[] interfaces = null;
	Class current = c;       
	boolean isMBean = false;
	Class inter = null;
	Class mbeanInterface = null;
	Class cl = null;
	boolean isDyn = false;
	
	while ((current != null) && !(isMBean)) {
	    interfaces = current.getInterfaces();   
	    isDyn = false;
	    for (int i=0;i<interfaces.length;i++)  {   
		inter = implementsMBean(interfaces[i], current.getName());
		if (inter != null) {   
		    isMBean = true;
		    mbeanInterface = inter;
		    cl = current;		 
		}
		if (isDynamic(interfaces[i])) {	   
		    isMBean = true;
		    isDyn = true;
		}
		//Error: The MBean implements the MBean and the Dynamic MBean interface
		if ((mbeanInterface != null) && (current.equals(cl)) && (isDyn)) {
		    throw new NotCompliantMBeanException(c.getName() + " implements javax.management.DynamicMBean and MBean interface");
		}
	    }
	    if (isMBean) break;
	    current = current.getSuperclass();
	}
	// Error: MBean does not implement javax.management.DynamicMBean or MBean interface
	if (!isMBean) throw new NotCompliantMBeanException(c.getName() + " does not implement the " + c.getName() + 
							   "MBean interface or the DynamicMBean interface");
	
	if (mbeanInterface != null) {
	    return (introspect(mbeanInterface));
	} else {
	    return null;
	}
    }

    
    /**
     * Basic method for testing if a given class is a dynamic MBean
     *
     * @param c The class to be tested
     * 
     * @return The MBean interface implemented by the MBean or null if the MBean 
     * is a Dynamic MBean.
     *
     */
    public static synchronized Class getMBeanInterface(Class c) {
	
	//Check if the MBean implements the MBean or the Dynamic MBean interface
	Class[] interfaces = null;
	Class current = c;       
	boolean isMBean = false;
	Class inter = null;
	Class mbeanInterface = null;
	Class cl = null;
	boolean isDyn = false;
	
	while ((current != null) && !(isMBean)) {
	    interfaces = current.getInterfaces();   
	    isDyn = false;
	    for (int i=0;i<interfaces.length;i++)  {	     
		inter = implementsMBean(interfaces[i], current.getName());
		if (inter != null) {
		    isMBean = true;
		    mbeanInterface = inter;
		    cl = current;		 
		}
		if (isDynamic(interfaces[i])) {	   
		    isMBean = true;
		    isDyn = true;
		} 	    
		if (isMBean) break;
	    }
	    current = current.getSuperclass();
	}
	return mbeanInterface;     
    }
    
    /*
     * ------------------------------------------
     *  PRIVATE METHODS
     * ------------------------------------------
     */
    
    
    /**
     * Discovers the getters, setters, operations of the class
     *
     * @param beanClass The XXMBean interface implemented by the tested class.
     *
     * @exception NotCompliantMBeanException The tested class is not a JMX compliant MBean
     */    
    private static MBeanInfo introspect(Class beanClass)  throws NotCompliantMBeanException {

	attributes = new java.util.Vector();
	operations = new java.util.Vector();
	
	Method methodList[] = null;
	methodList = beanClass.getMethods();
	
	// Now analyze each method.        
	for (int i = 0; i < methodList.length; i++) { 
	    Method method = methodList[i];
	    if (method == null) {
		continue;
	    }            
	    String name = method.getName();            
	    Class argTypes[] = method.getParameterTypes();
	    Class resultType = method.getReturnType();
	    int argCount = argTypes.length;
	    MBeanAttributeInfo attr = null;
	    MBeanOperationInfo oper = null;
	    
	    // if the method name is get(), set(), is() it is an operation 
	    if (name.equals("get") || name.equals("is") || name.equals("set")) {	
		try {
		    oper = new MBeanOperationInfo(operationDescription, method);
		} catch (Exception e) { e.printStackTrace();}
		operations.addElement(oper);
		continue;
	    } 
	    // if the method name does not start with "get", "set", "is" it is an operation 
	    if ((!name.startsWith("get")) &&  (!name.startsWith("set")) &&  (!name.startsWith("is"))) {
		try {
		    oper = new MBeanOperationInfo(operationDescription, method);
		} catch (Exception e) { e.printStackTrace();}
		operations.addElement(oper);
		continue;	    
	    }
	    // if the method has no parameters
	    if (argCount == 0) {
		// if the method has no parameters and starts with "get" and returns void it is an operation 
		if (name.startsWith("get")) {
		    if (resultType == Void.TYPE) {
			// operation
			try {
			    oper = new MBeanOperationInfo(operationDescription, method);
			} catch (Exception e) { e.printStackTrace();}
			operations.addElement(oper);
			continue;
		    }
		    
		    // if the method has no parameters and starts with "get" and does not return void it is a getter. 
		    try {
			attr = new MBeanAttributeInfo(name.substring(3), attributeDescription, method, null);
		    } catch (Exception e) {
			e.printStackTrace();
		    }		
		    if (testConsistency(attr)) {
			attributes.addElement(attr);
		    }
		    continue;
		}
		
		// if the method has no parameters and starts with "is" and returns boolean or Boolean it is a getter. 
		try {
		    if (((resultType == Class.forName("java.lang.Boolean") ) || (resultType == Boolean.TYPE)) && (name.startsWith("is"))) {
			try {
			    attr = new MBeanAttributeInfo(name.substring(2), attributeDescription, method, null);
			} catch (Exception e) { e.printStackTrace();}		
			if (testConsistency(attr)) {
			    attributes.addElement(attr);
			}
			continue;
		    }
		} catch (ClassNotFoundException e) {}
		// if the method has no parameters and starts with "set" it is an operation.
		if (name.startsWith("set")) {
		    // operation
		    try {
			oper = new MBeanOperationInfo(operationDescription, method);
		    } catch (Exception e) { e.printStackTrace();}
		    operations.addElement(oper);
		    continue;
		}	
	    } 
	    // if the method has one parameter	 
	    if (argCount == 1) {
		// if the method has one parameter and starts with "set" and returns void it is a setter.
		if (((resultType == Void.TYPE)) && (name.startsWith("set"))) {
		    // Simple setter.
		    try {
			attr = new MBeanAttributeInfo(name.substring(3), attributeDescription, null, method);
		    } catch (Exception e) { e.printStackTrace();}		
		    if (testConsistency(attr)) {
			attributes.addElement(attr);
		    }
		    continue;
		} else {
		    // if the method has one parameter and starts with "set" and does not return  void it is an operation.
		    try {
			oper = new MBeanOperationInfo(operationDescription, method);
		    } catch (Exception e) { e.printStackTrace();}	
		    operations.addElement(oper);
		    continue;
		}
	    } 
	    // if the method has more then one parameters it is an operation	 
	    if (argCount > 1) {
		// operation
		try {
		    oper = new MBeanOperationInfo(operationDescription, method);
		} catch (Exception e) { e.printStackTrace();}
		operations.addElement(oper);
		continue;		   	
	    }		    	    
	}
	return constructResult(); 
    }
        
    /**
     * Checks if the types and the signatures of getters/setters/operations are conform to the MBean
     * design patterns.Error cases:
     * 	-  It exposes a method void Y getXX() AND a method void setXX(Z) (parameter type mismatch) 
     * 	-  It exposes a method void setXX(Y) AND a method void setXX(Z) (parameter type mismatch) 
     *     -  It exposes a  boolean isXX() method AND a YY getXX() or a void setXX(Y).
     * Returns false if the attribute is already in attributes Vector
     */    
    private static boolean testConsistency(MBeanAttributeInfo attr)  throws NotCompliantMBeanException {
	
	for (Iterator i = attributes.iterator();i.hasNext();) {
	    MBeanAttributeInfo mb = (MBeanAttributeInfo) i.next();
	    if (mb.getName().compareTo(attr.getName()) == 0) {
		if ((attr.isReadable() && mb.isReadable()) && (attr.isIs() || mb.isIs())) {
		    throw new NotCompliantMBeanException("Getter is" + mb.getName() + " cannot co-exist with getter get" + attr.getName());
		}  
		if (mb.getType().compareTo(attr.getType()) != 0) {
		    if (mb.isWritable() && attr.isWritable()) {
			throw new NotCompliantMBeanException("Type mismatch between parameters of set" + mb.getName() + " methods");
		    } else {
			throw new NotCompliantMBeanException("Type mismatch between parameters of get or is" + mb.getName() + ", set" + mb.getName() + " methods");		    
		    }		  
		}
		if (attr.isReadable() && mb.isReadable()) {
		    return false;
		}
		if (attr.isWritable() && mb.isWritable()) {
		    return false;
		}
	    }   
	}
	return true;
    }

    /**
     * Discovers the constructors of the MBean
     */
    static MBeanConstructorInfo[] getConstructors () {
	Constructor[] consList = baseClass.getConstructors();
	MBeanConstructorInfo[] resultConstructors;
	java.util.Vector constructors = new java.util.Vector();
	
	// Now analyze each Constructor.        
	for (int i = 0; i < consList.length; i++) {
	    Constructor constructor = consList[i];    	    
	    MBeanConstructorInfo mc = null;
	    try {               
		mc = new MBeanConstructorInfo(constructorDescription, constructor);		     		                
	    } catch (Exception ex) {
		mc = null;
	    }
	    if (mc != null) {
		constructors.addElement(mc);
	    }
	}
	// Allocate and populate the result array.
	resultConstructors = new MBeanConstructorInfo[constructors.size()];        
	for (int i = 0; i < resultConstructors.length; i++) {
	    resultConstructors[i] = (MBeanConstructorInfo)constructors.elementAt(i);
	}
	return resultConstructors;     
    }
    
    /**
     * Constructs the MBeanInfo of the MBean.
     */
    private static MBeanInfo constructResult() {
	
	MBeanAttributeInfo[] resultAttributes;
	MBeanOperationInfo[] resultOperations;
	MBeanConstructorInfo[] resultConstructors;
	
	for (int i=0;i<attributes.size();i++) {
	    MBeanAttributeInfo bi = (MBeanAttributeInfo) attributes.elementAt(i);      
            for (int j=0;j<attributes.size();j++) {
		MBeanAttributeInfo mi = (MBeanAttributeInfo)attributes.elementAt(j);		
                if ((mi.getName().compareTo(bi.getName()) == 0) && (i != j)) {
		    try {
			attributes.setElementAt(new MBeanAttributeInfo(bi.getName(), bi.getType(), attributeDescription, true, true, bi.isIs()), j);
		    } catch (Exception e) {}
		    attributes.removeElementAt(i);
		}
	    }
	}
                
        resultAttributes = new MBeanAttributeInfo[attributes.size()];
        for (int i= 0;i<attributes.size();i++) {
            resultAttributes[i] = (MBeanAttributeInfo) attributes.elementAt(i);
        }
 
        resultOperations = new MBeanOperationInfo[operations.size()];
        for (int i= 0;i<operations.size();i++) {
            resultOperations[i] = (MBeanOperationInfo) operations.elementAt(i);
        }

	resultConstructors = getConstructors();     
        MBeanInfo resultMBeanInfo = new MBeanInfo(className, mbeanInfoDescription, resultAttributes, resultConstructors, resultOperations, null);
	return resultMBeanInfo;
    }

    /**
     * Returns the XXMBean interface or null if no such interface exists
     *
     * @param c The interface to be tested
     * @param clName The name of the class implementing this interface
     */
    static Class implementsMBean(Class c, String clName) {
	if (c.getName().compareTo(clName + "MBean") == 0) {
	    return c;
	}   
	Class[] interfaces = c.getInterfaces();
	for (int i = 0;i < interfaces.length; i++) {   
	    if (interfaces[i].getName().compareTo(clName + "MBean") == 0) {
		return interfaces[i];
	    }     
	}
	return null;
    }
    
    /**
     * Returns true if the MBean is dynamic, false otherwise
     * 
     * @param c The interface to be tested
     */
    static boolean isDynamic(Class c) {
	if (c.getName().compareTo("javax.management.DynamicMBean") == 0) {
	    return true;
	}   
	Class[] interfaces = c.getInterfaces();
	for (int i = 0;i < interfaces.length; i++) {   
	    if (interfaces[i].getName().compareTo("javax.management.DynamicMBean") == 0) {
		return true;
	    }     
	}
	return false;
    }
    
}

