/*
 * @(#)file      MBeanServerImpl.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.24
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package com.sun.management.jmx;

// java import
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.lang.reflect.InvocationTargetException; 
import java.lang.reflect.Method; 
import java.lang.reflect.Constructor;
import java.io.OptionalDataException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException ;

// RI import
import javax.management.*; 


/**
 * This is the base class for MBean manipulation on the agent side.. It
 * contains the methods necessary for the creation, registration, and
 * deletion of MBeans as well as the access methods for registered MBeans.
 * This is the core component of the JMX infrastructure.
 * <P>
 * Every MBean which is added to the MBean server becomes manageable: its attributes and operations
 * become remotely accessible through the connectors/adaptors connected to that MBean server. 
 * A Java object cannot be registered in the MBean server unless it is a JMX compliant MBean.
 * <P>
 * When an MBean is registered or unregistered in the MBean server an 
 * {@link javax.management.MBeanServerNotification MBeanServerNotification} 
 * Notification is emitted. To register an object as listener to MBeanServerNotifications
 * you should call the MBean server method {@link #addNotificationListener addNotificationListener} with <CODE>ObjectName</CODE>
 * the <CODE>ObjectName</CODE> of the {@link javax.management.MBeanServerDelegate MBeanServerDelegate}. 
 * This <CODE>ObjectName</CODE> is:
 * <BR>
 * <CODE>JMImplementation:type=MBeanServerDelegate</CODE>.
 *
 */
public class MBeanServerImpl implements MBeanServer { 


    /** The Metadata object used by the MBean Server */
    private transient MetaData meta = null;

    /** The Repository object used by the MBean Server */
    private transient RepositorySupport repository = null;

    /** The MBeanServerDelegate object representing the MBean Server */
    private transient MBeanServerDelegate MBeanServerDelegateObject = null;
    
    /** The MBeanServerDelegate object name */
    private transient ObjectName          MBeanServerDelegateObjectName = null;

    /** The list of class loader of the MBeanServer class */
    private transient ClassLoader loader = null;

    /** The default domain of the object names */
    private String domain = null;

    /** True if the repository perform queries, false otherwise */
    private boolean queryByRepo;

    /** The list of internal notification listeners */
    private transient java.util.Hashtable internalListener  = new java.util.Hashtable();
    
    /** The sequence number identifyng the notifications sent */
    private int sequenceNumber=0;
   
    /** The list of MBeanServer objects in the Java VM */
    private static java.util.ArrayList MBeanServerList = new java.util.ArrayList();   

    /** The name of this class to be used for tracing */
    private final String dbgTag = "MBeanServer";
     
    
    /**
     * Creates an MBeanServer with a standard default domain name.     
     * The default domain name is used as the domain part in the ObjectName
     * of MBeans, if no domain is specified by the user.
     * <p>
     * The standard default domain name is defined in {@link com.sun.management.jmx.ServiceName#DOMAIN ServiceName.DOMAIN}
     */
    public MBeanServerImpl()  { 
        initialize(new RepositorySupport(), null);
    } 

    /**
     * Creates an MBeanServer with the specified default domain name.
     * The default domain name is used as the domain part in the ObjectName
     * of MBeans if no domain is specified by the user.
     */
    public MBeanServerImpl(String domain)  { 
        initialize(new RepositorySupport(), domain);
    } 



       /**
     * Instantiates an object using the list of all class loaders registered
     * in the MBean server ({@link javax.management.loading.DefaultLoaderRepository Default Loader Repository}).
     * The object's class should have a public constructor.
     * It returns a reference to the newly created object.
     * The newly created object is not registered in the MBean server.
     *
     * @param className The class name of the object to be instantiated.    
     *
     * @return The newly instantiated object.    
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or the <CODE>java.lang.Exception</CODE> that 
     * occurred when trying to invoke the object's constructor.
     * @exception MBeanException The constructor of the object has thrown an exception
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null.
     *
     */
    public Object instantiate(String className) throws ReflectionException, MBeanException { 

        Class theClass;
        Object moi= null;

        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during object instantiation");
        }  
        try {
            theClass= findClass(className);
        }
        catch (ClassNotFoundException e) {
            try {
                theClass = DefaultLoaderRepository.loadClass(className);
            }
            catch (ClassNotFoundException ee) {
                throw new ReflectionException(ee, "The MBean class could not be loaded by the default loader repository");
            }
        }
        return (internal_instantiate(theClass));
    }


    /**
     * Instantiates an object using the class Loader specified by its <CODE>ObjectName</CODE>.
     * If the loader name is null, the ClassLoader that loaded the MBean Server will be used.
     * The object's class should have a public constructor.
     * It returns a reference to the newly created object.
     * The newly created object is not registered in the MBean server.
     *
     * @param className The class name of the MBean to be instantiated.    
     * @param loaderName The object name of the class loader to be used.
     *
     * @return The newly instantiated object.    
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or the <CODE>java.lang.Exception</CODE> that 
     * occurred when trying to invoke the object's constructor.
     * @exception MBeanException The constructor of the object has thrown an exception.
     * @exception InstanceNotFoundException The specified class loader is not registered in the MBaenServer. 
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null.
     *
     */
    public Object instantiate(String className, ObjectName loaderName) 
        throws ReflectionException, MBeanException, InstanceNotFoundException { 
             
        Class theClass;
        Object moi= null;

        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during object instantiation");
        }  

        if (loaderName == null) {
            try {
                theClass= findClass(className);
            } catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the MBeanServer's loader");
            }
        } else {            
            // try to identify the loader             
            try {
                theClass = findClass(className, loaderName);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the specified loader");
            }
        }
        return (internal_instantiate(theClass));

    }

     /**
     * Instantiates an object using the list of all class loaders registered
     * in the MBean server ({@link javax.management.loading.DefaultLoaderRepository Default Loader Repository}).
     * The object's class should have a public constructor.
     * The call returns a reference to the newly created object.
     * The newly created object is not registered in the MBean server.
     *
     * @param className The class name of the object to be instantiated.
     * @param params An array containing the parameters of the constructor to be invoked.
     * @param signature An array containing the signature of the constructor to be invoked.     
     *
     * @return The newly instantiated object.    
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or the <CODE>java.lang.Exception</CODE> that 
     * occurred when trying to invoke the object's constructor.  
     * @exception MBeanException The constructor of the object has thrown an exception
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null.
     *
     */    
    public Object instantiate(String className, Object params[], String signature[]) 
        throws ReflectionException, MBeanException { 


	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        Class theClass;

        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during object instantiation");
        }  
        try {
            theClass= findClass(className);
        }
        catch (ClassNotFoundException e) {
            try {
                theClass = DefaultLoaderRepository.loadClass(className);
            }
            catch (ClassNotFoundException ee) {
                throw new ReflectionException(ee, "The MBean class could not be loaded by the default loader repository");
            }
        }
        return (internal_instantiate(theClass, params, signature));
    }


     /**
     * Instantiates an object. The class loader to be used is identified by its object
     * name. If the object name of the loader is null, the ClassLoader that loaded the MBean server will be used.
     * The object's class should have a public constructor.
     * The call returns a reference to the newly created object.
     * The newly created object is not registered in the MBean server.
     *
     * @param className The class name of the object to be instantiated.
     * @param params An array containing the parameters of the constructor to be invoked.
     * @param signature An array containing the signature of the constructor to be invoked.     
     * @param loaderName The object name of the class loader to be used.
     *
     * @return The newly instantiated object.    
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or the <CODE>java.lang.Exception</CODE> that 
     * occurred when trying to invoke the object's constructor.  
     * @exception MBeanException The constructor of the object has thrown an exception
     * @exception InstanceNotFoundException The specified class loader is not registered in the MBean server. 
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null.
     *
     */       
    public Object instantiate(String className, ObjectName loaderName, Object params[], String signature[]) 
        throws ReflectionException, MBeanException, InstanceNotFoundException { 


	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        Class theClass;

        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during object instantiation");
        }  

        // The class loader service must be available               
        if (loaderName == null) {
            try {
                theClass= findClass(className);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the MBeanServer's loader");
            }     
        } else {
            // try to identify the loader         
            try {
                theClass = findClass(className, loaderName);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the specified loader");
            }
        }       
        return (internal_instantiate(theClass, params, signature));

    }

   /**
     * Instantiates and registers an MBean in the MBean server. 
     * The MBean server will use the {@link javax.management.loading.DefaultLoaderRepository Default Loader Repository}
     * to load the class of the MBean.
     * An object name is associated to the MBean.
     * If the object name given is null, the MBean can automatically provide its
     * own name by implementing the {@link javax.management.MBeanRegistration MBeanRegistration} interface. 
     * The call returns an <CODE>ObjectInstance</CODE> object representing the newly created MBean.
     *
     * @param className The class name of the MBean to be instantiated.    
     * @param name The object name of the MBean. May be null.    
     *
     * @return  An <CODE>ObjectInstance</CODE>, containing the <CODE>ObjectName</CODE> and the Java class name
     * of the newly instantiated MBean.    
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or a <CODE><CODE>java.lang.Exception</CODE></CODE> that occurred 
     * when trying to invoke the MBean's constructor.
     * @exception InstanceAlreadyExistsException The MBean is already under the control of the MBean server.
     * @exception MBeanRegistrationException The <CODE>preRegister</CODE> (<CODE>MBeanRegistration</CODE> interface) method of the MBean
     * has thrown an exception. The MBean will not be registered.    
     * @exception MBeanException The constructor of the MBean has thrown an exception
     * @exception NotCompliantMBeanException This class is not a JMX compliant MBean
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: 
     * The className passed in parameter is null, the <CODE>ObjectName</CODE> passed in parameter contains a pattern or no <CODE>ObjectName</CODE> is specified
     * for the MBean.
     *
     */
    public ObjectInstance createMBean(String className, ObjectName name)
        throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException { 


	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        ObjectName logicalName;
        Class theClass = null;;
        Object moi= null;
        
        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during object creation");
        }  

        // Test  object name   
        if ((name !=null) && (name.isPattern() == true)) {
            throw new RuntimeOperationsException(new IllegalArgumentException("invalid name->" + name.toString()), "Exception occured during object creation");
        }
        String nameMessage = "null";
        if (name!=null) {
            nameMessage = name.toString();
        }
        if (isTraceOn()) {
            trace(dbgTag, "createMBean", "ClassName = " + className + ",ObjectName = " + nameMessage);
        }
        
        try {
            theClass= findClass(className);
        }
        catch (ClassNotFoundException e) {
            try {
                theClass = DefaultLoaderRepository.loadClass(className);
            }
            catch (ClassNotFoundException ex) {
                throw new ReflectionException(ex, "The MBean class could not be loaded by the default loader repository");
            }
        }         

        // Check the JMX compliance of the class        
        meta.testCompliance(theClass); 

        moi = internal_instantiate(theClass);

        ObjectName res = null;
        if (moi instanceof MBeanRegistration) {
            res=preRegisterInvoker(moi,name);
        }
        // if name passed in parameter is null, consider the ObjectName returned by preRegister
        if (name == null) {
            logicalName = res;
        } else {
            logicalName = name;
        }

        if (logicalName!=null) {
            internal_addObject(moi, logicalName);
        } else {
            if (moi instanceof MBeanRegistration) {
                postRegisterInvoker(moi, false);
            }
            throw new RuntimeOperationsException(new IllegalArgumentException("No object name specified"), "Exception occured trying to register the MBean");
        }       

        if (moi instanceof MBeanRegistration) 
            postRegisterInvoker(moi, true);

        /**
         * Checks if the newly registered MBean is a ClassLoader
         * If so, it adds the  MBean in the default loader repository.
         */     
        if (moi instanceof ClassLoader) {
            DefaultLoaderRepository.addClassLoader((ClassLoader)moi);
        }       

	// if the MBean is a dynamic MBean ask its MBeanInfo for the class name
	if (moi instanceof DynamicMBean) {
	    try {
		MBeanInfo mbi = ((DynamicMBean)moi).getMBeanInfo();
		
		if (mbi == null) {
		    throw new NotCompliantMBeanException("The MBeanInfo returned is null");
		}
		
		className = mbi.getClassName();
		
		if (className == null) {
		    throw new NotCompliantMBeanException("The class Name returned is null");
		}
	    } catch (Exception e) {
		 throw new NotCompliantMBeanException("Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	}

        return(new ObjectInstance(logicalName, className));       
    }

    /**
     * Instantiates and registers an MBean in the MBean server. 
     * The class loader to be used is identified by its object  name. An object name is associated to the MBean. 
     * If the object name  of the loader is null, the ClassLoader that loaded the MBean server will be used.
     * If the MBean's object name given is null, the MBean can automatically provide its
     * own name by implementing the {@link javax.management.MBeanRegistration MBeanRegistration} interface. 
     * The call returns an <CODE>ObjectInstance</CODE> object representing the newly created MBean.
     *
     * @param className The class name of the MBean to be instantiated.    
     * @param name The object name of the MBean. May be null.    
     * @param loaderName The object name of the class loader to be used.
     *
     * @return  An <CODE>ObjectInstance</CODE>, containing the <CODE>ObjectName</CODE> and the Java class name
     * of the newly instantiated MBean.    
     *
     * @exception ReflectionException  Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or a <CODE>java.lang.Exception</CODE> that occurred 
     * when trying to invoke the MBean's constructor.
     * @exception InstanceAlreadyExistsException The MBean is already under the control of the MBean server.
     * @exception MBeanRegistrationException The <CODE>preRegister</CODE> (<CODE>MBeanRegistration</CODE>  interface) method of the MBean
     * has thrown an exception. The MBean will not be registered.    
     * @exception MBeanException The constructor of the MBean has thrown an exception
     * @exception NotCompliantMBeanException This class is not a JMX compliant MBean
     * @exception InstanceNotFoundException The specified class loader is not registered in the MBean server. 
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null,
     * the <CODE>ObjectName</CODE> passed in parameter contains a pattern or no <CODE>ObjectName</CODE> is specified for the MBean.
     */
    public ObjectInstance createMBean(String className, ObjectName name, ObjectName loaderName) 
        throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException,
               InstanceNotFoundException { 

        ObjectName logicalName;
        Class theClass;
        Object moi= null;

        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during MBean creation");
        }  

        // Test  object name   
        if ((name !=null) && (name.isPattern() == true)) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Invalid name->" + name.toString()), "Exception occured during MBean registration");
        }


        // The class loader service must be available               
        if (loaderName == null) {
            String nameMessage = "null";
            if (name!=null) {
                nameMessage = name.toString();
            }       
            if (isTraceOn()) {
                trace(dbgTag, "createMBean", "ClassName = " + className + ",ObjectName = " + nameMessage + ", loader null");
            }

            try {
                theClass= findClass(className);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the MBeanServer's loader");
            }
        } else {
            String nameMessage = "null";
            if (name!=null) {
                nameMessage = name.toString();
            }       
            if (isTraceOn()) {
                trace(dbgTag, "createMBean", "ClassName = " + className + ",ObjectName = " + nameMessage + ", LoaderName = " + loaderName);
            }      
            // try to identify the loader             
            try {
                theClass = findClass(className, loaderName);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the specified loader");
            }
        }

        // Check the JMX compliance of the class        
        meta.testCompliance(theClass); 

        moi = internal_instantiate(theClass);

        ObjectName res = null;
        if (moi instanceof MBeanRegistration) {
            res=preRegisterInvoker(moi,name);
        }
        // if name passed in parameter is null, consider the ObjectName returned by preRegister
        if (name == null) {
            logicalName = res;
        } else {
            logicalName = name;
        }

        if (logicalName!=null) {
            internal_addObject(moi, logicalName);
        } else {
            if (moi instanceof MBeanRegistration ) {
                postRegisterInvoker(moi, false);
            }
            throw new RuntimeOperationsException(new IllegalArgumentException("No object name specified"), "Exception occured trying to register the MBean");
        }       

        if (moi instanceof MBeanRegistration) 
            postRegisterInvoker(moi, true);

        /**
         * Checks if the newly registered MBean is a ClassLoader
         * If so, it adds the  MBean in the default loader repository.
         */     
        if (moi instanceof ClassLoader) {
            DefaultLoaderRepository.addClassLoader((ClassLoader)moi);
        }       

	// if the MBean is a dynamic MBean ask its MBeanInfo for the class name
	if (moi instanceof DynamicMBean) {
	    try {
		MBeanInfo mbi = ((DynamicMBean)moi).getMBeanInfo();
		
		if (mbi == null) {
		    throw new NotCompliantMBeanException("The MBeanInfo returned is null");
		}
		
		className = mbi.getClassName();
		
		if (className == null) {
		    throw new NotCompliantMBeanException("The class Name returned is null");
		}
	    } catch (Exception e) {
		 throw new NotCompliantMBeanException("Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	}

        return(new ObjectInstance(logicalName, className));       
    }


      /**
     * Instantiates and registers an MBean in the MBean server.
     * The MBean server will use the {@link javax.management.loading.DefaultLoaderRepository Default Loader Repository}
     * to load the class of the MBean.
     * An object name is associated to the MBean. 
     * If the object name given is null, the MBean can automatically provide its
     * own name by implementing the {@link javax.management.MBeanRegistration MBeanRegistration} interface.
     * The call returns an <CODE>ObjectInstance</CODE> object representing the newly created MBean.
     *
     * @param className The class name of the MBean to be instantiated.
     * @param name The object name of the MBean. May be null.
     * @param params An array containing the parameters of the constructor to be invoked.
     * @param signature An array containing the signature of the constructor to be invoked.     
     *
     * @return  An <CODE>ObjectInstance</CODE>, containing the <CODE>ObjectName</CODE> and the Java class name
     * of the newly instantiated MBean.
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or a <CODE>java.lang.Exception</CODE> that occurred 
     * when trying to invoke the MBean's constructor.
     * @exception InstanceAlreadyExistsException The MBean is already under the control of the MBean server.
     * @exception MBeanRegistrationException The <CODE>preRegister</CODE> (<CODE>MBeanRegistration</CODE>  interface) method of the MBean
     * has thrown an exception. The MBean will not be registered.
     * @exception MBeanException The constructor of the MBean has thrown an exception
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null,
     * the <CODE>ObjectName</CODE> passed in parameter contains a pattern or no <CODE>ObjectName</CODE> is specified for the MBean.
     *
     */
    public ObjectInstance createMBean(String className, ObjectName name, Object params[], String signature[]) 
        throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException  { 
        ObjectName logicalName;  
        Class theClass;
 
        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during MBean creation");
        }  

        // Test  object name   
        if ((name !=null) && (name.isPattern() == true)) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Invalid name->" + name.toString()), "Exception occured during MBean registration");
        }       
        String nameMessage = "null";
        if (name!=null) {
            nameMessage = name.toString();
        }
        if (isTraceOn()) {
            trace(dbgTag, "createMBean", "ClassName = " + className + ",ObjectName = " + nameMessage);
        }
        try {
            theClass= findClass(className);
        }
        catch (ClassNotFoundException e) {
            try {
                theClass = DefaultLoaderRepository.loadClass(className);
            }
            catch (ClassNotFoundException ex) {
                throw new ReflectionException(ex, "The MBean class could not be loaded by the default loader repository");
            }
        }                

        // Check the JMX compliance of the class        
        meta.testCompliance(theClass); 
        
        Object moi= internal_instantiate(theClass, params,  signature);
     
        ObjectName res = null;
        if (moi instanceof MBeanRegistration) {
            res=preRegisterInvoker(moi,name);
        }
        // if name passed in parameter is null, consider the ObjectName returned by preRegister
        if (name == null) {
            logicalName = res;
        } else {
            logicalName = name;
        }

        if (logicalName!=null) {
            internal_addObject(moi, logicalName);
        } else {
            if (moi instanceof MBeanRegistration ) {
                postRegisterInvoker(moi, false);
            }
            throw new RuntimeOperationsException(new IllegalArgumentException("No object name specified"), "Exception occured trying to register the MBean");
        }       

        if (moi instanceof MBeanRegistration) {
            postRegisterInvoker(moi, true);
        }

        /**
         * Checks if the newly registered MBean is a ClassLoader
         * If so, it adds the  MBean in the default loader repository.
         */     
        if (moi instanceof ClassLoader) {
            DefaultLoaderRepository.addClassLoader((ClassLoader)moi);
        }       

	// if the MBean is a dynamic MBean ask its MBeanInfo for the class name
	if (moi instanceof DynamicMBean) {
	    try {
		MBeanInfo mbi = ((DynamicMBean)moi).getMBeanInfo();
		
		if (mbi == null) {
		    throw new NotCompliantMBeanException("The MBeanInfo returned is null");
		}
		
		className = mbi.getClassName();
		
		if (className == null) {
		    throw new NotCompliantMBeanException("The class Name returned is null");
		}
	    } catch (Exception e) {
		 throw new NotCompliantMBeanException("Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	}

        return(new ObjectInstance(logicalName, className));
    }

   /**
     * Instantiates and registers an MBean in the MBean server. 
     * The class loader to be used is identified by its object
     * name. An object name is associated to the MBean. If the object name
     * of the loader is not specified, the ClassLoader that loaded the MBean server will be used.
     * If  the MBean object name given is null, the MBean can automatically provide its
     * own name by implementing the {@link javax.management.MBeanRegistration MBeanRegistration} interface. 
     * The call returns an <CODE>ObjectInstance</CODE> object representing the newly created MBean.
     *
     * @param className The class name of the MBean to be instantiated.
     * @param name The object name of the MBean. May be null.
     * @param params An array containing the parameters of the constructor to be invoked.
     * @param signature An array containing the signature of the constructor to be invoked.     
     * @param loaderName The object name of the class loader to be used.
     *
     * @return  An <CODE>ObjectInstance</CODE>, containing the <CODE>ObjectName</CODE> and the Java class name
     * of the newly instantiated MBean.
     *
     * @exception ReflectionException Wraps a <CODE>java.lang.ClassNotFoundException</CODE> or a <CODE>java.lang.Exception</CODE> that occurred 
     * when trying to invoke the MBean's constructor.     
     * @exception InstanceAlreadyExistsException The MBean is already under the control of the MBean server.
     * @exception MBeanRegistrationException The <CODE>preRegister</CODE> (<CODE>MBeanRegistration</CODE>  interface) method of the MBean
     * has thrown an exception. The MBean will not be registered.
     * @exception MBeanException The constructor of the MBean has thrown an exception
     * @exception InstanceNotFoundException The specified class loader is not registered in the MBean server. 
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The className passed in parameter is null,
     * the <CODE>ObjectName</CODE> passed in parameter contains a pattern or no <CODE>ObjectName</CODE> is specified for the MBean.
     *
     */
    public ObjectInstance createMBean(String className, ObjectName name, ObjectName loaderName, Object params[], String signature[]) 
        throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException,
               InstanceNotFoundException { 
        ObjectName logicalName;  
        Class theClass;
 
        if (className == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The class name cannot be null"), "Exception occured during MBean creation");
        }  

        // Test  object name   
        if ((name !=null) && (name.isPattern() == true)) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Invalid name->" + name.toString()), "Exception occured during MBean registration");
        }       
        // The class loader service must be available               
        if (loaderName == null) {
            String nameMessage = "null";
            if (name!=null) {
                nameMessage = name.toString();
            }
            if (isTraceOn()) {
                trace(dbgTag, "createMBean", "ClassName = " + className + ",ObjectName = " + nameMessage + " Loader name = null");
            }    
            try {
                theClass= findClass(className);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the MBeanServer's loader");
            }
        } else {
            String nameMessage = "null";
            if (name!=null) {
                nameMessage = name.toString();
            }
            if (isTraceOn()) {
                trace(dbgTag, "createMBean", "ClassName = " + className + ",ObjectName = " + nameMessage + ",Loader name = "+ loaderName.toString());
            }      
            // try to identify the loader         
            try {
                theClass = findClass(className, loaderName);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the specified loader");
            }
        }       

        // Check the JMX compliance of the class        
        meta.testCompliance(theClass); 

        Object moi= internal_instantiate(theClass, params,  signature);     

        ObjectName res = null;
        if (moi instanceof MBeanRegistration) {
            res=preRegisterInvoker(moi,name);
        }
        // if name passed in parameter is null, consider the ObjectName returned by preRegister
        if (name == null) {
            logicalName = res;
        } else {
            logicalName = name;
        }
        
        if (logicalName!=null) {
            internal_addObject(moi, logicalName);
        } else {
            if (moi instanceof MBeanRegistration ) {
                postRegisterInvoker(moi, false);
            }
            throw new RuntimeOperationsException(new IllegalArgumentException("No object name specified"), "Exception occured trying to register the MBean");
        }       
        
        if (moi instanceof MBeanRegistration) {
            postRegisterInvoker(moi, true);
        }

        /**
         * Checks if the newly registered MBean is a ClassLoader
         * If so, it adds the  MBean in the default loader repository.
         */     
        if (moi instanceof ClassLoader) {
            DefaultLoaderRepository.addClassLoader((ClassLoader)moi);
        }       

	// if the MBean is a dynamic MBean ask its MBeanInfo for the class name
	if (moi instanceof DynamicMBean) {
	    try {
		MBeanInfo mbi = ((DynamicMBean)moi).getMBeanInfo();
		
		if (mbi == null) {
		    throw new NotCompliantMBeanException("The MBeanInfo returned is null");
		}
		
		className = mbi.getClassName();
		
		if (className == null) {
		    throw new NotCompliantMBeanException("The class Name returned is null");
		}
	    } catch (Exception e) {
		 throw new NotCompliantMBeanException("Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	}

        return(new ObjectInstance(logicalName, className));
    }

      /**
     * Registers a pre-existing object as an MBean with the MBean server. If the object name given is
     * null, the MBean may automatically provide its own name by implementing the
     * {@link javax.management.MBeanRegistration MBeanRegistration}  interface.
     * The call returns an <CODE>ObjectInstance</CODE> object representing the registered MBean.
     *
     * @param object The  MBean to be registered as an MBean.     
     * @param name The object name of the MBean. May be null.
     *
     * @return  The <CODE>ObjectInstance</CODE> for the MBean that has been registered.     
     *
     * @exception InstanceAlreadyExistsException The MBean is already under the control of the MBean server.
     * @exception MBeanRegistrationException The <CODE>preRegister</CODE> (<CODE>MBeanRegistration</CODE>  interface) method of the MBean
     * has thrown an exception. The MBean will not be registered.
     * @exception NotCompliantMBeanException This object is not a JMX compliant MBean
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The object passed in parameter is null or
     * no object name is specified.
     *
     */
    public ObjectInstance registerMBean(Object object, ObjectName name) throws InstanceAlreadyExistsException, MBeanRegistrationException,
    NotCompliantMBeanException  { 

        ObjectName logicalName;

	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        Class theClass = object.getClass();   
	String className = theClass.getName();

        // Check the JMX compliance of the class        
        meta.testCompliance(theClass); 

        if (object == null) {    
            throw new RuntimeOperationsException(new IllegalArgumentException("Cannot add null object"), "Exception occured trying to register the MBean");
        }
        
        String nameMessage = "null";
        if (name!=null) {
            nameMessage = name.toString();
        }
        if (isTraceOn()) {
            trace(dbgTag, "registerMBean", "ObjectName = " + nameMessage);
        }          
        
        ObjectName res = null;
        if (object instanceof MBeanRegistration) {
            res=preRegisterInvoker(object,name);   
        }
   
        // if name passed in parameter is null, consider the ObjectName returned by preRegister
        if (name == null) {
            logicalName = res;
        } else {
            logicalName = name;
        }
        if (logicalName!=null) {
            internal_addObject(object, logicalName);
        } else {
            if (object instanceof MBeanRegistration ) {            
                postRegisterInvoker(object, false);        
            }
            throw new RuntimeOperationsException(new IllegalArgumentException("No object name specified"), "Exception occured trying to register the MBean");
        }       
        if (object instanceof MBeanRegistration)     
            postRegisterInvoker(object, true);

        /**
         * Checks if the newly registered MBean is a ClassLoader
         * If so, it adds the  MBean in the default loader repository.
         */     
        if (object instanceof ClassLoader) {
            DefaultLoaderRepository.addClassLoader((ClassLoader)object);
        }       

	// if the MBean is a dynamic MBean ask its MBeanInfo for the class name
	if (object instanceof DynamicMBean) {
	    try {
		MBeanInfo mbi = ((DynamicMBean)object).getMBeanInfo();
		
		if (mbi == null) {
		    throw new NotCompliantMBeanException("The MBeanInfo returned is null");
		}
		
		className = mbi.getClassName();
		
		if (className == null) {
		    throw new NotCompliantMBeanException("The class Name returned is null");
		}
	    } catch (Exception e) {
		 throw new NotCompliantMBeanException("Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	}

        return(new ObjectInstance(logicalName, className));
    } 

    /**
     * De-registers an MBean from the MBean server. The MBean is identified by
     * its object name. Once the method has been invoked, the MBean may
     * no longer be accessed by its object name.
     *
     * @param name The object name of the MBean to be de-registered.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     * @exception MBeanRegistrationException The preDeregister ((<CODE>MBeanRegistration</CODE>  interface) method of the MBean
     * has thrown an exception.
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The object name in parameter is null or 
     * the MBean you are when trying to de-register is the {@link javax.management.MBeanServerDelegate MBeanServerDelegate} MBean.
     *
     */
    public void unregisterMBean(ObjectName name) throws InstanceNotFoundException, MBeanRegistrationException  {    
        Object object;

        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Object name cannot be null"), "Exception occured trying to unregister the MBean");
        }  

        if (name.equals(MBeanServerDelegateObjectName)) {
            throw new RuntimeOperationsException(new IllegalArgumentException("The MBeanDelegate MBean cannot be unregistered"));
        }  
  
        synchronized(this) {
            object = repository.retrieve(name);
            if (object==null) {         
                if (isTraceOn()) {
                    trace("unregisterMBean", name+": Found no object");
                }           
                throw new InstanceNotFoundException(name.toString());          
            }   
            if (object instanceof MBeanRegistration) {
                preDeregisterInvoker(object);      
            }
            // Let the repository do the work.   
            try {
		repository.remove(name);            
            }
            catch (InstanceNotFoundException e) {
                throw e;
            }   
        
            // ---------------------
            // Send deletion event
            // ---------------------
            if (isTraceOn()) {
                trace("unregisterMBean", "Send delete notification of object " + name.getCanonicalName());
            }       
            sendNotification(MBeanServerNotification.UNREGISTRATION_NOTIFICATION, name ) ;

            if (object instanceof MBeanRegistration) {
                postDeregisterInvoker(object);      
            }
  
	    /**
	     * Checks if the unregistered MBean is a ClassLoader
	     * If so, it removes the  MBean from the default loader repository.
	     */

            if (object instanceof ClassLoader) {
                DefaultLoaderRepository.removeClassLoader((ClassLoader)object);
            }
        }
    } 

     /**
     * Gets the <CODE>ObjectInstance</CODE> for a given MBean registered with the MBean server.
     *
     * @param name The object name of the MBean.
     *
     * @return The <CODE>ObjectInstance</CODE> associated to the MBean specified by <VAR>name</VAR>.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     */
    public ObjectInstance getObjectInstance(ObjectName name) throws InstanceNotFoundException {
        Object obj = getMBean(name);
	String className = obj.getClass().getName();


	// if the MBean is a dynamic MBean ask its MBeanInfo for the class name
	if (obj instanceof DynamicMBean) {
	    try {
		MBeanInfo mbi = ((DynamicMBean)obj).getMBeanInfo();
		
		if (mbi == null) {
		    throw new JMRuntimeException("The MBeanInfo returned is null");
		}
		
		className = mbi.getClassName();
		
		if (className == null) {
		    throw new JMRuntimeException("The class Name returned is null");
		}
	    } catch (RuntimeException e) {
		throw new RuntimeMBeanException((RuntimeException)e, "Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	    catch (Error er) {
		throw new RuntimeErrorException((Error)er, "Error " + er.getClass().getName() + " :" + er.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
	    }
	}

        return new ObjectInstance(name, className);
    }
 /**
     * Gets MBeans controlled by the MBean server. This method allows any
     * of the following to be obtained: All MBeans, a set of MBeans specified
     * by pattern matching on the <CODE>ObjectName</CODE> and/or a Query expression, a
     * specific MBean. When the object name is null or no domain and key properties are specified,
     * all objects are to be selected (and filtered if a query is specified). It returns the
     * set of <CODE>ObjectInstance</CODE> objects (containing the <CODE>ObjectName</CODE> and the Java Class name) for
     * the selected MBeans.
     *
     * @param name The object name pattern identifying the MBeans to be retrieved. If
     * null or no domain and key properties are specified, all the MBeans registered will be retrieved.
     * @param query The query expression to be applied for selecting MBeans. If null
     * no query expression will be applied for selecting MBeans.
     *
     * @return  A set containing the <CODE>ObjectInstance</CODE> objects for the selected MBeans.
     * If no MBean satisfies the query an empty list is returned.
     *
     */
    public Set queryMBeans(ObjectName name, QueryExp query) {
        Set list = new HashSet(); 
        synchronized(this) {
            list =  repository.query(name, query);
        }
        // The repository performs the filtering
        if (queryByRepo) {
            return list;
        } else {
            // The filtering will be performed by the MBeanServer
            return(filterListOfObjects(list, query));
        }   
    } 

    /**
     * Gets the names of MBeans controlled by the MBean server. This method
     * enables any of the following to be obtained: The names of all MBeans,
     * the names of a set of MBeans specified by pattern matching on the
     * <CODE>ObjectName</CODE> and/or a Query expression, a specific MBean name (equivalent to
     * testing whether an MBean is registered). When the object name is
     * nullor no domain and key properties are specified, all objects are selected (and filtered if a
     * query is specified). It returns the set of ObjectNames for the MBeans
     * selected.
     *
     * @param name The object name pattern identifying the MBeans to be retrieved. If
     * null or no domain and key properties are specified, all the MBeans registered will be retrieved.
     * @param query The query expression to be applied for selecting MBeans. If null
     * no query expression will be applied for selecting MBeans.
     *
     * @return  A set containing the ObjectNames for the MBeans selected.
     * If no MBean satisfies the query, an empty list is returned.
     *
     */
    public Set queryNames(ObjectName name, QueryExp query) {

        HashSet result = new HashSet();    
        Set nos = queryMBeans(name, query);
        for (Iterator i  = nos.iterator(); i.hasNext(); ) {
            ObjectInstance oi     = (ObjectInstance) i.next();
            result.add(oi.getObjectName());     
        }
        return (result);        
    } 



    /**
     * Checks whether an MBean, identified by its object name, is already registered
     * with the MBean server.
     *
     * @param name The object name of the MBean to be checked.
     *
     * @return  True if the MBean is already registered in the MBean server, false otherwise.
     *
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The object name in parameter is null.
     *
     */
    public boolean isRegistered(ObjectName name)  { 
        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Object name cannot be null"), "Object name cannot be null");
        }  
        if (name.getDomain().equals("")) {
            try {
                name = new ObjectName(domain + name.toString());
            } catch (MalformedObjectNameException e) {
                if (isDebugOn()) {
                    debug("contains", "Unexpected MalformedObjectNameException");
                }       
            }
        }
        synchronized(this) {
            return (repository.contains(name));
        }
    } 


   
    /**
     * Returns the number of MBeans registered in the MBean server.
     */
    public Integer getMBeanCount()  { 
        return (repository.getCount());
    } 

    /**
     * Gets the value of a specific attribute of a named MBean. The MBean
     * is identified by its object name.
     *
     * @param name The object name of the MBean from which the attribute is to be retrieved.
     * @param attribute A String specifying the name of the attribute to be
     * retrieved.
     *
     * @return  The value of the retrieved attribute.
     *
     * @exception AttributeNotFoundException The attribute specified is not accessible in the MBean.
     * @exception MBeanException  Wraps an exception thrown by the MBean's getter.
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     * @exception ReflectionException  Wraps a <CODE>java.lang.Exception</CODE> thrown when trying to invoke the setter. 
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The object name in parameter is null or 
     * the attribute in parameter is null.
     */
    public Object getAttribute(ObjectName name, String attribute) throws
    MBeanException, AttributeNotFoundException, InstanceNotFoundException, ReflectionException { 

        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Object name cannot be null"), "Exception occured trying to invoke the getter on the MBean");
        }    
        if (attribute == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Attribute cannot be null"), "Exception occured trying to invoke the getter on the MBean");
        }    
        if (isTraceOn()) {
            trace("getAttribute", "Attribute= " + attribute + ", obj= " + name);
        }         
        Object instance = getMBean(name);
        return(getAttribute(instance, attribute));
    } 


    /**
     * Gets the value of a specific attribute of an MBean. 
     *
     * @param instance The MBean from which the attribute is to be retrieved.
     * @param attribute An String specifying the name of the attribute to be
     * retrieved.
     *
     * @return  The value of the retrieved attribute.
     *
     * @exception AttributeNotFoundException The specified attribute is not accessible in the MBean.
     * @exception MBeanException  Wraps an exception thrown by the MBean's getter.
     * @exception ReflectionException  Wraps a java.lang.Exception thrown while trying to invoke the getter.   
     */
  
    Object getAttribute(Object instance, String attribute)  throws
    MBeanException, AttributeNotFoundException, ReflectionException  {

        if (attribute == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Attribute name cannot be null"), "Exception occured trying to invoke the getter on the MBean");
        }  
     
        Class mbeanClass = meta.getMBeanInterface(instance.getClass());
        
        // Treatment for Dynamic MBean
        if (mbeanClass == null) {
            try {
                return ((javax.management.DynamicMBean)instance).getAttribute(attribute);
            } catch (RuntimeException e) {
                throw new RuntimeOperationsException(e, "RuntimeException thrown by the getAttribute method of the DynamicMBean " + " for the attribute " + attribute);
            } catch (Error e) {
                throw new RuntimeErrorException((Error)e, "Error thrown by the getAttribute method of the DynamicMBean " + " for the attribute " + attribute);                  
            }
        }

        // Query the metadata service to get the appropriate getter
        // of the object.  
        Method meth = null;
        meth = meta.findGetter(mbeanClass, attribute);
        if (meth == null) {
	    if (isTraceOn()) {
		trace("getAttribute", "Cannot find getter for "+attribute);
	    }                 
            throw new AttributeNotFoundException(attribute + " not accessible");
        }
        // Invoke the getter     
        if (isTraceOn()) {
            trace("getAttribute", "Invoke callback");
        }                 
        Object result= null;
        try {
            result = meth.invoke(instance,null);      
        } catch (InvocationTargetException e) {
            Throwable t = e.getTargetException();
            if (t instanceof RuntimeException) {
                throw new RuntimeMBeanException((RuntimeException)t, "Runtime Exception thrown in the getter for the attribute " + attribute); 
            } else if (t instanceof Error) {
                throw new RuntimeErrorException((Error) t ,"Error thrown in the getter for the attribute " + attribute);    
            } else {
                throw new MBeanException((Exception) t, "Exception thrown in the getter for the attribute " + attribute);    
            }
        } catch (RuntimeException e) {
            throw new RuntimeOperationsException(e, "Runtime exception thrown trying to invoke the getter for the attribute " + attribute);    
        }
        catch (IllegalAccessException e) {
            throw new ReflectionException(e, "Exception thrown trting to invoke the getter for the attribute " + attribute);    
        }
        catch (Error e) {
            throw new RuntimeErrorException((Error)e, "Error thrown trting to invoke the getter for the attribute " + attribute);    
        }
        if (isTraceOn()) {
            trace("getAttribute", attribute + "= " + result + "\n");
        }                 
        return result;   
    }

      /**
     * Enables the values of several attributes of a named MBean. The MBean
     * is identified by its object name.
     *
     * @param name The object name of the MBean from which the attributes are
     * retrieved.
     * @param attributes A list of the attributes to be retrieved.
     *
     * @return The list of the retrieved attributes.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     * @exception ReflectionException An exception occurred when trying to invoke the getAttributes method of a Dynamic MBean.
     * @exception RuntimeOperationsException Wrap a <CODE>java.lang.IllegalArgumentException</CODE>: The object name in parameter is null or 
     * attributes in parameter is null.
     *
     */
    public AttributeList getAttributes(ObjectName name, String[] attributes)
        throws InstanceNotFoundException, ReflectionException  { 
  
        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName name cannot be null"),"Exception occured trying to invoke the getter on the MBean");
        } 

        if (attributes == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Attributes cannot be null"), "Exception occured trying to invoke the getter on the MBean");
        } 
        AttributeList result = new AttributeList();
        // Check if we really have to work ...
        //
        if (attributes.length == 0)
            return result;
        if (isTraceOn()) {
            trace("getAttributes", "Object= " + name);
        }          
        Object obj = getMBean(name);

        Class mbeanClass = meta.getMBeanInterface(obj.getClass());

        // Treatment for Dynamic MBean
        if (mbeanClass == null) {
            try {
                return (((javax.management.DynamicMBean)obj).getAttributes(attributes));
            } catch (RuntimeException e) {
                throw new RuntimeOperationsException(e, "RuntimeException thrown trying by the getAttributes method of the DynamicMBean");
            } catch (Error e) {
                throw new RuntimeErrorException((Error)e, "Error thrown by the getAttributes method of the DynamicMBean ");                     
            }
        }
   
        // Go through the list of attributes
        //
        int maxLimit=  attributes.length;

        for (int i=0;i<maxLimit;i++) { 
            String elmt = (String)attributes[i];
            try {        
                Object value = getAttribute(obj, elmt);     
                result.add(new Attribute(elmt, value));
            }     
            catch (Exception excep) {
                if (isDebugOn()) {
                    debug("getAttributes", "Object= " + name);
                }          
            }
        }
        return result;
   
    } 

    /**
     * Sets the value of a specific attribute of a named MBean. The MBean
     * is identified by its object name.
     *
     * @param name The name of the MBean within which the attribute is to be set.
     * @param attribute The identification of the attribute to be set and the value it is to be set to.
     *
     * @return  The value of the attribute that has been set.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     * @exception AttributeNotFoundException The attribute specified is not accessible in the MBean.
     * @exception InvalidAttributeValueException The value specified for the attribute is not valid.
     * @exception MBeanException Wraps an exception thrown by the MBean's setter.
     * @exception ReflectionException  Wraps a <CODE>java.lang.Exception</CODE> thrown when trying to invoke the setter. 
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The object name in parameter is null or 
     * the attribute in parameter is null.
     */
    public void setAttribute(ObjectName name, Attribute attribute) throws
    InstanceNotFoundException, AttributeNotFoundException, InvalidAttributeValueException, MBeanException, 
    ReflectionException  { 

        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName name cannot be null"), "Exception occured trying to invoke the setter on the MBean");
        } 
   
        if (attribute == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Attribute cannot be null"), "Exception occured trying to invoke the setter on the MBean");
        } 
        if (isTraceOn()) {
            trace("setAttribute", "Object= " + name + ", attribute=" + attribute.getName());
        }          
        Object instance= getMBean(name);
        Object o = setAttribute(instance, attribute);

    }


    /**
     * Sets the value of a specific attribute of an MBean. 
     *
     * @param instance The MBean within which the attribute is to be set.
     * @param attribute The identification of the attribute to be set and the value it is to be set to.
     *
     * @return  The value of the attribute that has been set.
     *
     * @exception AttributeNotFoundException  The specified attribute is not accessible in the MBean.
     * @exception InvalidAttributeValueException The specified value for the attribute is not valid.
     * @exception MBeanException Wraps an exception thrown by the MBean's setter.
     * @exception ReflectionException  Wraps a java.lang.Exception thrown while trying to invoke the setter. 
     */
    Object setAttribute(Object instance, Attribute attribute) throws AttributeNotFoundException, InvalidAttributeValueException, MBeanException, ReflectionException {
        
        Class mbeanClass = meta.getMBeanInterface(instance.getClass());

        // Treatment for Dynamic MBean
        if (mbeanClass == null) {
            try {
                ((javax.management.DynamicMBean)instance).setAttribute(attribute);
                return null;
            } catch (RuntimeException e) {
                throw new RuntimeOperationsException(e, "RuntimeException thrown by the setAttribute " + attribute + "method of the Dynamic MBean");
            } catch (Error e) {
                throw new RuntimeErrorException((Error)e, "Error thrown by the setAttribute " + attribute + "method of the Dynamic MBean");                     
            }
        }
        
        Object result = null;
        Object value = attribute.getValue();
        value = checkUpdateContext(instance, value);

        // Query the metadata service to get the appropriate setter
        // of the object.
        Method meth = null;
	if (value == null) {
	    meth = meta.findSetter(mbeanClass, attribute.getName());     
	} else {
	    meth = meta.findSetter(mbeanClass, attribute.getName(), value.getClass());     
	}
        if (meth == null) {       
            // Check whether the type is a primitive one       
            Class primClass = meta.findPrimForClass(value);     
       
            if (primClass != null) {       
                meth = meta.findSetter(mbeanClass, attribute.getName(), primClass);     
            }
        }     
        if (meth == null) {
            // Try to check if the attribute name does correspond to a valid property       
            meth= meta.findSetter(mbeanClass, attribute.getName());
            if (meth == null) {
                throw new AttributeNotFoundException( attribute.getName() + " not accessible");
            }  else {
		Object v = attribute.getValue();
		if (v == null) {
		    throw new InvalidAttributeValueException("attribute= " + attribute.getName() + " value = null");
		} else {
		    throw new InvalidAttributeValueException("attribute= " + attribute.getName() + " value = " + attribute.getValue());
		}
            }
        }     
        // Invoke the setter     
        if (isTraceOn()) {
            trace("setAttribute", "Invoking the set method for " + attribute.getName());
        }   
     
        Object[] values = new Object[1];
        values[0] = value;

        try {
            result = meth.invoke(instance,values);

        } catch (IllegalAccessException e) {
            // Wrap the exception.            
            throw new ReflectionException(e, "IllegalAccessException occured trying to invoke the setter on the MBean");
        }
        catch (InvocationTargetException e) {
            Throwable t = e.getTargetException();
            if (t instanceof RuntimeException) {
                throw new RuntimeMBeanException((RuntimeException)t, "RuntimeException thrown in the MBean's setter");  
            } else if (t instanceof Error) {
                throw new RuntimeErrorException((Error) t, "Error thrown in the MBean's setter");   
            } else {
                throw new MBeanException((Exception) t, "Exception thrown in the MBean's setter");
            }
        }
        if (isTraceOn()) {
            trace("setAttribute", attribute.getName() + "= " + value);
        }         
        return value;
    }


      /**
     * Sets the values of several attributes of a named MBean. The MBean is
     * identified by its object name.
     *
     * @param name The object name of the MBean within which the attributes are to
     * be set.
     * @param attributes A list of attributes: The identification of the
     * attributes to be set and  the values they are to be set to.
     *
     * @return  The list of attributes that were set, with their new values.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     * @exception ReflectionException An exception occurred when trying to invoke the getAttributes method of a Dynamic MBean.
     * @exception RuntimeOperationsException Wraps a <CODE>java.lang.IllegalArgumentException</CODE>: The object name in parameter is null or 
     * attributes in parameter is null.
     *
     */
    public AttributeList setAttributes(ObjectName name, AttributeList attributes)
        throws InstanceNotFoundException, ReflectionException  { 

        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName name cannot be null"), "Exception occured trying to invoke the setter on the MBean");
        } 
        if (attributes == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("AttributeList  cannot be null"), "Exception occured trying to invoke the setter on the MBean");
        } 
        AttributeList result = new AttributeList();
        if (isTraceOn()) {
            trace("setAttributes", "Object= " + name);
        }            
        // Check if we really have to work ...
        //
        if ((attributes == null) || (attributes.isEmpty()))
            return result;

        Object obj = getMBean(name);

        Class mbeanClass = meta.getMBeanInterface(obj.getClass());

        // Treatment for Dynamic MBean
        if (mbeanClass == null) {
            try {
                return (((javax.management.DynamicMBean)obj).setAttributes(attributes));
            } catch (RuntimeException e) {
                throw new RuntimeOperationsException(e, "RuntimeException thrown by the setAttributes method of the Dynamic MBean");
            } catch (Error e) {
                throw new RuntimeErrorException((Error)e, "Error thrown trying by the setAttributes method of the Dynamic MBean");
            }
        }
        
        // Go through the list of attributes
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            Attribute attr = (Attribute) i.next();
            String id          = attr.getName();
            Object value       = attr.getValue();          
            try {
                Object newValue = null;                           
                newValue = setAttribute(obj, attr);  
                if (isTraceOn()) {
                    trace("setAttributes", "Updating the repository\n");
                }                                                
                result.add(new Attribute(id, newValue));
            } catch (Exception excep) {
                if (isDebugOn()) {
                    debug("setAttributes", "Unexpected exception occured: " + excep.getClass().getName());
                }
            }
        }
        return result;           
    }

    /**
     * Invokes an operation on an MBean.
     *
     * @param name The object name of the MBean on which the method is to be invoked.
     * @param operationName The name of the operation to be invoked.
     * @param params An array containing the parameters to be set when the operation is
     * invoked
     * @param signature An array containing the signature of the operation. The class objects will
     * be loaded using the same class loader as the one used for loading the MBean on which the operation was invoked.
     *
     * @return  The object returned by the operation, which represents the result ofinvoking the operation on the 
     * MBean specified.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.
     * @exception MBeanException  Wraps an exception thrown by the MBean's invoked method.
     * @exception ReflectionException  Wraps a <CODE>java.lang.Exception</CODE> thrown while trying to invoke the method.
     *
     */
    public Object invoke(ObjectName name, String operationName, Object params[],
                         String signature[]) throws InstanceNotFoundException, MBeanException, ReflectionException { 
        Object obj = getMBean(name);
        return(invoke(obj, operationName, params, signature));
    }
 
    /**
     * Invokes an operation on an MBean.
     *
     * @param instance The MBean on which the method is to be invoked.
     * @param operationName The name of the operation to be invoked.
     * @param params An array containing the parameters to be set when the operation is
     * invoked
     * @param signature An array containing the signature of the operation. The class objects will
     * be loaded using the same class loader as the one used for loading the
     * MBean on which the operation was invoked.
     *
     * @return  The object returned by the operation, which represents the result of
     * invoking the operation on the MBean specified.
     *
     * @exception MBeanException  Wraps an exception thrown by the MBean's invoked method.
     * @exception ReflectionException  Wraps a java.lang.Exception thrown while trying to invoke the method.
     */
    Object invoke(Object instance, String operationName, Object params[],
                  String signature[]) throws  MBeanException, ReflectionException {
     
        if (operationName == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Operation name  cannot be null"), "Exception occured trying to invoke the operation on the MBean");
        } 

        Class mbeanClass = meta.getMBeanInterface(instance.getClass());        

        // Treatment for Dynamic MBean
        if (mbeanClass == null) {
            try {
                return (((javax.management.DynamicMBean)instance).invoke(operationName, params, signature));
            } catch (RuntimeException e) {
                throw new RuntimeOperationsException(e, "RuntimeException thrown by the invoke method of the Dynamic MBean");
            } catch (Error e) {
                throw new RuntimeErrorException((Error)e, "Error thrown by the invoke in the invoke method of the Dynamic MBean");
            }
        }

        // Build the signature of the method
        //
        Class[] tab;
        if (signature == null)
            // The simple case !
            tab= null;
        else {
            // We need to have the class loader of the target object 
            //
            ClassLoader aloader= (ClassLoader) instance.getClass().getClassLoader();
            ClassLoader tmpLoader;
      
            int length= signature.length;
            tab= new Class[length];

            try {
                for (int i= 0; i < length; i++) {
                    // Start handling primitive types (int. boolean and so forth)
                    //
                    Class primCla = meta.findClassForPrim(signature[i]);
                    if (primCla != null) {
                        tab[i] = primCla;
                        continue;
                    }
                    // Ok we do not have a primitive type ! We need to build the signature of the method
                    // 
                    if (aloader != null) {
                        // We need to load the class through the class loader of the target object.
                        // 
                        tab[i]= aloader.loadClass(signature[i]);
                        tmpLoader= tab[i].getClassLoader();
                        if ((params[i] == null) || (tmpLoader == null)) 
                            continue;
            
                        // if the class is loaded from a different class loader than
                        //
                        if (!tmpLoader.equals(params[i].getClass().getClassLoader())) {     
                            params[i]= transferObject(params[i], aloader);
                        }
                    } else {
                        // Load through the default class loader
                        //
                        tab[i]= findClass(signature[i]);
                    }
                }
            } catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The parameter class could not be found");
            }
        }     
        // Query the metadata service to get the right method
        //    
        Method mth= meta.findMethod(mbeanClass, operationName, tab);
        
        if (mth == null) {
            if (isTraceOn()) {
                trace("invoke", operationName + " not found");
            }             
            throw new ReflectionException(new NoSuchMethodException(operationName), "The operation with name " + operationName + " could not be found");
        }        
        // invoke the method        
        if (isTraceOn()) {
            trace("invoke", "Invoking " + operationName);
        }           
        Object result=null;
        try {
            result=  mth.invoke(instance, params); 
        }
        catch (IllegalAccessException e) {
            throw new ReflectionException(e, "IllegalAccessException occured trying to invoke operation " + operationName);
        }
        catch (RuntimeException e) {
            throw new RuntimeOperationsException(e, "RuntimeException occured trying to invoke operation " + operationName);
        }        
        catch (InvocationTargetException e) {
            // Wrap the exception.         
            Throwable t = e.getTargetException();
            if (t instanceof RuntimeException) {
                throw new RuntimeMBeanException((RuntimeException)t, "RuntimeException thrown in operation " + operationName);     
            } else if (t instanceof Error) {
                throw new RuntimeErrorException((Error) t, "Error thrown in operation " + operationName);
            } else {
                throw new MBeanException((Exception) t, "Exception thrown in operation " + operationName);
            }
        }
        if (isTraceOn()) {
            trace("invoke", "Send the result");
        }      
        return (result);
    } 


    /**
     * Gets a specific MBean controlled by the MBeanServer.
     */
    private Object getMBean(ObjectName name)  throws InstanceNotFoundException {

        if (name == null) {
            throw new RuntimeOperationsException(new IllegalArgumentException("Object name cannot be null"), "Exception occured trying to get an MBean");
        } 
        if (name.getDomain().equals("")) {
            try {
                name = new ObjectName(domain + name.toString());
            } catch (MalformedObjectNameException e) {
                if (isDebugOn()) {
                    debug("getMBean", "Unexpected MalformedObjectNameException");
                }       
            }
        }
        Object obj = null;
        synchronized(this) {
            obj = repository.retrieve(name);    
            if (obj == null) {         
		if (isTraceOn()) {
		    trace("getMBean", name+": Found no object");
		}                       
		throw new InstanceNotFoundException(name.toString());          
            }  
        }
        return obj;
    }
   
    /**
     * Returns the default domain used for naming the MBean.
     * The default domain name is used as the domain part in the ObjectName
     * of MBeans if no domain is specified by the user.
     */
    public String getDefaultDomain()  { 
        return domain;
    } 

    /**
     * Enables to add a listener to a registered MBean.
     *
     * @param name The name of the MBean on which the listener should be added.
     * @param listener The listener object which will handle the notifications emitted by the registered MBean.
     * @param filter The filter object. If filter is null, no filtering will be performed before handling notifications.
     * @param handback The context to be sent to the listener when a notification is emitted.
     *
     * @exception InstanceNotFoundException The MBean name provided does not match any of the registered MBeans.
     */
    public void addNotificationListener(ObjectName name, NotificationListener listener, NotificationFilter filter, Object handback)
        throws InstanceNotFoundException{

	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        if (isTraceOn()) {
            trace("addNotificationListener", "obj= " + name);
        }          
        Object instance = getMBean(name);
        NotificationBroadcaster broadcaster = null ;

	if (!( instance instanceof NotificationBroadcaster)) {
            throw new RuntimeOperationsException(new IllegalArgumentException(name.getCanonicalName() ),  "The MBean " + name.getCanonicalName() +" does not implement the NotificationBroadcaster interface") ;
        }       
	broadcaster =  (NotificationBroadcaster) instance ;      

        // ------------------
        // Check listener 
        // ------------------
        if ( listener == null ) {
	    throw new RuntimeOperationsException(new IllegalArgumentException("Null listener"),"Null listener");
	}

        // ------------------
        // Looking for the corresponding, internal, listener
        // ------------------
        MBeanServerNotificationListener actualListener = (MBeanServerNotificationListener) internalListener.get(listener) ;
        if ( actualListener == null ) {
	    actualListener = new MBeanServerNotificationListener(listener, filter, handback, name) ;
	    internalListener.put(listener,actualListener) ;
	}
	
        // ------------------
        // add the listener
        // ------------------
	    broadcaster.addNotificationListener(actualListener, filter, handback) ;	
    }


     /**
     * Enables to add a listener to a registered MBean.
     *
     * @param name The name of the MBean on which the listener should be added.
     * @param listener The object name of the listener which will handle the notifications emitted by the registered MBean.
     * @param filter The filter object. If filter is null, no filtering will be performed before handling notifications.
     * @param handback The context to be sent to the listener when a notification is emitted.
     *    
     * @exception InstanceNotFoundException The MBean name of the notification listener or of the notification broadcaster
     * does not match any of the registered MBeans.
     *
     */
    public void addNotificationListener(ObjectName name, ObjectName listener, NotificationFilter filter, Object handback)
        throws InstanceNotFoundException{


	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        // ----------------
        // Get listener object
        // ----------------
        Object instance = getMBean(listener);
        if ( ! (instance instanceof NotificationListener )) {
	    throw new RuntimeOperationsException(new IllegalArgumentException(listener.getCanonicalName()), "The MBean " + listener.getCanonicalName() + "does not implement the NotificationListener interface") ;
        }

        // ----------------
        // Add a listener on a MBean
        // ----------------
        if (isTraceOn()) {
            trace("addNotificationListener", "obj= " + name + " listener= " + listener);
        }           
        addNotificationListener(name,(NotificationListener) instance, filter, handback) ;
    }


    /**
     * Enables to remove a listener from a registered MBean.
     *
     * @param name The name of the MBean on which the listener should be removed.
     * @param listener The listener object which will handle the notifications emitted by the registered MBean.
     * This method will remove all the information related to this listener.
     *
     * @exception InstanceNotFoundException The MBean name provided does not match any of the registered MBeans.
     * @exception ListenerNotFoundException The listener is not registered in the MBean.
     */
    public void removeNotificationListener(ObjectName name, NotificationListener listener )
        throws InstanceNotFoundException, ListenerNotFoundException {

        if (isTraceOn()) {
            trace("removeNotificationListener", "obj= " + name);
        }        
        Object instance = getMBean(name);

        NotificationBroadcaster broadcaster = null ;
        if ( instance instanceof NotificationBroadcaster ) {
            broadcaster =  (NotificationBroadcaster) instance ;
        } else {
            throw new RuntimeOperationsException(new IllegalArgumentException(name.getCanonicalName() ),  "The MBean " + name.getCanonicalName()+ " does not implement the NotificationBroadcaster interface") ;
        }

        // ------------------
        // Looking for the corresponding, internal, listener
        // ------------------
        MBeanServerNotificationListener actualListener = (MBeanServerNotificationListener) internalListener.get(listener) ;
        if ( actualListener == null )
	    {
                throw new ListenerNotFoundException("Unknown listener") ;
	    }
        internalListener.remove(listener) ;

        broadcaster.removeNotificationListener(actualListener) ;
    }

    /**
     * Enables to remove a listener from a registered MBean.
     *
     * @param name The name of the MBean on which the listener should be removed.
     * @param listener The object name of the listener which will handle the notifications emitted by the registered MBean.
     * This method will remove all the information related to this listener.
     *
     * @exception InstanceNotFoundException The MBean name provided does not match any of the registered MBeans.
     * @exception ListenerNotFoundException The listener is not registered in the MBean.
     */
    public void removeNotificationListener(ObjectName name, ObjectName listener) 
        throws InstanceNotFoundException, ListenerNotFoundException {

        // ----------------
        // Get listener object
        // ----------------
        Object instance ;
        try 
	    {
                instance = getMBean(listener);
	    }
        catch (InstanceNotFoundException e)
	    {
                throw new ListenerNotFoundException(e.getMessage()) ;
	    }

        if ( ! (instance instanceof NotificationListener )) {
            throw new RuntimeOperationsException( new IllegalArgumentException(listener.getCanonicalName() ),  "The MBean "+ name.getCanonicalName() + " does not implement the NotificationBroadcaster interface") ;
        }
        if (isTraceOn()) {
            trace("removeNotificationListener", "obj= " + name + " listener= " + listener);
        }               
        removeNotificationListener(name,(NotificationListener) instance) ;
    }

     /**
     * This method discovers the attributes and operations that an MBean exposes
     * for management.
     *
     * @param name The name of the MBean to analyze
     *
     * @return  An instance of <CODE>MBeanInfo</CODE> allowing the retrieval of all attributes and operations
     * of this MBean.
     *
     * @exception IntrospectionException An exception occurs during introspection.
     * @exception InstanceNotFoundException The MBean specified is not found.
     * @exception ReflectionException An exception occurred when trying to invoke the getMBeanInfo of a Dynamic MBean.
     */
    public MBeanInfo getMBeanInfo(ObjectName name) throws
    InstanceNotFoundException, IntrospectionException, ReflectionException { 


	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        MBeanInfo bi = null;
        Object moi = getMBean(name);

        Class mbeanClass = meta.getMBeanInterface(moi.getClass());        

        // Dynamic MBean treatment
        if (mbeanClass == null) {
            try {
                bi = (MBeanInfo)((javax.management.DynamicMBean)moi).getMBeanInfo();
            } catch (RuntimeException r) {
                throw new RuntimeMBeanException((RuntimeException)r, "Runtime Exception thrown by getMBeanInfo method of Dynamic MBean");
            }
            catch (Error e ) {
                throw new RuntimeErrorException((Error)e, "Error thrown by getMBeanInfo method of Dynamic MBean");
            }
            if (bi != null) {
                return (MBeanInfo) bi.clone();
            } else {
                throw new JMRuntimeException("Dynamic MBean " + name + " returns null MBeanInfo");
            }
            
        } else {     
            MBeanInfo mbi = meta.getMBeanInfo(moi.getClass());     
            return new MBeanInfo(mbi.getClassName(), mbi.getDescription(), mbi.getAttributes(),
                                 mbi.getConstructors(), mbi.getOperations(), findNotifications(moi));
        }
    }

    /** Returns true if the MBean specified is an instance of the specified class, false otherwise.
     * 
     * @param name The <CODE>ObjectName</CODE> of the MBean.
     * @param className The name of the class.
     *
     * @return true if the MBean specified is an instance of the specified class, false otherwise.
     *
     * @exception InstanceNotFoundException The MBean specified is not registered in the MBean server.          
     */
    public boolean isInstanceOf(ObjectName name, String className) throws InstanceNotFoundException {

	Object obj = getMBean(name);
	Class c = null;
	try {
	    c = findClass(className);
	} catch (ClassNotFoundException e) {
	    return false;
	}
	if (c.isInstance(obj)) {
	    return true;
	} else {
	    return false;
	}
    }

   /**
     * De-serializes a byte array in the context of the class loader 
     * of an MBean.
     *
     * @param name The name of the MBean whose class loader should be used for the de-serialization.
     * @param data The byte array to be de-sererialized.
     *
     * @return  The de-serialized object stream.
     *
     * @exception InstanceNotFoundException The MBean specified is not found.          
     * @exception OperationsException Any of the usual Input/Output related exceptions.
     *
     */
    public ObjectInputStream deserialize(ObjectName name, byte[] data)
        throws InstanceNotFoundException, OperationsException {

        // Check parameter validity    
        if (data == null) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Null data passed in parameter");
        }
        if (data.length == 0) {
	    throw new  RuntimeOperationsException(new IllegalArgumentException(), "Empty data passed in parameter");
        }
 
        // Get the MBean
        Object instance = getMBean(name);        
      
        // Object deserialization      
        ByteArrayInputStream bIn;
        ObjectInputStream    objIn;
        String               typeStr;

        bIn   = new ByteArrayInputStream(data);
        try {
            objIn = new ObjectInputStreamWithLoader(bIn,instance.getClass().getClassLoader());
        } catch (IOException e) {
            throw new OperationsException("An IOException occured trying to de-serialize the data");
        }
 
        return objIn;
    }


    /**
     * De-serializes a byte array in the context of a given MBean class loader.
     * The class loader is the one that loaded the class with name "className".
     *
     * @param name The name of the class whose class loader should be used for the de-serialization.
     * @param data The byte array to be de-sererialized.
     *
     * @return  The de-serialized object stream.
     *
     * @exception OperationsException Any of the usual Input/Output related exceptions.
     * @exception ReflectionException The specified class could not be loaded by the default loader repository     
     *
     */
    public ObjectInputStream deserialize(String className, byte[] data)
        throws OperationsException, ReflectionException {

        // Check parameter validity
        if (data == null) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Null data passed in parameter");
        }
        if (data.length == 0) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Empty data passed in parameter");
        }
        if (className == null) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Null className passed in parameter");
        }       

        // Load the class using the default loader repository    
        Class theClass = null;
        try {
            theClass= findClass(className);
        }
        catch (ClassNotFoundException e) {
            try {
                theClass = DefaultLoaderRepository.loadClass(className);
            }
            catch (ClassNotFoundException ee) {
                throw new ReflectionException(ee, "The MBean class could not be loaded by the default loader repository");
            }
        }
 
        // Object deserialization
        ByteArrayInputStream bIn;
        ObjectInputStream    objIn;
        String               typeStr;

        bIn   = new ByteArrayInputStream(data);
        try {
            objIn = new ObjectInputStreamWithLoader(bIn,theClass.getClassLoader());
        } catch (IOException e) {
            throw new OperationsException("An IOException occured trying to de-serialize the data");
        }
 
        return objIn;
    }

   
    
    /**
     * De-serializes a byte array in the context of a given MBean class loader.
     * The class loader is the one that loaded the class with name "className".
     * The name of the class loader to be used for loading the specified class is specified.
     * If null, the MBean Server's class loader will be used. 
     *
     * @param name The name of the class whose class loader should be used for the de-serialization.
     * @param data The byte array to be de-sererialized.
     * @param loaderName The name of the class loader to be used for loading the specified class.
     * If null, the MBean Server's class loader will be used.
     *
     * @return  The de-serialized object stream.
     *
     * @exception InstanceNotFoundException The specified class loader MBean is not found.          
     * @exception OperationsException Any of the usual Input/Output related exceptions.
     * @exception ReflectionException The specified class could not be loaded by the specified class loader.
     *
     */
    public ObjectInputStream deserialize(String className, ObjectName loaderName, byte[] data)
        throws InstanceNotFoundException, OperationsException, ReflectionException {

        // Check parameter validity
        if (data == null) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Null data passed in parameter");
        }
        if (data.length == 0) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Empty data passed in parameter");
        }
        if (className == null) {
            throw new  RuntimeOperationsException(new IllegalArgumentException(), "Null className passed in parameter");
        }       
        Class theClass = null;
        if (loaderName == null) {
            // Load the class using the agent class loader
            try {
                theClass = findClass(className);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the agent class loader");
            }
        } else {
            // Get the class loader MBean
            Object instance = getMBean(loaderName);     

            if (!(instance instanceof ClassLoader)) {
                throw new RuntimeOperationsException(new IllegalArgumentException(loaderName.toString() + " is not a ClassLoader"));
            }
            
            // Load the class using the class loader MBean
            try {
                theClass = ((ClassLoader)instance).loadClass(className);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "The MBean class could not be loaded by the " + loaderName.toString() + " class loader");
            }
        }
 
        // Object deserialization
        ByteArrayInputStream bIn;
        ObjectInputStream    objIn;
        String               typeStr;
        
        bIn   = new ByteArrayInputStream(data);
        try {
            objIn = new ObjectInputStreamWithLoader(bIn,theClass.getClassLoader());
        } catch (IOException e) {
            throw new OperationsException("An IOException occured trying to de-serialize the data");
        }
        
        return objIn;
    }
    
  
    /**
     * Instantiates an object given its class, using its empty constructor.
     * The call returns a reference to the newly created object.
     */
    private Object internal_instantiate(Class theClass) throws ReflectionException, MBeanException {
        Object moi = null;


	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        Constructor cons = meta.findConstructor(theClass, null);
        if (cons == null) {
            throw new ReflectionException(new NoSuchMethodException("No such constructor"));
        }
        // Instantiate the new object   
        try {     
            moi= cons.newInstance(null);
        }
        catch (InvocationTargetException e) {
            // Wrap the exception.         
            Throwable t = e.getTargetException();
            if (t instanceof RuntimeException) {        
                throw new RuntimeMBeanException((RuntimeException)t, "RuntimeException thrown in the MBean's empty constructor");       
            } else if (t instanceof Error) {
                throw new RuntimeErrorException((Error) t, "Error thrown in the MBean's empty constructor");    
            } else {
                throw new MBeanException((Exception) t, "Exception thrown in the MBean's empty constructor");  
            }
        }
        catch (NoSuchMethodError error) {
            throw new ReflectionException(new NoSuchMethodException("No constructor"), "No such constructor");
        }
        catch (InstantiationException e) {
            throw new ReflectionException(e, "Exception thrown trying to invoke the MBean's empty constructor");
        }
        catch (IllegalAccessException e) {
            throw new ReflectionException(e, "Exception thrown trying to invoke the MBean's empty constructor");
        }
        catch (IllegalArgumentException e) {
            throw new ReflectionException(e, "Exception thrown trying to invoke the MBean's empty constructor");
        }
        return moi;

    }

    /**
     * Instantiates an object given its class, the parameters and signature of its constructor
     * The call returns a reference to the newly created object.
     */
    private Object internal_instantiate(Class theClass, Object params[], String signature[])
        throws ReflectionException, MBeanException {
        // Instantiate the new object        

	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        Class[] tab;
        Object moi= null;
        try {     
            // Build the signature of the constructor             
            if (signature == null) {
                // The simple case !       
                tab= null;
            } else {
                int length= signature.length;
                tab= new Class[length];
                ClassLoader aloader= (ClassLoader) theClass.getClassLoader();
                ClassLoader tmpLoader;
                for (int i= 0; i < length; i++) {
                    // Start handling primitive types (int. boolean and so forth)           
                    Class primCla = meta.findClassForPrim(signature[i]);
                    if (primCla != null) {
                        tab[i] = primCla;
                        continue;
                    }
                    // Ok we do not have a primitive type ! We need to build the signature of the method            
                    if (aloader != null) {
                        // We need to load the class through the class loader of the target object.             
                        tab[i]= aloader.loadClass(signature[i]);
                        tmpLoader= tab[i].getClassLoader();
                        if ((params[i] == null) || (tmpLoader == null)) 
                            continue;           
                        // if the class is loaded from a different class loader than            
                        if (!tmpLoader.equals(params[i].getClass().getClassLoader())) {
                            params[i]= transferObject(params[i], aloader);
                        }
                    } else {
                        // Load through the default class loader                
                        tab[i]= findClass(signature[i]);
                    }
                }
            }
        } catch (ClassNotFoundException e) {
            throw new ReflectionException(e, "The constructor parameter classes could not be loaded");
        }

        // Exception IllegalArgumentException raised in Jdk1.1.8
        catch (IllegalArgumentException e) {
            throw new ReflectionException(e, "The constructor parameter classes could not be loaded");
        }
    
        // Query the metadata service to get the right constructor          
        Constructor cons = null;
        cons= meta.findConstructor(theClass, tab);
        
        if (cons == null) {
            throw new ReflectionException(new NoSuchMethodException("No such constructor"));
        }
        try {
            moi = cons.newInstance(params);     
        } 
        catch (NoSuchMethodError error) {
            throw new ReflectionException(new NoSuchMethodException("No such constructor found"), "No such constructor" );
        }
        catch (InstantiationException e) {
            throw new ReflectionException(e, "Exception thrown trying to invoke the MBean's constructor");
        }
        catch (IllegalAccessException e) {
            throw new ReflectionException(e, "Exception thrown trying to invoke the MBean's constructor");
        }
        catch (InvocationTargetException e) {
            // Wrap the exception.         
            Throwable th = e.getTargetException();
            if (th instanceof RuntimeException) {
                throw new RuntimeMBeanException((RuntimeException)th, "RuntimeException thrown in the MBean's constructor");    
            } else if (th instanceof Error) {
                throw new RuntimeErrorException((Error) th, "Error thrown in the MBean's constructor");   
            } else {
                throw new MBeanException((Exception) th, "Exception thrown in the MBean's constructor");
            }
        }       
        return moi;
    }
    
    
    /**
     * Invokes the  preRegister method  of an MBean that implements MBeanRegistration
     */
    private ObjectName preRegisterInvoker(Object moi, ObjectName name) throws InstanceAlreadyExistsException, MBeanRegistrationException {
   
        ObjectName logicalName=null;
        ObjectName newName=null;
      
        try {
            newName=(ObjectName)((MBeanRegistration)moi).preRegister(this, name);
        }
        catch (RuntimeException e) {
                throw new RuntimeMBeanException((RuntimeException)e, "RuntimeException thrown in preRegister method");       
	}
	catch (Error er) {      
                throw new RuntimeErrorException((Error) er, "Error thrown in preRegister method");
	}       
         catch (MBeanRegistrationException r){
	     throw (MBeanRegistrationException)r;
	 }
	 catch (Exception ex) {
	     throw new MBeanRegistrationException((Exception) ex, "Exception thrown in preRegister method");
	 }      
        
        if (name!=null) 
            logicalName=name;
        else
            logicalName=newName;       

        return logicalName;
    }

    /**
     * Invokes the  postRegister method  of an MBean that implements MBeanRegistration
     */
    private void postRegisterInvoker(Object moi, boolean registrationDone) {
 
        try {
            ((MBeanRegistration)moi).postRegister(new Boolean(registrationDone));
        }
        catch (RuntimeException e) {
                throw new RuntimeMBeanException((RuntimeException)e,  "RuntimeException thrown in postRegister method");   
	}
	catch (Error er)  {
	    throw new RuntimeErrorException((Error) er,  "Error thrown in postRegister method");       
	}       
        catch (Exception ee) {
        }
    }
  
    /**
     * Invokes the  preDeregister method  of an MBean that implements MBeanRegistration
     */
    private void preDeregisterInvoker(Object moi) throws MBeanRegistrationException {
    
        try {
            ((MBeanRegistration)moi).preDeregister();
        }
        catch (RuntimeException e) {
                throw new RuntimeMBeanException((RuntimeException) e,  "RuntimeException thrown in preDeregister method");     
	}
	catch (Error er) {         
	    throw new RuntimeErrorException((Error) er,  "Error thrown in preDeregister method");     
	}
	catch (MBeanRegistrationException t){
	    throw (MBeanRegistrationException)t;   
	}    
        catch (Exception ex) {
	    throw new MBeanRegistrationException((Exception)ex,  "Exception thrown in preDeregister method"); 
	}              
    }

    /**
     * Invokes the  postDeregister method  of an MBean that implements MBeanRegistration
     */
    private void postDeregisterInvoker(Object moi) {
      
        try {
            ((MBeanRegistration)moi).postDeregister();
        }
        catch (RuntimeException e) {    
                throw new RuntimeMBeanException((RuntimeException)e, "RuntimeException thrown in postDeregister method"); 
	}
	catch (Error er) {    
                throw new RuntimeErrorException((Error) er, "Error thrown in postDeregister method"); 
	}      
        catch (Exception ex) {
        }
        
    }


    /**
     * Adds a MBean in the repository
     */
    private void internal_addObject(Object object, ObjectName logicalName) throws InstanceAlreadyExistsException {

	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        if (isTraceOn()) {
            trace("internal_addObject", "Name= " + logicalName);
        }
        if (logicalName.getDomain().equals("")) {
            try {
                logicalName = new ObjectName(domain + logicalName.toString());
            } catch (MalformedObjectNameException e) {
                if (isDebugOn()) {
                    debug("internal_addObject", "Unexpected MalformedObjectNameException");
                }       
            }
        }
        // Let the repository do the work.
        
        synchronized(this) {
            try {
                repository.addMBean(object, logicalName);
            }
            catch (InstanceAlreadyExistsException e) {
                if (object instanceof MBeanRegistration ) {                     
                    postRegisterInvoker(object,false);       
                }
                throw e;
            }          
        }
        // ---------------------
        // Send create event
        // ---------------------
        if (isTraceOn()) {
            trace("addObject", "Send create notification of object " + logicalName.getCanonicalName());
        }         
       
        sendNotification(MBeanServerNotification.REGISTRATION_NOTIFICATION, logicalName ) ;
    }
   
    /**
     * Performs the necessary initializations for the MBeanServer.
     * Creates and registers the MetaData service and the MBeanServer identification MBean
     */
    private void initialize(RepositorySupport repository, String domain) {

	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------

        this.repository = repository;
    
        if ((domain == null) || (domain.equals(""))) {
            //this.domain = InetAddress.getLocalHost().getHostName();
            this.domain = repository.getDefaultDomain();
        }
        else {
            this.domain = domain;
        }
  
        // Set the domain name within the repository
        repository.setDefaultDomain(this.domain);
        if (isTraceOn()) {
            trace("new", "Initializing domain " + this.domain);
        }         
        // Check whether or not the repository is able to perform
        // the query service.
        //
        try {
            queryByRepo = repository.isFiltering();
        } catch (Exception e) {
            queryByRepo = false;
        }
        // Create the MetaData service  
        meta=new MetaData();
        
        // Create the MBeanServer identification MBean
        try {
            MBeanServerDelegateObject     = new MBeanServerDelegate () ;
            MBeanServerDelegateObjectName = new  ObjectName(ServiceName.DELEGATE) ;
                
            this.registerMBean(MBeanServerDelegateObject, MBeanServerDelegateObjectName ) ;     
        }
        catch (JMException e) {
            if (isDebugOn()) {
                debug("new", "Unexpected exception occured: " + e.getClass().getName());       
            }   
            
        }
        ClassLoader myLoader = this.getClass().getClassLoader();
        if (myLoader != null) {
            DefaultLoaderRepository.addClassLoader(myLoader);
        }
	DefaultLoaderRepository.addClassLoader(ClassLoader.getSystemClassLoader());
    }


    /**
     * Gets the class for the specified class name  
     */
    private Class findClass(String className) 
        throws ClassNotFoundException {
   
        Class theClass = null;  
        ClassLoader myLoader = this.getClass().getClassLoader();
        if (myLoader != null) {
            theClass = myLoader.loadClass(className);
        } else {
            theClass = Class.forName(className);                
        }
        return theClass;
    }

    /**
     * Gets the class for the specified class name using the specified class loader
     */

    private Class findClass(String className, ObjectName aLoader) 
        throws ClassNotFoundException, InstanceNotFoundException  {
    
        if (aLoader == null)
            return findClass(className);     
        // Retrieve the class loader from the repository         
        Object loader = null;
        synchronized(this) {
            loader = repository.retrieve(aLoader);
        }
        if (loader == null) {
            throw new InstanceNotFoundException("The loader named " + className + " is not registered in the MBeanServer");
        }     
        if (!(loader instanceof ClassLoader)) {
            throw new RuntimeOperationsException(new IllegalArgumentException(), aLoader.toString() + " is not a ClassLoader");
        }
 
        // Ok now we have the service. Let it do the work !              
        Class theClass = ((ClassLoader)loader).loadClass(className);
        return theClass;
    }

    /**
     * Load a class using the same class loader as a target object.
     */
    private Class findClass(Object target, String className) throws ReflectionException {
     
        ClassLoader aloader= target.getClass().getClassLoader();
        if (aloader == null) {
            try {
                return findClass(className);
            }
            catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "Failed to load class className using the MBean's class loader");
            }
     
        } else {
            try {
                return aloader.loadClass(className);
            } catch (ClassNotFoundException e) {
                throw new ReflectionException(e, "Failed to load class className using the MBean's class loader");
            }
        }
    }


    /**
     * The method checks if the context of the operation is ok, i.e if all the
     * objects are comming from the same class loader. If not, the method will
     * update the context.
     */
    private Object checkUpdateContext(Object target, Object value) {
 
        // Simple check, we never know ...
        //
        if (value == null)
            return null;
        ClassLoader t_loader= target.getClass().getClassLoader();
        if (t_loader == null)
            return value;
        ClassLoader v_loader= value.getClass().getClassLoader();
        if (v_loader == null)
            // The value comes from the system class loader. It is ok.
            return value;
        if (t_loader.equals(v_loader))
            // The two classes come from the same class loader. It is ok.
            return value;
        else 
            // We need to "tranfer" the value within the context of the class loader
            // of the object on which we have to apply the operation (target)
            //
            return transferObject(value, t_loader);
    }


    /**
     * Transfer a given object into the context of a different class loader
     */
    private Object transferObject(Object obj, ClassLoader aloader) {
        
        // Serialize the parameter        
        ByteArrayOutputStream arrayOut= new ByteArrayOutputStream();
    
        try {
            ObjectOutputStream out= new ObjectOutputStream(arrayOut);
            out.writeObject(obj);
            out.flush();
        } catch (Exception e) {
            return obj;
        } 
    
        // Start deserialization        
        try {
            ByteArrayInputStream  arrayIn= new ByteArrayInputStream(arrayOut.toByteArray());
            ObjectInputStreamWithLoader in= new ObjectInputStreamWithLoader (arrayIn, aloader);
            return in.readObject();
        } catch (Exception e) {
            return obj;
        }
    }  


    /**
     * Applies the specified queries to the set of objects
     */
    private Set filterListOfObjects(Set list, QueryExp query) {
        Set result = new HashSet();

        // No query ...
        if (query == null ) {
            for (Iterator i  = list.iterator(); i.hasNext(); ) {
                NamedObject no = (NamedObject) i.next();
		Object obj = no.getObject();
		String className = obj.getClass().getName();

		// if the MBean is a dynamic MBean ask its MBeanInfo for the class name		
		if (obj instanceof DynamicMBean) {
		    try {
			MBeanInfo mbi = ((DynamicMBean)obj).getMBeanInfo();
			
			if (mbi == null) {
			    throw new JMRuntimeException("The MBeanInfo returned is null");
			}
			className = mbi.getClassName();
		
			if (className == null) {
			    throw new JMRuntimeException("The class Name returned is null");
			}
		    } catch (RuntimeException e) {
			throw new RuntimeMBeanException((RuntimeException)e, "Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
		    }
		    catch (Error er) {
			throw new RuntimeErrorException((Error)er, "Error " + er.getClass().getName() + " :" + er.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
		    }
		}		
		result.add(new ObjectInstance(no.getName(), className));
            }
        } else {
    
            // Access the filter  
            for (Iterator i  = list.iterator(); i.hasNext(); ) {
                NamedObject no = (NamedObject) i.next();
                Object object = no.getObject();
		String className = object.getClass().getName();          
                boolean res = false;
		query.setMBeanServer(this);
                try {
                    res = query.apply(no.getName());
                } catch (Exception e) {
                    res = false;
                }
                if (res) {

		    // if the MBean is a dynamic MBean ask its MBeanInfo for the class name		
		    if (object instanceof DynamicMBean) {
			try {
			    MBeanInfo mbi = ((DynamicMBean)object).getMBeanInfo();
			    
			    if (mbi == null) {
				throw new JMRuntimeException("The MBeanInfo returned is null");
			    }
			    className = mbi.getClassName();
			    
			    if (className == null) {
				throw new JMRuntimeException("The class Name returned is null");
			    }
			} catch (RuntimeException e) {
			    throw new RuntimeMBeanException((RuntimeException)e, "Exception " + e.getClass().getName() + " :" + e.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
			}
			catch (Error er) {
			    throw new RuntimeErrorException((Error)er, "Error " + er.getClass().getName() + " :" + er.getMessage() + " thrown by the getMBeanInfo method of Dynamic MBean");
			}
		    }		
		    result.add(new ObjectInstance(no.getName(), className));		    
                } 
            }
        }
	return result;  
    }

    private void sendNotification(String NotifType, ObjectName name)
    {

	// ------------------------------ 
	// TIME BOMB HERE
	// ------------------------------
        // ---------------------
        // Create notification
        // ---------------------
        synchronized(this)
            {
                sequenceNumber = sequenceNumber + 1 ;
            }
        if (isTraceOn()) {
            trace("sendNotification", "Incr sequenceNumber = " + sequenceNumber);
        }        
        MBeanServerNotification notif = new MBeanServerNotification(NotifType,MBeanServerDelegateObjectName,sequenceNumber, name) ;
        MBeanServerDelegateObject.sendNotification(notif) ;
    }

    /**
     * Returns the MBeanNotificationInfo of the MBeans that implement
     * the NotificationBroadcaster interface.
     */
    private MBeanNotificationInfo[] findNotifications(Object moi) {
        
        if (moi instanceof javax.management.NotificationBroadcaster) {       
            MBeanNotificationInfo[] mbn = ((NotificationBroadcaster)moi).getNotificationInfo();
            if (mbn == null) {
                return new MBeanNotificationInfo[0];
            }
            MBeanNotificationInfo[] result = new MBeanNotificationInfo[mbn.length];
            for (int i = 0; i < mbn.length; i++) {
                result[i] = (MBeanNotificationInfo) mbn[i].clone();
            }
            return result;
        }
        return new MBeanNotificationInfo[0];
    }


    // TRACES & DEBUG
    //---------------

    private boolean isTraceOn() {
        return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MBEANSERVER);
    }

    private void trace(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MBEANSERVER, clz, func, info);
    }

    private void trace(String func, String info) {
        trace(dbgTag, func, info);
    }

    private boolean isDebugOn() {
        return Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_MBEANSERVER);
    }

    private void debug(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_MBEANSERVER, clz, func, info);
    }

    private void debug(String func, String info) {
        debug(dbgTag, func, info);
    }
  
}
