/*
 * @(#)file      RepositorySupport.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.35
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package com.sun.management.jmx;


// java import
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Set;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Vector;
import java.util.Iterator;

// RI import
import javax.management.* ; 
 

/**
 * The RepositorySupport  implements the Repository interface.
 * This repository does not support persistency.
 */
class RepositorySupport  implements  Repository { 


  /**
   * The structure for storing the objects is very basic.
   * A Hashtable is used for storing the different domains.
   * For each domain, a hashtable contains the instances
   */
   private Hashtable domainTb;

   /**
   * Number of elements contained in the Repository
   */
   private int nbElements = 0;

  /**
   * Domain name of the server the repository is attached to.
   * It is quicker to store the information in the repository rather
   * than querying the framework each time the info is required.
   */
   private String domain = ServiceName.DOMAIN;

    /** The name of this class to be used for tracing */
   private String dbgTag = "Repository";

   
    /**
     * The default constructor.
     */
    public RepositorySupport() {
	domainTb= new Hashtable();
    }	
    
    
    /**
     * The purpose of this method is to provide a unified way to provide whatever
     * configuration information is needed by the specific underlying implementation
     * of the repository.
     *
     * @param configParameters An list containing the configuration parameters needed by the specific
     * Repository Service implementation.
     */
    public void setConfigParameters(ArrayList configParameters) {
	return;
    } 

    /**
     * Indicates whether or not the Repository Service supports filtering. If
     * the Repository Service does not support filtering, the MBean Server
     * will perform filtering.
     *
     * @return  true if filtering is supported, false otherwise.
     */
    public boolean isFiltering() {
	// Let the MBeanServer perform the filtering !     
	return false;
    }
    
    /**
     * Stores an MBean associated with its object name in the repository.
     *
     *@param object MBean to be stored in the repository.
     *@param name MBean object name.
     *
     */
    public void addMBean(Object object, ObjectName name) throws InstanceAlreadyExistsException {
	
	if (isTraceOn()) {
	    trace("addMBean", "name=" + name);
	}  
	// Validate name for an object     
	if (name.isPattern() == true) {
	    throw new RuntimeOperationsException(new IllegalArgumentException("Repository: "+ name.toString()));
	}
	// Extract the domain name. 
	String dom = name.getDomain();     
	// Test if the domain is JMImplementation
	try {      
	    if ((!(query(new ObjectName("JMImplementation:*"), null).isEmpty())) && (dom.equals("JMImplementation"))){
		throw new RuntimeOperationsException(new IllegalArgumentException("Repository: domain name cannot be JMImplementation"));
	    }	         
	}
	catch (MalformedObjectNameException e) {
	} 
	
	if (dom.equals("")) {
	    dom= domain;
	}
	
	// If domain not already exists, add it to the hash table
	Hashtable moiTb= (Hashtable) domainTb.get(dom);
	if (moiTb == null) {
	    addNewDomMoi(object, dom, name);
	    return;
	}
	// Look at the instance now      
	Object elmt= (Object) moiTb.get(name);
	if (elmt != null) {       
	    throw new InstanceAlreadyExistsException(name.toString());
	} else {
	    nbElements++;
	    moiTb.put(name, object);
	}      
	return;      
    } 
    
    /**
     * Checks whether an MBean of the name specified is already stored in
     * the repository.
     *
     * @param name name of the MBean to find.
     *
     * @return  true if the MBean is stored in the repository, false otherwise.
     *
     */
    public boolean contains(ObjectName name) {
	if (isTraceOn()) {
	    trace("contains", "name=" + name);
	}  
	Object res= retrieve(name);
	return ((res==null) ? false : true);
    }
    
    /**
     * Retrieves the MBean of the name specified from the repository. The
     * object name must match exactly.
     *
     * @param name name of the MBean to retrieve.
     *
     * @return  The retrieved MBean if it is contained in the repository, null otherwise.
     *
     */
    public Object retrieve(ObjectName name) {
	if (isTraceOn()) {
	    trace("retrieve", "name=" + name);
	}  
	if (name.isPattern() == true) {
	    return null;
	}      
	// Extract the domain name.       
	String dom= name.getDomain();
	if (dom.equals("")) {
	    dom= domain;
	}
	Hashtable moiTb= (Hashtable) domainTb.get(dom);
	if (moiTb == null) {
	    return null;
	}  
	Object elmt= (Object) moiTb.get(name);
	return elmt;
    } 
    
    /**
     * Selects and retrieves the list of MBeans whose names match the specified
     * object name pattern and which match the specified query expression (optionally).
     *
     * @param name The name of the MBean(s) to retrieve - may be a specific object or
     * a name pattern allowing multiple MBeans to be selected.
     * @param query query expression to apply when selecting objects - this parameter will
     * be ignored when the Repository Service does not support filtering.
     *
     * @return  The list of MBeans selected. There may be zero, one or many MBeans returned
     * in the set.
     *
     */
    public Set query(ObjectName name, QueryExp query) {
	
	HashSet result = new HashSet();
	
	// The following filter cases are considered :
	// null, "", "*:*"" :  names in all domains
	// ":*" : names in defaultDomain
	// "domain:*" : names in the specified domain
	// "domain:[key=value], *"
	
	if (name == null) {
	    try {
		name=new ObjectName("*:*");
	    }
	    catch (MalformedObjectNameException e) {
	    }
	}
	ObjectName tmp=null;
	try {
	    tmp = new ObjectName("*:*");
	} catch (MalformedObjectNameException e) {
	}
	
	// if no pattern is contained in the name 
	if (!name.isPattern()) {       
	    try {
		Object o = retrieve(name);
		if (o != null) {
		    result.add(new NamedObject(name, o));
		}
	    }
	    catch (RuntimeOperationsException e ) {
	    }
	}
	//all  names in all domains
	else if  (name.equals(tmp)) {      
	    for(Enumeration e = domainTb.elements(); e.hasMoreElements();) {	
		Hashtable moiTb = (Hashtable) e.nextElement();
		for (Enumeration enumi = moiTb.keys(); enumi.hasMoreElements();) {	 
		    ObjectName key = (ObjectName) enumi.nextElement();
		    result.add(new NamedObject(key, moiTb.get(key)));
		}
	    }	  	 
	}      
	// all names in default domain
	else if (name.getDomain().equals("") && (name.getKeyPropertyListString().equals(""))) {		 
	    for(Enumeration e = domainTb.elements(); e.hasMoreElements();) {	   
		Hashtable moiTb = (Hashtable) e.nextElement();
		for (Enumeration enumi = moiTb.keys(); enumi.hasMoreElements();) {	     
		    ObjectName key = (ObjectName) enumi.nextElement();
		    // if the name's domain is the default domain
		    if (key.getDomain().equals(domain)) {	
			result.add(new NamedObject(key, moiTb.get(key)));
		    }
		}
	    }
	}
	else {
	    // Pattern matching in the domain name (*, ?)	 
	    Vector doms = findDomainToQuery(name);
	    String theDom = null;	 
	    for (Iterator i = doms.iterator();i.hasNext();) {
		theDom = (String)i.next();
		// all names in a specified domain
		if (name.getKeyPropertyListString().equals("")) {
		    Hashtable moiTb = (Hashtable) domainTb.get(theDom);
		    if (moiTb == null) {
			return result;
		    }
		    for (Enumeration enumi = moiTb.keys(); enumi.hasMoreElements();) {	     
			ObjectName key = (ObjectName) enumi.nextElement();
			result.add(new NamedObject(key, moiTb.get(key)));
		    }
		}
		else {
		    for(Enumeration e = domainTb.elements(); e.hasMoreElements();) {	 
			Hashtable moiTb = (Hashtable) e.nextElement();
			for (Enumeration enumi = moiTb.keys(); enumi.hasMoreElements();) {	     
			    ObjectName key = (ObjectName) enumi.nextElement();
			    // if the name's domain is the specified domain
			    if (key.getDomain().equals(theDom)) {	    
				// Look for the property list
				Hashtable toSearch = name.getKeyPropertyList(); 
				Hashtable list = key.getKeyPropertyList();
				boolean inResult = false;
				// if we have pattern "propertylist,*"
				if (name.isPropertyPattern()) {
				    inResult = true;
				    for(Enumeration enump = toSearch.keys(); enump.hasMoreElements();) {
					String prop = (String) enump.nextElement();	     
					// test if the value for a given key property is the same in the two hash tables
					if (!toSearch.get(prop).equals(list.get(prop))) {
					    inResult = false;
					    break;
					}	       
				    }
				} else {
				    // if we have pattern "propertylist" only				
				    if (list.size() == toSearch.size()) {
					inResult = true;
					for(Enumeration enum = toSearch.keys(); enum.hasMoreElements();) {
						String prop = (String) enum.nextElement();
						// test if the value for a given key property is the same in the two hash tables
						if (!toSearch.get(prop).equals(list.get(prop))) {
							inResult = false;
							break;
						}
					}					
				    }				    
				}
				// if all couples (property, value) are contained 
				if (inResult) {	     
				    result.add(new NamedObject(key, moiTb.get(key))); 
				}
			    } /* if */	     
			} /* for Enum enumi */
		    }/* for Enum e */
		} /* else */     	   
	    } /* for */
	}/* else */
	
	return result;     
    }
    
    /**
     * Removes an MBean from the repository.
     *
     * @param name name of the MBean to remove.
     *
     * @exception InstanceNotFoundException The MBean does not exist in the repository.
     */
    public void remove(ObjectName name) throws InstanceNotFoundException {
	if (isTraceOn()) {
	    trace("remove", "name=" + name);
	}  
	// Check if the object is known      
	if (contains(name) == false) {
	    throw new InstanceNotFoundException(name.toString());
	}
	String dom= name.getDomain();
	Hashtable moiTb= (Hashtable) domainTb.get(dom);
	if (moiTb == null) {
	    throw new InstanceNotFoundException(name.toString());
	}
	Object elmt= (Object) moiTb.get(name);
	if (elmt == null) {
	    throw new InstanceNotFoundException(name.toString());
	}
	// That's it we have it !      
	moiTb.remove(name);
	nbElements--;
    } 
    
    /**
     * Gets the number of MBeans stored in the repository.
     *
     * @return  Number of MBeans.
     */
    
    public Integer getCount() {
	return new Integer(nbElements);
    }
    
    /**
     * Gets the name of the domain currently used by default in the repository.
     *
     * @return  A string giving the name of the default domain name.
     */
    public  String getDefaultDomain() {
	return domain;
    }
    
    /**
     * Sets the name of the domain currently used by default in the repository.
     *
     * @param domain the default domain name.
     */
    public void setDefaultDomain(String domain) { 
	this.domain= domain;
    }
    

    private void addNewDomMoi(Object object, String dom, ObjectName name) {
	Hashtable moiTb= new Hashtable();
	domainTb.put(dom, moiTb);
	moiTb.put(name, object);
	nbElements++;
    }
    
    private Vector findDomainToQuery(ObjectName name) {
	Vector matchDoms = new Vector();
	String dom = name.getDomain();     
	if ((dom.indexOf("*") != 0) || (dom.indexOf("?") != 0)) {
	    for (Enumeration enumi = domainTb.keys(); enumi.hasMoreElements();) {	     
		String key = (String) enumi.nextElement();
		if (wildmatch(key, dom)) {
		    matchDoms.addElement(key); 
		}
	    }
	} else {
	    matchDoms.addElement(dom);
	}	 
	return matchDoms;
    }
    
    /*
     * Tests whether string s is matched by pattern p.
     * Supports "?", "*" each of which may be escaped with "\";
     * Not yet supported: internationalization; "\" inside brackets.<P>
     * Wildcard matching routine by Karl Heuer.  Public Domain.<P>
     */
    private boolean wildmatch(String s, String p) {
	char c;
	int si = 0, pi = 0;
	int slen = s.length();
	int plen = p.length();
	
	while (pi < plen) {            // While still string
	    c = p.charAt(pi++);
	    if (c == '?') {
		if (++si > slen)
		    return false;
	    } else if (c == '*') {        // Wildcard
		if (pi >= plen)
		    return true;
		do {
		    if (wildmatch(s.substring(si), p.substring(pi)))
			return true;
		} while (++si < slen);
		return false;           
	    } else {
		if (si >= slen || c != s.charAt(si++))
		    return false;
	    }
	}
	return (si == slen);
    }
    
    
    // TRACES & DEBUG
    //---------------
    
    private boolean isTraceOn() {
	return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MBEANSERVER);
    }
     
    private void trace(String clz, String func, String info) {
	Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MBEANSERVER, clz, func, info);
    }
    
    private void trace(String func, String info) {
        trace(dbgTag, func, info);
    }
    
    private boolean isDebugOn() {
        return Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_MBEANSERVER);
    }

    private void debug(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_MBEANSERVER, clz, func, info);
    }

    private void debug(String func, String info) {
        debug(dbgTag, func, info);
    }

 }
