/*
 * @(#)file      Trace.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.26
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package com.sun.management.jmx;

// java import
//
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Properties;
import java.util.Iterator;

// jmx import
//
import javax.management.*;

/**
 * This class is a static class and is used to get JMX internal runtime information.
 *
 * <P>Two levels of information can be specified:
 * <P>1) trace. It provides information to tell what JMX is doing, this information is used to help
 * a user to develop a JMX application.
 * <P>2) debug. It provides information to help a JMX developper to diagnose the JMX implementation.
 *
 * <P>A user can specify information types to select information. The following types are specified:
 * <P>1) INFO_ADAPTOR_HTML Information from HTML adaptor.
 * <P>2) INFO_MBEANSERVER. Information from a MBean Server.
 * <P>3) INFO_MLET. Information from a MLet service.
 * <P>4) INFO_MONITOR. Information from a Monitor.
 * <P>5) INFO_TIMER. Information from a Timer.
 * <P>6) INFO_MISC. Information from all other classes except those specified types.
 * <P>7) INFO_ALL. Information from all classes.
 * <P>8) INFO_RELATION. Information from Relation Service.
 * <P>8) INFO_MODELMBEAN. Information from the Model MBean components.
 *
 * <P>This class uses notification machanism to distribute the information. To register a listener,
 * a TraceFilter object should be provided to do filtering.
 *
 * There are two ways to receive trace information:
 *
 * <UL><LI> adding a listener with a filter in the code. It is possible to have more than one
 * listeners but with different filters.</LI>

 * <LI>specifying system properties in the command to start the Java
 * interpreter when you run a class. In this case, the code of the class must
 * include a call to the
 * {@link #parseTraceProperties parseTraceProperties} method of this class.
 * <P> To specifiy trace level, add -DLEVEL_TRACE or -DLEVEL_DEBUG to you command line. By default, the level is set to LEVEL_TRACE.
 * <P> To specifiy trace types, add the selected types to the command line as: -DINFO_MLET -DINFO_TIMER to
 * select mlet and timer information, it is possible to add more than one type. By default, the type is set to INFO_ALL.
 * <P> To specify output, add -DTRACE_OUTPUT=name_of_outfile to the command line,
 * name_of_outfile is a file name to write message. By default, the message is written to screen.
 * <P>
 * If at least of the three properties is specified in the command, all other
 * which are not specified will be set to default value.
 * </LI></UL>
 */
public class Trace {
    /**
     * Information level defined for trace level.
     * The formation will be provided to help development of JDMK applications.
     */
    public static final int LEVEL_TRACE = 1;

    /**
     * Information level defined for debug level.
     * The formation will be provided to help diagnosy. Selecting this level
     * will result to select the LEVEL_TRACE too.
     */
    public static final int LEVEL_DEBUG = 2;

    /**
     * Information type defined for MBean Server information.
     */
    public static final int INFO_MBEANSERVER = 1;

    /**
     * Information type defined for MLet service information.
     */
    public static final int INFO_MLET = 2;

    /**
     * Information type defined for Monitor information.
     */
    public static final int INFO_MONITOR = 4;

    /**
     * Information type defined for Timer information.
     */
    public static final int INFO_TIMER = 8;

    /**
     * Information type defined for HTML adaptor information.
     */
    public static final int INFO_ADAPTOR_HTML	= 16;

    /**
     * Information type defined for all other classes.
     */  
    public static final int INFO_MISC = 32;

    /**
     * Information type defined for Relation Service.
     */  
    public static final int INFO_RELATION = 64;

    /**
     * Information type defined for Model MBean
     */
    public static final int INFO_MODELMBEAN = 128;

    /**
     * Information type defined to represent all types defined.
     */
    public static int INFO_ALL = 
	INFO_MBEANSERVER |
	INFO_MLET |
	INFO_MONITOR |
	INFO_ADAPTOR_HTML |
	INFO_TIMER |
	INFO_MISC |
	INFO_RELATION |
	INFO_MODELMBEAN;

    // --------------
    // public methods
    // --------------

    /**
     * Verify whether the specified info level and the info type are selected by a listener.
     *
     * <P>It is strongly recommanded to call this method before sending an information
     * to this Trace class.
     *
     * @param level the level of trace information.
     * @param type the type of the trace information.
     */
    public static boolean isSelected(int level, int type) {
	boolean ret = false;

	// check levels
	if (level == 1 && selectedLevels != 0) {
	    ret = true;
	} else {
	    ret = ((selectedLevels & level) != 0);
	}

	// check types
	ret &= ((selectedTypes & type) != 0);

	return ret;
    }

    /**
     * Parses the system properties specified at the command-line to determine
     * the specified trace conditions
     */
    public static void parseTraceProperties() throws IOException {
	int lv = 0;
	int ty = 0;
		
	Properties p = System.getProperties();

	String outFile = p.getProperty("TRACE_OUTPUT");

	// get selected levels
	if (p.getProperty("LEVEL_TRACE") != null)
	    lv |= LEVEL_TRACE;

	if (p.getProperty("LEVEL_DEBUG") != null)
	    lv |= LEVEL_DEBUG;

	// get selected types
	if (p.getProperty("INFO_ADAPTOR_HTML") != null)
	    ty |= INFO_ADAPTOR_HTML;

	if (p.getProperty("INFO_MBEANSERVER") != null)
	    ty |= INFO_MBEANSERVER;

	if (p.getProperty("INFO_MLET") != null)
	    ty |= INFO_MLET;

	if (p.getProperty("INFO_MONITOR") != null)
	    ty |= INFO_MONITOR;

	if (p.getProperty("INFO_TIMER") != null)
	    ty |= INFO_TIMER;

	if (p.getProperty("INFO_MISC") != null)
	    ty |= INFO_MISC;

	if (p.getProperty("INFO_ALL") != null)
	    ty = INFO_ALL;

	if (p.getProperty("INFO_RELATION") != null)
	    ty = INFO_RELATION;
	
	if (p.getProperty("INFO_MODELMBEAN") != null)
	    ty = INFO_MODELMBEAN;

	// set default values if necessary
	if (outFile != null || lv != 0 || ty != 0) {
	    if (lv == 0)
		lv = LEVEL_TRACE;

	    if (ty == 0)
		ty = INFO_ALL;

	    TraceFilter filter = new TraceFilter(lv, ty);

	    if (outFile != null && !outFile.equals("")) {
		addNotificationListener(new TraceListener(outFile), filter, null);
	    } else {
		addNotificationListener(new TraceListener(), filter, null);
	    }
	}
    }

    /**
     * Send a new information to this Trace class
     *
     * @param level the level of trace information to be sent.
     * @param type the type of trace information to be sent.
     * @param className the name of the class from which the trace information is from.
     * @param methodName the name of the method from which the trace information is from.
     * @param info the trace information to be sent.
     * @return false if the level and the type are not selected.
     */
    public static boolean send(int level,
			       int type,
			       String className,
			       String methodName,
			       String info) {

	if (!(isSelected(level, type)))
	    return false;

	// construct a TraceNotification object
	TraceNotification notif = new TraceNotification("Trace",
							getSeq(type),
							getSeq(0),
							level,
							type,
							className,
							methodName,
							info,
							null);

	// send out
	distribute(notif);

	return true;
    }

    /**
     * Send an exception to this Trace class.
     *
     * @param level the level of trace information to be sent.
     * @param type the type of trace information to be sent.
     * @param className the name of the class from which the trace information is from.
     * @param methodName the name of the method from which the trace information is from.
     * @param exception exception sent as the trace information.
     */
    public static boolean send(int level,
			       int type,
			       String className,
			       String methodName,
			       Throwable exception) {
	if (!(isSelected(level, type)))
	    return false;

	// construct a TraceNotification object
	TraceNotification notif = new TraceNotification("Trace",
							getSeq(type),
							getSeq(0),
							level,
							type,
							className,
							methodName,
							null,
							exception);

	// send out
	distribute(notif);

	return true;
    }

    /**
     * Add a listener with filtering conditions
     *
     * @param listener the listener to be added to receive trace notification.
     * @param f the filter used to select a trace notification. This filter should be
     * an instance of the class TraceFilter or its sub-class. If the filter is set to null
     * all trace information with any type and any level will be send to the listener.
     * @param handback the object sent back to listener.
     * @exception IllegalArgumentException Thrown if the specified filter is not an instance
     * of the class TraceFilter or its sub-classes, or no listener is specified.
     */
    public static void addNotificationListener(NotificationListener listener,
					       NotificationFilter f,
					       Object handback) throws IllegalArgumentException {
	if (listener == null)
	    throw new IllegalArgumentException("A listener should be specified.");

	TraceFilter filter;
	if (f == null) {
	    filter = new TraceFilter(LEVEL_DEBUG, INFO_ALL);
	} else if (f instanceof TraceFilter) {
	    filter = (TraceFilter)f;
	} else {
	    throw new IllegalArgumentException("The filter should be an instance of the class TraceFilter.");
	}

	listeners.add(new ListenerInfo(listener, filter, handback));

	// update selected types
	updateSelected(filter.getLevels(), filter.getTypes());
    }

    /**
     * Add a listener with default filtering conditions:
     * all types but only trace level are selected.
     * 
     * @param listener the listener to be added to receive trace notification.
     * @param handback the object will send back to listener.
     * @exception IllegalArgumentException Thrown if the listener is null.
     */
    public static void addNotificationListener(TraceListener listener,
					       Object handback) throws IllegalArgumentException {
	if (listener == null)
	    throw new IllegalArgumentException("A listener should be specified.");

	TraceFilter filter = new TraceFilter(LEVEL_TRACE, INFO_ALL);
	listeners.add(new ListenerInfo(listener, new TraceFilter(LEVEL_TRACE, INFO_ALL), handback));

	// update selected levels and types
	//
	updateSelected(LEVEL_TRACE, INFO_ALL);
    }

    /**
     * Remove a listener
     *
     * @param listener the listener to be removed.
     */
    public static void removeNotificationListener(NotificationListener listener) {
	// remove the listener from the list
	//
	synchronized(listeners) {
	    for (Iterator i = listeners.iterator(); i.hasNext();) {
		ListenerInfo lsInfo = (ListenerInfo)i.next();
		if (lsInfo.listener == listener) {
		    listeners.remove(i);
	            break;
		}
	    }
	}

	// update selected levels and types
	//
	synchronized(listeners) {
	    for (Iterator i = listeners.iterator(); i.hasNext();) {
		ListenerInfo ls = (ListenerInfo)i.next();
		updateSelected(ls.filter.getLevels(), ls.filter.getTypes());
	    }
	}
    }

    /**
     * Remove all listeners.
     */
    public static void removeAllListeners() {
	listeners.clear();
	selectedLevels = 0;
	selectedTypes = 0;
    }

    // -----------------
    // protected methods
    // -----------------
    /**
     * check if specified types or levels are allowed.
     *
     * @param levels the levels to be checked.
     @ @param types the types to be checked.
    */
    static boolean selectedCheck(int levels, int types) {
	//check levels
	boolean ret = (levels == 1 || levels == 2 || levels == 3);

	// check types
	ret &= ((types | INFO_ALL) == INFO_ALL);

	return ret;
    }

    // get the string representation of a level
    //
    static String getLevel(int level) {
	String ret = null;

	switch (level) {
	case LEVEL_TRACE :
	    ret = "LEVEL_TRACE";
	    break;
	case LEVEL_DEBUG :
	    ret = "LEVEL_DEBUG";
	    break;
	default :
	    ret = "Unknown level";
	    break;
	}

	return ret;
    }

    // get string representation of a type
    //
    static String getType(int type) {

        String ret;

        switch (type) {
        case INFO_ADAPTOR_HTML :
            ret = "INFO_ADAPTOR_HTML";
            break;
        case INFO_MBEANSERVER :
            ret = "INFO_MBEANSERVER";
            break;
        case INFO_MLET :
            ret = "INFO_MLET";
            break;
        case INFO_MONITOR :
            ret = "INFO_MONITOR";
            break;
        case INFO_TIMER :
            ret = "INFO_TIMER";
            break;
        case INFO_MISC :
            ret = "INFO_MISC";
            break;
        case INFO_RELATION :
            ret = "INFO_RELATION";
            break;
	case INFO_MODELMBEAN :
	    ret = "INFO_MODELMBEAN";
	    break;
        default:
		ret = UNKOWNTYPE;
		break;
	}
 
        return ret;
    }

    // ---------------
    // private classes
    // ---------------
    // used to save the info about a listener
    //
    private static class ListenerInfo {
	public NotificationListener listener;
	public TraceFilter filter;
	public Object handback;

	public ListenerInfo(NotificationListener listener,
			    TraceFilter filter,
			    Object handback) {

	    this.listener = listener;
	    this.filter = filter;
	    this.handback = handback;
	}
    }

    // protected
    protected static String getRIType(int type) {
	return getType(type);
    }

    // ---------------
    // private methods
    // ---------------
    // distributes the trace info to listeners
    //
    private static void distribute(TraceNotification notif) {
	synchronized(listeners) {
	    for (Iterator i = listeners.iterator(); i.hasNext();) {
		ListenerInfo lsInfo = (ListenerInfo)i.next();
		if (lsInfo.filter == null ||
		    lsInfo.filter.isNotificationEnabled(notif)) {
		    try {
			lsInfo.listener.handleNotification(notif, lsInfo.handback);
		    } catch (Exception e) {
			// what to do? remove that listener?
			removeNotificationListener(lsInfo.listener);
			if (Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_MISC)) {
			    send(Trace.LEVEL_DEBUG, Trace.INFO_MISC, "Trace", "distribute", e);
			}
		    }
		}
	    }
	}
    }

    // update selected
    private static void updateSelected(int levels, int types) {
	// update selected levels
	selectedLevels |= levels;

	// update selected types
	selectedTypes |= types;
    }

    // get a sequence number for a specific type of trace
    //
    private static long getSeq(int type) {
	Integer ty = new Integer(type);

	Long seq;
	long numb;
	seq = (Long)type_seq.get(ty);
	if (seq == null) {
	    seq = new Long(0);
	    type_seq.put(ty, seq);
	}

	numb = seq.longValue();
	if (numb == Long.MAX_VALUE) {
	    numb = 1;
	} else {
	    numb += 1;
	}

	type_seq.put(ty, new Long(numb));

	return numb;
    }
	
    // -----------------
    // private variables
    // -----------------

    private static ArrayList listeners = new ArrayList();

    private static int selectedLevels = 0;
    private static int selectedTypes = 0;

    private static HashMap type_seq = new HashMap(20);

    protected static final String UNKOWNTYPE = "Unknown type";
}
