/*
 * @(#)file      MBeanAttributeInfo.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.16
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management;


/**
 * Describes an MBean attribute exposed for management.
 */
public class MBeanAttributeInfo extends MBeanFeatureInfo implements java.io.Serializable, Cloneable  { 


    /**
     * The attribute read method.
     */
    private transient java.lang.reflect.Method readMethod = null;
    
    /**
     * The attribute write method.
     */
    private transient java.lang.reflect.Method writeMethod = null;
    
    /**
     * The actual attribute type.
     */
    private String attributeType = null;
    
    
    /**
     * The attribute write right.
     */
    private boolean isWrite = false;
    
    /**
     * The attribute read right.
     */
    private boolean isRead = false;
    
    /**
     * Indicates if this method is a "is"
     */
    private boolean is = false;

   
    /**
     * Constructs an <CODE>MBeanAttributeInfo</CODE> object.
     *
     * @param name The name of the attribute.
     * @param type The type or class name of the attribute.
     * @param description A human readable description of the attribute.
     * @param isReadable True if the attribute has a getter method, false otherwise.
     * @param isWritable True if the attribute has a setter method, false otherwise.
     * @param is True if this attribute has an "is" getter, false otherwise.
     *
     */    
    public MBeanAttributeInfo(String name,
			      String type,
			      String description,
			      boolean isReadable,
			      boolean isWritable,
			      boolean isIs) {
	
	super(name, description); 
	this.attributeType = type;
	this.isRead= isReadable;
	this.isWrite = isWritable;
	if (isIs && !isReadable) {
	    throw new IllegalArgumentException("Cannot have an \"is\" getter for a non-readable attribute.");
	}
	if (isIs && (!type.equals("java.lang.Boolean") && (!type.equals("boolean")))) {   
	    throw new IllegalArgumentException("Cannot have an \"is\" getter for a non-boolean attribute.");
	}
	this.is = isIs;
	
    }
    
    /**
     * This constructor takes the name of a simple attribute, and Method
     * objects for reading and writing the attribute.
     *   
     * @param name The programmatic name of the attribute.
     * @param description A human readable description of the attribute.
     * @param getter The method used for reading the attribute value.
     *          May be null if the property is write-only.
     * @param setter The method used for writing the attribute value.
     *          May be null if the attribute is read-only.
     * @exception IntrospectionException There is a consistency problem in the definition of this attribute.
     */
    public MBeanAttributeInfo(String name, String description, java.lang.reflect.Method getter, 
			      java.lang.reflect.Method setter) throws IntrospectionException {
        super(name, description);
	this.readMethod = getter;
	this.writeMethod = setter;
	findAttributeType();
	findAttributeRights();
    }

    /**
     * Package-private duplicate constructor.
     * <P>
     * This must isolate the new object from any changes to the old object.
     */ 
    MBeanAttributeInfo(MBeanAttributeInfo old) {
	super(old.name, old.description);
        attributeType = old.attributeType;
        isRead = old.isRead;        
        isWrite = old.isWrite ;
	is = old.is;
    }

    /**
     * Creates and returns a copy of this object.
     */
    public Object clone () {
        return (new MBeanAttributeInfo(this));
    }
    
    /**
     * Returns the class name of the attribute.  
     */
    public String getType() {
	return attributeType;
    }
    
    /**
     * Whether the value of the attribute can be read.
     *
     * @return True if the attribute can be read, false otherwise.
     */
    public boolean isReadable() {
	return isRead;
    }
    
    /**
     * Whether new values can be written to the attribute.
     *
     * @return True if the attribute can be written to, false otherwise.
     */
    public boolean isWritable() {
	return isWrite;
    }
    
    /**
     * Indicates if this attribute has an "is" getter
     */
    public  boolean isIs() {
	return is;
    }
 

    /**
     * Sets the read/write rights of the attribute.
     */
    private void findAttributeRights() {	
        if (readMethod != null) {
            isRead = true;
	    if (readMethod.getName().startsWith("is")) {
		is = true;
	    }		
        }
        if (writeMethod != null) {
            isWrite = true;
        }	
    }

    /**
     * Finds the type of the attribute.     
    */

     private void findAttributeType()  throws IntrospectionException {
	 Class type = null;
	 try {
            attributeType = null;
            
            if (readMethod != null) {
    	        if (readMethod.getParameterTypes().length != 0) {
    	            throw new IntrospectionException("bad read method arg count");
    	        }
    	        type = readMethod.getReturnType();
    	        if (type == Void.TYPE) {
    	            throw new IntrospectionException("read method " + readMethod.getName() + " returns void");
    	        }
            }
            
            if (writeMethod != null) {
    	        Class params[] = writeMethod.getParameterTypes();
    	        if (params.length != 1) {
    	            throw new IntrospectionException("bad write method arg count");
    	        }
    	        if (type != null && type != params[0]) {
    	            throw new IntrospectionException("type mismatch between read and write methods");
    	        }
    	        type = params[0];
            }
            
        } catch (IntrospectionException ex) {
            throw ex;
        }
	attributeType = type.getName();
    }

 }
