/*
 * @(#)file      MBeanConstructorInfo.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.10
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management;


/**
 * Describes a constructor exposed by an MBean.
 */
public class MBeanConstructorInfo extends MBeanFeatureInfo implements java.io.Serializable, Cloneable  { 

 
    /**
     * The method.  
     */
    private transient java.lang.reflect.Constructor constructor = null;     
        
    /**
     * The signature of the method, that is, the class names of the arguments.
     */
    private MBeanParameterInfo[] signature = null;
    
    
    /**
     * Constructs an <CODE>MBeanConstructorInfo</CODE> object.
     *
     * @param description A human readable description of the operation.
     * @param constructor The <CODE>java.lang.reflect.Constructor</CODE> object describing the MBean constructor.
     */
    public MBeanConstructorInfo(String description, java.lang.reflect.Constructor constructor) {
	super(constructor.getName(), description);
	this.constructor = constructor;
	findMethodParameterTypes();
    }

    /**
     * Constructs an <CODE>MBeanConstructorInfo</CODE> object.
     *
     * @param name The name of the constructor.
     * @param description A human readable description of the constructor.
     * @param signature <CODE>MBeanParameterInfo</CODE> objects describing the parameters(arguments) of the constructor.
     */
    public MBeanConstructorInfo(String	   name, 
				 String	    description,
				 MBeanParameterInfo[]	    signature) {
	
	super(name, description);
	this.signature = signature;
    }

    /**
     * Package-private duplicate constructor.
     * <P>
     * This must isolate the new object from any changes to the old object.
     */
    
    MBeanConstructorInfo(MBeanConstructorInfo old) {
	super(old.name, old.description);
	signature = old.signature;
    }

    /**
     * Creates and returns a copy of this object.
     */
    public Object clone () {
           return  (new MBeanConstructorInfo(this)) ;
    }
    
    /**
     * Returns the signature of the method, that is, information on the operation's arguments.
     */
    public MBeanParameterInfo[] getSignature() {
	if (signature != null) {
	    int len = signature.length;
	    MBeanParameterInfo[] res = new MBeanParameterInfo[len];
	    for (int i = 0; i < len; i++) {	
		res[i] =  new MBeanParameterInfo(signature[i]);
	    }
	    return res;
	} else {
	    return new MBeanParameterInfo[0];
	}
     }


    /**
     * Sets the action signature and parameters.
     * <P>
     * All primitive type parameters (int, ..etc.) are converted to native type
     * parameters (java.lang.Integer, ..etc.).
     */
    private void findMethodParameterTypes() {
	String[] sigs = findSignatures(constructor.getParameterTypes());
	if (sigs != null) {
	    int len = sigs.length;
	    signature = new MBeanParameterInfo[len];
	    for (int i = 0; i < len; i++) {	
		signature[i] =  new MBeanParameterInfo("", sigs[i], "");
	    }       
	}	
    }

    /**
     * Converts the array of classes to an array of class signatures.
     */  
    private String[] findSignatures(Class[] clz) {
        String signers[] = new String[clz.length];
        for (int i = 0; i < clz.length; i++) {
            signers[i] = findSignature(clz[i]);
        }
        return signers;
    }

    /**
     * Converts the class to a class signature.
     */  
    private String findSignature(Class clz) {
        return clz.getName();
    }


 }

