/*
 * @(#)file      MBeanInfo.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.16
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management;


/**
 * Describes the management interface exposed by an MBean; 
 * that is, the set of attributes and operations which are available for management operations.
 */
public class MBeanInfo  implements Cloneable, java.io.Serializable  {
 

    /**
     * The human readable description of the class.     
     */
    private String description = null;
        
    /**
     * Tells if this is an open MBean.  
     */
    private boolean isOpen = false;
    
    /**
     * The MBean qualified name.
     */
    private String className = null;
    
    /**
     * The MBean attribute descriptors.
     */
    private MBeanAttributeInfo[] attributes = null;
    
    /**
     * The MBean operation descriptors.
     */
    private MBeanOperationInfo[] operations = null;

     /**
     * The MBean constructor descriptors.
     */
     private MBeanConstructorInfo[] constructors = null;

    /**
     * The MBean notification descriptors.
     */
     private MBeanNotificationInfo[] notifications = null;


    /**
     * Constructs an <CODE>MBeanInfo</CODE>.
     *   
     * @param name The name of the Java class of the MBean described by this <CODE>MBeanInfo</CODE>.
     * @param description A human readable description of the MBean (optional).
     * @param attributes   The list of exposed attributes of the MBean.    
     * @param constructors   The list of public constructors of the MBean.
     * @param operations   The list of operations of the MBean.
     * @param notifications   The list of notifications emitted by an MBean.
     *
     */
    public MBeanInfo(String	className, 
		     String	description,	
		     MBeanAttributeInfo[] attributes, 		
		     MBeanConstructorInfo[] constructors,
		     MBeanOperationInfo[] operations,
		     MBeanNotificationInfo[]	notifications) {
	
	this.className = className;
	this.description = description;
	this.attributes = attributes;
	this.operations = operations;
	this.constructors = constructors;
	this.notifications = notifications;
    }

    /*
     * Package-private duplicate constructor.
     * <P>
     * This must isolate the new object from any changes to the old object.
     */
    MBeanInfo(MBeanInfo old) {
	
	className = old.className;
	description = old.description;
	
	if (old.attributes != null) {
	    int len = old.attributes.length;
	    attributes = new MBeanAttributeInfo[len];
	    for (int i = 0; i < len; i++) {	
		attributes[i] = (MBeanAttributeInfo) (old.attributes[i]).clone();    	        
	    }
	}	    
	if (old.operations != null) {
	    int len = old.operations.length;
	    operations = new MBeanOperationInfo[len];
	    for (int i = 0; i < len; i++) {
		operations[i] = (MBeanOperationInfo) (old.operations[i]).clone();
	    }
	}
	if (old.constructors != null) {
	    int len = old.constructors.length;
	    constructors = new MBeanConstructorInfo[len];
	    for (int i = 0; i < len; i++) {
		constructors[i] = (MBeanConstructorInfo) (old.constructors[i]).clone();
	    }
	}
	if (old.notifications != null) {
	    int len = old.notifications.length;
	    notifications = new MBeanNotificationInfo[len];
	    for (int i = 0; i < len; i++) {
		notifications[i] = (MBeanNotificationInfo) (old.notifications[i]).clone();
	    }
	}
    }
    
    /**
     * Creates and returns a copy of this object.
     */
     public Object clone () {
           return  new MBeanInfo(this) ;
     }

    
    /**
     * Returns the name of the Java class of the MBean described by
     * this <CODE>MBeanInfo</CODE>.
     */    
    public String getClassName()  { 
	return className;
    } 
    
    /**
     * Returns a human readable description of the MBean.
     */    
    public String getDescription()  { 
	return description;
    } 
    
    /**
     * Returns the list of  attributes exposed for management. 
     * Each attribute is described by an <CODE>MBeanAttributeInfo</CODE> object.
     *
     * @return  An array of <CODE>MBeanAttributeInfo</CODE> objects.
     */    
    public MBeanAttributeInfo[] getAttributes()   { 
	if (attributes != null) {
	    int len = attributes.length;
	    MBeanAttributeInfo[] res = new MBeanAttributeInfo[len];
	    for (int i = 0; i < len; i++) {	
		res[i] = (MBeanAttributeInfo) (attributes[i]).clone();	        
	    }
	    return res;
	} else {
	    return new MBeanAttributeInfo[0];
	}
    } 

    /**
     * Returns the list of operations  of the MBean. 
     * Each operation is described by an <CODE>MBeanOperationInfo</CODE> object.
     *
     * @return  An array of <CODE>MBeanOperationInfo</CODE> objects.
     */    
    public MBeanOperationInfo[] getOperations()  { 
	if (operations != null) {
	    int len = operations.length;
	    MBeanOperationInfo[] res = new MBeanOperationInfo[len];
	    for (int i = 0; i < len; i++) {	
		res[i] = (MBeanOperationInfo) (operations[i]).clone(); 	        
	    }
	    return res;
	} else {
	    return new MBeanOperationInfo[0];
	}
    } 

    /**
     * Returns the list of the public constructors  of the MBean. 
     * Each constructor is described by an <CODE>MBeanConstructorInfo</CODE> object.
     *
     * @return  An array of <CODE>MBeanConstructorInfo</CODE> objects.
     */
    public MBeanConstructorInfo[] getConstructors()  { 
	if (constructors != null) {
	    int len = constructors.length;
	    MBeanConstructorInfo[] res = new MBeanConstructorInfo[len];
	    for (int i = 0; i < len; i++) {	
		res[i] = (MBeanConstructorInfo) (constructors[i]).clone();   	        	
	    }
	    return res;
	} else {
	    return new MBeanConstructorInfo[0];
	}
    } 

    /**
     * Returns the list of the notifications emitted by the MBean. 
     * Each notification is described by an <CODE>MBeanNotificationInfo</CODE> object.
     *
     * @return  An array of <CODE>MBeanNotificationInfo</CODE> objects.
     */
    public MBeanNotificationInfo[] getNotifications()  { 
	if (notifications != null) {
	    int len = notifications.length;
	    MBeanNotificationInfo[] res = new MBeanNotificationInfo[len];
	    for (int i = 0; i < len; i++) {	
		res[i] = (MBeanNotificationInfo) (notifications[i]).clone();       
	    }
	    return res;
	} else {
	    return new MBeanNotificationInfo[0];
	}
    } 

 }
