/*
 * @(#)file      MBeanOperationInfo.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.13
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management;


/**
 * Describes a management operation exposed by an MBean.
 */
public class MBeanOperationInfo extends MBeanFeatureInfo implements java.io.Serializable, Cloneable  { 

    
    /**
     * Indicates that the operation is read-like,
     * it basically returns information.
     */
    public static final int INFO = 0;
    
    /**
     * Indicates that the operation is a write-like,
     * and would modify the MBean in some way, typically by writing some value 
     * or changing a configuration.
     */
    public static final int ACTION = 1 ;
    
    /**
     * Indicates that the operation is a read/write-like.
     */
    public static final int ACTION_INFO = 2;
    
    /**
     * Indicates that the operation has an "unknown" nature.
     */
    public static final int UNKNOWN = 3;
    
    
    /**
     * The method.  
     */
    private transient java.lang.reflect.Method method = null;     
            
    /**
     * The method's return value.
     */
    private String type = null;
         
    /**
     * The signature of the method, that is, the class names of the arguments.
     */
    private MBeanParameterInfo[] signature = null;
        
    /**
     * Returns the impact of the method, one of
     * <CODE>INFO</CODE>, <CODE>ACTION</CODE>, <CODE>ACTION_INFO</CODE>, <CODE>UNKNOWN</CODE>.     
     */
    private int impact = MBeanOperationInfo.UNKNOWN;
    
    
    /**
     * Constructs an <CODE>MBeanOperationInfo</CODE> object.
     *
     * @param method The <CODE>java.lang.reflect.Method</CODE> object describing the MBean operation.
     * @param description A human readable description of the operation.
     */
    public MBeanOperationInfo(String description, java.lang.reflect.Method method) {
	super(method.getName(), description);
	this.method = method;
	findMethodReturnType();
	findMethodParameterTypes();
    }
    
    /**
     * Constructs an <CODE>MBeanOperationInfo</CODE> object.
     *
     * @param name The name of the method.
     * @param description A human readable description of the operation.
     * @param signature <CODE>MBeanParameterInfo</CODE> objects describing the parameters(arguments) of the method.
     * @param type The type of the method's return value.
     * @param impact The impact of the method, one of <CODE>INFO, ACTION, ACTION_INFO, UNKNOWN</CODE>.
     */
    public MBeanOperationInfo(String name, 
			      String description,
			      MBeanParameterInfo[] signature, 
			      String type,
			      int impact) {
	super(name, description);
	this.signature = signature;
	this.type = type;
	this.impact = impact;
    }
    
    /**
     * Package-private duplicate constructor.
     * <P>
     * This must isolate the new object from any changes to the old object.
     */
    MBeanOperationInfo(MBeanOperationInfo old) {
	super(old.name, old.description);
	signature = old.signature;
	type = old.type;
	impact = old.impact;
    }

    
    /**
     * Creates and returns a copy of this object.
     */
    public Object clone () {
        return  (new MBeanOperationInfo(this)) ;
    }
    
    /**
     * Returns the description of the method's return value.
     */
    public String getReturnType() {
	return type;
    }
     
    /**
     * Returns the signature of the method, that is, information on the operation's arguments.
     */
    public MBeanParameterInfo[] getSignature() {
	if (signature != null) {
	    int len = signature.length;
	    MBeanParameterInfo[] res = new MBeanParameterInfo[len];
	    for (int i = 0; i < len; i++) {	
		res[i] =  new MBeanParameterInfo(signature[i]);
	    }
	    return res;
	} else {
	    return new MBeanParameterInfo[0];
	}
     }
    
    /**
     * Returns the impact of the method, one of
     * <CODE>INFO</CODE>, <CODE>ACTION</CODE>, <CODE>ACTION_INFO</CODE>, <CODE>UNKNOWN</CODE>.
     */
    public int getImpact() {
	return impact;
     }

    
    /**
     * Sets the action signature and parameters.
     * <P>
     */
    private void findMethodParameterTypes() {
	String[] sigs = findSignatures(method.getParameterTypes());
	if (sigs != null) {
	    int len = sigs.length;
	    signature = new MBeanParameterInfo[len];
	    for (int i = 0; i < len; i++) {	
		signature[i] =  new MBeanParameterInfo("", sigs[i], "");
	    }       
	}	
    }

    /**
     * Converts the array of classes to an array of class signatures.
     */  
    private String[] findSignatures(Class[] clz) {
        String signers[] = new String[clz.length];
        for (int i = 0; i < clz.length; i++) {
            signers[i] = findSignature(clz[i]);
        }
        return signers;
    }

    /**
     * Converts the class to a class signature.
     */  
    private String findSignature(Class clz) {
        return clz.getName();
    }


    /**
     * Sets the action type.
     * <P>
     */
    private void findMethodReturnType() {
        type = method.getReturnType().getName();
    }

 }

