/*
 * @(#)file      NotificationBroadcasterSupport.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.30
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management;


/**
 * Provides an implementation of {@link javax.management.NotificationBroadcaster NotificationBroadcaster}
 * interface.
 * It could be used as a super class of MBean to deal with notifications.
 */
public class NotificationBroadcasterSupport implements NotificationBroadcaster { 


    /**
     * Listener hastable containing the hand-back objects.
     */  
    private java.util.Hashtable handbackTable = new java.util.Hashtable();
    
    /**
     * Listener hastable containing the filter objects.
     */  
    private java.util.Hashtable filterTable = new java.util.Hashtable();
    
    /**
     * Adds a listener to a registered MBean.
     *
     * @param listener The listener object which will handle the notifications emitted by the registered MBean.
     * @param filter The filter object. If filter is null, no filtering will be performed before handling notifications.
     * @param handback An opaque object to be sent back to the listener when a notification is emitted. This object
     * cannot be used by the Notification broadcaster object. It should be resent unchanged with the notification
     * to the listener.
     *
     * @exception IllegalArgumentException Listener parameter is null.
     */
    public synchronized void addNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback)  {

        // ---------------------------
        // Check listener
        // ---------------------------
        if (listener == null) {
            throw new java.lang.IllegalArgumentException ("Listener can't be null") ;
        }

        // ---------------------------
        // looking for listener in handbackTable
        // ---------------------------
        java.util.Vector handbackList = (java.util.Vector) handbackTable.get(listener) ; 
        java.util.Vector filterList = (java.util.Vector) filterTable.get(listener) ; 
        if ( handbackList == null ) {
	    handbackList = new java.util.Vector() ;
	    filterList = new java.util.Vector() ;
	    handbackTable.put(listener, handbackList) ;
	    filterTable.put(listener, filterList) ;
        }

        // ---------------------------
        // Add the handback and the filter
        // ---------------------------
        handbackList.addElement(handback) ;
        filterList.addElement(filter) ;
    }

    /**
     * Removes a listener from a registered MBean.
     *
     * @param name The name of the MBean on which the listener should be removed.
     * @param listener The listener object which will handle the notifications emitted by the registered MBean.
     * This method will remove all the information related to this listener.
     *
     * @exception ListenerNotFoundException The listener is not registered in the MBean.
     */
    public synchronized void removeNotificationListener(NotificationListener listener)
        throws ListenerNotFoundException {

        // ---------------------------
        // looking for listener in handbackTable
        // ---------------------------
        java.util.Vector handbackList = (java.util.Vector) handbackTable.get(listener) ; 
        java.util.Vector filterList = (java.util.Vector) filterTable.get(listener) ; 
        if (handbackList == null) {
            throw new ListenerNotFoundException("listener") ;
        }

        // ---------------------------
        // If handback is null, remove the listener entry
        // ---------------------------
        handbackTable.remove(listener) ;
        filterTable.remove(listener) ;
    }

    /**
     * Returns a NotificationInfo object contaning the name of the Java class of the notification
     * and the notification types sent.  
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[0];
    }  


    /**
     * Enables an MBean to send a notification.
     *   
     *  @param notification The notification to send.
     */
    public synchronized void sendNotification(Notification notification) {
        
        // ---------------------------
        // loop on listener
        // ---------------------------
        for(java.util.Enumeration k = handbackTable.keys(); k.hasMoreElements(); ) {
            
            NotificationListener listener = (NotificationListener) k.nextElement();

            // ---------------------------
            // Get the associated handback list and the associated filter list
            // ---------------------------
            java.util.Vector handbackList = (java.util.Vector) handbackTable.get(listener) ;
            java.util.Vector filterList = (java.util.Vector) filterTable.get(listener) ;

            // ---------------------------
            // loop on handback
            // ---------------------------
            java.util.Enumeration f = filterList.elements();
            for(java.util.Enumeration h = handbackList.elements(); h.hasMoreElements(); ) {                
                Object handback = h.nextElement();
                NotificationFilter filter = (NotificationFilter)f.nextElement();
                if ((filter == null) || 
                    ((filter != null) && (filter.isNotificationEnabled(notification)))) {
                    listener.handleNotification(notification,handback) ;
                }
            }
        }
    }
 
}
