/*
 * @(#)file      ObjectName.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.22
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management;


// java import
import java.io.Serializable;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;


/**
 * Represents the object name of an MBean. An instance of this class can be used to represent: 
 * <ul>
 * <li> An object name 
 * <li> An object name pattern, within the context of a query
 * </ul>
 */
public class ObjectName implements Serializable { 

    /**
     * Domain name.
     */
    private String domain = new String();

    /**
     * Table that contains all the pairs (key,value) for this ObjectName.
     */
    private Hashtable propertyList = new Hashtable();

    /**
     * Contains the properties in the same order
     * as the user specified them, if the constructor
     * ObjectName(String) was used to create this object.
     */
    private String propertyListString = new String();

    /**
     * Canonical name.
     */
    private String canonicalName = new String();

    /**
     * boolean that declares if this ObjectName contains a pattern
     */
    private boolean pattern = false;

    /**
     * boolean that declares if this ObjectName contains a pattern on the 
     * key property list
     */

    private boolean propertyPattern = false;


    /**
     * Allows an object name to be created from the given string. If the
     * string does not have the right format a MalformedObjectNameException should
     * be raised.
     *
     * @param name  A string representation of the object name.
     *
     * @exception MalformedObjectNameException The string passed as a parameter does not have the right format.
     */
    public ObjectName(String name) throws MalformedObjectNameException { 
        String domain = new String();
	Hashtable table = new Hashtable();
	ArrayList properties = new ArrayList();
	int index;

	// The name cannot be null     
	if (name == null)
	  throw new MalformedObjectNameException("ObjectName: Name cannot be null");

	// Test if the name is empty
	if (name.equals(" ") || name.equals("")) {
	  name="*:*";
	}
	// Test if the name contains a name pattern.       
	if ((name.indexOf("*") != -1) || (name.indexOf("?") != -1)) {
	  this.pattern=true;
	}    

	// Get the domain
	index = name.indexOf(":");     
	if (index == -1) {
	  throw new MalformedObjectNameException("ObjectName: domain part must be specified");
	} 
	domain = name.substring(0, index);
	name = name.substring(index+1, name.length());	   
	if (name.length() == 0) {
	    throw new MalformedObjectNameException("ObjectName: Key properties cannot be null");
	}
	
	// Get the key properties      
	StringTokenizer st, st2;
	String item, key, equ, val;
	// Process first property
	st = new StringTokenizer(name, ",", true);
	item = st.nextToken();
	st2 = new StringTokenizer(item, "=", true);
	try {
	  key = st2.nextToken();
	  equ = st2.nextToken();
	  val = st2.nextToken();
	  if (!isValid(val)) {
	      throw new  MalformedObjectNameException("ObjectName: Invalid value -> " + val);
	  }
	  if (st2.hasMoreTokens()) {
	    throw new MalformedObjectNameException("ObjectName: Invalid (key,value) pair -> " + item);
	  }  
	  properties.add(key);
	  table.put(key,val);	  
	} catch(MalformedObjectNameException e) {
	    throw e;
	}
	catch(Exception e) {
	    if (!pattern) {
		throw new MalformedObjectNameException("ObjectName: Invalid (key,value) pair -> " + item);
	    } 
	    if (item.compareTo("*") != 0) {
		throw new MalformedObjectNameException("ObjectName: Invalid pattern -> " + item);
	    } 
	    propertyPattern = true;	        
	}
		 
	// If we have multiple key properties, process them
	while(st.hasMoreTokens()) {
	  try {
	    item = st.nextToken();
	    if (!(item.equals(","))) {
	      throw new MalformedObjectNameException("ObjectName: Invalid property list");
	    }
	    item = st.nextToken();
	  } catch(Exception e) {
	    throw new MalformedObjectNameException("ObjectName: Invalid property list");
	  }
	   st2 = new StringTokenizer(item, "=", true);
	   try {
	     key = st2.nextToken();
	     equ = st2.nextToken();
	     val = st2.nextToken();
	     if (!isValid(val)) {
		 throw new  MalformedObjectNameException("ObjectName: Invalid value -> " + val);
	     }
	     if (st2.hasMoreTokens()) {
	       throw new MalformedObjectNameException("ObjectName: Invalid (key,value) pair -> " + item);
	     }
	     if (table.containsKey(key)) {
		 throw new MalformedObjectNameException("ObjectName: key -> " + key  + " already exists");
	     } else {
		 properties.add(key);
		 table.put(key,val);
	     }	     
	   } catch(MalformedObjectNameException e) {
	       throw e;
	   }
	   catch(Exception e) {
	       if (!pattern) {
		   throw new MalformedObjectNameException("ObjectName: Invalid (key,value) pair -> " + item);	       
	       } 
	       if (item.compareTo("*") != 0) {
		   throw new MalformedObjectNameException("ObjectName: Invalid pattern -> " + item);
	       } else {
		   this.propertyPattern = true;
	       }	       
	   }
	}   
	initObjectName(domain, table, properties);
    }

    /**
     * Allows an object name to be created using only one key property. If
     * the parameters do not have the right format a MalformedObjectNameException should
     * be raised.
     *
     * @param domain  The domain part of the object name.     
     * @param key  The attribute in the key property of the object name.     
     * @param value  The value in the key property of the object name.
     *
     * @exception MalformedObjectNameException The string passed as a parameter does not have the right format.
     */
    public ObjectName(String domain, String key, String value) throws MalformedObjectNameException { 
        Hashtable table = new Hashtable();

	try {
	  table.put(key, value);
	} catch(NullPointerException e) {
	    throw new MalformedObjectNameException("ObjectName: Neither the key nor the value can be null");
	}	
	initObjectName(domain, table, null);
    }

    /**
     * Allows an object name with several key properties to be created. If
     * the parameters do not have the right format a MalformedObjectNameException should
     * be raised.
     *
     * @param domain  The domain part of the object name.
     * @param table  A hash table containing one or more search keys. Each search key
     * is an attribute-value pair.
     *
     * @exception MalformedObjectNameException The string passed as a parameter does not have the right format.
     */
    public ObjectName(String domain, Hashtable table) throws MalformedObjectNameException { 
	if (table == null) {
	    throw new MalformedObjectNameException("ObjectName: Hashtable is null");
	}	
	if (table.isEmpty()) {
	    throw new MalformedObjectNameException("ObjectName: Empty Hashtable");
	}
        initObjectName(domain, table, null);
    } 


    /**
     * Compares the current object name with another object name.
     *
     * @param object  The object name that the current object name is to be compared with.
     *
     * @return  True if the two object names are equal, otherwise false.
     */
    public boolean equals(Object object)  {   
       if (!(object instanceof ObjectName)) {
	 return false;
       }
       ObjectName val = (ObjectName) object;  
       return canonicalName.equals(val.getCanonicalName());     
   }

    /**
     * Returns a hash code for this object name.
     *
     */   
    public int hashCode() {
        return canonicalName.hashCode();
    }

    /**
     * Returns a string representation of the object name.
     */
    public String toString()  {   
        String result  = domain + ":";
	if (propertyListString.length() > 0) {
	  result += propertyListString;       
	}
	return result;
    }

    /**
     * Checks whether the object name is a pattern for a query.
     *
     * @return  True if the name is a pattern, otherwise false.
     */
    public boolean isPattern() {    
        return pattern;
    }
 

    /**
    * Returns the canonical form of the name; that is, a string representation where
    * the properties are sorted in lexical order.
    */
    public String getCanonicalName()  {    
        return canonicalName;            
   } 

    /**
     * Returns the domain part.
     */    
    public String getDomain()  {       
        return domain;        
    } 
   
    /**
    * Obtains the value of a key property in a key property.
    *
    * @param property The property whose value is to be obtained.
    *
    * @return  The value of the property.
    */
    public String getKeyProperty(String property)  {      
        return (String)propertyList.get(property);          
   }     

    /**
    * Returns a pointer to the list of key properties.
    */
    public Hashtable getKeyPropertyList()  { 
        return propertyList;
    }     

    /**
     * Returns a string representation of the list of key properties specified
     * at creation time.
     */
    public String getKeyPropertyListString()  { 
        return propertyListString;
    } 

    /**
     * Returns a string representation of the list of key properties, in
     * which the key properties are sorted in lexical order. This is used
     * in lexicographic comparisons performed in order to select MBeans based
     * on their key property list.
     */
    public String getCanonicalKeyPropertyListString()  { 
         int propertyListStart = domain.length() +  1;

         try {
           return new String(canonicalName.substring(propertyListStart, canonicalName.length()));
         } catch(Exception e) {
	   return new String();
	 } 
    }

    
    /**
     * Checks whether the object name is a pattern on the key properties.
     *
     * @return  True if the name is a pattern, otherwise false.
     */
     public boolean isPropertyPattern() {    
        return propertyPattern;
    }


    /**
     * Initialise the ObjectName
     */    
    private void initObjectName(String domain, Hashtable table, ArrayList properties) throws MalformedObjectNameException {

        if ((domain == null) || (table == null))
	  throw new MalformedObjectNameException("ObjectName: Invalid object name");
	if (isDomain(domain)) {
	    setDomain(domain);	    
	} else {
	  throw new MalformedObjectNameException("ObjectName: Invalid domain -> " + domain);
	}
	if (properties == null) {
	    for (Enumeration e = table.keys(); e.hasMoreElements();) {	    	
		String key = null;
		try {	
		    key = (String) e.nextElement();
		} catch (ClassCastException ee) {
		    throw new MalformedObjectNameException("ObjectName: Invalid key : Not a String");
		}	    
		if (!isValid(key)) {
		    throw new MalformedObjectNameException("ObjectName: Invalid key -> " + key);
		}
		Object prop = table.get(key);
		if (!(prop instanceof java.lang.String)) {
		    throw new MalformedObjectNameException("ObjectName: Invalid value : Not a String");
		}
		if (!isValid((String)prop)) {
		    throw new MalformedObjectNameException("ObjectName: Invalid value -> " + prop);
		}	  
		setProperty(key, (String)prop);	  
	    }
	} else {
	    for (Iterator i = properties.iterator();i.hasNext();)  {	    	
		String key = null;
		try {	
		    key = (String) i.next();
		} catch (ClassCastException ee) {
		    throw new MalformedObjectNameException("ObjectName: Invalid key : Not a String");
		}	    
		if (!isValid(key)) {
		    throw new MalformedObjectNameException("ObjectName: Invalid key -> " + key);
		}
		Object prop = table.get(key);
		if (!(prop instanceof java.lang.String)) {
		    throw new MalformedObjectNameException("ObjectName: Invalid value : Not a String");
		}
		if (!isValid((String)prop)) {
		    throw new MalformedObjectNameException("ObjectName: Invalid value -> " + prop);
		}	  
		setProperty(key, (String)prop);	  
	    }
	}
    }

    /**
     * Sets the domain name.
     */
    private void setDomain(String domain) {
        this.domain = domain;
	canonicalName += domain + ":";
    }

    /**
     * Add a new property to the propertyList hashtable
     */  
    private void setProperty(String property, String value) { 
        // Add entry to the propertyList hashtable
        propertyList.put(property, value);     
        // Update the propertyListString
	String commaProperty = ","+propertyListString;
	int locateProperty =  commaProperty.indexOf(","+property+"=");
	if (locateProperty == -1) {
	  if (propertyListString.length() > 0) {
	    propertyListString += ",";
	  }
	  propertyListString += property+"="+value;
	} else {
	  String newProperty = propertyListString.substring(0, locateProperty);
	  newProperty += property+"="+value;
	  int locateRemProp = propertyListString.indexOf(",", locateProperty);
	  if (locateRemProp != -1) {
	    newProperty += propertyListString.substring(locateRemProp);
	  }
	  propertyListString = newProperty;
	}	
	// Insert the new property in the canonicalName string
	// so that the properties are still sorted
	int propertyStart = domain.length() + 1;
	if (canonicalName.length() == propertyStart) {
	  canonicalName += property +"="+value;
	} else {
	  StringBuffer newCanonical = new StringBuffer(canonicalName.substring(0, propertyStart));
	  String canonicalProp = canonicalName.substring(propertyStart);
	  StringTokenizer st = new StringTokenizer(canonicalProp, ",=", false);
	  boolean first = true;
	  int compProp = -1;
	  while (st.hasMoreTokens()) {
	    String oneProp = st.nextToken();
	    String oneValue = st.nextToken();
	    if (first) {
	      first = false;
	    } else {
	      newCanonical.append(",");
	    }
	    compProp = oneProp.compareTo(property);
	    if (compProp < 0) {
	      newCanonical.append(oneProp + "="+ oneValue);
	    } else {
	      newCanonical.append(property + "=" + value);
	      if (compProp > 0) {
		newCanonical.append(",");
		newCanonical.append(oneProp +"=" + oneValue);
	      }
	      break;
	    }
	  }
	  if (compProp >= 0) {
	    while (st.hasMoreTokens()) {
	      String oneProp = st.nextToken();
	      String oneValue = st.nextToken();
	      newCanonical.append("," + oneProp + "="+ oneValue);
	    }
	  } else {
	    newCanonical.append("," + property + "=" + value);
	  }
	  canonicalName = newCanonical.toString();
	}
    }
   
    /**
     * Check if domain is a valid domain
     */
    private boolean isDomain(String domain) {
        StringTokenizer st = new StringTokenizer(domain, ":=,", true);
	int nt = st.countTokens();
	if (nt == 0) {
	  return true;
	}
	if (nt == 1) {
	  String item = st.nextToken();
	  if (!(item.equals(":") || item.equals("=") || item.equals(","))) {
	    return true;
	  }
	}
	return false;
    }
   
    /**
     * Check if property is a valid key for a hashtable
     */
    private boolean isValid(String property) {
        StringTokenizer st = new StringTokenizer(property, ":=,*?", true);
	int nt = st.countTokens();
	if (nt == 1) {
	  String item = st.nextToken();
	  if (!(item.equals(":") || item.equals("=") || item.equals(",") 
		|| item.equals("*") || item.equals("?"))) {
	    return true;
	  }
	}
	return false;
    }

 }

