/*
 * @(#)file      DefaultLoaderRepository.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.15
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.loading;


// Java import
import java.io.Serializable;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;

// RI import
import com.sun.management.jmx.Trace ;

/**
 * Keeps the list of Class Loaders registered in the MBean Server.
 * It provides the necessary methods to load classes using the registered Class Loaders.
 *
 */
public class DefaultLoaderRepository implements Serializable {
    

    /**
     * List of class loaders
     */ 
    protected static Vector loaders= new Vector();
  
    /**
     * List of valid search
     */
    private static Hashtable search= new Hashtable();
    
    private static final String dbgTag = "DefaultLoaderRepository";


    /**
     * Go through the list of class loaders and try to load the requested class.
     * The method will stop as soon as the class is found. If the class
     * is not found the method will throw a <CODE>ClassNotFoundException</CODE>
     * exception.
     *
     * @param className The name of the class to be loaded.
     *
     * @exception ClassNotFoundException The specified class could not be found.
     */
    public static Class loadClass(String className) throws ClassNotFoundException {
	return  loadClass((Vector) loaders.clone(), className);
    }

    
    /**
     * Go through the list of class loaders but exclude the given class loader, then try to load 
     * the requested class.
     * The method will stop as soon as the class is found. If the class
     * is not found the method will throw a <CODE>ClassNotFoundException</CODE>
     * exception.
     *
     * @param className The name of the class to be loaded.
     * @param loader The class loader to be excluded.
     *
     * @exception ClassNotFoundException The specified class could not be found.
     */

    public static Class loadClassWithout(ClassLoader loader,String className) throws ClassNotFoundException {	
	if (isTraceOn()) {
	    trace("loadClassWithout", className + "\twithout " + loader);
	} 
	
	// Ok work with a clone !
	//
	Vector listToUse= (Vector) loaders.clone();
	
	if (loader == null) {
	    return loadClass(listToUse, className);
	}
	
	// Remove the loader if it is in the list ...
	//
	listToUse.removeElement(loader);
	
	startValidSearch(loader, className);
	try {
	    return loadClass(listToUse, className);
	} finally {
	    stopValidSearch(loader, className);
	}
    }
    
    
    private static Class loadClass(Vector list, String className) throws ClassNotFoundException {
        // From now on we do not care about what can happen to the original list
        // of class loaders ...
        //
        for(Iterator i = list.iterator(); i.hasNext();) {
            try {       
		ClassLoader cl = (ClassLoader)i.next();
		if (isTraceOn()) {
		    trace("loadClass", "trying loader = " + cl);
		}     
		if (cl instanceof javax.management.loading.MLet) {
		    return ((javax.management.loading.MLet)cl).loadClassCalledFromDLR(className);
		} else {
		    return  cl.loadClass(className);
		}
            } catch(ClassNotFoundException e) {}
        }
        throw new ClassNotFoundException(className);
    }

    private static synchronized void startValidSearch(ClassLoader aloader,String className) 
        throws ClassNotFoundException {
        // Check if we have such a current search
        //
        Vector excluded= (Vector) search.get(className);
        if ((excluded!= null) && (excluded.contains(aloader))) {
	    if (isTraceOn()) {
		trace("startValidSearch", "already requested loader=" + aloader + " class= " + className);
	    }            
            throw new ClassNotFoundException(className);
        }
      
        // Add an entry
        //
        if (excluded == null) {
            excluded= new Vector(1);
            search.put(className, excluded);
        }         
        excluded.addElement(aloader);
	if (isTraceOn()) {
	    trace("startValidSearch", "loader=" + aloader + " class= " + className);
	}     
    }

    private static synchronized void stopValidSearch(ClassLoader aloader,String className) {
    
        // Retrieve the search.
        //
        Vector excluded= (Vector) search.get(className);
        if (excluded!= null) {
            excluded.removeElement(aloader);
	    if (isTraceOn()) {
		trace("stopValidSearch", "loader=" + aloader + " class= " + className);
	    }               
	}
    }
  
    // TRACES & DEBUG
    //---------------

    private static boolean isTraceOn() {
        return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MBEANSERVER);
    }

    private static void trace(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MBEANSERVER, clz, func, info);
    }

    private static void trace(String func, String info) {
        trace(dbgTag, func, info);
    }

    private static boolean isDebugOn() {
        return Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_MBEANSERVER);
    }

    private static void debug(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_MBEANSERVER, clz, func, info);
    }

    private static void debug(String func, String info) {
        debug(dbgTag, func, info);
    }
 
 }
