/*
 * @(#)file      DescriptorSupport.java
 * @(#)author    IBM Corp.
 * @(#)version   1.8
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */

package javax.management.modelmbean;

import java.util.*;
import javax.management.RuntimeOperationsException;
import javax.management.MBeanException;
import javax.management.Descriptor;

import com.sun.management.jmx.Trace;


/**
 * This class represents the metadata set for a ModelMBean element.
 * A descriptor is part of the ModelMBeanInfo, ModelMBeanNotificationInfo, ModelMBeanAttributeInfo,
 * ModelMBeanConstructorInfo, and ModelMBeanParameterInfo.
 * <P>
 * A descriptor consists of a collection of fields.  Each field is in fieldname=fieldvalue format.
 * Field names are not case sensitive, case will be preserved on field values.
 * <P>
 * All field names and values are not predefined. New fields can be defined and added by any program.
 * Some fields have been predefined for consistency of implmentation and support by the ModelMBeanInfo
 * ModelMBean*Info, and ModelMBean classes.
 *<P>
 */

public class DescriptorSupport
	 implements javax.management.Descriptor
{
	/* Input string is an xml encoded string to the public, maybe anything internally */

	HashMap descriptor;

	private static final int DEFAULT_SIZE = 20;
	public String currClass = "DescriptorSupport";


	/** 
	 * Descriptor default constructor. 
	 * Default initial descriptor size is 20.  It will grow as needed.
	 */
	public DescriptorSupport()
	{
		if (tracing())
		{
			trace("Descriptor()","Constructor"); 
		}
		descriptor = new HashMap(DEFAULT_SIZE);
	}

	/** 
	 * Descriptor constructor.
	 * Takes as parameter the initial capacity of the Map that stores the descriptor fields. 
	 * Capacity will grow as needed.
	 *
	 * @param initNumFields The initial capacity of the Map that stores the descriptor fields.
	 * @exception RuntimeOperationsException for illegal value for maxNumFields ( <= 0)
	 */
	public DescriptorSupport(int initNumFields) 
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("Descriptor(maxNumFields = " + initNumFields + ")","Constructor"); 
		}
		if (initNumFields <= 0)
		{
			if (tracing())
			{
				trace("Descriptor(maxNumFields)","Illegal arguments: initNumFields <= 0");
			}
			throw new RuntimeOperationsException(
												new IllegalArgumentException("Descriptor field limit is invalid"),
												("Exception occured trying to construct a descriptor"));
		}
		descriptor = new HashMap(initNumFields);
	}

	/** 
	 * Descriptor constructor taking a Descriptor as parameter.
	 * Creates a new descriptor initialized to the values of the descriptor passed in parameter.
	 * 
	 * @param  inDesc the descriptor to be used to initialize the constructed descriptor. If it is 
	 * null or contains no descriptor fields, an empty Descriptor will be created.    
	 * @exception RuntimeOperationsException for illegal value for inDescr.  inDescr cannot be
	 *              null.  If the descriptor fails for any reason, this exception will be thrown.
	 */
	public DescriptorSupport(DescriptorSupport inDescr)
	{
		if (tracing())
		{
			trace("Descriptor(Descriptor)","Constructor"); 
		}
		if ((inDescr == null) || (inDescr.descriptor == null))
		{
			descriptor = new HashMap(DEFAULT_SIZE);
		} else
		{
			descriptor = new HashMap(inDescr.descriptor);
		}
	}


	/** 
	 * Descriptor constructor taking an XML String.
	 * This method is part of this implementation and not required by the JMX Specification.
	 *
	 * @param inStr An XML-formatted string used to populate this Descriptor
	 * The format will be: 
	 *  <Descriptor>
	 *  <field name="name" value="current_descriptor_name"></field>
	 *  <field name="descriptorType" value="current_descriptor_type></field>
	 *  ...
	 *  </Descriptor>
	 *
	 * All fields values will be created as Strings.  If the field values are not
	 * Strings, the programmer will have to reset or convert these fields correctly.
	 * Fields which are not String objects will have toString() called on them
	 * to create the value. You will not be able to reconstruct these
	 * objects unless they have been specifically setup to support toString() into
	 * a meaningfull format and have a matching constructor that accepts a String
	 * in the same format.
		 *
	 */
	public DescriptorSupport(String inStr)
	throws MBeanException, RuntimeOperationsException, XMLParseException {
		/* parse an XML-formatted string and populate internal structure with it */
		if (tracing())
		{
			trace("Descriptor(String ='" + inStr + "'","Constructor");
		}
		if (inStr == null)
		{
			if (tracing())
			{
				trace("Descriptor(String = null)","Illegal arguments");
			}
			throw new RuntimeOperationsException(
												new IllegalArgumentException("Descriptor is null or invalid"),
												("Exception occured trying to construct a descriptor"));
		}
		// parse xmlstring into structures
		descriptor = new HashMap(DEFAULT_SIZE);	 // create dummy descriptor: should have same size 
		// as number of fields in xmlstring
		// loop through structures and put them in descriptor

		StringTokenizer st = new StringTokenizer(inStr, "<> \t\n\r\f");

		boolean inFld = false;
		boolean inDesc = false;
		String fieldName = null;
		String fieldValue = null;

		while (st.hasMoreTokens())
		{  // loop through tokens

			String tok = st.nextToken();

			if (tok.equalsIgnoreCase("FIELD"))
			{
				inFld = true;
			} else if (tok.equalsIgnoreCase("/FIELD"))
			{
				if ((fieldName != null) && (fieldValue != null))
				{
					setField(fieldName.toLowerCase(), fieldValue);
				}
				fieldName = null;
				fieldValue = null;
				inFld = false;
			}

			else if (tok.equalsIgnoreCase("DESCRIPTOR"))
			{
				inDesc = true;
			} else if (tok.equalsIgnoreCase("/DESCRIPTOR"))
			{
				inDesc = false;
				fieldName = null;
				fieldValue = null;
				inFld = false;
			}

			else if (inFld && inDesc)
			{  // want kw=value, eg, name="myname" value="myvalue"
				int eq_separator = tok.indexOf("=");
				if (eq_separator > 0)
				{
					String kwPart = tok.substring(0,eq_separator);
					String valPart = tok.substring(eq_separator+1);
					if (kwPart.equalsIgnoreCase("NAME"))
						fieldName = new String(valPart);
					else if (kwPart.equalsIgnoreCase("VALUE"))
						fieldValue = new String(valPart);
					else
					{  // xml parse exception
						throw new XMLParseException("expected a field value, received '" + tok + "'");
					}
				} else
				{ // xml parse exception
					throw new XMLParseException("expected keyword=value, received '" + tok + "'");
				}
			}
		}  // while tokens

		if (tracing())
		{
			trace("Descriptor(XMLString)","Exit"); 
		}
	}

	/** 
	 * Constructor taking field names and field values.  The array and array elements cannot be null.
	 *
	 * @param fieldNames String array of field names.  No elements of this array can be null.
	 * @param fieldValues Object array of the corresponding field values. 
	 *                  Elements of the array can be null. The fieldValue must be valid for the fieldName.
	 *  
	 * Note: array sizes of parameters should match. If both arrays are null or empty, then
	 * an empty descriptor is created.  No exception is thrown. 
	 *
	 * @exception RuntimeOperationsException for illegal value for field Names or field Values.
	 *              The array lengths must be equal. 
	 *              If the descriptor construction fails for any reason, this exception will be thrown.
	 * 
	 */
	public DescriptorSupport(String[] fieldNames, Object[] fieldValues) throws RuntimeOperationsException {
		if (tracing())
		{
			trace("Descriptor(fieldNames, fieldObjects)","Constructor");
		}
		if (((fieldNames == null) && (fieldValues == null)) || 
			((fieldNames.length == 0) && (fieldValues.length == 0)))
		{
			descriptor = new HashMap(DEFAULT_SIZE);
			return;
		}
		if ((fieldNames == null) || (fieldValues == null) || 
			(fieldNames.length != fieldValues.length))
		{
			if (tracing())
			{
				trace("Descriptor(String[],Object[])","Illegal arguments"); 
			}

			throw new RuntimeOperationsException(
												new IllegalArgumentException("FieldNames or FieldValues are null or invalid"),
												("Exception occured trying to construct a descriptor"));
		}

		/* populate internal structure with fields */
		descriptor = new HashMap(fieldNames.length);
		// a field value can be "null"

		for (int i=0; i < fieldNames.length; i++)
		{
			// setField will throw an exception if a fieldName is be null.
			// the fieldName and fieldValue will be validated in setField. 
			setField(fieldNames[i], fieldValues[i]);
		}
		if (tracing())
		{
			trace("Descriptor(fieldNames, fieldObjects)","Exit");
		}
	}

	/** 
	 * Constructor taking fields in the <i>fieldName=fieldValue</i> format.
	 *
	 * @param fields String array of with each element containing a field name and value.
	 * If this array is null or empty, then the default constructor will be executed. Null strings
	 * or empty strings will be ignored.
	 *  
	 * Note: each string should be of the form <i>fieldName=fieldValue</i>,
	 * with no imbedded blanks or other punctuation.
	 *
	 * @exception RuntimeOperationsException for illegal value for field Names or field Values.
	 *              The field must contain an "=". "=fieldValue", "fieldName", and "fieldValue" are illegal.
	 *              FieldName cannot be null.  "fieldName=" will cause the value to be null.   
	 *              If the descriptor construction fails for any reason, this exception will be thrown.
	 * 
	*/
	public DescriptorSupport(String[] fields)
	{
		if (tracing())
		{
			trace("Descriptor(fields)","Constructor");
		}
		if (( fields == null ) || ( fields.length == 0))
		{
			descriptor = new HashMap(DEFAULT_SIZE);
			return;
		}

		descriptor = new HashMap(fields.length);

		for (int i=0; i < fields.length; i++)
		{
			if ((fields[i] == null) || (fields[i].equals("")))
			{
				continue;
			}
			int eq_separator = fields[i].indexOf("=");
			if (eq_separator <= 0) // illegal if no = or is first character
			{
				if (tracing())
				{
					trace("Descriptor(String[])",
						  "Illegal arguments: field does not have '=' as a name and value separator");
				}
				throw 
					 new RuntimeOperationsException(
												   new IllegalArgumentException("Field in invalid format: no equals sign"),
												   ("Exception occured trying to construct a descriptor"));
			}

			String fieldName = fields[i].substring(0,eq_separator);
			String fieldValue = null;
			if (eq_separator < fields[i].length())	 // = is not in last character
			{
				fieldValue = fields[i].substring(eq_separator+1);
			}

			if ((fieldName == null) || (fieldName.equals("")))
			{
				if (tracing())
				{
					trace("Descriptor(String[])","Illegal arguments: fieldName is null");
				}

				throw new RuntimeOperationsException(
													new IllegalArgumentException("Field in invalid format: no equals sign"),
													("Exception occured trying to construct a descriptor"));
			}
			setField(fieldName,fieldValue);
		}
		if (tracing())
		{
			trace("Descriptor(fields)","Exit"); 
		}
	}

	// Implementation of the Descriptor interface
	/**
	* returns the string value for a specific fieldname
	* @param fieldName The field name in question; if not found, null is returned.
	*
	* @return String Field value
	*
	* @exception RuntimeOperationsException for illegal value (null or empty string) for field Names.
	*              
	*/


	public Object getFieldValue(String inFieldName)
	throws RuntimeOperationsException {

		if ((inFieldName == null) || (inFieldName.equals("")))
		{
			if (tracing())
			{
				trace("getField()","Illegal arguments: null field name."); 
			}
			throw new RuntimeOperationsException(
												new IllegalArgumentException("Fieldname requested is null"),
												("Exception occured trying to get a field from a descriptor"));
		}
		Object retValue = descriptor.get(inFieldName.toLowerCase());
		if (tracing())
		{
			trace("getField(" + inFieldName + ")","Returns '" + retValue + "'"); 
		}
		return(retValue);
	}

	/**         
	 * Sets the string value for a specific fieldname.  The value must be valid for the field.
	 * If the field does not exist, it is added. if it does exist, the value is replaced.
	 *
	 * @param String fieldName: The field name to be set. Must not be null or empty string.
	 * @param String fieldValue: The field value to be set for the field name.  Can be null.
	 *
	 * @exception RuntimeOperationsException for illegal value for field Names.
	 *              
	 */


	public void setField(String inFieldName, Object fieldValue) throws RuntimeOperationsException {

		// field value can be null
		if ((inFieldName == null) || (inFieldName.equals("")))
		{
			if (tracing())
			{
				trace("setField(String,String)","Illegal arguments: null field name");
			}

			throw new RuntimeOperationsException(
												new IllegalArgumentException("Fieldname to be set is null"),
												("Exception occured trying to set a field from a descriptor"));
		}

		String fieldName = inFieldName.toLowerCase();
		if (tracing())
		{
			trace("setField(fieldName, fieldValue)",
				  ("Entry: setting '" + fieldName + "' to '" + fieldValue + "'.")); 
		}

		if (validateField(fieldName, fieldValue))
		{
			descriptor.put(fieldName,fieldValue);
		} else
		{
			if (tracing())
			{
				trace("setField(fieldName,FieldValue)","Illegal arguments");
			}

			throw new RuntimeOperationsException(
												new IllegalArgumentException("Field value invalid: " 
																			 + fieldName + "=" + fieldValue),
												("Field " + fieldName + "=" + fieldValue + " is invalid. Exception occured trying to set a field from a descriptor"));
		}

		if (tracing())
		{
			trace("setField(fieldName, fieldValue)","Exit"); 
		}
	}
	/**
	 * Returns all the fields in the descriptor 
	 * 
	 * @return String array of fields in the format <i>fieldName=fieldValue</i>.  If there
	 * are no fields in the descriptor, then String[0] is returned. If a fieldValue is an
	 * object then the toString() method is called on it and its returned value is used
	 * as the value for the field enclosed in parens.
	 *
	 * @exception RuntimeOperationsException for illegal value for field Names or Values.
	 *
	 */

	public String[] getFields()
	{
		if (tracing())
		{
			trace("getFields()","Entry"); 
		}
		if (descriptor == null)
		{
			return new String[0];
		}
		int numberOfEntries = descriptor.size();
		if (numberOfEntries == 0)
		{
			return new String[0];
		}

		String[] responseFields = new String[numberOfEntries];
		Set returnedSet = descriptor.entrySet();

		if (returnedSet == null)
		{
			if (tracing())
			{
				trace("getFields()","No fields found in descriptor");
			}
			return new String[0];
		}

		int i = 0;
		Object currValue = null;
		Map.Entry currElement = null;

		if (tracing())
		{
			trace("getFields()","Returning " + numberOfEntries + " fields"); 
		}
		for (Iterator iter = returnedSet.iterator(); iter.hasNext(); i++)
		{
			currElement = (Map.Entry) iter.next();

			if (currElement == null)
			{
				if (tracing())
				{
					trace("getFields()","Element is null");
				}
			} else
			{
				currValue = currElement.getValue();
				if (currValue == null)
				{
					responseFields[i] = new String((currElement.getKey()).toString() + "=");
				} else
				{
					if (currValue instanceof java.lang.String)
					{
						responseFields[i] = new String((currElement.getKey()).toString() + 
													   "=" + (currValue).toString());
					} else
					{
						responseFields[i] = new String((currElement.getKey()).toString() + 
													   "=(" + (currValue).toString() + ")");
					}
				}
			}
		}

		if (tracing())
		{
			trace("getFields()","Exit"); 
		}

		return responseFields;
	}

	/**
	 * Returns all the fields names in the descriptor.
	 * 
	 * @return String array of fields names. If the descriptor is empty, you will get
	 *          an empty array.
	 *
	 */


	public String[] getFieldNames()
	{
		if (tracing())
		{
			trace("getFieldNames()","Entry"); 
		}
		if (descriptor == null)
		{
			return new String[0];
		}
		int numberOfEntries = descriptor.size();

		String[] responseFields = new String[numberOfEntries];
		Set returnedSet = descriptor.entrySet();

		if (( returnedSet == null ) || (numberOfEntries == 0))
		{
			if (tracing())
			{
				trace("getFieldNames()","no descriptor fields found");
			}
			return new String[0];
		}

		int i = 0;

		if (tracing())
		{
			trace("getFieldNames()","Returning " + numberOfEntries + " fields"); 
		}

		for (Iterator iter = returnedSet.iterator(); iter.hasNext(); i++)
		{
			Map.Entry currElement = (Map.Entry) iter.next();

			if (( currElement == null ) || (currElement.getKey() == null))
			{
				if (tracing())
				{
					trace("getFieldNames()","Field is null");
				}
			} else
			{
				responseFields[i] = new String((currElement.getKey()).toString());                
			}
		}

		if (tracing())
		{
			trace("getFieldNames()","Exit"); 
		}

		return responseFields;
	}


	/**
	 * Returns all the field values in the descriptor as an array of Objects. The
	 * retuned values are in the same order as the fieldNames String array parameter.
	 *
	 * @param fieldNames String array of the names of the fields that the values
	 * should be returned for.  If the array is empty then an empty array will be 
	 * returned.  If the array is 'null' then all values will be returned.  If a field 
	 * name in the array does not exist, then null is returned for the matching array
	 * element being returned.
	 *
	 * @return Object array of field values. If the descriptor is empty, you will get
	 *          an empty array.
	 *
	 */
	public Object[] getFieldValues(String[] fieldNames)
	{
		if (tracing())
		{
			trace("getFieldValues(fieldNames)","Entry"); 
		}
		// if fieldNames == null return all values
		// if fieldNames is String[0] return no values

		if ((descriptor == null) || 
			((fieldNames != null) && (fieldNames.length == 0)))
		{
			return new Object[0];
		}

		int numberOfEntries = descriptor.size();
		Set returnedSet = descriptor.entrySet();

		if (( returnedSet == null ) || (numberOfEntries == 0))
		{
			if (tracing())
			{
				trace("getFieldValues()","no descriptor fields found");
			}
			return new Object[0];
		}

		Object[] responseFields;
		if (fieldNames != null)
		{
			responseFields = new Object[fieldNames.length];	// room for selected
		} else
		{
			responseFields = new Object[numberOfEntries];  // room for all
		}

		int i = 0;

		if (tracing())
		{
			trace("getFieldValues()","Returning " + numberOfEntries + " fields"); 
		}

		if (fieldNames == null)
		{
			for (Iterator iter = returnedSet.iterator(); iter.hasNext(); i++)
			{
				Map.Entry currElement = (Map.Entry) iter.next();

				if (( currElement == null ) || (currElement.getKey() == null))
				{
					if (tracing())
					{
						trace("getFieldValues()","Field is null");
					}
					responseFields[i] = null;
				} else
				{
					responseFields[i] = new String((String)currElement.getValue());                
				}
			}
		} else
		{
			for (i=0; i < fieldNames.length; i++)
			{
				if ((fieldNames[i] == null) || (fieldNames[i].equals("")))
				{
					responseFields[i] = null;
				} else
				{
					responseFields[i] = getFieldValue(fieldNames[i]);
				}
			}
		}


		if (tracing())
		{
			trace("getFieldValues()","Exit"); 
		}

		return responseFields;
	}
	/** 
	 * Sets all Fields in the list to the new value in with the same index
	 * in the fieldValue array.  Array sizes must match.  
	 * The field value will be validated before it is set.  
	 * If it is not valid, then an exception will be thrown.
	 * If the arrays are empty, then no change will take effect.
	 * 
	 * @param fieldNames String array of field names. The array and array elements cannot be null.
	 * @param fieldValues Object array of the corresponding field values.  The array cannot be null.
	 *                      Elements of the array can be null.
	 *  
	 * @exception RuntimeOperationsException for illegal value for field Names or field Values.
	 *              Niether can be null.  The array lengths must be equal. 
	 *              If the descriptor construction fails for any reason, this exception will be thrown.
	 *
	 */

	public void setFields(String[] fieldNames, Object[] fieldValues) throws RuntimeOperationsException {

		if (tracing())
		{
			trace("setFields(fieldNames, ObjectValues)","Entry"); 
		}
		if (((fieldNames == null) && (fieldValues == null)) ||
			((fieldNames.length == 0) && (fieldValues.length == 0)))
		{
			/* does nothing */
			return;
		}

		if ((fieldNames == null) || (fieldValues == null) || 
			(fieldNames.length != fieldValues.length))
		{
			if (tracing())
			{
				trace("Descriptor.setFields(String[],Object[])","Illegal arguments"); 
			}

			throw new RuntimeOperationsException(
												new IllegalArgumentException("FieldNames and FieldValues are null or invalid"),
												("Exception occured trying to set object fields a descriptor"));
		}

		for (int i=0; i < fieldNames.length; i++)
		{
			if (( fieldNames[i] == null) || (fieldNames[i].equals("")))
			{
				if (tracing())
				{
					trace("Descriptor.setFields(String[],Object[])",
						  "Null field name encountered at " + i + " element"); 
				}

				throw new RuntimeOperationsException(
													new IllegalArgumentException("FieldNames is null or invalid"),
													("Exception occured trying to set object fields a descriptor"));
			}
			setField(fieldNames[i], fieldValues[i]);
		}
		if (tracing())
		{
			trace("Descriptor.setFields(fieldNames, fieldObjects)","Exit"); 
		}

	}
	/**
	* Returns a new Descriptor which is a duplicate of the Descriptor.
	*
	* @exception RuntimeOperationsException for illegal value for field Names or field Values.
	*              If the descriptor construction fails for any reason, this exception will be thrown.
	*/

	public Object clone() throws RuntimeOperationsException {
		if (tracing())
		{
			trace("Descriptor.clone()","Executed"); 
		}
		return(new DescriptorSupport(this));
	}

	/**
	* Removes a field from the descriptor
	*
	* @param fieldName String name of the field to be removed.
	* If the field is not found no exception is thrown.
	*/
	public void removeField(String fieldName)
	{
		if ((fieldName == null) || (fieldName.equals("")))
		{
			return;
		}
		descriptor.remove(fieldName);
		return;
	}
	/**
	* Returns true if fieldValues are checked to be sure they are legal for the fieldNames.
	*
	* This implementation returns false if:
	*	name or descriptorType fields are null
	*	class, role, getMethod and setMethod, if defined, must not be null
	*	persistPeriod, currencyTimeLimit, lastUpdatedTimeStamp, lastReturnedTimeStamp 
	*		must numerics greater than -1.
	*	iterable, log, export must be t, f, true, or false
	* 	visibility must be between 1 and 4
	*	severity must be betwen 0 and 5
	*	persistPolicy must be OnUpdate,OnTimer,NoMoreOftenThan,Always,or Never is must not
	*				be case sensitive.
	* 	
	*
	* @exception RuntimeOperationsException If the validity checking fails for any reason, this exception will be thrown.
	*/

public boolean isValid() throws RuntimeOperationsException {
		if (tracing())
		{
			trace("Descriptor.isValid()","Executed"); 
		}
		// verify that the descriptor is valid, by iterating over each field... 

		Set returnedSet = descriptor.entrySet();

		if (returnedSet == null)   // null descriptor, not valid
		{
			if (tracing())
			{
				trace("Descriptor.isValid()","returns false (null set)");
			}
			return false;
		}
		// must have a name and descriptor type field
		String thisName = (String)(this.getFieldValue("name"));
		String thisDescType = (String)(getFieldValue("descriptorType"));

		if ((thisName == null) || (thisDescType == null) || 
			(thisName.equals("")) || (thisDescType.equals("")))
		{
			return false;
		}

		// According to the descriptor type we should validate the fields contained

		for (Iterator iter = returnedSet.iterator(); iter.hasNext();)
		{
			Map.Entry currElement = (Map.Entry) iter.next();

			if (currElement != null)
			{
				if (currElement.getValue() != null)
				{
					// validate the field valued...
					if (validateField((currElement.getKey()).toString(),
									  (currElement.getValue()).toString()))
					{
						continue;
					} else
					{
						if (tracing()) trace("isValid()","Field " + (currElement.getKey()).toString() + "=" + (currElement.getValue()).toString() + " is not valid");
						return false;
					}
				}
			}
		}

		// fell through, all fields OK

		if (tracing())
		{
			trace("Descriptor.isValid()","returns true");
		}
		return true;
	}


	// worker routine for isValid()
	// name is not null
	// descriptorType is not null
	// getMethod and setMethod are not null
	// persistPeriod is numeric
	// currencyTimeLimit is numeric
	// lastUpdatedTimeStamp is numeric
	// iterable is T or F
	// visibility is 1-4
	// severity is 0-5
	// log is T or F
	// role is not null
	// class is not null
	// lastReturnedTimeStamp is numeric


	private boolean validateField(String fldName, Object fldValue)
	{
		if ((fldName == null) || (fldName.equals("")))
			return false;
		String SfldValue = "";
		boolean isAString = false;
		if ((fldValue != null) && (fldValue instanceof java.lang.String))
		{
			SfldValue = (String) fldValue;
			isAString = true;
		}

		if (fldName.equalsIgnoreCase("Name") ||
			fldName.equalsIgnoreCase("DescriptorType") ||
			fldName.equalsIgnoreCase("SetMethod") ||
			fldName.equalsIgnoreCase("GetMethod") ||
			fldName.equalsIgnoreCase("Role") ||
			fldName.equalsIgnoreCase("Class"))
		{
			if (( fldValue != null ) && (isAString))
				return true;
			else
				return false;
		} else if (fldName.equalsIgnoreCase("visibility"))
		{
			int v;
			if ((fldValue != null) && (isAString))
			{
				v = toNumeric(SfldValue); 
			} else if ((fldValue != null) && (fldValue instanceof java.lang.Integer))
			{
				v = ((Integer)fldValue).intValue();
			} else return false;

			if (v >= 1 &&  v <= 4)
				return true;
			else
				return false;
		} else if (fldName.equalsIgnoreCase("severity"))
		{

			int v;
			if ((fldValue != null) && (isAString))
			{
				v = toNumeric(SfldValue); 
			} else if ((fldValue != null) && (fldValue instanceof java.lang.Integer))
			{
				v = ((Integer)fldValue).intValue();
			} else return false;

			if (v >= 0 &&  v <= 5)
				return true;
			else
				return false;
		}

		else if (fldName.equalsIgnoreCase("PersistPolicy"))
		{
			if (((fldValue != null) && (isAString)) &&
				( SfldValue.equalsIgnoreCase("OnUpdate") || 
				  SfldValue.equalsIgnoreCase("OnTimer") || 
				  SfldValue.equalsIgnoreCase("NoMoreOftenThan") || 
				  SfldValue.equalsIgnoreCase("Always") || 
				  SfldValue.equalsIgnoreCase("Never") ))
				return true;
			else
				return false;
		}

		else if (fldName.equalsIgnoreCase("PersistPeriod") || 
				 fldName.equalsIgnoreCase("CurrencyTimeLimit") ||
				 fldName.equalsIgnoreCase("LastUpdatedTimeStamp") ||
				 fldName.equalsIgnoreCase("LastReturnedTimeStamp"))
		{

			int v;
			if ((fldValue != null) && (isAString))
			{
				v = toNumeric(SfldValue); 
			} else if ((fldValue != null) && (fldValue instanceof java.lang.Integer))
			{
				v = ((Integer)fldValue).intValue();
			} else return false;

			if (v >= -1)
				return true;
			else
				return false;
		}

		else if (fldName.equalsIgnoreCase("ReadOnly")  ||
				 fldName.equalsIgnoreCase("log") ||
				 fldName.equalsIgnoreCase("Iterable"))
		{
			if ((fldValue instanceof java.lang.Boolean) ||
				(isAString && 
				 (SfldValue.equalsIgnoreCase("T") ||
				  SfldValue.equalsIgnoreCase("true") ||
				  SfldValue.equalsIgnoreCase("F") ||
				  SfldValue.equalsIgnoreCase("false") )))
				return true;
			else
				return false;
		}


		// default to true, it is a field we aren't validating (user etc.)
		return true;
	}



	/**
	 * @return <pv>an XML String representing the descriptor.</pv>
	 *      The XML string will be formatted for each field with
	 *      "<field> NAME=name VALUE=fieldValue TYPE=String|Object|ClassName></field>"
	 *
	 * The format will be: 
	 *  <Descriptor>
	 *  <field name="name" value="current_descriptor_name"></field>
	 *  <field name="descriptorType" value="current_descriptor_type></field>
	 *  ...
	 *  </Descriptor>
	 *
	 * Fields which are not String objects will have toString() called on them
	 * to create the value. We do not guarantee that you can reconstruct these
	 * objects unles they have been specifically setup to support toString() into
	 * a meaningfull format and have a matching constructor that accepts a String
	 * in the same format.

	 *
	 * @exception RuntimeOperationsException for illegal value for field Names or field Values.
	 *              If the XML formated string construction fails for any reason, 
	 *              this exception will be thrown.
	 */

	public String toXMLString()
	{
		if (tracing())
		{
			trace("Descriptor.toXMLString()","Executed"); 
		}

		String respStr = new String("<Descriptor>");

		Set returnedSet = descriptor.entrySet();

		if (returnedSet == null)
		{
			if (tracing())
			{
				trace("Descriptor.toXMLString()","returnedSet is null"); 
			}
			return null;
		}

		int i = 0;
		for (Iterator iter = returnedSet.iterator(); iter.hasNext(); i++)
		{
			Map.Entry currElement = (Map.Entry) iter.next();

			if (currElement == null)
			{
				if (tracing())
				{
					trace("Descriptor.toXMLString()","Element is null"); 
				}
			} else
			{
				if (currElement.getValue() == null)
				{
					respStr += ("<field name=\"" + (currElement.getKey()).toString() + 
								"\" value=\"null\"></field>");
				} else
				{
					respStr += ("<field name=\"" + (currElement.getKey()).toString() + 
								"\" value=\"" + (currElement.getValue()).toString() + "\"></field>");
				}
			}
		}


		respStr += "</Descriptor>";

		if (tracing())
		{
			trace("Descriptor.toXMLString()","Returns " + respStr); 
		}

		return respStr;
	}

	/**
	* @return <pv>a human readable string representing the descriptor</pv>
	*       The string will be in the format of 
	*       "fieldName=fieldValue,fieldName2=fieldValue2,..."
	*
	* @exception RuntimeOperationsException for illegal value for field Names or field Values.
	*              If the descriptor string fails for any reason, this exception will be thrown.
	*/
	public String toString() 
	{
		if (tracing())
		{
			trace("Descriptor.toString()","Entry"); 
		}

		String respStr = "";
		String[] fields = getFields();

		if (tracing())
		{
			trace("Descriptor.toString()","Printing " + fields.length + " fields"); 
		}

		if ((fields == null) || (fields.length == 0))
		{
			if (tracing())
			{
				trace("Descriptor.toString()","Empty Descriptor");
			}
			return null;
		}

		for (int i=0; i < fields.length; i++)
		{
			if (i == (fields.length - 1))
			{
				respStr = respStr.concat(fields[i]);
			} else
			{
				respStr = respStr.concat(fields[i] + ", ");
			}
		}

		if (tracing())
		{
			trace("Descriptor.toString()","Exit returning " + respStr); 
		}

		return respStr;
	}


	// utility to convert to int, returns -1 if bogus.

	private int toNumeric(String inStr)
	{
		int result = -1;
		try
		{
			result = java.lang.Integer.parseInt(inStr);
		} catch (Exception e)
		{
			return -1;
		}
		return result;
	}


	// Trace and debug functions

	private boolean tracing()
	{
		return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_ALL); 
	}

	private void trace(String inClass, String inMethod, String inText)
	{
		Trace.send(Trace.LEVEL_TRACE, Trace.INFO_ALL, inClass, inMethod, Integer.toHexString(this.hashCode()) + " " + inText); 
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}

}






































