/*
 * @(#)file      ModelMBeanAttributeInfo.java
 * @(#)author    IBM Corp.
 * @(#)version   1.10
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */

package javax.management.modelmbean;

import javax.management.Descriptor;
import javax.management.DescriptorAccess;
import javax.management.*;
import java.lang.reflect.*;

import com.sun.management.jmx.Trace;         

/**
 * The ModelMBeanAttributeInfo object describes an attribute of the ModelMBean.
 * It is a subclass of MBeanAttributeInfo with the addition of an associated Descriptor
 * and an implementation of the DescriptorAccess interface.
 * <P>
 * The fields in the descriptor are defined, but not limited to, the following: <P>
 * <PRE>
 * name           : attribute name  
 * descriptorType : must be "attribute"   
 * value          : current value for attribute 
 * default        : default value for attribute 
 * displayName    : name of attribute to be used in displays 
 * getMethod      : name of operation descriptor for get method  
 * setMethod      : name of operation descriptor for set method 
 * protocolMap    : object which implements the ProtocolMap interface: map of protocol names and protocol hints 
 * persistPolicy  : Update|OnTime|NoMoreOftenThan|Always|Never  
 * persistPeriod  : seconds - frequency of persist cycle. Used when persistPolicy is"OnTime" or "NoMoreOftenThan".  
 * currencyTimeLimit : how long value is valid, <0 never, =0 always, >0 seconds  
 * lastUpdatedTimeStamp : when value was set  
 * iterable       : T - object value supports Iterable interface, F - does not support Iterable interface       
 * visibility     : 1-4 where 1: always visible 4: rarely visible  
 * presentationString : xml formatted string to allow presentation of data 
 * </PRE>
 * The default descriptor contains the name and descriptorType fields.
 */


public class ModelMBeanAttributeInfo extends MBeanAttributeInfo 
	 implements DescriptorAccess, Cloneable
{

	private Descriptor attrDescriptor = createDefaultDescriptor();
	private String currClass = "ModelMBeanAttributeInfo";

	/**
	 * Constructs a ModelMBeanAttributeInfo object with a default descriptor.
	 *
	 * @param name The name of the attribute.
	 * @param description A human readable description of the attribute. Optional.
	 * @param getter The method used for reading the attribute value.
	 *          May be null if the property is write-only.
	 * @param setter The method used for writing the attribute value.
	 *          May be null if the attribute is read-only.
	 * @exception IntrospectionException There is a consistency problem in the definition of this attribute.
	 *
	 * The descriptor must be set before this  
	 * ModelMBeanAttributeInfo instance can be used.
	 */

	public ModelMBeanAttributeInfo(String name,
								   String description,
								   Method getter,
								   Method setter) 
	throws javax.management.IntrospectionException {
		super(name, description, getter, setter);
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo(" + name + ",String,Method,Method)","Entry");
		}

		attrDescriptor = createDefaultDescriptor();
		// put getter and setter methods in operations list 
		// create default descriptor

	}

	/**
	 * Constructs a ModelMBeanAttributeInfo object. 
	 *
	 * @param name The name of the attribute.
	 * @param description A human readable description of the attribute. Optional.
	 * @param getter The method used for reading the attribute value.
	 *          May be null if the property is write-only.
	 * @param setter The method used for writing the attribute value.
	 *          May be null if the attribute is read-only.
	 * @param descriptor An instance of Descriptor containing the appropriate metadata
	 *                   for this instance of the Attribute. If it is null or invalid then
	 *			a default desriptor will be created.
	 *
	 * @exception IntrospectionException There is a consistency problem in the definition of this attribute.
	 *
	 * The descriptor must be set before this  
	 * ModelMBeanAttributeInfo instance can be used.
	 */

	public ModelMBeanAttributeInfo(String name,
								   String description,
								   Method getter,
								   Method setter,
								   Descriptor descriptor) 
	throws javax.management.IntrospectionException {

		super(name, description, getter, setter);
		// put getter and setter methods in operations list
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo(" + name + ", String, Method, Method, Descriptor)","Entry");
		}
		if (descriptor == null)
		{
			attrDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(descriptor))
			{
				attrDescriptor = (Descriptor) descriptor.clone();
			} else
			{  // exception
				attrDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	 * Constructs a ModelMBeanAttributeInfo object with a default descriptor.
	 *
	 * @param name The name of the attribute
	 * @param type The type or class name of the attribute
	 * @param description A human readable description of the attribute.
	 * @param isReadable True if the attribute has a getter method, false otherwise.
	 * @param isWritable True if the attribute has a setter method, false otherwise.
	 * @param isIs True if the attribute has an "is" getter, false otherwise.
	 *
	 */   
	public ModelMBeanAttributeInfo(String name,
								   String type,
								   String description,
								   boolean isReadable,
								   boolean isWritable,
								   boolean isIs)
	{

		super(name, type, description, isReadable, isWritable, isIs);
		// create default descriptor
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo(" + name + ",String,String,boolean,boolean)","Entry");
		}
		attrDescriptor = createDefaultDescriptor();

	}
	/**
	 * Constructs a ModelMBeanAttributeInfo object with a default descriptor.
	 *
	 * @param name The name of the attribute
	 * @param type The type or class name of the attribute
	 * @param description A human readable description of the attribute.
	 * @param isReadable True if the attribute has a getter method, false otherwise.
	 * @param isWritable True if the attribute has a setter method, false otherwise.
	 * @param isIs True if the attribute has an "is" getter, false otherwise.
	 * @param descriptor An instance of Descriptor containing the appropriate metadata
	 *                   for this instance of the Attribute. If it is null or invalid then
	 *			a default desriptor will be created.
	 *
	 */       
	public ModelMBeanAttributeInfo(String name,
								   String type,
								   String description,
								   boolean isReadable,
								   boolean isWritable,
								   boolean isIs,
								   Descriptor descriptor)
	{
		super(name, type, description, isReadable, isWritable, isIs);
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo(" + name + ",String,String,boolean,boolean,Descriptor)","Entry");
		}
		if ((descriptor != null) && isValid(descriptor))
		{
			attrDescriptor = (Descriptor) descriptor.clone();
		} else
		{
			attrDescriptor =  createDefaultDescriptor();
			throw new IllegalArgumentException("Invalid descriptor passed in parameter");
		}
	}

	/**
	 * Constructs a new ModelMBeanAttributeInfo object from this ModelMBeanAttributeInfo Object.
	 * A default descriptor will be created.
	 *
	 * @param inInfo the ModelMBeanAttributeInfo to be duplicated
	 */

	public ModelMBeanAttributeInfo(ModelMBeanAttributeInfo inInfo)
	{
		super(inInfo.getName(), 
			  inInfo.getType(), 
			  inInfo.getDescription(), 
			  inInfo.isReadable(), 
			  inInfo.isWritable(),
			  inInfo.isIs());
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo(ModelMBeanAttributeInfo)","Entry");
		}
		Descriptor newDesc = inInfo.getDescriptor();

		//Descriptor newDesc = inInfo.attrDescriptor;

		if ((newDesc != null) && (isValid(newDesc)))
		{
			attrDescriptor = newDesc;
		} else
		{
			attrDescriptor = createDefaultDescriptor();
		}
	}
	/** 
	 * Gets a copy of the associated Descriptor for the ModelMBeanAttributeInfo
	 *
	 * @return Descriptor associated with the ModelMBeanAttributeInfo object.
	 */

	public Descriptor getDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo.getDescriptor()","Entry");
		}
		if (attrDescriptor == null)
		{
			attrDescriptor = createDefaultDescriptor();
		}
		return((Descriptor)attrDescriptor.clone());
	}


	/** 
	* Sets associated Descriptor (full replace) for the ModelMBeanAttributeDescriptor.
	* If the new Descriptor is null, then the associated Descriptor reverts to 
	* a default descriptor.  The Descriptor is validated before it is assigned.  If
	* the new Descriptor is invalid, then an IllegalArgumentException is thrown.
	*
	* @param Descriptor inDescriptor - replaces the Descriptor associated with the 
	* ModelMBeanAttributeInfo
	*
	*/
	public void setDescriptor(Descriptor inDescriptor)
	{
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo.setDescriptor()","Executed for " + inDescriptor.toString());
		}
		if (inDescriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanAttributeInfo.setDescriptor()",
					  "Received null for new descriptor value, setting descriptor to default values");
			}
			attrDescriptor = createDefaultDescriptor();
		} else
		{

			if (isValid(inDescriptor))
			{
				attrDescriptor =  (Descriptor) inDescriptor.clone();
			} else
			{
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	* Creates and returns a new ModelMBeanAttributeInfo which is a duplicate of this ModelMBeanAttributeInfo.
	*
	* @exception RuntimeOperationsException for illegal value for field Names or field Values.
	*              If the descriptor construction fails for any reason, this exception will be thrown.
	*/

	public Object clone()
	{
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo.clone()","Entry");
		}
		return(new ModelMBeanAttributeInfo(this));
	}

	/**
	* Returns a human readable version of the ModelMBeanAttributeInfo instance
	*/
	public String toString()
	{
		String retStr = new String("ModelMBeanAttributeInfo: " + this.getName() +
								   " ; Description: " + this.getDescription() +
								   " ; Types: " + this.getType() + 
								   " ; isReadable: " + this.isReadable() +
								   " ; isWritable: " + this.isWritable() + 
								   " ; Descriptor: " + this.getDescriptor());
		return retStr;
	}

	/** 
	* Creates and returns a Descriptor with default values set: 
	*  descriptorType=attribute,name=this.getName(),displayName=this.getName(),
	* 		persistPolicy=never,visibility=1,iterable=F
	*/

	private Descriptor createDefaultDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanAttributeInfo.createDefaultDescriptor()","Entry");
		}
		return new DescriptorSupport(new String[] {"descriptorType=attribute",
											  ("name=" + this.getName()),
											  ("displayName=" + this.getName()),
											  "iterable=F"});
	}
	/**
	* Tests that the descriptor is valid and adds appropriate default fields not already
	* specified. Field values must be correct for field names. 
	* Descriptor must have the same name as the attribute,the descriptorType field must be "attribute",
	* The following fields will be defaulted if they are not already set:
	* displayName=this.getName(),persistPolicy=never,visibility=1,iterable=F
	*/
	private boolean isValid(Descriptor inDesc)
	{
		// name and descriptor type must be correct
		// add in displayName, persistPolicy, visibility, iterable if they apply
		boolean results=true;
		String badField="none";
		if (inDesc == null)
		{
			badField="nullDescriptor";
			results = false;
		}

		else if (!inDesc.isValid())
		{	 // checks for empty descriptors, null,
			// checks for empty name and descriptorType adn valid values for fields.
			badField="inValidDescriptor";
			results = false;
		}

		else if (! ((String)inDesc.getFieldValue("name")).equalsIgnoreCase(this.getName()))
		{
			badField="name";
			results = false;
		} else
		{

			if (! ((String)inDesc.getFieldValue("descriptorType")).equalsIgnoreCase("attribute"))
			{
				badField="desriptorType";
				results = false;
			} else if ((inDesc.getFieldValue("displayName")) == null)
			{
				inDesc.setField("displayName",this.getName());
			} else if ((inDesc.getFieldValue("iterable")) == null)
			{
				inDesc.setField("iterable","F");
			}
		}
		if (tracing()) trace("isValid()",("Returning " + results + ": Invalid field is " + badField));
		return results;
	}

	// SUN Trace and debug functions
	private boolean tracing()
	{
		//        return false;
		return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN);
	}

	private void trace(String inClass, String inMethod, String inText)
	{
		//        System.out.println("TRACE: " + inClass + ":" + inMethod + ": " + inText);
		Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN, inClass, 
				   inMethod,  Integer.toHexString(this.hashCode()) + " " + inText); 
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}

}
