/*
 * @(#)file      ModelMBeanConstructorInfo.java
 * @(#)author    IBM Corp.
 * @(#)version   1.10
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */

package javax.management.modelmbean;

import javax.management.Descriptor;       
import javax.management.DescriptorAccess; 
import javax.management.*;
import java.lang.reflect.*;

import com.sun.management.jmx.Trace;         

/**
 * The ModelMBeanConstructorInfo object describes a constructor of the ModelMBean.
 * It is a subclass of MBeanConstructorInfo with the addition of an associated Descriptor
 * and an implementation of the DescriptorAccess interface.
 * <P>
 * <PRE>
 * The fields in the descriptor are defined, but not limited to, the following: <P>
 * name           : operation name 
 * descriptorType : must be "operation"  
 * role           : must be "constructor"
 * displayName    : human readable name of constructor 
 * class          : class where method is defined (fully qualified) 
 * visibility            : 1-4 where 1: always visible 4: rarely visible 
 * presentationString :  xml formatted string to describe how to present operation 
 *</PRE>
 * PersistencePolicy and CurrencyTimeLimit fields are not valid for the constructor.
 * The default constructo will have the name, descriptorType, and role fields.
 */

public class ModelMBeanConstructorInfo extends MBeanConstructorInfo 
	 implements DescriptorAccess, Cloneable
{

	private Descriptor consDescriptor = createDefaultDescriptor();
	private String currClass = "ModelMBeanConstructorInfo";


	/**
	* Constructs a MBeanConstructorInfo object with a default descriptor.
	*
	* @param description A human readable description of the constructor.
	* @param method The java.lang.reflect.Method object describing the MBean constructor. 
	*/

	public ModelMBeanConstructorInfo(String description,
						 Constructor constructorMethod)
	{
		super(description, constructorMethod);

		if (tracing())
		{
			trace("ModelMBeanConstructorInfo(String, Method)","Executed");
		}
		consDescriptor = createDefaultDescriptor();

		// put getter and setter methods in constructors list 
		// create default descriptor

	}

	/**
	* Constructs a MBeanConstructorInfo object.
	*
	* @param description A human readable description of the constructor.
	* @param method The java.lang.reflect.Method object describing the ModelMBean constructor.
	* @param descriptor An instance of Descriptor containing the appropriate metadata
	*                   for this instance of the ModelMBeanConstructorInfo. If it is null or invalid then
	 *			a default desriptor will be created.
	*/

	public ModelMBeanConstructorInfo(String description,
					 	 Constructor constructorMethod,
					 	 Descriptor descriptor)
	{

		super(description, constructorMethod);
		// put getter and setter methods in constructors list
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo(String, Method, Descriptor)","Executed");
		}
		if (descriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanConstructorInfo(String, Method, Descriptor)","Descriptor passed in is null, setting descriptor to default values");
			}

			consDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(descriptor))
			{
				consDescriptor = (Descriptor) descriptor.clone();
			} else
			{  // exception
				consDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");

			}

		}
	}
	/**
	* Constructs a ModelMBeanConstructorInfo object with a default descriptor.
	*
	* @param name The name of the constructor.
	* @param description A human readable description of the constructor.
	* @param signature MBeanParameterInfo object array describing the parameters(arguments) of the constructor.
	*/

	public ModelMBeanConstructorInfo(String name,
									 String description,
									 MBeanParameterInfo[] signature)
	{

		super(name, description, signature);
		// create default descriptor
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo(String, String, MBeanParameterInfo[])","Executed");
		}
		consDescriptor = createDefaultDescriptor();
	}
	/**
	* Constructs a MBeanConstructorInfo object.
	*
	* @param name The name of the constructor.
	* @param description A human readable description of the constructor.
	* @param signature MBeanParameterInfo objects describing the parameters(arguments) of the constructor.
	* @param descriptor An instance of Descriptor containing the appropriate metadata
	*                   for this instance of the MBeanConstructorInfo. If it is null or invalid then
	*			a default desriptor will be created.
	*/

	public ModelMBeanConstructorInfo(String name,
									 String description,
									 MBeanParameterInfo[] signature,
									 Descriptor descriptor)
	{
		super(name, description, signature);
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo(String, String, MBeanParameterInfo[], Descriptor)","Executed");
		}
		if (descriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanConstructorInfo(String, Method, Descriptor)","Descriptor passed in is null, setting descriptor to default values");
			}
			consDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(descriptor))
			{
				consDescriptor = (Descriptor) descriptor.clone();
			} else
			{  // exception
				consDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");

			}
		}
	}

	/**
	 * Constructs a new ModelMBeanConstructorInfo object from this ModelMBeanConstructor Object.
	 *
	 * @param inInfo the ModelMBeanConstructorInfo to be duplicated
	 *
	 */
	ModelMBeanConstructorInfo(ModelMBeanConstructorInfo old)
	{
		super(old.name, old.description, old.getSignature());
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo(ModelMBeanConstructorInfo)","Executed");
		}
		if (old.consDescriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanConstructorInfo(String, Method, Descriptor)","Existing descriptor passed in is null, setting new descriptor to default values");
			}
			consDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(consDescriptor))
			{
				consDescriptor = (Descriptor) old.consDescriptor.clone();
			} else
			{  // exception
				consDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");

			}

		}
	}

	/**
	* Creates and returns a new ModelMBeanConstructorInfo which is a duplicate of this ModelMBeanConstructorInfo.
	*
	*/
	public Object clone ()
	{
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo.clone()","Executed");
		}
		return(new ModelMBeanConstructorInfo(this)) ;
	}

	/**                                                                             
	 * Returns a copy of the associated Descriptor
	 *
	 * @return Descriptor associated with the ModelMBeanConstructorInfo object.
	 */


	public Descriptor getDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo.getDescriptor()","Executed");
		}
		if (consDescriptor == null)
		{
			consDescriptor = createDefaultDescriptor();
		}
		return((Descriptor)consDescriptor.clone());
	}
	/** 
	* Sets associated Descriptor (full replace) of ModelMBeanConstructorInfo.
	* If the new Descriptor is null, then the associated Descriptor reverts to 
	* a default descriptor.  The Descriptor is validated before it is assigned.  If
	* the new Descriptor is invalid, then an IllegalArgumentException is thrown.
	*
	* @param Descriptor inDescriptor - replaces the Descriptor associated with the 
	* ModelMBeanConstructor.
	*
	*/
	public void setDescriptor(Descriptor inDescriptor)
	{
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo.setDescriptor()","Executed");
		}
		if (inDescriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanConstructorInfo(String, Method, Descriptor)","Descriptor passed in is null, setting descriptor to default values");
			}
			consDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(inDescriptor))
			{
				consDescriptor = (Descriptor) inDescriptor.clone();
			} else
			{
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	* Returns a string containing the entire contents of the ModelMBeanConstructorInfo in human readable form.
	*/
	public String toString()
	{
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo.toString()","Executed");
		}
		String retStr = new String("ModelMBeanConstructorInfo: " + this.getName() +
								   " ; Description: " + this.getDescription() +
								   " ; Descriptor: " + this.getDescriptor() +
								   " ; Signature: ");
		MBeanParameterInfo[] pTypes = this.getSignature();
		for (int i=0; i < pTypes.length; i++)
		{
			retStr = retStr.concat((pTypes[i]).getType() + ", ");
		}
		return retStr;
	}
	/**
	* Creates default descriptor for constructor as follows:
	* descriptorType=operation,role=constructor,
	* name=this.getName(),displayname=this.getName(),visibility=1
	*/
	private Descriptor createDefaultDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo.createDefaultDescriptor()","Executed");
		}
		return new DescriptorSupport(new String[] {"descriptorType=operation",
											  "role=constructor",
											  ("name=" + this.getName()),
											  ("displayname=" + this.getName())});
	}

	/**
	* Tests that the descriptor is valid and adds appropriate default fields not already
	* specified. Field values must be correct for field names. 
	* Descriptor must have the same name as the operation,the descriptorType field must 
	* be "operation", the role field must be set to "constructor".
	* The following fields will be defaulted if they are not already set:
	* displayName=this.getName(),persistPolicy=never,visibility=1
	*/
	private boolean isValid(Descriptor inDesc)
	{
		boolean results = true;
		String badField="none";
		// if name != this.getName
		// if (descriptorType != operation)
		// look for displayName, persistPolicy, visibility and add in
		if (inDesc == null)
		{
			badField="nullDescriptor";
			results = false;
		}

		else if (!inDesc.isValid())
		{	 // checks for empty descriptors, null,
			// checks for empty name and descriptorType adn valid values for fields.
			badField="invalidDescriptor";
			results = false;
		}

		else
		{
			if (! ((String)inDesc.getFieldValue("name")).equalsIgnoreCase(this.getName()))
			{
				badField="name";
				results = false;
			}
			if (! ((String)inDesc.getFieldValue("descriptorType")).equalsIgnoreCase("operation"))
			{
				badField="descriptorType";
				results = false;
			}
			if (inDesc.getFieldValue("role") == null)
			{
				inDesc.setField("role","constructor");
			}
			if (! ((String)inDesc.getFieldValue("role")).equalsIgnoreCase("constructor"))
			{
				badField = "role";
				results = false;
			} else if ((inDesc.getFieldValue("displayName")) == null)
			{
				inDesc.setField("displayName",this.getName());
			}
		}
		if (tracing()) trace("isValid()",("Returning " + results + ": Invalid field is " + badField));
		return results;
	}

	// SUN Trace and debug functions
	private boolean tracing()
	{
		//        return false;
		return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN);
	}

	private void trace(String inClass, String inMethod, String inText)
	{
		//        System.out.println("TRACE: " + inClass + ":" + inMethod + ": " + inText);
		Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN, inClass, 
				   inMethod,  Integer.toHexString(this.hashCode()) + " " + inText); 
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}

}
