/*
 * @(#)file      ModelMBeanInfo.java
 * @(#)author    IBM Corp.
 * @(#)version   1.8
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */


package javax.management.modelmbean;

import javax.management.Descriptor;
import javax.management.DescriptorAccess;
import javax.management.*;
import javax.management.RuntimeOperationsException;
import javax.management.MBeanException;

/**
 * This interface is the implemented by the ModelMBeanInfo for every ModelMBean. An implementation of this interface 
 * must be shipped with every JMX Agent. 
 * <P>
 * Java resources wishing to be manageable instatiate the ModelMBean using the MBeanServer's 
 * createMBean method.  The resource then sets the ModelMBeanInfo and Descriptors for the ModelMBean
 * instance. The attributes, operations, and notifications exposed via the ModelMBeanInfo for the 
 * ModelMBean comprise the management interface and are accessible
 * from Mbeans, connectors/adapters like other MBeans. Through the Descriptors, values and methods in
 * the managed application can be defined and mapped to attributes and operations of the ModelMBean.
 * This mapping can be defined during development in a file or dynamically and
 * programmatically at runtime. 
 * <P>
 * Every ModelMBean which is instantiated in the MBeanServer becomes manageable: 
 * its attributes, operations, and notifications
 * become remotely accessible through the connectors/adaptors connected to that MBeanServer. 
 * A Java object cannot be registered in the MBeanServer unless it is a JMX compliant MBean.
 * By instantiating a ModelMBean, resources are guaranteed that the MBean is valid.
 *
 * MBeanException and RuntimeOperatiosException must be thrown on every public method.  This allows 
 *  for wrappering exceptions from distributed communications (RMI, EJB, etc.)
 */

public interface ModelMBeanInfo
{


	/**
	 * returns a Descriptor array consisting of all 
	 * Descriptors for the ModelMBeanInfo including the MBean Descriptor, attribute Descriptors, 
	 * operation Descriptors, constructor Descriptors, and notification Descriptors.
	 *
	 * @param inDescriptorType value of descriptorType field that must be set for the descriptor
	 * to be returned.  Must be "mbean", "attribute", "operation", "constructor", or "notification".
	 * If it is null then all types will be returned.
	 *
	 * @return Descriptor array containing all descriptors for the ModelMBean
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception 
	 *
	 */


	public Descriptor[] getDescriptors(String inDescriptorType)
	throws MBeanException, RuntimeOperationsException;

	/**
	* adds or replaces descriptors in the ModelMBeanInfo.
	*
	* @param inDescriptors The descriptors to be set in the ModelMBeanInfo. Null
	* elements of the list will be ignored.  All descriptors must have name and descriptorType fields.
	*
	* @return void
	*
	* @exception RuntimeOperationsException Wraps exceptions for illegal or null arguments
	*
	*/


	public void setDescriptors(Descriptor[] inDescriptors)
	throws MBeanException, RuntimeOperationsException;

	/**
	 * returns a Descriptor requested by name and descriptorType.
	 *
	 * @param String descriptorName The name of the descriptor.
	 * @param String descriptorType The type of the descriptor being requested.  If this is null
	 * then all types are searched. Valid types are 'attribute', 'constructor', 'operation',
	 * and 'notification'. This value must be equal to the 'descriptorType' field in the
	 * descriptor that is returned.
	 *
	 * @return Descriptor containing the descriptor for the ModelMBean with the same name and descriptorType.
	 * If a descriptor is not found, null is returned.
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps exceptions for invalid input name or type. 
	 *
	 */

	public Descriptor getDescriptor(String inDescriptorName, String inDescriptorType)
	throws MBeanException, RuntimeOperationsException;

	/**
	 * adds or replaces descriptors in all the info arrays for the ModelMBean.
	 *
	 * @param Descriptor The descriptor to be set in the ModelMBean. It must NOT be null.
	 * All descriptors must have name and descriptorType fields.
	 * @param String descriptorType The type of the descriptor being set. If this is null
	 * then the descriptorType field in the descriptor is used. If specified 
	 * this value must be set in the descriptorType field for the descriptor
	 * to be returned.  Must be "mbean","attribute", "operation", "constructor", or "notification".
	 *
	 * @return void
	 *
	 * @exception RuntimeOperationsException Wraps exceptions for illegal or null arguments
	 *
	 */

	public void setDescriptor(Descriptor inDescriptor, String inDescriptorType)
	throws MBeanException, RuntimeOperationsException;

	/**
	* returns a MBean's descriptor.  This descriptor contains
	* metadata about the MBean and default policies for persistence and caching for the entire MBean.
	* Policies may be overridden by descriptors associated with attribute, constructors, or operations
	* with the same fieldNames.
	*
	* @exception MBeanException Wraps another exception 
	* @exception RuntimeOperationsException Wraps another exception 
	*
	*/


	public Descriptor getMBeanDescriptor()
	throws MBeanException, RuntimeOperationsException;

	/**
	* sets the ModelMBean's descriptor (fully replace).  
	* This descriptor contains
	* metadata about the MBean and default policies for persistence and caching for the 
	* entire MBean. 
	* Policies may be overridden by descriptors associated with attribute, constructors, or operations
	* with the same fieldNames.
	* This operation does a complete replacement of the descriptor, no merging is done.
	* The default MBeanDescriptor is 
	*
	* @exception MBeanException Wraps another exception 
	* @exception RuntimeOperationsException Wraps another exception 
	*
	*/


	public void setMBeanDescriptor(Descriptor inDescriptor)
	throws MBeanException, RuntimeOperationsException;


	/**
	 * returns a ModelMBeanAttributeInfo requested by name.
	 *
	 * @param String inName The name of the ModelMBeanAttributeInfo to get.
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception for invalid attribute
	 * name or ModelMBeanAttributeInfo to be returned.
	 *
	 */

	public ModelMBeanAttributeInfo getAttribute(String inName)
	throws MBeanException, RuntimeOperationsException;


	/**
	 * returns a ModelMBeanOperationInfo requested by name.
	 *
	 * @param String inName The name of the ModelMBeanOperationInfo to get.
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception for invalid attribute
	 * name or ModelMBeanAttributeInfo to be returned.
	 *
	 */

	public ModelMBeanOperationInfo getOperation(String inName)
	throws MBeanException, RuntimeOperationsException;


	/**
	 * returns a ModelMBeanNotificationInfo requested by name.
	 *
	 * @param String inName The name of the ModelMBeanNotificationInfo to get.
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception 
	 *
	 */

	public ModelMBeanNotificationInfo getNotification(String inName)
	throws MBeanException, RuntimeOperationsException;

	public java.lang.Object clone();
	public MBeanAttributeInfo[] getAttributes(); 

	public java.lang.String getClassName(); 
	public MBeanConstructorInfo[] getConstructors(); 
	public java.lang.String getDescription(); 
	public MBeanNotificationInfo[] getNotifications(); 
	public MBeanOperationInfo[] getOperations(); 

}
