/*
 * @(#)file      ModelMBeanInfoSupport.java
 * @(#)author    IBM Corp.
 * @(#)version   1.7
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */



package javax.management.modelmbean;

import javax.management.Descriptor;
import javax.management.DescriptorAccess;
import javax.management.modelmbean.ModelMBeanInfo;
import javax.management.*;

import java.io.Serializable;
import com.sun.management.jmx.Trace;
import javax.management.MBeanException;
import javax.management.RuntimeOperationsException;

/**
 * This class represents the meta data for ModelMBeans.  Descriptors have been added on the meta data objects.
 * <P>
 * Java resources wishing to be manageable instatiate the ModelMBean using the MBeanServer's 
 * createMBean method.  The resource then sets the ModelMBeanInfo and Descriptors for the ModelMBean
 * instance. The attributes and operations exposed via the ModelMBeanInfo for the ModelMBean are accessible
 * from Mbeans, connectors/adapters like other MBeans. Through the Descriptors, values and methods in
 * the managed application can be defined and mapped to attributes and operations of the ModelMBean.
 * This mapping can be defined during development in a file or dynamically and
 * programmatically at runtime. 
 * <P>
 * Every ModelMBean which is instantiated in the MBeanServer becomes manageable: 
 * its attributes and operations
 * become remotely accessible through the connectors/adaptors connected to that MBeanServer. 
 * A Java object cannot be registered in the MBeanServer unless it is a JMX compliant MBean.
 * By instantiating a ModelMBean, resources are guaranteed that the MBean is valid.
 *
 * MBeanException and RuntimeOperationsException must be thrown on every public method.  This allows 
 *  for wrappering exceptions from distributed communications (RMI, EJB, etc.)
 */

public class ModelMBeanInfoSupport extends MBeanInfo 
	 implements ModelMBeanInfo, java.io.Serializable
{

	private Descriptor modelMBeanDescriptor = null;
	private String currClass = "ModelMBeanInfoSupport";

	private MBeanAttributeInfo[] mmbAttributes= null;

	private MBeanConstructorInfo[] mmbConstructors= null;

	private MBeanNotificationInfo[] mmbNotifications= null;

	private MBeanOperationInfo[] mmbOperations= null;

	private static final String ATTR = "attribute";
	private static final String OPER = "operation";
	private static final String NOTF = "notification";
	private static final String CONS = "constructor";
	private static final String MMB = "mbean";
	private static final String ALL = "all";

	/**
	* constructs a ModelMBeanInfoSupport which is a duplicate of the one passed in.
	*
	* @param ModelMBeanInfo mbi - the ModelMBeanInfo instance from which the ModelMBeanInfo
	* being created is initialized.
	*/
	public ModelMBeanInfoSupport(ModelMBeanInfo  mbi)
	{
		super(mbi.getClassName(),
			  mbi.getDescription(),
			  mbi.getAttributes(),
			  mbi.getConstructors(),
			  mbi.getOperations(),
			  mbi.getNotifications());

		mmbAttributes = mbi.getAttributes();
		mmbConstructors = mbi.getConstructors();
		mmbOperations = mbi.getOperations();
		mmbNotifications = mbi.getNotifications();

		try
		{
			Descriptor mbeandescriptor = mbi.getMBeanDescriptor();

			if ((mbeandescriptor != null) && isValidDescriptor(mbeandescriptor))
			{
				if (tracing()) trace("ModelMBeanInfo(ModelMBeanInfo)","ModelMBeanDescriptor is valid, cloning Descriptor *" + mbeandescriptor.toString() + "*");
				modelMBeanDescriptor = (Descriptor) mbeandescriptor.clone();
			} else
			{
				if (tracing()) trace("ModelMBeanInfo(ModelMBeanInfo)",
									 "ModelMBeanDescriptor in ModelMBeanInfo is null or invalid, setting to default value");
				modelMBeanDescriptor = createDefaultDescriptor();
			}
		} catch (MBeanException mbe)
		{
			modelMBeanDescriptor = createDefaultDescriptor();
			if (tracing()) trace("ModelMBeanInfo(ModelMBeanInfo)","Could not get modelMBeanDescriptor, setting to default value");
		}

		if (tracing())
		{
			trace("ModelMBeanInfo(ModelMBeanInfo)","Executed");
		}
	}

	/**
	 * Creates a ModelMBeanInfoSupport with the provided information, but the descriptor is a default.
	 * The default descriptor is: name=mbeanName,descriptorType=mbean, 
	 * displayName=this.getClassName(), persistPolicy=never,log=F,export=F,visiblity=1
	 *
	 * @param className - classname of ModelMBeanInfo
	 * @param description - human readable description of the ModelMBean
	 * @param ModelMBeanAttributeInfo[] attributes - array of ModelMBeanAttributeInfo objects which have descriptors
	 * @param ModelMBeanConstructorInfo[] constructors - array of ModelMBeanConstructorInfo objects which have descriptor
	 * @param ModelMBeanOperationInfo[] operations - array of ModelMBeanOperationInfo objects which have descriptor
	 * @param ModelMBeanNotificationInfo[] notifications - array of ModelMBeanNotificationInfo objects which have descriptor
	 */
	public ModelMBeanInfoSupport(String className, 
								 String description,    
								 ModelMBeanAttributeInfo[] attributes,      
								 ModelMBeanConstructorInfo[] constructors,
								 ModelMBeanOperationInfo[] operations,
								 ModelMBeanNotificationInfo[] notifications)
	{
		super(className, description, attributes, constructors, operations, notifications);
		mmbAttributes = attributes;
		mmbConstructors = constructors;
		mmbOperations = operations;
		mmbNotifications = notifications;
		modelMBeanDescriptor = createDefaultDescriptor();
		if (tracing())
		{
			trace("ModelMBeanInfo(String,String,ModelMBeanAttributeInfo[],ModelMBeanConstructorInfo[],ModelMBeanOperationInfo[],ModelMBeanNotificationInfo[])",
				  "Executed");
		}
	}

	/**
	 * Creates a ModelMBeanInfoSupport with the provided information, but the descriptor is a default.
	 *
	 * @param className - classname of ModelMBeanInfo
	 * @param description - human readable description of the ModelMBean
	 * @param ModelMBeanAttributeInfo[] attributes - array of ModelMBeanAttributeInfo objects which have descriptors
	 * @param ModelMBeanConstructorInfo[] constructors - array of ModelMBeanConstructorInfo objects which have descriptor
	 * @param ModelMBeanOperationInfo[] operations - array of ModelMBeanOperationInfo objects which have descriptor
	 * @param ModelMBeanNotificationInfo[] notifications - array of ModelMBeanNotificationInfo objects which have descriptor
	 * @param MBeanDescriptor - descriptor to be used as the MBeanDescriptor containing mbean wide policy. if the
	 *		descriptor is invalid or null, a default descriptor will be constructed.
	 * The default descriptor is: name=mbeanName,descriptorType=mbean, 
	 * displayName=this.getClassName(), persistPolicy=never,log=F,export=F,visiblity=1.
	 * If the descriptor does not contain all these fields, they will be added with these default values.
	 */

	public ModelMBeanInfoSupport(String    className, 
								 String description,    
								 ModelMBeanAttributeInfo[] attributes,      
								 ModelMBeanConstructorInfo[] constructors,
								 ModelMBeanOperationInfo[] operations,
								 ModelMBeanNotificationInfo[] notifications,
								 Descriptor mbeandescriptor)
	{
		super(className, description, attributes, constructors, operations, notifications);
		mmbAttributes = attributes;
		mmbConstructors = constructors;
		mmbOperations = operations;
		mmbNotifications = notifications;
		if ((mbeandescriptor != null) && (isValidDescriptor(mbeandescriptor)))
		{
			if (tracing())
				trace("ModelMBeanInfo(String,String,ModelMBeanAttributeInfo[],ModelMBeanConstructorInfo[],ModelMBeanOperationInfo[],ModelMBeanNotificationInfo[],Descriptor)",
					  "MBeanDescriptor is not null and is valid");

			modelMBeanDescriptor = (Descriptor) mbeandescriptor.clone();
		} else
		{
			if (tracing())
				trace("ModelMBeanInfo(String,String,ModelMBeanAttributeInfo[],ModelMBeanConstructorInfo[],ModelMBeanOperationInfo[],ModelMBeanNotificationInfo[],Descriptor)",
					  "MBeanDescriptor is null or invalid, setting default descriptor");

			modelMBeanDescriptor = createDefaultDescriptor();
		}
		if (tracing())
		{
			trace("ModelMBeanInfo(String,String,ModelMBeanAttributeInfo[],ModelMBeanConstructorInfo[],ModelMBeanOperationInfo[],ModelMBeanNotificationInfo[],Descriptor)",
				  "Executed");
		}
	}
	/**
	* returns a new ModelMBeanInfoSupport object that is a copy of this ModelMBeanInfoSupport.
	*/
	public Object clone()
	{
		return(new ModelMBeanInfoSupport(this));
	}
	/**
	 * Returns a DescriptorList.
	 * It contains all Descriptors for the ModelMBean including the attributeDescriptors, 
	 * operationDescriptors, and notificationDescriptors.
	 *
	 * @return Descriptor array containing all descriptors for the ModelMBean
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception 
	 *
	 */

	public Descriptor[] getDescriptors(String inDescriptorType)
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getDescriptors()","Entry");
		}

		if ((inDescriptorType == null) || (inDescriptorType == ""))
		{
			inDescriptorType = "all";
		}
		// if no descriptors of that type, will return empty array
		Descriptor[] retList;  

		if (inDescriptorType.equals(MMB))
		{
			retList = new Descriptor[1];
			retList[0] = modelMBeanDescriptor;
		}
		if (inDescriptorType.equals(ATTR))
		{
			MBeanAttributeInfo[] attrList = mmbAttributes;
			int numAttrs = 0;
			if (attrList != null) numAttrs = attrList.length;

			retList = new Descriptor[numAttrs];
			for (int i=0; i < numAttrs; i++)
			{
				retList[i] = (((ModelMBeanAttributeInfo) attrList[i]).getDescriptor());
			}
		} else if (inDescriptorType.equals(OPER))
		{
			MBeanOperationInfo[] operList = mmbOperations; 
			int numOpers = 0;
			if (operList != null) numOpers = operList.length;

			retList = new Descriptor[numOpers];
			for (int i=0; i < numOpers; i++)
			{
				retList[i] = (((ModelMBeanOperationInfo) operList[i]).getDescriptor());
			}
		} else if (inDescriptorType.equals(CONS))
		{
			MBeanConstructorInfo[] consList =  mmbConstructors;	
			int numCons = 0;
			if (consList != null) numCons = consList.length;

			retList = new Descriptor[numCons];
			for (int i=0; i < numCons; i++)
			{
				retList[i] = (((ModelMBeanConstructorInfo) consList[i]).getDescriptor());
			}
		} else if (inDescriptorType.equals(NOTF))
		{
			MBeanNotificationInfo[] notifList = mmbNotifications; 
			int numNotifs = 0;
			if (notifList != null) numNotifs = notifList.length;

			retList = new Descriptor[numNotifs];
			for (int i=0; i < numNotifs; i++)
			{
				retList[i] = (((ModelMBeanNotificationInfo) notifList[i]).getDescriptor());
			}
		} else if (inDescriptorType.equals(ALL))
		{

			MBeanAttributeInfo[] attrList = mmbAttributes; 
			int numAttrs = 0;
			if (attrList != null) numAttrs = attrList.length;

			MBeanOperationInfo[] operList = mmbOperations; 
			int numOpers = 0;
			if (operList != null) numOpers = operList.length;

			MBeanConstructorInfo[] consList = mmbConstructors; 
			int numCons = 0;
			if (consList != null) numCons = consList.length;

			MBeanNotificationInfo[] notifList = mmbNotifications; 
			int numNotifs = 0;
			if (notifList != null) numNotifs = notifList.length;


			retList = new Descriptor[numAttrs + numCons + numOpers + numNotifs];
			int j=0;
			for (int i=0; i < numAttrs; i++)
			{
				retList[j] = (((ModelMBeanAttributeInfo) attrList[i]).getDescriptor());
				j++;
			}
			for (int i=0; i < numCons; i++)
			{
				retList[j] = (((ModelMBeanConstructorInfo)consList[i]).getDescriptor());
				j++;
			}
			for (int i=0; i < numOpers; i++)
			{
				retList[j] = (((ModelMBeanOperationInfo)operList[i]).getDescriptor());
				j++;
			}
			for (int i=0; i < numNotifs; i++)
			{
				retList[j] = (((ModelMBeanNotificationInfo)notifList[i]).getDescriptor());
				j++;
			}
		} else
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor Type is invalid"),
												 ("Exception occured trying find the descriptors of the MBean"));
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getDescriptors()","Exit");
		}

		return retList;
	}

	/**
	* Adds or replaces descriptors in all the ModelMBeanInfoSupport for the ModelMBean.
	*
	* @param DescriptorList The descriptors to be set in the ModelMBean. 
	*      All descriptors must have name and descriptorType fields and be valid.
	*      If the descriptor list is empty, no sets will occur.  If it is null, an exception
	*      will be thrown. Null elements of the list will be ignored.  
	*
	*
	* @exception RuntimeOperationsException Wraps exceptions for illegal or null arguments
	*
	*/


	public void setDescriptors(Descriptor[] inDescriptors)
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.setDescriptors(Descriptor[])","Entry");
		}
		if (inDescriptors==null)
		{ // throw RuntimeOperationsException - invalid descriptor
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor list is invalid"),
												 ("Exception occured trying set the descriptors of the MBeanInfo"));
		}
		if (inDescriptors.length == 0)
		{ // empty list, no-op
			return;
		}
		for (int j=0; j < inDescriptors.length; j++)
		{
			setDescriptor(inDescriptors[j],null);
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.setDescriptors(Descriptor[])","Exit");
		}

	}
	/**
	 * Returns a Descriptor requested by name
	 *
	 * @param String descriptorName The name of the descriptor.
	 * @return Descriptor containing the descriptor for the ModelMBean with the same name and descriptorType.
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception 
	 *
	 */

	public Descriptor getDescriptor(String inDescriptorName)
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getDescriptor(String)","Entry");
		}
		return(getDescriptor(inDescriptorName, null));
	}
	/**
	 * Returns a Descriptor requested by name and descriptorType.
	 *
	 * @param String descriptorName The name of the descriptor.
	 * @param String descriptorType The type of the descriptor being requested.  If this is null
	 * a RuntimeOperationsException will be thrown.
	 * Type must match that in the 'descriptorType' field on the Descriptor.  It must be "mbean", 
	 * "attribute", "operation", or "notification".
	 *
	 * @return Descriptor containing the descriptor for the ModelMBean with the same name and descriptorType.
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception 
	 *
	 */

	public Descriptor getDescriptor(String inDescriptorName, String inDescriptorType)
	throws MBeanException, RuntimeOperationsException {
		Descriptor retDescriptor = null;
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getDescriptor(String,String)","Entry");
		}

		if (inDescriptorName==null)
		{
			// throw RuntimeOperationsException - invalid descriptor
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor is invalid"),
												 ("Exception occured trying set the descriptors of the MBeanInfo"));
		}

		if (inDescriptorType == null)
		{ // thow RuntimeOperationsException - invalid descriptor type
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor Type is invalid"),
												 ("Exception occured trying set the descriptors of the MBean"));
		}

		if (inDescriptorType.equals(MMB))
		{
			retDescriptor = ((Descriptor) modelMBeanDescriptor.clone());
		} else if (inDescriptorType.equals(ATTR) || (inDescriptorType == null) ||
				   (retDescriptor != null))
		{
			ModelMBeanAttributeInfo attr = getAttribute(inDescriptorName);
			if (attr != null)
			{
				retDescriptor = attr.getDescriptor();
			}
		} else if (inDescriptorType.equals(OPER) || (inDescriptorType == null) ||
				   (retDescriptor != null))
		{
			ModelMBeanOperationInfo oper = getOperation(inDescriptorName);
			if (oper != null)
			{
				retDescriptor = oper.getDescriptor();
			}
		} else if (inDescriptorType.equals(CONS) || (inDescriptorType == null) ||
				   (retDescriptor != null))
		{
			ModelMBeanConstructorInfo oper = getConstructor(inDescriptorName);
			if (oper != null)
			{
				retDescriptor = oper.getDescriptor();
			}
		} else if (inDescriptorType.equals(NOTF) || (inDescriptorType == null) ||
				   (retDescriptor != null))
		{
			ModelMBeanNotificationInfo notif = getNotification(inDescriptorName);
			if (notif != null)
			{
				retDescriptor = notif.getDescriptor();
			}

		} else
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor Type is invalid"),
												 ("Exception occured trying find the descriptors of the MBean"));
		}

		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getDescriptor(String,String)","Exit");
		}

		return retDescriptor;
	}

	/**
	 * Adds or replaces a descriptor in the ModelMBeanInfoSupport.
	 *
	 * @param inDescriptor The descriptor to be set in the ModelMBean. 
	 *
	 *      If the descriptor to be set is null, no sets will be done.
	 * All descriptors must have name and descriptorType fields.
	 * @param String descriptorType The type of the descriptor being requested.  If this is null then
	 * the value of the 'descriptorType' field in the inDescriptor will be used.
	 * Type must match that in the 'descriptorType' field on the Descriptor.  It must be "mbean", 
	 * "attribute", "operation", or "notification".
	 *
	 *
	 * @exception RuntimeOperationsException Wraps exceptions for illegal or null arguments
	 *
	 */

	public void setDescriptor(Descriptor inDescriptor, String inDescriptorType)
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.setDescriptor(Descriptor,String)","Entry");
		}

		if (inDescriptor==null)
		{ // throw RuntimeOperationsException - invalid descriptor
			if (tracing())
				trace("ModelMBeanInfoSupport.setDescriptor(Descriptor,String)","Descriptor is null");
			return;
		}

		if (inDescriptorType == null)
		{
			inDescriptorType = (String) inDescriptor.getFieldValue("descriptorType");

			if (inDescriptorType == null)
			{ // thow RuntimeOperationsException - invalid descriptor type
				throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor Type is invalid"),
													 ("Exception occured trying set the descriptors of the MBean"));
			}
		}

		String inDescriptorName = (String) inDescriptor.getFieldValue("name");
		if (inDescriptorName == null)
		{ // thow RuntimeOperationsException - invalid descriptor type
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor Name is invalid"),
												 ("Exception occured trying set the descriptors of the MBean"));
		}

		if (inDescriptorType.equals(MMB))
		{
			setMBeanDescriptor(inDescriptor);
		} else if (inDescriptorType.equals(ATTR))
		{

			//MBeanAttributeInfo[] attrList =  this.getAttributes();

			MBeanAttributeInfo[] attrList =  mmbAttributes;
			int numAttrs = 0;
			if (attrList != null) numAttrs = attrList.length;

			for (int i=0; i < numAttrs; i++)
			{
				if (inDescriptorName.equals(attrList[i].getName()))
				{
					((ModelMBeanAttributeInfo)attrList[i]).setDescriptor(inDescriptor);
					if (tracing())
					{
						trace("ModelMBeanInfoSupport.setDescriptor","setting descriptor to " + inDescriptor.toString());
						trace("ModelMBeanInfoSupport.setDescriptor","local: AttributeInfo descriptor is " + ((ModelMBeanAttributeInfo)attrList[i]).getDescriptor().toString());
						trace("ModelMBeanInfoSupport.setDescriptor","modelMBeanInfo: AttributeInfo descriptor is " + this.getDescriptor(inDescriptorName,"attribute").toString());
					}
				}
			}
		} else if (inDescriptorType.equals(OPER))
		{
			MBeanOperationInfo[] operList =  mmbOperations;	
			int numOpers = 0;
			if (operList != null) numOpers = operList.length;

			for (int i=0; i < numOpers; i++)
			{
				if (inDescriptorName.equals(operList[i].getName()))
				{
					((ModelMBeanOperationInfo)operList[i]).setDescriptor(inDescriptor);
				}
			}
		} else if (inDescriptorType.equals(CONS))
		{
			MBeanConstructorInfo[] consList =  mmbConstructors;	
			int numCons = 0;
			if (consList != null) numCons = consList.length;

			for (int i=0; i < numCons; i++)
			{
				if (inDescriptorName.equals(consList[i].getName()))
				{
					((ModelMBeanConstructorInfo)consList[i]).setDescriptor(inDescriptor);
				}
			}
		} else if (inDescriptorType.equals(NOTF))
		{
			MBeanNotificationInfo[] notifList =  mmbNotifications;
			int numNotifs = 0;
			if (notifList != null) numNotifs = notifList.length;

			for (int i=0; i < numNotifs; i++)
			{
				if (inDescriptorName.equals(notifList[i].getName()))
				{
					((ModelMBeanNotificationInfo)notifList[i]).setDescriptor(inDescriptor);
				}
			}
		} else
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Descriptor Type is invalid"),
												 ("Exception occured trying set the descriptors of the MBean"));
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.setDescriptor(Descriptor,String)","Exit");
		}

	}

	/**
	 * Returns a ModelMBeanAttributeInfo requested by name.
	 *
	 * @param String inName The name of the ModelMBeanAttributeInfo to get.
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception for invalid attribute
	 * name or ModelMBeanAttributeInfo to be returned.
	 *
	 */

	public ModelMBeanAttributeInfo getAttribute(String inName)
	throws MBeanException, RuntimeOperationsException {
		ModelMBeanAttributeInfo retInfo = null;
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getAttributeInfo(String)","Entry");
		}
		if (inName == null)
		{ // thow RuntimeOperationsException - invalid argument
			if (tracing()) trace("ModelMBeanInfoSupport.getAttributeInfo(String)","Attribute name must not be null");
			return retInfo;
		}
		MBeanAttributeInfo[] attrList = mmbAttributes;
		int numAttrs = 0;
		if (attrList != null) numAttrs = attrList.length;

		for (int i=0; (i < numAttrs) && (retInfo == null); i++)
		{
			if (tracing())
			{
				trace("ModelMBeanInfoSupport.getAttribute","this.getAttributes() MBeanAttributeInfo Array " + i + ":" + ((ModelMBeanAttributeInfo)attrList[i]).getDescriptor().toString());
				trace("ModelMBeanInfoSupport.getAttribute","this.mmbAttributes MBeanAttributeInfo Array " + i + ":" + ((ModelMBeanAttributeInfo)mmbAttributes[i]).getDescriptor().toString());
			}
			if (inName.equals(attrList[i].getName()))
			{
				retInfo = ((ModelMBeanAttributeInfo)attrList[i].clone());
			}
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getAttribute()","Exit");
		}

		return retInfo;
	}

	/**
	 * Returns a ModelMBeanOperationInfo requested by name.
	 *
	 * @param String inName The name of the ModelMBeanOperationInfo to get.
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception for invalid attribute
	 * name or ModelMBeanAttributeInfo to be returned.
	 *
	 */

	public ModelMBeanOperationInfo getOperation(String inName)
	throws MBeanException, RuntimeOperationsException {
		ModelMBeanOperationInfo retInfo = null;
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getOperation(String)","Entry");
		}
		if (inName == null)
		{ // thow RuntimeOperationsException - invalid argument
			if (tracing()) trace("ModelMBeanInfoSupport.getOperation(String)","Operation name must not be null");
			return retInfo;
		}
		MBeanOperationInfo[] operList = mmbOperations; //this.getOperations();
		int numOpers = 0;
		if (operList != null) numOpers = operList.length;

		for (int i=0; (i < numOpers) && (retInfo == null); i++)
		{
			if (inName.equals(operList[i].getName()))
			{
				retInfo = ((ModelMBeanOperationInfo) operList[i].clone());
			}
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getOperation(String)","Exit");
		}

		return retInfo;
	}

	public ModelMBeanConstructorInfo getConstructor(String inName)
	throws MBeanException, RuntimeOperationsException {
		ModelMBeanConstructorInfo retInfo = null;
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getConstructor(String)","Entry");
		}
		if (inName == null)
		{ // thow RuntimeOperationsException - invalid argument
			if (tracing()) trace("ModelMBeanInfoSupport.getConstructor(String)","Constructor name must not be null");
			return retInfo;
		}
		MBeanConstructorInfo[] consList = mmbConstructors; //this.getConstructors();
		int numCons = 0;
		if (consList != null) numCons = consList.length;

		for (int i=0; (i < numCons) && (retInfo == null); i++)
		{
			if (inName.equals(consList[i].getName()))
			{
				retInfo = ((ModelMBeanConstructorInfo) consList[i].clone());
			}
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getConstructor(String)","Exit");
		}

		return retInfo;
	}
	/**
	 * Returns a ModelMBeanNotificationInfo requested by name.
	 *
	 * @param String inName The name of the ModelMBeanNotificationInfo to get.
	 *
	 * @exception MBeanException Wraps another exception 
	 * @exception RuntimeOperationsException Wraps another exception 
	 *
	 */

	public ModelMBeanNotificationInfo getNotification(String inName)
	throws MBeanException, RuntimeOperationsException {
		ModelMBeanNotificationInfo retInfo = null;
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getNotification(String)","Entry");
		}
		if (inName == null)
		{ // thow RuntimeOperationsException - invalid argument
			if (tracing()) trace("ModelMBeanInfoSupport.getNotification(string)","Notification name must not be null");
			return retInfo;
		}
		MBeanNotificationInfo[] notifList = mmbNotifications; //this.getNotifications();
		int numNotifs = 0;
		if (notifList != null) numNotifs = notifList.length;

		for (int i=0; (i < numNotifs) && (retInfo == null); i++)
		{
			if (inName.equals(notifList[i].getName()))
			{
				retInfo = ((ModelMBeanNotificationInfo) notifList[i].clone());
			}
		}
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getNotification(String)","Exit");
		}

		return retInfo;
	}

	/* MBeanInfo support */
	/**
	* Returns the list of  attributes exposed for management as ModelMBeanAttributeInfo objects.
	* This method signature must specify MBeanAttributeInfo to be compliant with the 
	* the DynamicMBean interface.  
	* @return  MBeanAttributeInfo object array.
	*/    
	public MBeanAttributeInfo[] getAttributes()
	{
		if (mmbAttributes == null) return new MBeanAttributeInfo[0];
		MBeanAttributeInfo[] attrs = new MBeanAttributeInfo[mmbAttributes.length];
		for (int i = 0; i < mmbAttributes.length; i++)
		{
			attrs[i] = (MBeanAttributeInfo) (mmbAttributes[i]).clone();         
		}
		return attrs;

	} 

	/**
	* Returns the list of operations exposed for management as ModelMBeanOperationInfo objects.
	* This method signature must specify MBeanOperationInfo to be compliant with the 
	* the DynamicMBean interface.  
	* @return  MBeanOperationInfo object array.
	*/    
	public MBeanOperationInfo[] getOperations()
	{
		if (mmbOperations == null) return new MBeanOperationInfo[0];
		MBeanOperationInfo[] opers = new MBeanOperationInfo[mmbOperations.length];
		for (int i = 0; i < mmbOperations.length; i++)
		{
			opers[i] = (MBeanOperationInfo) (mmbOperations[i]).clone();         
		}
		return opers;
	} 

	/**
	* Returns the list of constructors exposed for management as ModelMBeanConstructorInfo objects.
	* This method signature must specify MBeanConstructorInfo to be compliant with the 
	* the DynamicMBean interface.  
	* @return  MBeanConstructorInfo object array.
	*/    
	public MBeanConstructorInfo[] getConstructors()
	{

		if (mmbConstructors == null) return new MBeanConstructorInfo[0];
		MBeanConstructorInfo[] cons = new MBeanConstructorInfo[mmbConstructors.length];
		for (int i = 0; i < mmbConstructors.length; i++)
		{
			cons[i] = (MBeanConstructorInfo) (mmbConstructors[i]).clone();          
		}
		return cons;
	} 

	/**
	* Returns the list of notifications exposed for management as ModelMBeanNotificationInfo objects.
	* This method signature must specify MBeanNotificationInfo to be compliant with the 
	* the DynamicMBean interface.  
	* @return  MBeanNotificationInfo object array.
	*/    
	public MBeanNotificationInfo[] getNotifications()
	{

		if (mmbNotifications == null) return new MBeanNotificationInfo[0];
		MBeanNotificationInfo[] notifs = new MBeanNotificationInfo[mmbNotifications.length];
		for (int i = 0; i < mmbNotifications.length; i++)
		{
			notifs[i] = (MBeanNotificationInfo) (mmbNotifications[i]).clone();          
		}
		return notifs;
	} 


	/**
	* Returns the ModelMBean's descriptor which contains mbean wide policies.  This descriptor contains
	* metadata about the MBean and default policies for persistence and caching.
	* <P> <PRE>
	* The fields in the descriptor are defined, but not limited to, the following: 
	* name           : mbean name  
	* descriptorType : must be "mbean"   
	* displayName    : name of attribute to be used in displays 
	* persistPolicy  : OnUpdate|OnTimer|NoMoreOftenThan|Always|Never  
	* persistLocation : The fully qualified directory name where the MBean should be persisted (if appropriate)
	* persistFile    : File name into which the MBean should be persisted
	* persistPeriod  : seconds - frequency of persist cycle for OnTime and NoMoreOftenThan PersistPolicy 
	* persistLocation : directory name in which to store mbeans that support the PersistentMBean interface
	* currencyTimeLimit : how long value is valid, <0 never, =0 always, >0 seconds  
	* log            : where t: log all notifications f: log no notifications
	* logfile        : fully qualified filename to log events to
	* visibility     : 1-4 where 1: always visible 4: rarely visible
	* export         : name to be used to export/expose this MBean so that it is findable by
	*                  other JMX Agents.   
	* presentationString : xml formatted string to allow presentation of data to be associated with the MBean. 
	* </PRE>
	* <P>
	* The default descriptor is: name=mbeanName,descriptorType=mbean, displayName=this.getClassName(), 
	*	persistPolicy=never,log=F,export=F,visiblity=1
	* If the descriptor does not contain all these fields, they will be added with these default values.
	*
	* @exception MBeanException Wraps another exception 
	* @exception RuntimeOperationsException Wraps another exception 
	*
	*/

	public Descriptor getMBeanDescriptor() 
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.getMBeanDescriptor()","Executed");
		}
		if (modelMBeanDescriptor == null)
		{
			return null;
		}
		if (tracing()) trace("ModelMBeanInfoSupport.getMBeanDesriptor()", "Returning " + modelMBeanDescriptor.toString());
		return((Descriptor) modelMBeanDescriptor.clone());
	}

	/**
	* Sets the ModelMBean's descriptor.  This descriptor contains default, mbean wide 
	* metadata about the MBean and default policies for persistence and caching. This operation
	* does a complete replacement of the descriptor, no merging is done. If the descriptor to
	* set to is invalid then the existing or default descriptor will be preserved.
	* The default descriptor is: name=mbeanName,descriptorType=mbean, displayName=this.getClassName(), 
	*	persistPolicy=never,log=F,export=F,visiblity=1
	* If the descriptor does not contain all these fields, they will be added with these default values.
	*
	* See getMBeanDescriptor method javadoc for description of valid field names.
	*
	* @exception MBeanException Wraps another exception 
	* @exception RuntimeOperationsException Wraps another exception for bad descriptor. 
	*
	*/

	public void setMBeanDescriptor(Descriptor inMBeanDescriptor)
	throws MBeanException, RuntimeOperationsException {
		if (tracing())
		{
			trace("ModelMBeanInfoSupport.setMBeanDescriptor(Descriptor)","Executed");
		}
		if ((inMBeanDescriptor == null) ||  (modelMBeanDescriptor.equals("")) ||
			(!isValidDescriptor(inMBeanDescriptor) ))
		{
			if (tracing()) trace("ModelMBeanInfoSupport.setMBeanDescriptor(Descriptor)","MBean Descriptor is not valid");
			modelMBeanDescriptor = createDefaultDescriptor();                              
		} else
		{
			modelMBeanDescriptor = (Descriptor) inMBeanDescriptor.clone();

		}
	}

	/* The default descriptor is:
	* name=mbeanName,descriptorType=mbean, displayName=this.getClassName(), 
	*	persistPolicy=never,log=F,export=F,visiblity=1
	*/
	private Descriptor createDefaultDescriptor()
	{

		Descriptor dftDesc = null;
		dftDesc = new DescriptorSupport(new String[] {("name=" + this.getClassName()),
												 "descriptorType=mbean",
												 ("displayName=" + this.getClassName()),
												 "persistPolicy=never",
												 "log=F",
												 "export=F",
												 "visibility=1"});
		return dftDesc;
	}
	/*
	* Validates the ModelMBeanDescriptor 
	* If the descriptor does not contain all these fields, they will be added with these default values.
	* name=mbeanName,descriptorType=mbean, displayName=this.getClassName(), 
	*	persistPolicy=never,log=F,export=F,visiblity=1
	*
	* Will return false if the MBeanDescriptor has a null name or descriptorType.
	*/
	private boolean isValidDescriptor(Descriptor inDesc)
	{
		boolean results = true;
		String badField = "none";
		// if name != mbi.getClassName
		// if (descriptorType != mbean)
		// look for displayName, persistPolicy, logging, export, visibility and add in
		if (tracing()) trace("isValidDescriptor","Validating descriptor: " + inDesc.toString());
		if (inDesc == null)
		{
			badField = "nullDescriptor";
			results = false;
		}

		else if (!inDesc.isValid())
		{	 // checks for empty descriptors, null,
			// checks for empty name and descriptorType adn valid values for fields.
			badField="InvalidDescriptor";           
			results = false;
		}

		if (((String)inDesc.getFieldValue("name")) == null)
		{
			badField="name";
			results = false;
		} else if (! ((String)inDesc.getFieldValue("descriptorType")).equalsIgnoreCase(MMB))
		{
			badField="descriptorType";
			results = false;
		} else if ((inDesc.getFieldValue("displayName")) == null)
		{
			inDesc.setField("displayName",this.getClassName());
		} else if ((inDesc.getFieldValue("persistPolicy")) == null)
		{
			inDesc.setField("persistPolicy","never");
		} else if ((inDesc.getFieldValue("log")) == null)
		{
			inDesc.setField("log","F");
		} else if ((inDesc.getFieldValue("export")) == null)
		{
			inDesc.setField("export","F");
		} else if ((inDesc.getFieldValue("visiblity")) == null)
		{
			inDesc.setField("visibility","1");
		}
		if (tracing()) trace("isValidDescriptor","returning " + results + ": Invalid field is " + badField);
		return results;
	}

	// SUN Trace and debug functions
	private boolean tracing()
	{
		return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN);
	}

	private void trace(String inClass, String inMethod, String inText)
	{
		Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN, inClass, 
				   inMethod, Integer.toHexString(this.hashCode()) + " " + inText); 
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}


}


