/*
 * @(#)file      ModelMBeanNotificationBroadcaster.java
 * @(#)author    IBM Corp.
 * @(#)version   1.8
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */

package javax.management.modelmbean;

import javax.management.*;

import javax.management.InstanceNotFoundException;

/**                    
 * This interface must be implemented by the ModelMBeans. An implementation of this interface 
 * must be shipped with every JMX Agent. 
 * <P>
 * Java resources wishing to be manageable instatiate the ModelMBean using the MBeanServer's 
 * createMBean method.  The resource then sets the ModelMBeanInfo (with Descriptors) for the ModelMBean
 * instance. The attributes and operations exposed via the ModelMBeanInfo for the ModelMBean are accessible
 * from Mbeans, connectors/adapters like other MBeans. Through the ModelMBeanInfo Descriptors, values and methods in
 * the managed application can be defined and mapped to attributes and operations of the ModelMBean.
 * This mapping can be defined during development in an XML formatted file or dynamically and
 * programmatically at runtime. 
 * <P>
 * Every ModelMBean which is instantiated in the MBeanServer becomes manageable: 
 * its attributes and operations
 * become remotely accessible through the connectors/adaptors connected to that MBeanServer. 
 * A Java object cannot be registered in the MBeanServer unless it is a JMX compliant MBean.
 * By instantiating a ModelMBean, resources are guaranteed that the MBean is valid.
 * <P>
 * MBeanException and RuntimeOperatiosException must be thrown on every public method.  This allows 
 * for wrappering exceptions from distributed communications (RMI, EJB, etc.).  These exceptions do
 * not have to be thrown by teh implementation except in the scenarios described in the specification 
 * and javadoc.
 */

public interface ModelMBeanNotificationBroadcaster extends NotificationBroadcaster
{

	/** 
	 * Sends a Notification which is passed in to the registered Notification listeners on 
	 * the ModelMBean as a jmx.modelmbean.general notification.
	 *
	 * @param Notification ntfyObj - The notification which is to be passed to the 'handleNotification' method
	 *   of the listener object.
	 *
	 * @exception MBeanException The initializer of the object has thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: 
	 *       The Notification object passed in parameter is null or invalid.
	 *
	 */

	public void sendNotification(Notification ntfyObj)
	throws MBeanException, RuntimeOperationsException;

	/** 
	 * Sends a Notification which contains the text string that is passed in
	 * to the registered Notification listeners on the ModelMBean.
	 *
	 * @param ntfyText The text which is to be passed in the Notification to the 'handleNotification'
	 * method of the listener object.  
	 * the constructed Notification will be:
	 *   type        "jmx.modelmbean.general"
	 *   source      this ModelMBean instance
	 *   sequence    1
	 *
	 *
	 * @exception MBeanException The initializer of the object has thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: 
	 *       The Notification text string passed in parameter is null or invalid.
	 *
	 */
	public void sendNotification(String ntfyText)
	throws MBeanException, RuntimeOperationsException;

	/** 
	 * Sends an attributeChangeNotification which is passed in to the registered 
	 * attributeChangeNotification listeners on the ModelMBean.
	 *
	 * @param notification The notification which is to be passed to the 'handleNotification' method
	 *   of the listener object.
	 *
	 * @exception MBeanException The initializer of the object has thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: 
	 *       The Notification object passed in parameter is null or invalid.
	 *
	 */
	public void sendAttributeChangeNotification(AttributeChangeNotification ntfyObj)
	throws MBeanException, RuntimeOperationsException;


	/** 
	 * Sends an attributeChangeNotification which contains the old value and new value for the
	 * attribute to the registered AttributeChangeNotification listeners on the ModelMBean.
	 * <P>
	 * @param oldValue The origional value for the Attribute
	 * @param newValue The current value for the Attribute
	 *<P>
	 * <PRE>  
	 * The constructed attributeChangeNotification will be:
	 *   type        "jmx.attribute.change"
	 *   source      this ModelMBean instance
	 *   sequence    1
	 *   attributeName oldValue.getName()
	 *   attributeType oldValue's class
	 *   attributeOldValue oldValue.getValue()
	 *   attributeNewValue newValue.getValue()
	 * </PRE>
	 *
	 * @exception MBeanException to wrapper implementation exceptions
	 * @exception RuntimeOperationsException to wrapper IllegalArgumentExceptions.                                                                                 
	 *
	 */
	public void sendAttributeChangeNotification(Attribute inOldVal, Attribute inNewVal)
	throws MBeanException, RuntimeOperationsException;

	/** These interface should be implemented by any MBean wishing to broadcast AttributeChangeNotifications
	* to a separate set of listeners that other notifications would go to.  All other notifications should
	* go to listeners who registered using the NotificationBroadcaster interface.  
	* <P>
	* Use of this interface a convenience and is optional.  AttributeChangeNotifications can be sent to all Notification
	* listeners simply by using the NotificationBroadcaster interface alone.
	* <P>
	* Imlementers of this interface must also implement the NotificationBroadcaster interface.
	* ModelMBeans must implement this interface.
	*/
	/**
	 * Registers an object which implements the NotificationListener interface as a listener
	 * for AttributeChangeNotifications.  This
	 * object's 'handleNotification()' method will be invoked when any attributeChangeNotification is issued through
	 * or by the MBean.  This does not include other Notifications.  They must be registered
	 * for independently. An AttributeChangeNotification will be generated for this attributeName.
	 *   
	 * @param listener The listener object which will handles notifications emitted by the registered MBean.
	 * @param attributeName The name of the MBean attribute for which to receive change notifications.
	 *      If null, then all attribute changes will cause an attributeChangeNotification to be issued. 
	 * @param handback The context to be sent to the listener with the notification when a notification is emitted. 
	 *
	 * @exception IllegalArgumentException Listener is null or attributeName is null.
	 *
	 */
	public void addAttributeChangeNotificationListener(NotificationListener inlistener,
													   String inAttributeName, 
													   Object inhandback)
	throws MBeanException, RuntimeOperationsException, IllegalArgumentException;


	/**
	 * Removes a listener for attributeChangeNotifications from the MBean.
	 *   
	 * @param listener The listener name which was handling notifications emitted by the registered MBean. 
	 * This method will remove all information related to this listener.
	 * @param attributeName The attribute for which the listener no longer wants to receive attributeChangeNotifications.
	 *   
	 * @exception ListenerNotFoundException The couple (listener,handback) is not registered in the MBean.
	 * The exception message contains either "listener", "handback" or the object name  depending on which object cannot be found.
	 *
	 */
	public void removeAttributeChangeNotificationListener(NotificationListener inlistener,
														  String inAttributeName)
	throws MBeanException, RuntimeOperationsException, ListenerNotFoundException;

}
