/*
 * @(#)file      ModelMBeanNotificationInfo.java
 * @(#)author    IBM Corp.
 * @(#)version   1.10
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */

package javax.management.modelmbean;    

import javax.management.Descriptor;       
import javax.management.DescriptorAccess; 
import javax.management.*;

import com.sun.management.jmx.Trace;         

/**
 * The ModelMBeanNotificationInfo object describes a notification emitted by a ModelMBean.
 * It is a subclass of MBeanNotificationInfo with the addition of an associated Descriptor
 * and an implementation of the Descriptor interface.
 * <P>
 * <PRE>
 * The fields in the descriptor are defined, but not limited to, the following: <P>
 * name           : notification name <P>
 * descriptorType : must be "notification"   <P> 
 * severity       : 1-5 where 1: fatal 2: severe 3: error 4: warn 5: info  <P>
 * messageID      : unique key for message text (to allow translation,analysis)  <P>
 * messageText    : text of notification   <P>
 * log            : T - log message F - do not log message   <P>
 * logfile        : string fully qualified file name appropriate for operating system<P>
 * visibility     : 1-4 where 1: always visible 4: rarely visible   <P>
 * presentationString : xml formatted string to allow presentation of data  <P>
 * </PRE>
 * The default descriptor contains the name, descriptorType, and severity=5 fields.
 */

public class ModelMBeanNotificationInfo extends MBeanNotificationInfo 
	 implements DescriptorAccess, Cloneable
{

	/* default descriptor for MBeanNotificationInfo*/
	private Descriptor ntfyDescriptor = createDefaultDescriptor();

	private String currClass = "ModelMBeanNotificationInfo";

	/**
	 * Constructs a ModelMBeanNotificationInfo object with a default descxriptor.
	 *
	 * @param notifType The notification type string (in dot notation).
	 * @param name The name of the Notification class.
	 * @param description A human readable description of the Notification. Optional.
	 * <P>
	 * The descriptor must be set before this  
	 * ModelMBeanNotificationInfo instance can be used.
	 */

	public ModelMBeanNotificationInfo(String[] notifTypes,
									  String name,
									  String description)
	{

		super(notifTypes, name, description);
		// create default descriptor
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo(String[],String,String)","Executed");
		}
		ntfyDescriptor = createDefaultDescriptor();

	}
	/**
	 * Constructs a ModelMBeanNotificationInfo object.
	 *
	 * @param notifType The notification type string (in dot notation).
	 * @param name The name of the Notification class.
	 * @param description A human readable description of the Notification. Optional.
	 * @param descriptor An instance of Descriptor containing the appropriate metadata
	 *                   for this instance of the MBeanNotificationInfo.If it is null or invalid then
	 *					a default desriptor will be created.
	 *
	 */
	public ModelMBeanNotificationInfo(String[] notifTypes,
									  String name,
									  String description, 
									  Descriptor descriptor)
	{
		super(notifTypes, name, description);
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo(String[],String,String,Descriptor)","Executed");
		}
		if (descriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanNotificationInfo()","Received null for new descriptor value, setting descriptor to default values");
			}
			ntfyDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(descriptor))
			{
				ntfyDescriptor = (Descriptor) descriptor.clone();
			} else
			{  // exception
				ntfyDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	 * Constructs a new ModelMBeanNotificationInfo object from this ModelMBeanNotfication Object.
	 *
	 * @param inInfo the ModelMBeanNotificationInfo to be duplicated
	 *
	 */
	public ModelMBeanNotificationInfo(ModelMBeanNotificationInfo inInfo)
	{
		super(inInfo.getNotifTypes(), 
			  inInfo.getName(), 
			  inInfo.getDescription());
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo(ModelMBeanNotificationInfo)","Executed");
		}
		Descriptor newDesc = inInfo.getDescriptor();
		if (newDesc == null)
		{
			if (tracing())
			{
				trace("ModelMBeanNotificationInfo(ModelMBeanNotificationInfo)","Received null for new descriptor value, setting descriptor to default values");
			}
			ntfyDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(newDesc))
			{
				ntfyDescriptor = (Descriptor) newDesc.clone();
			} else
			{  // exception
				ntfyDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	* Creates and returns a new ModelMBeanNotificationInfo which is a duplicate of this ModelMBeanNotificationInfo.
	*
	*/

	public Object clone ()
	{
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo.clone()","Executed");
		}

		return(new ModelMBeanNotificationInfo(this));
	}

	/** 
	* Returns a copy of the associated Descriptor for the ModelMBeanNotificationInfo
	*
	* @return Descriptor associated with the ModelMBeanNotificationInfo object.
	*/

	public Descriptor getDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo.getDescriptor()","Executed");
		}
		if (ntfyDescriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanNotificationInfo.getDescriptor()","Received null for new descriptor value, setting descriptor to default values");
			}
			ntfyDescriptor = createDefaultDescriptor();
		}

		return((Descriptor)ntfyDescriptor.clone());
	}
	/** 
	* Sets associated Descriptor (full replace)  for the ModelMBeanNotificationInfo
	* If the new Descriptor is null, then the associated Descriptor reverts to 
	* a default descriptor.  The Descriptor is validated before it is assigned.  If
	* the new Descriptor is invalid, then an IllegalArgumentException is thrown.
	*
	* @param Descriptor inDescriptor - replaces the Descriptor associated with the 
	* ModelMBeanNotification interface
	*
	*/

	public void setDescriptor(Descriptor inDescriptor)
	{
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo.setDescriptor(Descriptor)","Executed");
		}
		if (inDescriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanNotificationInfo.setDescriptor()","Received null for new descriptor value, setting descriptor to default values");
			}
			ntfyDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(inDescriptor))
			{
				ntfyDescriptor = (Descriptor) inDescriptor.clone();
			} else
			{
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/** 
	* Returns a human readable string containing ModelMBeanNotificationInfo
	*
	* @exception RuntimeOperationsException for illegal value for field Names or field Values.
	*              If the descriptor string fails for any reason, this exception will be thrown.
	*/

	public String toString()
	{
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo.toString()","Executed");
		}

		String retStr = new String("ModelMBeanNotificationInfo: " + this.getName() +
								   " ; Description: " + this.getDescription() +
								   " ; Descriptor: " + this.getDescriptor() +
								   " ; Types: ");
		String[] nTypes = this.getNotifTypes();
		for (int i=0; i < nTypes.length; i++)
		{
			retStr = retStr.concat(nTypes[i] + ", ");
		}
		return retStr;
	}
	/**
   * Creates default descriptor for notification as follows:
   * descriptorType=notification,
   * name=this.getName(),displayname=this.getName(),log=F,severity=5,visibility=1
   */

	private Descriptor createDefaultDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanNotificationInfo.createDefaultDescriptor()","Executed");
		}
		return new DescriptorSupport(new String[] {"descriptorType=notification",
											  ("name=" + this.getName()),
											  ("displayName=" + this.getName()),
											  "severity=5"});
	}
	/**
	* Tests that the descriptor is valid and adds appropriate default fields not already
	* specified. Field values must be correct for field names. 
	* Descriptor must have the same name as the notification,the descriptorType field must be "notification",
	* The following fields will be defaulted if they are not already set:
	* displayName=this.getName(),displayname=this.getName(),log=F,severity=5,visibility=1
	*/

	private boolean isValid(Descriptor inDesc)
	{
		boolean results = true;
		String badField = "none";
		// if name != this.getName
		// if (descriptorType != notification)
		// look for displayName, persistPolicy, visibility and add in

		if (inDesc == null)
		{
			badField="nullDescriptor";
			return false;
		}
		if (!inDesc.isValid())
		{	 // checks for empty descriptors, null,
			// checks for empty name and descriptorType adn valid values for fields.
			badField="invalidDescriptor";
			results = false;
		} else
		{

			if (! ((String)inDesc.getFieldValue("name")).equalsIgnoreCase(this.getName()))
			{
				badField="name";
				results = false;
			}
			if (! ((String)inDesc.getFieldValue("descriptorType")).equalsIgnoreCase("notification"))
			{
				badField="descriptorType";
				results = false;
			}
			if ((inDesc.getFieldValue("displayName")) == null)
			{
				inDesc.setField("displayName",this.getName());
			}
			if ((inDesc.getFieldValue("severity")) == null)
			{
				inDesc.setField("severity","5");
			}
		}
		if (tracing()) trace("isValid()",("Returning " + results + ": Invalid field is " + badField));
		return results;
	}


	// SUN Trace and debug functions
	private boolean tracing()
	{
		//      return true;
		return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN);
	}

	private void trace(String inClass, String inMethod, String inText)
	{
		//      System.out.println("TRACE: " + inClass + ":" + inMethod + ": " + inText);
		Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN, inClass,
				   inMethod,  Integer.toHexString(this.hashCode()) + " " + inText); 
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}

}
