/*
 * @(#)file      ModelMBeanOperationInfo.java
 * @(#)author    IBM Corp.
 * @(#)version   1.10
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */


package javax.management.modelmbean;


import javax.management.Descriptor;       
import javax.management.DescriptorAccess; 
import javax.management.*;
import java.lang.reflect.*;

import com.sun.management.jmx.Trace;         


/**
 * The ModelMBeanOperationInfo object describes a management operation of the ModelMBean.
 * It is a subclass of MBeanOperationInfo with the addition of an associated Descriptor
 * and an implementation of the DescriptorAccess interface.
 * <P>
 * <PRE>
 * The fields in the descriptor are defined, but not limited to, the following: 
 * name           : operation name 
 * descriptorType : must be "operation"    
 * class          : class where method is defined (fully qualified) 
 * role           : getter, setter, operation, constructor          
 * targetObject   : object on which to execute this method          
 * targetType     : type of object reference for targetObject: 
 *                  ObjectReference|EJBHandle|IOR|RMIReference 
 * lastReturnedValue : cached value for operation                   
 * currencyTimeLimit : how long cached value is valid               
 * lastReturnedTimeStamp : when cached value was set                
 * visibility            : 1-4 where 1: always visible 4: rarely visible 
 * presentationString :  xml formatted string to describe how to present operation 
 * </PRE>
 * The default descriptor will have name, descriptorType, and role fields set.
 */

public class ModelMBeanOperationInfo extends MBeanOperationInfo 
	 implements DescriptorAccess
{

	private Descriptor operDescriptor = createDefaultDescriptor();

	private String currClass = "ModelMBeanOperationInfo";

	/**
	 * Constructs a ModelMBeanOperationInfo object with a default descriptor.
	 *
	 * @param method The java.lang.reflect.Method object describing the MBean operation.
	 * @param description A human readable description of the operation.
	 */

	public ModelMBeanOperationInfo(String description,
								   Method operationMethod)
	{
		super(description, operationMethod);
		// put getter and setter methods in operations list 
		// create default descriptor
		if (tracing())
		{
			trace("ModelMBeanOperationInfo(String,Method)","Executed");
		}
		operDescriptor = createDefaultDescriptor();


	}

	/**
	 * Constructs a ModelMBeanOperationInfo object.
	 *
	 * @param method The java.lang.reflect.Method object describing the MBean operation.
	 * @param description A human readable description of the operation.
	 * @param descriptor An instance of Descriptor containing the appropriate metadata
	 *                   for this instance of the MBeanNotificationInfo.If it is null or invalid then
	 *			a default desriptor will be created.
	 */

	public ModelMBeanOperationInfo(String description,
								   Method operationMethod,
								   Descriptor descriptor)
	{

		super(description, operationMethod);
		// put getter and setter methods in operations list
		if (tracing())
		{
			trace("ModelMBeanOperationInfo(String,Method,Descriptor)","Executed");
		}
		if (descriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanOperationInfo()","Received null for new descriptor value, setting descriptor to default values");
			}
			operDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(descriptor))
			{
				operDescriptor = (Descriptor) descriptor.clone();
			} else
			{
				operDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	* Constructs a ModelMBeanOperationInfo object with a default descriptor.
	*
	* @param name The name of the method.
	* @param description A human readable description of the operation.
	* @param signature MBeanParameterInfo objects describing the parameters(arguments) of the method.
	* @param type The type of the method's return value.
	* @param impact The impact of the method, one of INFO, ACTION, ACTION_INFO, UNKNOWN.
	*/

	public ModelMBeanOperationInfo(String name,
								   String description,
								   MBeanParameterInfo[] signature, 
								   String type,
								   int impact)
	{

		super(name, description, signature, type, impact);
		// create default descriptor
		if (tracing())
		{
			trace("ModelMBeanOperationInfo(String,String,MBeanParameterInfo[],String,int)","Executed");
		}
		operDescriptor = createDefaultDescriptor();

	}

	/**
	* Constructs a ModelMBeanOperationInfo object.
	*
	* @param name The name of the method.
	* @param description A human readable description of the operation.
	* @param signature MBeanParameterInfo objects describing the parameters(arguments) of the method.
	* @param type The type of the method's return value.
	* @param impact The impact of the method, one of INFO, ACTION, ACTION_INFO, UNKNOWN.
	* @param descriptor An instance of Descriptor containing the appropriate metadata
	*                   for this instance of the MBeanOperationInfo.If it is null or invalid then
	*			a default desriptor will be created.
	*/

	public ModelMBeanOperationInfo(String name,
								   String description,
								   MBeanParameterInfo[] signature, 
								   String type,
								   int impact,
								   Descriptor descriptor)
	{
		super(name, description, signature, type, impact);
		if (tracing())
		{
			trace("ModelMBeanOperationInfo(String,String,MBeanParameterInfo[],String,int,Descriptor)","Executed");
		}
		if (descriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanOperationInfo()","Received null for new descriptor value, setting descriptor to default values");
			}
			operDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(descriptor))
			{
				operDescriptor = (Descriptor) descriptor.clone();
			} else
			{
				operDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}

		}
	}

	/**
	 * Constructs a new ModelMBeanOperationInfo object from this ModelMBeanOperation Object.
	 *
	 * @param inInfo the ModelMBeanOperationInfo to be duplicated
	 *
	 */

	public ModelMBeanOperationInfo(ModelMBeanOperationInfo inInfo)
	{
		super(inInfo.getName(), 
			  inInfo.getDescription(), 
			  inInfo.getSignature(),
			  inInfo.getReturnType(), 
			  inInfo.getImpact());
		if (tracing())
		{
			trace("ModelMBeanOperationInfo(ModelMBeanOperationInfo)","Executed");
		}
		Descriptor newDesc = inInfo.getDescriptor();
		if (newDesc == null)
		{
			operDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(newDesc))
			{
				operDescriptor = (Descriptor) newDesc.clone();
			} else
			{
				operDescriptor = createDefaultDescriptor();
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}

		}
	}

	/**
	* Creates and returns a new ModelMBeanOperationInfo which is a duplicate of this ModelMBeanOperationInfo.
	*
	*/

	public Object clone ()
	{
		if (tracing())
		{
			trace("ModelMBeanOperationInfo.clone()","Executed");
		}
		return(new ModelMBeanOperationInfo(this)) ;
	}

	/** 
	 * Returns a copy of the associated Descriptor of the ModelMBeanOperationInfo
	 *
	 * @return Descriptor associated with the ModelMBeanOperationInfo object.
	 */

	public Descriptor getDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanOperationInfo.getDescriptor()","Executed");
		}
		if (operDescriptor == null)
		{
			operDescriptor = createDefaultDescriptor();
		}

		return((Descriptor) operDescriptor.clone());
	}

	/**
	 * Sets associated Descriptor (full replace) for the ModelMBeanOperationInfo
	 * If the new Descriptor is null, then the associated Descriptor reverts to 
	 * a default descriptor.  The Descriptor is validated before it is assigned.  If
	 * the new Descriptor is invalid, then an IllegalArgumentException is thrown.
	 *
	 * @param Descriptor inDescriptor - replaces the Descriptor associated with the 
	 * ModelMBeanOperation.
	 *
	 */
	public void setDescriptor(Descriptor inDescriptor)
	{
		if (tracing())
		{
			trace("ModelMBeanOperationInfo.setDescriptor(Descriptor)","Executed");
		}
		if (inDescriptor == null)
		{
			if (tracing())
			{
				trace("ModelMBeanOperationInfo.setDescriptor()","Received null for new descriptor value, setting descriptor to default values");
			}
			operDescriptor = createDefaultDescriptor();
		} else
		{
			if (isValid(inDescriptor))
			{
				operDescriptor = (Descriptor) inDescriptor.clone();
			} else
			{
				throw new IllegalArgumentException("Invalid descriptor passed in parameter");
			}
		}
	}

	/**
	* Returns a string containing the entire contents of the ModelMBeanOperationInfo in human readable form.
	*/
	public String toString()
	{
		if (tracing())
		{
			trace("ModelMBeanConstructorInfo.toString()","Executed");
		}
		String retStr = new String("ModelMBeanOperationInfo: " + this.getName() +
								   " ; Description: " + this.getDescription() +
								   " ; Descriptor: " + this.getDescriptor() +
								   " ; ReturnType: " + this.getReturnType() +
								   " ; Signature: ");
		MBeanParameterInfo[] pTypes = this.getSignature();
		for (int i=0; i < pTypes.length; i++)
		{
			retStr = retStr.concat((pTypes[i]).getType() + ", ");
		}
		return retStr;
	}
	/**
	* Creates default descriptor for operation as follows:
	* descriptorType=operation,class=this.getClass,role=operation,
	* name=this.getName(),displayname=this.getName(),persistPolicy=never,visibility=1
	*/
	private Descriptor createDefaultDescriptor()
	{
		if (tracing())
		{
			trace("ModelMBeanOperationInfo.createDefaultDescriptor()","Executed");
		}
		return new DescriptorSupport(new String[] {"descriptorType=operation",
											  ("name=" + this.getName()),
											  "class=" + (this.getClass()).toString(),
											  "role=operation",
											  ("displayname=" + this.getName())});
	}


	/**
	* Tests that the descriptor is valid and adds appropriate default fields not already
	* specified. Field values must be correct for field names. 
	* Descriptor must have the same name as the operation,the descriptorType field must 
	* be "operation"
	* The following fields will be defaulted if they are not already set:
	* role=operation,displayName=this.getName(),persistPolicy=never,visibility=1
	*/
	private boolean isValid(Descriptor inDesc)
	{
		boolean results = true;
		String badField = "none";
		// if name != this.getName
		// if (descriptorType != operation)
		// look for displayName, persistPolicy, visibility and add in
		if (inDesc == null)
		{
			results = false;
		}

		else if (!inDesc.isValid())
		{	 // checks for empty descriptors, null,
			// checks for empty name and descriptorType adn valid values for fields.
			results = false;
		} else
		{
			if (! ((String)inDesc.getFieldValue("name")).equalsIgnoreCase(this.getName()))
			{
				results = false;
			}
			if (! ((String)inDesc.getFieldValue("descriptorType")).equalsIgnoreCase("operation"))
			{
				results = false;
			}
			if (inDesc.getFieldValue("role") == null)
			{
				inDesc.setField("role","operation");
			}
			if ((inDesc.getFieldValue("displayName")) == null)
			{
				inDesc.setField("displayName",this.getName());
			}
		}
		if (tracing()) trace("isValid()",("Returning " + results + ": Invalid field is " + badField));
		return results;
	}

	// SUN Trace and debug functions
	private boolean tracing()
	{
		//      return false;
		return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN);
	}

	private void trace(String inClass, String inMethod, String inText)
	{
		//      System.out.println("TRACE: " + inClass + ":" + inMethod + ": " + inText);
		Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN, inClass,
				   inMethod,  Integer.toHexString(this.hashCode()) + " " + inText); 
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}

}
