/*
 * @(#)file      RequiredModelMBean.java
 * @(#)author    IBM Corp.
 * @(#)version   1.6
 * @(#)date      00/04/11
 */
/*
 * Copyright IBM Corp. 1999-2000.  All rights reserved.
 * 
 * The program is provided "as is" without any warranty express or implied,
 * including the warranty of non-infringement and the implied warranties of
 * merchantibility and fitness for a particular purpose. IBM will not be
 * liable for any damages suffered by you or any third party claim against 
 * you regarding the Program.
 */



package javax.management.modelmbean;

/* java imports */

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.*;

import java.util.Iterator;
import java.util.Date;
import java.util.Properties;
import java.util.TreeMap;
import java.util.Map;

import java.io.File;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import javax.management.*;
import javax.management.MBeanException;
import javax.management.RuntimeOperationsException;
import javax.management.InstanceNotFoundException;

import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationListener;
import javax.management.AttributeChangeNotification;

import javax.management.Descriptor;
import javax.management.DescriptorAccess;

import com.sun.management.jmx.Trace;


/**
 * This class is the implementation of a ModelMBean. An appropriate implementation of a ModelMBean
 * must be shipped with every JMX Agent and the class must be named RequiredModelMBean.
 * <P>
 * Java resources wishing to be manageable instatiate the RequiredModelMBean using the MBeanServer's
 * createMBean method.  The resource then sets the MBeanInfo and Descriptors for the RequiredModelMBean
 * instance. The attributes and operations exposed via the ModelMBeanInfo for the ModelMBean are accessible
 * from Mbeans, connectors/adapters like other MBeans. Through the Descriptors, values and methods in
 * the managed application can be defined and mapped to attributes and operations of the ModelMBean.
 * This mapping can be defined in an XML formatted file or dynamically and
 * programmatically at runtime.
 * <P>
 * Every RequiredModelMBean which is instantiated in the MBeanServer becomes manageable:
 * its attributes and operations
 * become remotely accessible through the connectors/adaptors connected to that MBeanServer.
 * A Java object cannot be registered in the MBeanServer unless it is a JMX compliant MBean.
 * By instantiating a RequiredModelMBean, resources are guaranteed that the MBean is valid.
 *
 * MBeanException and RuntimeOperatiosException must be thrown on every public method.  This allows
 *  for wrappering exceptions from distributed communications (RMI, EJB, etc.)
 */


public class RequiredModelMBean implements ModelMBean	{

	/*************************************/
	/* attributes                        */
	/*************************************/
	ModelMBeanInfo modelMBeanInfo;

	/* Notification broadcaster for any notification to be sent from the application
	 * through the RequiredModelMBean.  */
	private NotificationBroadcasterSupport generalBroadcaster = null;

	/* Notification broadcaster for attribute change notifications */
	private NotificationBroadcasterSupport attributeBroadcaster = null;

	private Object managedResource = null;	   // handle, name, or reference for instance on which
	// the actual invoke and operations will be
	// executed
	private String managedResourceType = "objectreference";	// ObjectReference | IOR | EJBHandle | MBeanName
	private PrintStream logOut = null;
	private String currClass = "RequiredModelMBean";

	/*************************************/
	/* constructors                      */
	/*************************************/
	/**
	 * Constructs an <CODE>RequiredModelMBean</CODE> with an empty ModelMBeanInfo.
	 *
	 * After the RequiredModelMBean's MBeanInfo and Descriptors are customized, the RequiredModelMBean
	 * should be registered with the MBeanServer.
	 * <P>
	 *
	 * @exception MBeanException The constructor of the object has thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException.
	 *
	 */

	public RequiredModelMBean()
	throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("RequiredModelMBean()","Entry and Exit");
		}
		try
		{
			load();
		} catch (Exception e)
		{
		}
		modelMBeanInfo = createDefaultModelMBeanInfo();

	}

	/**
	 * Constructs a RequiredModelMBean object using ModelMBeanInfo passed in.
	 * The RequiredModelMBean must be instantiated, but not registered with the MBeanServer.
	 * After the RequiredModelMBean's MBeanInfo and Descriptors are customized, the RequiredModelMBean
	 * should be registered with the MBeanServer.
	 * <P>
	 *
	 * @param mbi The ModelMBeanInfo object to be used by the RequiredModelMBean.
	 *
	 *
	 * @exception MBeanException The constructor of the object has thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException:
	 *       The MBeanInfo passed in parameter is null or invalid.
	 *
	 */

	public RequiredModelMBean(ModelMBeanInfo mbi)
	throws MBeanException, RuntimeOperationsException
	{


		if (tracing())
		{
			trace("RequiredModelMBean(MBeanInfo)","Entry");
		}
		try
		{
			load();
		} catch (Exception e)
		{
		}

		setModelMBeanInfo(mbi);
		if (tracing())
		{
			trace("RequiredModelMBean(MBeanInfo)","Exit");
		}
	}


	/*************************************/
	/* initializers                      */
	/*************************************/


	/**
	 * Initializes a RequiredModelMBean object using ModelMBeanInfo passed in.
	 * The RequiredModelMBean should be instantiated, but not registered with the MBeanServer.
	 * After the RequiredModelMBean's MBeanInfo and Descriptors are customized, the RequiredModelMBean
	 * should be registered with the MBeanServer.
	 * <P>
	 * @param mbi The MBeanInfo object to be used by the RequiredModelMBean.
	 *
	 *
	 * @exception MBeanException The constructor of the MBeanInfo has return null
	 *           or thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException:
	 *       The MBeanInfo passed in parameter is null or invalid.
	 *
	 */
	public void setModelMBeanInfo(ModelMBeanInfo mbi)
	throws MBeanException, RuntimeOperationsException

	{
		if (tracing())
		{
			trace("setModelMBeanInfo(ModelMBeanInfo)","Entry");
		}

		if ((mbi == null) || !(isValidModelMBeanInfo(mbi)))
		{
			if (tracing())	trace("ModelMBeanInfo is null or invalid, setting default",
								  "Exception occured trying to initialize the MBeanInfo of the MBean");

			throw new RuntimeOperationsException(new IllegalArgumentException("ModelMBeanInfo must not be null or invalid, default MBeanInfo created."),
												 ("Exception occured trying to initialize the ModelMBeanInfo of the RequiredModelMBean"));

		}

		if (tracing())
		{
			trace("setModelMBeanInfo(ModelMBeanInfo)","Setting ModelMBeanInfo to " + printModelMBeanInfo(mbi));
			trace("setModelMBeanInfo(ModelMBeanInfo)", "ModelMBeanInfo notifications has " + (mbi.getNotifications()).length + " elements");
		}

		//MBeanNotificationInfo[] oldNotifies = mbi.getNotifications();
		//if (tracing()) trace("setModelMBeanInfo(ModelMBeanInfo)","retrieved ModelMBeanNotifications");
		//ModelMBeanNotificationInfo[] newNotifies = addModelMBeanAttributeChangeNotification(oldNotifies);
		modelMBeanInfo = (ModelMBeanInfo)mbi.clone();

		if (modelMBeanInfo == null)
		{
			throw new MBeanException(new IllegalArgumentException("Constructed MBeanInfo is null"),
									 "Exception occured trying to construct the MBeanInfo for the MBean");
		}
		if (tracing()) trace("setModelMBeanInfo(ModelMBeanInfo)","set mbeanInfo to: " + printModelMBeanInfo(modelMBeanInfo));

		if (tracing())
		{
			trace("setModelMBeanInfo(ModelMBeanInfo)","Exit");
		}
	}


	/**
	 * Sets the instance handle of the object against which to execute all methods in this RequiredModelMBean
	 * management interface (ModelMBeanInfo and Descriptors). This setting can be overridden by setting
	 * the 'targetObject' field of the ModelMBeanOperationInfo's descriptor.
	 *
	 * @param mr Object that is the managed resource
	 * @param mr_type The type of reference for the managed resource.  Can be: ObjectReference,
	 *               Handle, IOR, EJBHandle, RMIReference.
	 *               If the MBeanServer cannot process the mr_type passed in, an exception will be
	 *               thrown.
	 *
	 *
	 * @exception MBeanException The initializer of the object has thrown an exception.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException:
	 *       The managed resource or managed resoure type passed in parameter is null or invalid.
	 * @exception InstanceNotFoundException The managed resource object could not be found
	 * @exception InvalidTargetObjectType The managed resource type cannot be processed by the
	 * RequiredModelMBean or JMX Agent.
	 */

	public void setManagedResource(Object mr, String mr_type)
	throws MBeanException, RuntimeOperationsException,
		 InstanceNotFoundException, InvalidTargetObjectTypeException

	{
		if (tracing())
		{
			trace("setManagedResource(Object,String)","Entry");
		}

		if (mr == null)
		{
			if (tracing())
			{
				trace("setManagedResource","ManagedResource is null");
			}

			throw new RuntimeOperationsException(new IllegalArgumentException("Managed resource object must not be null"),
												 ("Exception occured trying to set the managed resource for the MBean"));

		} else
		{
			// check that mr is an instance of an object else
			// throw Invalid Target Object Type Exception
			if (!(mr instanceof java.lang.Object))
			{
				mr_type = "not Object";
				throw new InvalidTargetObjectTypeException(mr_type);
			}

			// check that the mr_type is supported by this JMXAgent
			// only "objectReference" is supported
			if ((mr_type == null) || (! mr_type.equalsIgnoreCase("objectReference")))
			{
				if (tracing())
				{
					trace("RequiredModelMBean:setManagedResource","Managed Resouce Type is not supported: " + mr_type);
				}
				throw new InvalidTargetObjectTypeException(mr_type);
			}
			if (tracing())
			{
				trace("RequiredModelMBean:setManagedResource","Managed Resouce is valid");
			}
			managedResource = mr;
		}
		if (tracing())
		{
			trace("setManagedResource(Object, String)", "Exit");
		}
	}

	/*
	 * adds the attributeChangeEvent which is issued by the RequiredModelMBean to the MBeanInfo
	 * MBeanNotificationInfo array passed into the method.
	 * @param inMBNI originating from the constructor or initializer.
	 */
	private ModelMBeanNotificationInfo[]
		 addModelMBeanAttributeChangeNotification(MBeanNotificationInfo[] inMBNI)
	{
		if (tracing()) trace("addModelMBeanAttributeChangeNotification(MBeanNotificationInfo)","entry");
		DescriptorSupport attrChangeDesc = new DescriptorSupport(new String[] {"name=jmx.ModelMBean.General",
																		  "descriptorType=notification",
																		  "severity=5",
																		  "MessageId=JMXMMBAC"});

		ModelMBeanNotificationInfo attributeChangeNotificationInfo =
			 new ModelMBeanNotificationInfo((new String[] {"jmx.attribute.change"}),
											"AttributeChange",
											"ModelMBean Attribute Change Event",
											attrChangeDesc);
		if (tracing()) trace("addModelMBeanAttributeChangeNotification","ModelMBeanAttributeChangeNotification created: " + attributeChangeNotificationInfo);


		ModelMBeanNotificationInfo [] outMBNI;

		if (tracing())
		{
			trace("addModelMBeanAttributeChangeNotification(MBeanNotificationInfo)",
				  "Adding attributeChangeNotificationInfo: name:" + attributeChangeNotificationInfo.getName() +
				  " desc:" + attributeChangeNotificationInfo.getDescription() +
				  " notifs:" + (attributeChangeNotificationInfo.getNotifTypes())[0] );
		}

		// inMBNI can be null
		if ((inMBNI == null) || (inMBNI.length == 0))
		{
			if (tracing())
			{
				trace("addModelMBeanAttributeChangeNotification(MBeanNotificationInfo)",
					  "No MBeanInfo Notifications: adding attributeChangeNotificationInfo");
			}
			outMBNI = new ModelMBeanNotificationInfo[1];
			outMBNI[0] = attributeChangeNotificationInfo;
		} else
		{
			outMBNI = new ModelMBeanNotificationInfo[inMBNI.length + 1];
			if (tracing())
			{
				trace("addModelMBeanAttributeChangeNotification(MBeanNotificationInfo)",
					  inMBNI.length + " MBeanInfo Notifications: adding attributeChangeNotificationInfo to make " + outMBNI.length + " notifications");
			}
			for (int i=0; i < inMBNI.length; i++)
			{
				if (tracing())
				{
					trace("addModelMBeanAttributeChangeNotification(MBeanNotificationInfo)",
						  "Moving NotificationInfo " + i + ": name:" + (inMBNI[i]).getName() +
						  " desc:" + (inMBNI[i]).getDescription() +
						  " notifs:" + ((inMBNI[i]).getNotifTypes())[0] );
				}

				outMBNI[i] = (ModelMBeanNotificationInfo)((ModelMBeanNotificationInfo)(inMBNI[i])).clone();
			}
			outMBNI[outMBNI.length-1] = attributeChangeNotificationInfo;
		}
		if (tracing())
		{
			trace("addModelMBeanAttributeChangeNotification(MBeanNotificationInfo)","Exit");
		}
		return outMBNI;
	}


	/*************************************/
	/* accessors and modifiers           */
	/*************************************/

	/*************************************/
	/* persistence support               */
	/*************************************/

	/**
	 * Create the MBeanInfo from a serialized copy of this object.
	 */
	private void createMBeanInfoFromFile(FileInputStream fis)
	throws java.io.IOException, MBeanException
	{
		try
		{
			// Create the object input stream
			ObjectInputStream ois = new ObjectInputStream(fis);

			// Read the serialized object
			ModelMBeanInfo mmbInfo = (ModelMBeanInfo) ois.readObject();

			// Close
			ois.close();

			// Run initialization with the new MBeanInfo object
			setModelMBeanInfo(mmbInfo);
		}

		catch (ClassNotFoundException cnfe)
		{
			// Throw MBeanException
			throw new MBeanException(cnfe, "Could not load serialized ModelMBeanInfo.");
		}
	}

	/**
	 * The load method instantiates this RequiredModelMBean instance with the data found for
	 * the RequiredModelMBean in the persistent store.  The data loaded will include a serialized ModelMBeanInfo.
	 * If there are object references in the descriptors which are not serializeable,
	 * then persistence will fail.
	 *<P>
	 * For the reference implementation, it looks for a file with the same name
	 * as the RequiredModelMBean. It checks the RequiredModelMBean MBeanDescriptor for a
	 * 'PersistLocation' field which would contain the directory where the RequiredModelMBean's
	 * whould be stored. It also looks for a 'PersistName' field which contains the file
	 * name into which the ModelBMean should be persisted.  If these fields are not provided,
	 * then persistence is not implemented.
	 *
	 * This method should be called during construction or inialization of this instance,
	 * and before the RequiredModelMBean is registered with the MBeanServer.
	 *
	 *
	 * @exception MBeanException Wraps another exception or persistence is not supported.
	 * @exception RuntimeOperationsException Wraps exceptions from the persistence mechanism.
	 * @exception InstanceNotFoundException Could not find or load this MBean from persistent
	 *                                      storage.
	*/
	public void load()
	throws MBeanException, RuntimeOperationsException, InstanceNotFoundException
	{
		if (tracing()) trace("load()","Entry");

		Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();

		// get directory
		String persistDir = (String) mmbDesc.getFieldValue("persistLocation");
		String persistName = (String) mmbDesc.getFieldValue("persistName");

		String currName = (String) mmbDesc.getFieldValue("name");
		if ((persistDir == null) || (persistName == null))
		{
			if (tracing()) trace("load()","Persistence not supported for this MBean");
			return;
		}

		// generate filename from mbean name
		String mbeanFileName = new String(persistDir + File.separator + persistName);
		try
		{
			// look for file of that name
			FileInputStream fis = new FileInputStream(mbeanFileName);
			createMBeanInfoFromFile(fis);
		} catch (FileNotFoundException fnfe)
		{
			if (tracing()) trace("load()", "Persistent MBean file was not found");
		} catch (IOException ioe)
		{
			if (tracing()) trace("load()","IO Exception loading MBean from persistent store");
		}

		if (tracing()) trace("load()","Exit");

	}

	/*
	*  analyzes descriptor fields to see if the MBean or Attribute should be
	*  persisted now
	*
	*  returns true if 'persistPolicy' field is:
	*   != "never"
	*   = "always"
	*   = "onTimer" and now > 'lastPersistTime' + 'persistPeriod'
	*   = "NoMoreOftenThan" and now > 'lastPersistTime' + 'persistPeriod'
	*
	*  otherwise it returns false when 'persistPolicy' field is:
	*    = "never"
	*    = "onUpdate"
	*    = "onTimer" && now < 'lastPersistTime' + 'persistPeriod'
	*
	*/

	private boolean persistItNow(Descriptor mmbDesc)
	{

		boolean persistItNowResponse = false;
		String MBeanPersistPolicy = null;

		if (tracing())
			trace("persistItNow(Descriptor)","Entry");

		if (mmbDesc == null)
		{
			if (tracing())
				trace("persistItNow(Descriptor)","Input Descriptor null, no persist policy");
			return persistItNowResponse;
		}

		MBeanPersistPolicy = (String) mmbDesc.getFieldValue("persistPolicy");

		if ((MBeanPersistPolicy == null) || (MBeanPersistPolicy.equalsIgnoreCase("never")))
		{
			/* don't persist */
			persistItNowResponse = false;
		} else if (MBeanPersistPolicy.equalsIgnoreCase("onUpdate") ||
				   MBeanPersistPolicy.equalsIgnoreCase("always"))
		{
			persistItNowResponse = true;
		} else if (MBeanPersistPolicy.equalsIgnoreCase("onTimer") ||
				   MBeanPersistPolicy.equalsIgnoreCase("noMoreOftenThan"))
		{
			String MBeanExpTime =  (String) mmbDesc.getFieldValue("persistPeriod");
			if ((MBeanExpTime == null) || (MBeanExpTime.length() == 0))
			{
				MBeanExpTime = "0";
			}
			int MBeanPersistPeriod = (new Integer(MBeanExpTime)).intValue();
			String MBeanTStamp =  (String) mmbDesc.getFieldValue("lastPersistedTimeStamp");
			int MBeanLastPersisted = 0;
			if (MBeanTStamp != null)
				MBeanLastPersisted = (new Integer(MBeanTStamp)).intValue();
			if ((new Date()).getTime() >= (MBeanLastPersisted + MBeanPersistPeriod))
			{
				if (MBeanPersistPolicy.equalsIgnoreCase("onTimer"))
				{
					persistItNowResponse = true;
					/* store will update the lastPersistedTimeStamp
					   of all attributeDescriptors and will reset the timer
					   mbean if there is one.  if there is not timer mbean
					   functionality, then onTimer will only be checked when
					   something is updated.  The onTimer logic in store() may
					   chose not to persist it just yet */
				} else
				{ /* its noMoreOftenThan */
					persistItNowResponse = true;
				}
			}
		}


		if (tracing())
			trace("persistItNow(Descriptor)","Exit with " + persistItNowResponse);
		return persistItNowResponse;
	}


	/*
	 * Write the ModelMBeanInfo object out to a file as a serialized object.
	 */
	private void writeMBeanInfoToFile(FileOutputStream fos)
	throws java.io.IOException, MBeanException
	{
		try
		{
			// Create the object output stream
			ObjectOutputStream oos = new ObjectOutputStream(fos);

			// Write the serialized object
			oos.writeObject(modelMBeanInfo);

			// Flush and close output stream
			oos.flush();
			oos.close();
		}

		catch (java.io.NotSerializableException nse)
		{
			// Throw MBeanException
			throw new MBeanException(nse, "Could not store ModelMBeanInfo because it contained an object that was not serializable: " + nse.toString() + ".");
		}
	}

	/**
	 * Captures the current state of this RequiredModelMBean instance and
	 * writes it out to the persistent store.  The state stored could include
	 * attribute and operation values.
	 *
	 * Persistance policy from the attribute descriptor is used to guide execution
	 * of this method.  If no persistence policy is defined in the descriptor then
	 * the persistence policy (if there is one) defined in the MBeanDescriptor is
	 * used.
	 * <PRE>
	 * Store the MBean if 'persistPolicy' field is:
	 *   != "never"
	 *   = "always"
	 *   = "onTimer" and now > 'lastPersistTime' + 'persistPeriod'
	 *   = "NoMoreOftenThan" and now > 'lastPersistTime' + 'persistPeriod'
	 *
	 * Do not store the MBean if 'persistPolicy' field is:
	 *    = "never"
	 *    = "onUpdate"
	 *    = "onTimer" && now < 'lastPersistTime' + 'persistPeriod'
	 *</PRE>  <P>
	 * The data stored will include a serialized ModelMBeanInfo.
	 * If there are object references in the descriptors which are not serializeable,
	 * then persistence will fail.
	 *<P>
	 * For the reference implementation, it looks for a file with the same name
	 * as the RequiredModelMBean. It checks the RequiredModelMBean MBeanDescriptor for a
	 * 'PersistLocation' field which would contain the directory where the RequiredModelMBean's
	 * whould be stored. It also looks for a 'PersistName' field which contains the file
	 * name into which the RequiredModelBMean should be persisted.  If these fields are not provided,
	 * then persistence is not implemented.
	 *
	 * @exception MBeanException Wraps another exception or persistence is not supported
	 * @exception RuntimeOperationsException Wraps exceptions from the persistence mechanism
	 * @exception InstanceNotFoundException Could not find/access the persistant store
	*/
	public void store()
	throws MBeanException, RuntimeOperationsException, InstanceNotFoundException
	{
		if (tracing())
		{
			trace("store()","Entry");
		}
		/* Store in file with name = mbean name in jmx directory */
		/* persist attributes:
		   persist if persistPolicy != never ||
		   persistPolicy == always ||
		   persistPolicy == onTimer && now > lastPersistTime + persistPeriod ||
		   persistPolicy == NoMoreOftenThan && now > lastPersistTime + persistPeriod
		   don't persist if persistPolicy == never ||
		   persistPolicy == onUpdate ||
		   persistPolicy = onTimer && now < lastPersistTime + persistPeriod
		*/

		/* set lastPersistTime on attribute */
		/* persist ModelMBeanInfo */
		/* persist descriptors */

		/* check to see if should be persisted */
		boolean MBeanPersistItNow = false;
		if (modelMBeanInfo == null)
		{

			throw new RuntimeOperationsException(new IllegalArgumentException("ModelMBeanInfo must not be null"),
												 ("Exception occured trying to set the store data for the RequiredModelMBean"));
		}
		Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();

		MBeanPersistItNow = persistItNow(mmbDesc);

		if (MBeanPersistItNow == true)
		{
			try
			{
				// get directory
				String persistDir =  (String) mmbDesc.getFieldValue("persistLocation");
				String persistName =  (String) mmbDesc.getFieldValue("persistName");

				String currName = (String) mmbDesc.getFieldValue("name");
				// persistName = currName.getHashValue().toString()
				if ((persistDir == null) || (persistName == null))
				{
					if (tracing()) trace("load()","Persistence not supported for this MBean");
					return;
				}

				// generate filename from mbean name
				String persistFileName = new String(persistDir + File.separator + persistName);


				// Open file
				FileOutputStream fos = new FileOutputStream(persistFileName, false);
				if (tracing()) trace("store()","Writing MBean to file " + persistFileName);

				writeMBeanInfoToFile(fos);

			} catch (Exception e)
			{
				if (tracing()) trace("store()","Exception storing MBean into file for RequiredModelMBean " + e.getClass() + ":" + e.getMessage());
				e.printStackTrace();

			}
			// Persist metadata so can recreate the ModelMBeanInfo
			// If ModelMBeanAttributeInfo then
			//      check persistPolicy
			//      if persistItNow,
			//          persist the attribute value

		}

		if (tracing())
		{
			trace("store()","Exit");
		}


	}

	/*************************************/
	/* DynamicMBean Interface            */
	/*************************************/
	/*
	 * The resolveForCacheValue method checks the descriptor passed in to see if there is a valid
	 * cached value in the descriptor.
	 * The valid value will be in the 'value' field if there is one.
	 * If the 'currencyTimeLimit' field in the descriptor is:
	 *   <0 Then the value is not cached and is never valid.  Null is returned.
	 *       The 'value' and 'lastUpdatedTimeStamp' fields are cleared.
	 *   =0 Then the value is always cached and always valid.  The 'value' field is returned.
	 *       The 'lastUpdatedTimeStamp' field is not checked.
	 *   >0 Represents the number of seconds that the 'value' field is valid.
	 *       The 'value' field is no longer valid when 'lastUpdatedTimeStamp' + 'currencyTimeLimit' > Now.
	 *       When 'value' is valid, 'valid' is returned.
	 *       When 'value' is no longer valid then null is returned and ;value' and 'lastUpdatedTimeStamp'
	 *       fields are cleared.
	 *
	 */

	private Object resolveForCacheValue(Descriptor descr)
	throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("resolveForCacheValue(Descriptor)","Entry");
		}
		Object response = null;
		boolean resetValue = false, returnCachedValue = true;
		long currencyPeriod = 0;

		if (descr == null)
		{
			if (tracing())
			{
				trace("resolveForCacheValue(Descriptor)","Input Descriptor is null");
			}
			return response;
		}
		if (tracing())
		{
			trace("resolveForCacheValue(Descriptor)","descriptor is " + descr.toString());
		}

		Descriptor mmbDescr = modelMBeanInfo.getMBeanDescriptor();
		if (mmbDescr == null)
		{
			if (tracing())
			{
				trace("resolveForCacheValue(Descriptor)","MBean Descriptor is null");
			}
			//return response;
		}

		String expTime = (String) descr.getFieldValue("currencyTimeLimit");
		if ((expTime == null) && (mmbDescr != null))
		{
			expTime = (String) mmbDescr.getFieldValue("currencyTimeLimit");
		}

		if (expTime != null)
		{
			if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," currencyTimeLimit: " + expTime);
			// convert seconds to milliseconds for time comparison
			currencyPeriod = ((new Long(expTime)).longValue()) * 1000000;
			/* if currencyTimeLimit is 0 then value is never current or cached */
			if (currencyPeriod < 0)
			{
				returnCachedValue = true;
				resetValue = false;
				if (tracing()) trace("RequiredModelMBean.resolveForCacheValue",currencyPeriod + ": always valid Cache");
				/* if currencyTimeLimit is -1 then value is always current */
			} else if (currencyPeriod == 0)
			{
				returnCachedValue = false;
				resetValue = false;
				if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," never valid Cache");
			} else
			{
				String tStamp = (String) descr.getFieldValue("lastUpdatedTimeStamp");
				if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," lastUpdatedTimeStamp: " + tStamp);
				if (tStamp == null)
				{
					tStamp = "0";
				}
				long lastTime = (new Long(tStamp)).longValue();
				if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," currencyPeriod:" + currencyPeriod + " lastUpdatedTimeStamp:" + lastTime);

				long now = (new Date()).getTime();
				if (now < (lastTime + currencyPeriod))
				{
					returnCachedValue = true;
					resetValue = false;
					if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," timed valid Cache for " + now + " < " + (lastTime + currencyPeriod));
				} else
				{ /* value is expired */
					returnCachedValue = false;
					resetValue = true;
					if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," timed expired cache for " + now + " > " + (lastTime + currencyPeriod));
				}
			}

			if (tracing()) trace("RequiredModelMBean.resolveForCacheValue","returnCachedValue:" + returnCachedValue + " resetValue: " + resetValue);
			if (returnCachedValue == true)
			{
				Object currValue = descr.getFieldValue("value");
				if (currValue != null)
				{ /* error/validity check return value here */
					response = currValue;
					/* need to cast string cached value to type */
					if (tracing()) trace("RequiredModelMBean.resolveForCacheValue","valid Cache value: " + currValue);
					
				} else
				{
					response = null;
					if (tracing()) trace("RequiredModelMBean.resolveForCacheValue"," no Cached value");
				}
			}

			if (resetValue == true)
			{ /* value is not current, so remove it */
				descr.removeField("lastUpdatedTimeStamp");
				descr.removeField("value");
				response = null;
				modelMBeanInfo.setDescriptor(descr,null);
				if (tracing()) trace("RequiredModelMBean.resolveForCacheValue","reset cached value to null");
			}
		}
		if (tracing())
		{
			trace("resolveForCache(Descriptor)","Exit");
		}
		return response;


	}

	/**
	 * Returns the attributes, operations, and notifications that this RequiredModelMBean exposes
	 * for management.
	 *
	 * @return  An instance of ModelMBeanInfo allowing retrieval all attributes, operations, and Notifications
	 * of this MBean.
	 *
	 */

	public MBeanInfo getMBeanInfo()
	{
		// throws MBeanException, RuntimeOperationsException, InstanceNotFoundException
		if (tracing())
		{
			trace("getMBeanInfo()","Entry and Exit");
		}
		if (modelMBeanInfo == null)
		{
			if (tracing()) trace("getMBeanInfo()","modelMBeanInfo is null");
			modelMBeanInfo = createDefaultModelMBeanInfo();
			//return new ModelMBeanInfo(" ", "", null, null, null, null);
		}
		if (tracing())
		{
			trace("RequiredModelMBean.getMBeanInfo()","ModelMBeanInfo is " + modelMBeanInfo.getClassName() + " for " + modelMBeanInfo.getDescription());
			trace("RequiredModelMBean.getMBeanInfo()",printModelMBeanInfo(modelMBeanInfo));
		}
		return((MBeanInfo) ((ModelMBeanInfo)modelMBeanInfo).clone());

	}

	private String printModelMBeanInfo(ModelMBeanInfo info)
	{
		String retStr = "";
		if (info == null)
		{
			if (tracing()) trace("printModelMBeanInfo(ModelMBeanInfo)",
								 "ModelMBeanInfo to print is null, printing local ModelMBeanInfo");
			info = modelMBeanInfo;
		}
		retStr = retStr.concat("\nMBeanInfo for ModelMBean is:");
		retStr = retStr.concat("\nCLASSNAME: \t"+ info.getClassName());
		retStr = retStr.concat("\nDESCRIPTION: \t"+ info.getDescription());
		try
		{
			retStr = retStr.concat("\nMBEAN DESCRIPTOR: \t"+ info.getMBeanDescriptor());
		} catch (Exception e)
		{
			retStr = retStr.concat("\nMBEAN DESCRIPTOR: \t" + " is invalid");
		}

		retStr = retStr.concat("\nATTRIBUTES");
		MBeanAttributeInfo[] attrInfo = info.getAttributes();
		if ((attrInfo != null) && (attrInfo.length>0))
		{
			for (int i=0; i<attrInfo.length; i++)
			{
				retStr = retStr.concat(" ** NAME: \t"+ attrInfo[i].getName());
				retStr = retStr.concat("    DESCR: \t"+ attrInfo[i].getDescription());
				retStr = retStr.concat("    TYPE: \t"+ attrInfo[i].getType() +
									   "\tREAD: "+ attrInfo[i].isReadable() +
									   "\tWRITE: "+ attrInfo[i].isWritable());
				retStr = retStr.concat("    DESCRIPTOR: " + (((ModelMBeanAttributeInfo)attrInfo[i]).getDescriptor()).toString());
			}
		} else retStr = retStr.concat(" ** No attributes **");
		retStr = retStr.concat("\nCONSTRUCTORS");
		MBeanConstructorInfo[] constrInfo = info.getConstructors();
		if ((constrInfo != null) && (constrInfo.length > 0 ))
		{
			for (int i=0; i<constrInfo.length; i++)
			{
				retStr = retStr.concat(" ** NAME: \t"+ constrInfo[i].getName());
				retStr = retStr.concat("    DESCR: \t"+ constrInfo[i].getDescription());
				retStr = retStr.concat("    PARAM: \t"+ constrInfo[i].getSignature().length +" parameter(s)");
				retStr = retStr.concat("    DESCRIPTOR: " + (((ModelMBeanConstructorInfo)constrInfo[i]).getDescriptor()).toString());
			}
		} else retStr = retStr.concat(" ** No Constructors **");
		retStr = retStr.concat("\nOPERATIONS");
		MBeanOperationInfo[] opInfo = info.getOperations();
		if ((opInfo != null) && (opInfo.length>0))
		{
			for (int i=0; i<opInfo.length; i++)
			{
				retStr = retStr.concat(" ** NAME: \t"+ opInfo[i].getName());
				retStr = retStr.concat("    DESCR: \t"+ opInfo[i].getDescription());
				retStr = retStr.concat("    PARAM: \t"+ opInfo[i].getSignature().length +" parameter(s)");
				retStr = retStr.concat("    DESCRIPTOR: " + (((ModelMBeanOperationInfo)opInfo[i]).getDescriptor()).toString());
			}
		} else retStr = retStr.concat(" ** No operations ** ");
		retStr = retStr.concat("\nNOTIFICATIONS");
		MBeanNotificationInfo[] notifInfo = info.getNotifications();
		if ((notifInfo != null) && (notifInfo.length>0))
		{
			for (int i=0; i<notifInfo.length; i++)
			{
				retStr = retStr.concat(" ** NAME: \t"+ notifInfo[i].getName());
				retStr = retStr.concat("    DESCR: \t"+ notifInfo[i].getDescription());
				retStr = retStr.concat("    DESCRIPTOR: " + (((ModelMBeanNotificationInfo)notifInfo[i]).getDescriptor()).toString());
			}
		} else retStr = retStr.concat(" ** No notifications **");
		retStr = retStr.concat(" ** ModelMBean: End of MBeanInfo ** ");
		return retStr;
	}
	private void echo(String txt)
	{
		trace("echo(txt)",txt);
	}


	/**
	 * Invokes a method on or through a RequiredModelMBean and returns the result of the method execution.
	 * <PRE>
	 * The last value returned by an operation may be cached in the operation's descriptor which
	 * is in the ModelMBeanOperationInfo's descriptor.
	 * The valid value will be in the 'value' field if there is one.
	 * If the 'currencyTimeLimit' field in the descriptor is:
	 *  null
	 *  <0 Then the value is not cached and is never valid.  Null is returned.
	 *      The 'value' and 'lastUpdatedTimeStamp' fields are cleared.
	 *  =0 Then the value is always cached and always valid.  The 'value' field is returned.
	 *      The 'lastUpdatedTimeStamp' field is not checked.
	 *  >0 Represents the number of seconds that the 'value' field is valid.
	 *      The 'value' field is no longer valid when 'lastUpdatedTimeStamp' + 'currencyTimeLimit' > Now.
	 *      When 'value' is valid, 'valid' is returned.
	 *      When 'value' is no longer valid then null is returned and ;value' and 'lastUpdatedTimeStamp'
	 *      fields are cleared.
	 * 
	 * Note: For this implementation: if the cached value is not a String, then the object must have
	 * a constructor which accepts a string in the same format as the objects toString() method creates.
	 * </pre>
	 *
	 * @param operationName The name of the method to be invoked. The name can be the fully qualified
	 * method name including the classname, or just the method name if the classname is defined in the 'class'
	 * field of the operation descriptor.
	 * @param params An array containing the parameters to be set when the operation is
	 * invoked
	 * @param signature An array containing the signature of the operation. The class objects will
	 * be loaded using the same class loader as the one used for loading the MBean on which the operatoion was invoked.
	 *
	 * @return  The object returned by the method, which represents the result of invoking the method on the
	 * specified managed resource.
	 *
	 * @exception MBeanException  Wraps an exception thrown by the MBean's invoked method.
	 * @exception ReflectionException  Wraps an java.lang.Exception thrown while trying to invoke the method.
	 *
	 */

	public Object invoke(String opName, Object[] opArgs, String[] sig)
	throws MBeanException, ReflectionException
	//    throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("invoke(String, Object[], String[])","Entry");
		}

		if (opName == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Method name must not be null"),
												 "An exception occured while trying to invoke a method on a RequiredModelMBean");
		}

		ModelMBeanOperationInfo opInfo = null;
		Descriptor opDescr = null;

		Object response = null;
		Method mmbOpHandle = null;
		Class[] mmbSig = null;
		Object[] mmbArgs = opArgs;
		boolean[] mmbClassFlag = null;
		String opClassName="";
		String opMethodName=opName;

		// Parse for class name and method
		int opSplitter = opName.lastIndexOf(".");
		if (opSplitter > 0)
		{
			opClassName = opName.substring(0,opSplitter);
			opMethodName = opName.substring(opSplitter+1);
		}
		opSplitter = opMethodName.indexOf("(");
		if (opSplitter > 0)
		{
			opMethodName = opMethodName.substring(0,opSplitter);
		}
		if (tracing())
		{
			trace("invoke(String, Object[], String[])","Finding operation " + opName + " as " + opMethodName);
		}
		// Create signature for method for local object invoke
		boolean mmbLocalInvoke = true;

		if ((opArgs != null) && (opArgs.length != 0))
		{
			// intercept ModelMBean ifc calls and delegate to local methods first
			mmbArgs = opArgs;
			mmbSig = new Class[opArgs.length];
			mmbClassFlag = new boolean[opArgs.length];
			for (int i=0; i < opArgs.length; i++)
			{
				try
				{

					if (tracing()) trace("invoke(String, Object[], String[])","finding class for " + sig[i]);

					// set primitives in signature
					if ((sig[i]).equals(Boolean.TYPE.toString()))
						mmbSig[i] = Boolean.TYPE;
					else if ((sig[i]).equals(Integer.TYPE.toString()))
						mmbSig[i] = Integer.TYPE;
					else if ((sig[i]).equals(Character.TYPE.toString()))
						mmbSig[i] = Character.TYPE;
					else if ((sig[i]).equals(Float.TYPE.toString()))
						mmbSig[i] = Float.TYPE;
					else if ((sig[i]).equals(Long.TYPE.toString()))
						mmbSig[i] = Long.TYPE;
					else if ((sig[i]).equals(Double.TYPE.toString()))
						mmbSig[i] = Double.TYPE;
					else if ((sig[i]).equals(Byte.TYPE.toString()))
						mmbSig[i] = Byte.TYPE;
					else if ((sig[i]).equals(Short.TYPE.toString()))
						mmbSig[i] = Short.TYPE;
					else
					{
						// use default class loader
						mmbClassFlag[i] = true;
						mmbSig[i] = Class.forName(sig[i]);
					}

					if (tracing())
					{
						String mmbArgOut;
						if (mmbArgs[i] == null)	mmbArgOut = "null";
						else mmbArgOut = mmbArgs[i].toString();
						trace("invoke(String, Object[], String[])","invoke method found a valid argument: " + mmbSig[i].toString() + " is " + mmbArgOut);
					}

				} catch (ClassNotFoundException cnfl)
				{
					if (tracing())
					{
						trace("invoke(String, Object[], String[])",opName + " method " + opMethodName + "is not in RequiredModelMBean class:" + cnfl.getMessage());
						trace("invoke(String, Object[], String[])","The parameter class could not be found in default class loader");
						// we'll try this again with the class loader of the target object
						mmbClassFlag[i] = true;
						mmbSig[i] = null;
						mmbLocalInvoke = false;
					}
				}  // for
			}

		} else
		{	// set empty args and signature
			mmbSig = new Class[0];
			mmbArgs = new Object[0];
		}

		// Get method handle
		if (mmbLocalInvoke)
		{
			try
			{
				mmbOpHandle = (this.getClass()).getMethod(opMethodName, mmbSig);
			} catch (NoSuchMethodException nsml)
			{
				mmbLocalInvoke = false;
			}
		}

		// invoke method within the RequiredModelMBean object
		if (( mmbOpHandle != null ) || mmbLocalInvoke)
		{
			try
			{	 // do local invoke
				if (tracing()) trace("invoke(String, Object[], String[])",opName + " being invoked on ModelMBean");
				response = mmbOpHandle.invoke(this, mmbArgs);
				if (tracing())
				{
					if (response == null)
						trace("invoke(String, Object[], String[])","invoke returning null response from local invoke of " + opName);
					else
						trace("invoke(String, Object[], String[])","invoke returning non-null response from local invoke of " + opName);
				}
				return response;
			} catch (RuntimeErrorException ree)
			{
				throw new RuntimeOperationsException(ree, "RuntimeException occured in RequiredModelMBean while trying to invoke operation " + opName);

			} catch (RuntimeException re)
			{
				throw new RuntimeOperationsException(re, "RuntimeException occured in RequiredModelMBean while trying to invoke operation " + opName);
			} catch (IllegalAccessException iae)
			{
				throw new ReflectionException(iae, "IllegalAccessException occured in RequiredModelMBean while trying to invoke operation " + opName);
			} catch (InvocationTargetException ite)
			{
				Throwable mmbTargEx = ite.getTargetException();
				if (mmbTargEx instanceof RuntimeException)
				{
					throw new MBeanException ((RuntimeException)mmbTargEx, "RuntimeException thrown in RequiredModelMBean while trying to invoke operation " + opName);
					//} else if ( mmbTargEx instanceof Error )
					//{
					//    throw new MBeanException ((Error)mmbTargEx, "Error thrown in RequiredModelMBean while trying to invoke operation " + opName);
				} else if (mmbTargEx instanceof ReflectionException)
				{
					throw (ReflectionException) mmbTargEx;
				} else
				{
					throw new MBeanException ((Exception)mmbTargEx, "Exception thrown in RequiredModelMBean while trying to invoke operation " + opName);
				}
			} catch (Error err)
			{
				throw new RuntimeErrorException((Error)err, "Error occured in RequiredModelMBean while trying to invoke operation " + opName);
			} catch (Exception e)
			{
				throw new ReflectionException(e, "Exception occured in RequiredModelMBean while trying to invoke operation " + opName);
			}
		}


		// do invoke on another object from descriptors and mbeaninfo data
		if (tracing()) trace("invoke(String, Object[], String[])",opName + " is not on RequiredModelMBean, looking in descriptor for " + opMethodName);
		opInfo = modelMBeanInfo.getOperation(opName); // or should it be opMethodName?
		if (opInfo == null)
		{
			throw new MBeanException(new ServiceNotFoundException("operation " + opName + " execution not supported from descriptor data"),
									 "An exception occured in RequiredModelMBean while trying to invoke an operation");
		}
		opDescr = opInfo.getDescriptor();
		if (opDescr == null)
		{	// throw...
			throw new MBeanException(new ServiceNotFoundException("operation " + opName + " execution not supported from descriptor data"),
									 "An exception occured in RequiredModelMBean while trying to invoke an operation");
		} else
		{
			/* return current cached value */
			response = resolveForCacheValue(opDescr);

			if (response == null)
			{ /* no caching to run invoke */
				if (tracing()) trace("invoke(String, Object[], String[])","No cached value returned for operation");

				// get target object
				Object currObj = (Object) opDescr.getFieldValue("targetObject");
				if (currObj != null)
				{
					if (tracing()) trace("invoke(String, Object[], String[])","Found target object in descriptor");
					String currType = (String) opDescr.getFieldValue("targetObjectType");
					if ((currType == null) || (!currType.equalsIgnoreCase("objectReference")))
					{
						throw new MBeanException(new InvalidTargetObjectTypeException(currType),
												 "An exception occured while trying to invoke an operation on a descpriptor provided target");
						//throw ModelMBean.targetObjectTypeNotSupported;
					}
					if (tracing()) trace("invoke(String, Object[], String[]","target object is a valid type");
				} else
				{

					if (managedResource != null)
					{
						if (tracing()) trace("invoke(String, Object[], String[])","managedResource for invoke found");
						currObj = managedResource;
					} else
					{
						if (tracing()) trace("invoke(String, Object[], String[])","err managedResource for invoke is null");
						//throw ModelMBean.noTargetObjectForInvoke;
					}
				}

				// recreate sig with class loader of target object

				/* Have to use MethodDescriptor of mbeaninfo and strip off last token */
				/* note: this means that the opName is the fully qualified class.method name */
				/* or it has to have a "class" field in descriptor */
				/* or the class of the target object will be used */
				Class opClass = null;
				try
				{
					if (tracing()) trace("invoke(String, Object[], String[])", "getting class for operation");
					// get target object class name
					if ((opClassName == null) || opClassName.equals(""))
					{
						opClassName = (String) opDescr.getFieldValue("class");
						if ((opClassName == null) || (opClassName.equals("")))
							opClass = currObj.getClass();
						else
							opClass	= Class.forName(opClassName);
					} else
					{
						opClass = Class.forName(opClassName);
					}
					if (opClass == null)
					{
						throw new MBeanException(new Exception("A valid class could not be found for " + opMethodName),
												 "Exception occured while trying to find class for method of invoke");
					}
				} catch (ClassNotFoundException cnf)
				{
					throw new ReflectionException(cnf, "The target object class " + opClassName + " could not be found");
				}

				ClassLoader mmbLoader = null;
				try

				{
					if (tracing()) trace("invoke(String, Object[], String[])", "setting signature with correct class loaders for operation");
					// get signature with correct objects class loader
					mmbLoader = (currObj.getClass()).getClassLoader();

					if (mmbLoader != null)
					{
						for (int i=0; i < mmbSig.length; i++)
						{
							if (mmbClassFlag[i])
							{
								mmbSig[i] = mmbLoader.loadClass(sig[i]);
							}
						}
					}
				} catch (ClassNotFoundException cnf)
				{
					throw new ReflectionException(cnf, "The parameter class could not be found");
				}


				// get method handle
				if (tracing()) trace("invoke(String, Object[], String[])","Looking for operations class " + opClassName + " method " + opMethodName);

				Method opHandle;
				try
				{

					opHandle = opClass.getMethod(opMethodName, mmbSig);
					if (opHandle == null)
					{
						throw new ReflectionException(new Exception("null method handle"),"Retrieved null method handle for method");
					}
					if (tracing()) trace("invoke(String, Object[], String[])","Retrieved valid method handle for " + opClassName + "." + opMethodName);
				} catch (NoSuchMethodException nsm)
				{
					throw new ReflectionException(nsm, "The method " + opClassName + "." + opMethodName + " could not be found");
				}

				// invoke method on target object
				try
				{
					response = opHandle.invoke((Object)currObj, mmbArgs);
					if (tracing()) trace("invoke(String, Object[], String[])","invoke done for " + opName);

					if ((response != null) && (opDescr != null))
					{
						//if (tracing()) trace("response:" + response + " opDescr:" + opDescr.toString());
						/* check for validity of response class */
						/* update cached value */

						Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();

						String ctl = (String) opDescr.getFieldValue("currencyTimeLimit");
						if ((ctl == null) && (mmbDesc != null))
						{
							ctl = (String) mmbDesc.getFieldValue("currencyTimeLimit");
						}
						if ((ctl != null) && !(ctl.equals("0")))
						{
							opDescr.setField("value", response);
							opDescr.setField("lastUpdatedTimeStamp", (new Long((new Date()).getTime())).toString());
							opInfo.setDescriptor(opDescr);

							modelMBeanInfo.setDescriptor(opDescr,"operation");
							if (tracing())
							{
								trace("invoke(String,Object[],Object[])","new descriptor is " + opDescr.toString());

								trace("invoke()","OperationInfo descriptor is " + opInfo.getDescriptor().toString());
								trace("invoke()","OperationInfo descriptor is " + modelMBeanInfo.getDescriptor(opName,"operation").toString());
							}

						}
						if (tracing()) trace("invoke(String, Object[], String[])","invoke retrieved " + response.toString());
					}
				} catch (RuntimeErrorException ree)
				{
					throw new RuntimeOperationsException(ree, "RuntimeException occured in managed resource while trying to invoke operation " + opName);

				} catch (RuntimeException re)
				{
					throw new RuntimeOperationsException(re, "RuntimeException occured in managed resource while trying to invoke operation " + opName);
				} catch (IllegalAccessException iae)
				{
					throw new ReflectionException(iae, "IllegalAccessException occured in managed resource while trying to invoke operation " + opName);
				} catch (Error err)
				{
					throw new RuntimeErrorException((Error)err, "Error occured in managed resource while trying to invoke operation " + opName);
				} catch (InvocationTargetException ite)
				{
					Throwable mmbTargEx = ite.getTargetException();
					if (mmbTargEx instanceof RuntimeException)
					{
						throw new RuntimeMBeanException ((RuntimeException)mmbTargEx, "RuntimeException thrown in managed resource while trying to invoke operation " + opName);
					} else if (mmbTargEx instanceof Error)
					{
						throw new RuntimeErrorException ((Error)mmbTargEx, "Error thrown in managed resource while trying to invoke operation " + opName);
					} else if (mmbTargEx instanceof ReflectionException)
					{
						throw (ReflectionException) mmbTargEx;
					} else
					{
						throw new MBeanException ((Exception)mmbTargEx, "Exception thrown in managed resource while trying to invoke operation " + opName);
					}
				} catch (Exception e)
				{
					throw new ReflectionException(e, "Exception occured in managed resource while trying to invoke operation " + opName);
				}

			}		// response ! cached

			/* make sure response class matches type field */
			/* I don't think we need to do this after all ..
			if (tracing()) trace("invoke checking return type");
			String respType = opInfo.getReturnType();
			String responseClass = "void";
			if ( !respType.equals("void") && !respType.equals("null") )
			{
				responseClass = (response.getClass()).getName();
			}

			if ( ! respType.equals(responseClass) )
			{
				if (tracing()) trace("invoke: response class missmatch, returning null");
				// throw exception, didn't get back right attribute type
				response = null;
			} */

		}  // opDescr

		if (tracing())
		{
			trace("invoke(String, Object[], String[])","Exit");
		}
		return response;
	}

	/**
	 * Returns the value of a specific attribute defined for this ModelMBean.
	 * <PRE>
	 * The last value returned by an attribute may be cached in the attribute's descriptor.
	 * The valid value will be in the 'value' field if there is one.
	 * If the 'currencyTimeLimit' field in the descriptor is:
	 *   null
	 *   <0 Then the value is not cached and is never valid.  Null is returned.
	 *       The 'value' and 'lastUpdatedTimeStamp' fields are cleared.
	 *   =0 Then the value is always cached and always valid.  The 'value' field is returned.
	 *       The 'lastUpdatedTimeStamp' field is not checked.
	 *   >0 Represents the number of seconds that the 'value' field is valid.
	 *       The 'value' field is no longer valid when 'lastUpdatedTimeStamp' + 'currencyTimeLimit' > Now.
	 *       When 'value' is valid, 'valid' is returned.
	 *       When 'value' is no longer valid then null is returned and ;value' and 'lastUpdatedTimeStamp'
	 *       fields are cleared.
	 *
	 *
	 * </PRE>
	 * If there is no valid cached value then the 'getMethod' field in the attributes descriptor
	 * is analyzed.  If 'getMethod' contains the name of a valid operation descriptor, then the
	 * method described by the operation descriptor is executed.  The response from the method
	 * is returned as the value of the attribute.  If the operation fails or the response value is
	 * not of the same type as the attribute, an exception will be thrown.
	 *
	 * If currencyTimeLimit is > 0, then the value of the attribute is cached in the attribute descriptor's
	 * 'value' field and the 'lastUpdatedTimeStamp' field is set to the current time stamp.
	 *
	 * @param attributeName A String specifying the name of the attribute to be
	 * retrieved. It must match the name of a ModelMBeanAttributeInfo.
	 *
	 * @return  The value of the retrieved attribute from the descriptor 'value' field or from the
	 * invokation of the operation in the 'getMethod' field of the descriptor.
	 *
	 * @exception AttributeNotFoundException The specified attribute is not accessible in the MBean.
	 * @exception MBeanException  Wraps an exception thrown by the MBean's getter.
	 * @exception ReflectionException  Wraps an java.lang.Exception thrown while trying to invoke the setter.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: The object name in parameter is null or
	 * the attribute in parameter is null.
	 */


	public Object getAttribute(String attrName)
	throws AttributeNotFoundException, MBeanException, ReflectionException
	{
		if (attrName == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("attributeName must not be null"),
												 "Exception occured trying to get attribute of a RequiredModelMBean");
		}
		if (tracing())
		{
			trace("getAttribute(String)","Entry with" + attrName);
		}
		/* Check attributeDescriptor for getMethod */
		ModelMBeanAttributeInfo attrInfo=null;

		Descriptor attrDescr=null;

		Object response = null;
		try
		{
			if (modelMBeanInfo == null)
			{
				throw new AttributeNotFoundException("getAttribute failed: ModelMBeanAttributeInfo not found for " + attrName);
			}
			attrInfo = modelMBeanInfo.getAttribute(attrName);
			Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();

			if (attrInfo == null)
			{
				throw new AttributeNotFoundException("getAttribute failed: ModelMBeanAttributeInfo not found for " + attrName);
			}
			attrDescr = attrInfo.getDescriptor();
			if (attrDescr != null)
			{
				response = resolveForCacheValue(attrDescr);
				/* return current cached value */
				if (tracing()) trace("getAttribute(String)","*** cached value is " + response);
				if (response == null)
				{ /* no cached value, run getMethod */
				    if (tracing()) trace("getAttribute(String)","**** cached value is null - getting getMethod");
					String attrGetMethod = (String)(attrDescr.getFieldValue("getMethod"));
					if (attrGetMethod != null)
					{	/* run method from operations descriptor */
						if (tracing()) trace("getAttribute(String)","invoking a getMethod for " +  attrName);
						Object getResponse = invoke(attrGetMethod,
													new Object[] {},
													new String[] {});
						if (getResponse != null)
						{ /* error/validity check return value here */
							if (tracing()) trace("getAttribute(String)","got a non-null response from getMethod\n");
							response = getResponse;
							/* change cached value in attribute descriptor */
							String ctl = (String) attrDescr.getFieldValue("currencyTimeLimit");
							if ((ctl == null) && (mmbDesc != null))
							{
								ctl = (String) mmbDesc.getFieldValue("currencyTimeLimit");
							}
							if ((ctl != null) && !(ctl.equals("0")))
							{
								if (tracing())
								{
									trace("getAttribute(String)","setting cached value and lastUpdatedTime in descriptor");
								}
								attrDescr.setField("value", response);
								attrDescr.setField("lastUpdatedTimeStamp",
												   (new Long((new Date()).getTime())).toString());
								attrInfo.setDescriptor(attrDescr);  
								modelMBeanInfo.setDescriptor(attrDescr,"attribute");
								if (tracing())
								{
									trace("getAttribute(String)","new descriptor is " + attrDescr.toString());
								
									trace("setAttribute()","local: AttributeInfo descriptor is " + attrInfo.getDescriptor().toString());
									trace("setAttribute()","modelMBeanInfo: AttributeInfo descriptor is " + modelMBeanInfo.getDescriptor(attrName,"attribute").toString());
								}
							}
						} else
						{  /* response was invalid or really returned null */
							if (tracing()) trace("getAttribute(String)","got a null response from getMethod\n");
							response = null;
						}
					} else
					{ /* not getMethod so return default value */
						if (tracing()) trace("getAttribute(String)","could not find getMethod for " + attrName + ", returning default value");

						response = attrDescr.getFieldValue("default");
						// !! cast response to right class
					}
				}
				/* make sure response class matches type field */
				String respType = attrInfo.getType();
				if (response != null)
				{
					String responseClass = response.getClass().getName();
					if (! respType.equals(responseClass))
					{
						if (tracing()) trace("getAttribute(String)","wrong response type");
						/* throw exception, didn't get back right attribute type */
						throw new MBeanException(new InvalidAttributeValueException("Wrong value type recieved for get attribute"),
												 "An exception occured while trying to get an attribute value through a RequiredModelMBean");
						//response = null;
					}
				}

			} else
			{
				if (tracing()) trace("getAttribute(String)","getMethod failed " + attrName + " not in attributeDescriptor\n");
				throw new MBeanException(new InvalidAttributeValueException("Unable to resolve attribute value, no getMethod defined in descriptor for attribute"),
										 "An exception occured while trying to get an attribute value through a RequiredModelMBean");
			}
		} catch (MBeanException mbe)
		{
			throw mbe;
		} catch (Exception e)
		{
			if (tracing())
			{
				trace("getAttribute(String)","getMethod failed with " + e.getMessage() + " exception type " + (e.getClass()).toString());
				e.printStackTrace();
			}
			throw new MBeanException(e,"An exception occured while trying to get an attribute value: " + e.getMessage());
		}
		if (tracing())
		{
			trace("getAttribute(String)","Exit");
		}
		return response;
	}

	/**
	 * Returns the values of several attributes in the ModelMBean.
	 * Executes a getAttribute for each attribute name in the attrNames array passed in.
	 *
	 * @param attributes A String array of names of the attributes to be retrieved.
	 *
	 * @return The array  of the retrieved attributes.
	 *
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: The object name in parameter is null or
	 * attributes in parameter is null.
	 *
	 */
	public AttributeList getAttributes(String[] attrNames)
	{
		if (tracing())
		{
			trace("getAttributes(String[])","Entry");
		}
		AttributeList responseList = null;
		if (attrNames == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("attributeNames must not be null"),
												 "Exception occured trying to get attributes of a RequiredModelMBean");
		}
		responseList = new AttributeList();
		for (int i = 0; i < attrNames.length; i++)
		{
			try
			{
				responseList.add(getAttribute(attrNames[i]));
			} catch (Exception e)
			{
				// eat exceptions because interface doesn't have an exception on it
				//throw new MBeanException(e,"exception for getting attribute " + attrNames[i]);
			}
		} 

		if (tracing())
		{
			trace("getAttributes(String[])","Exit");
		}
		return responseList;
	}

	/**
	 * Sets the value of a specific attribute of a named ModelMBean.
	 *
	 * If the 'setMethod' field of the attribute's descriptor
	 * contains the name of a valid operation descriptor, then the
	 * method described by the operation descriptor is executed.  The response from the method
	 * is set as the value of the attribute in the descriptor.  If the operation fails or the response value is
	 * not of the same type as the attribute, an exception will be thrown.
	 * <PRE>
	 * If currencyTimeLimit is > 0, then the new value for the attribute is cached in the attribute descriptor's
	 * 'value' field and the 'lastUpdatedTimeStamp' field is set to the current time stamp.
	 * 
	 *
	 * If the persist field of the attribute's descriptor is not null then
	 * Persistance policy from the attribute descriptor is used to guide storing the attribute in a
	 * persistenant store.
	 * Store the MBean if 'persistPolicy' field is:
	 *   != "never"
	 *   = "always"
	 *   = "onUpdate"
	 *   = "onTimer" and now > 'lastPersistTime' + 'persistPeriod'
	 *   = "NoMoreOftenThan" and now > 'lastPersistTime' + 'persistPeriod'
	 *
	 * Do not store the MBean if 'persistPolicy' field is:
	 *   = "never"
	 *   = "onTimer" && now < 'lastPersistTime' + 'persistPeriod'
	 *   = "NoMoreOftenThan" and now < 'lastPersistTime' + 'persistPeriod'
	 * </PRE>
	 * @param attribute The Attribute instance containing the name of the attribute to be set and the value it is to be set to.
	 *
	 *
	 * @exception AttributeNotFoundException The specified attribute is not accessible in the MBean.
	 * @exception InvalidAttributeValueException The specified value for the attribute is not valid.
	 * @exception MBeanException Wraps an exception thrown by the MBean's setter.
	 * @exception ReflectionException  Wraps an java.lang.Exception thrown while trying to invoke the setter.
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: The object name in parameter is null or
	 * the attribute in parameter is null.
	 */


	public void setAttribute(Attribute attribute)
	throws AttributeNotFoundException, InvalidAttributeValueException, MBeanException, ReflectionException
	{
		if (tracing())
		{
			trace("setAttribute()","Entry");
		}
		if (attribute == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("attribute must not be null"),
												 "Exception occured trying to set an attribute of a RequiredModelMBean");
		}
		/* run setMethod if there is one */
		/* return cached value if its current */
		/* set cached value in descriptor and set date/time */
		/* send attribute change Notification */
		/* check persistence policy and persist if need be */
		String attrName = attribute.getName();
		Object attrValue = attribute.getValue();

		try
		{
			ModelMBeanAttributeInfo attrInfo = modelMBeanInfo.getAttribute(attrName);
			if (attrInfo == null)
			{
				throw new AttributeNotFoundException("setAttribute failed: " + attrName + " is not found ");
			}
			Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();
			Descriptor attrDescr = attrInfo.getDescriptor();

			if (attrDescr != null)
			{
				if (!attrInfo.isWritable())
				{
					throw new AttributeNotFoundException("setAttribute failed: " + attrName + " is not writable ");
				} else
				{
					Object setResponse = null;
					String attrSetMethod = (String)(attrDescr.getFieldValue("setMethod"));
					String attrType = (String)(attrInfo.getType());
					Object currValue = this.getAttribute(attrName);
					Attribute oldAttr = new Attribute(attrName, currValue);
					if (tracing()) trace("setAttribute()","changing attribute" + attrName + " from " + currValue.toString() + " to " + attrValue);

					if (attrSetMethod != null)
					{	/* run method from operations descriptor */
						setResponse = invoke(attrSetMethod,
											 (new Object[] {attrValue}),
											 (new String[] {attrType}) );
					}
					/* change cached value */

					String ctl = (String) attrDescr.getFieldValue("currencyTimeLimit");
					if ((ctl == null) && (mmbDesc != null))
					{
						ctl = (String) mmbDesc.getFieldValue("currencyTimeLimit");
					}

					if ((ctl != null) && !(ctl.equals("0")))
					{
						if (tracing()) trace("setAttribute()","setting cached value of " + attrName + " to " + attrValue);
						attrDescr.setField("value", attrValue);
						String currtime = (new Long((new Date()).getTime())).toString();
						attrDescr.setField("lastUpdatedTimeStamp", currtime);
						attrInfo.setDescriptor(attrDescr);

						modelMBeanInfo.setDescriptor(attrDescr,"attribute");
						if (tracing())
						{
							trace("setAttribute()","new descriptor is " + attrDescr.toString());

							trace("setAttribute()","AttributeInfo descriptor is " + attrInfo.getDescriptor().toString());
                                                        trace("setAttribute()","AttributeInfo descriptor is " + modelMBeanInfo.getDescriptor(attrName,"attribute").toString());
						}

					}
					if (tracing()) trace("setAttribute()","sending sendAttributeNotification");

					sendAttributeChangeNotification(oldAttr,attribute);

					/* check to see if should be persisted */
					boolean persistMBeanNow = false;

					String persistPolicy = (String) attrDescr.getFieldValue("persistPolicy");
					if (persistPolicy != null)
					{  // if attribute policy
						boolean persistAttributeNow = persistItNow(attrDescr);
						if (persistAttributeNow == true)
						{
							//store(attrName);
							store();
						}
					}		// if attribute policy
					else
					{  // else check global mbean policy
						if (mmbDesc != null)
						{
							persistMBeanNow = persistItNow(mmbDesc);
							if (persistMBeanNow == true)
							{
								store();
							}
						}
					}  // else check global mbean policy
				} // else writeable
			} // if descriptor
			else
			{	  // else no descriptor
				if (tracing()) trace("setAttribute(String)","setMethod failed " + attrName + " not in attributeDescriptor\n");
				throw new MBeanException(new InvalidAttributeValueException("Unable to resolve attribute value, nosetMethod defined in descriptor for attribute"),
										 "An exception occured while trying to set an attribute value through a RequiredModelMBean");
			}		   // else no descriptor
		}		 // try
		catch (InstanceNotFoundException e)
		{
			if (tracing()) trace("setAttribute(String)","setMethod failed with " + e.getMessage() + "\n");
			throw new MBeanException(e,"An exception occured while trying to set an attribute value: " + e.getMessage());
		}
		if (tracing())
		{
			trace("setAttribute(Attribute)","Exit");
		}
	}

	/**
	 * Sets the values of an array of attributes of this ModelMBean.
	 * Executes the setAttribute() method for each attribute in the list.
	 *
	 * @param attributes A list of attributes: The identification of the
	 * attributes to be set and  the values they are to be set to.
	 *
	 * @return  The array of attributes that were set, with their new values in Attribute instances.
	 *
	 * @exception RuntimeOperationsException Wraps an IllegalArgumentException: The object name in parameter is null or
	 * attributes in parameter is null.
	 *
	 */


	public AttributeList setAttributes(AttributeList attributes)
	{

		if (tracing())
		{
			trace("setAttributes(AttributeList)","Entry");
		}
		AttributeList responseList = null;
		String[] attrNames = null;
		if (attributes == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("attributes must not be null"),
												 "Exception occured trying to set attributes of a RequiredModelMBean");
		}
		attrNames = new String[attributes.size()];

		int i=0;
		for (Iterator iptr = attributes.iterator(); iptr.hasNext(); i++)
		{
			try
			{
				attrNames[i] = (((Attribute)iptr).getName());
				setAttribute((Attribute) iptr.next());
			} catch (Exception e)
			{
				// eat exceptions because none allowed on this interface
				// really should rethrow as MBeanException
			}
		}
		responseList = getAttributes(attrNames);  // set returning Attribute List
		if (tracing())
		{
			trace("setdAttributes(AttributeList)","Exit");
		}
		return attributes;
	}


	private boolean isValidModelMBeanInfo(ModelMBeanInfo mbi)
	{
		if (mbi == null)
		{
			if (tracing()) trace("isValidModelMBeanInfo(mbi)","ModelMBeanInfo is invalid - null");
			return false;
		}
		return true;
	}

	private ModelMBeanInfo createDefaultModelMBeanInfo()
	{
		return(new ModelMBeanInfoSupport((this.getClass().getName()),"Default ModelMBean", null, null, null, null));
	}
	/*************************************/
	/* NotificationBroadcaster Interface */
	/*************************************/


	private void writeToLog(String logFileName, String logEntry)
	{
		if (tracing()) trace("writeToLog()","Notification Logging to " + logFileName + ": " + logEntry);
		if ((logFileName == null) || (logEntry == null))
		{
			if (tracing()) trace("writeToLog()","Bad input parameters");
			return;
		}
		if (logOut == null)
		{
			try
			{
				FileOutputStream fos = new FileOutputStream(logFileName, true);
				logOut = new PrintStream(fos);
				if (tracing()) trace("writeToLog()","Successfully opened log " + logFileName);
			} catch (java.io.FileNotFoundException fnfe)
			{
				if (tracing()) trace("writeToLog","Could not find Notification log " + logFileName);
			} catch (Exception e)
			{
				if (tracing()) trace("writeToLog", "Exception occurred for:" + e.getClass() + ": " + e.getMessage());
			}

		}
		if (logOut != null)
		{
			logOut.println(logEntry);
		}
	}
	/**
	 * Registers an object which implements the NotificationListener interface as a listener.  This
	 * object's 'handleNotification()' method will be invoked when any notification is issued through
	 * or by the ModelMBean.  This does not include attributeChangeNotifications.  They must be registered
	 * for independently.
	 *
	 * @param listener The listener object which will handles notifications emitted by the registered MBean.
	 * @param filter The filter object. If null, no filtering will be performed before handling notifications.
	 * @param handback The context to be sent to the listener with the notification when a notification is emitted.
	 *
	 * @exception IllegalArgumentException The listener could not be added.
	 *
	 */

	public void addNotificationListener(NotificationListener inlistener,
										NotificationFilter infilter,
										Object inhandback)
	throws java.lang.IllegalArgumentException
	{
		if (tracing())
		{
			trace("addNotificationListener(NotificationListener, NotificationFilter, Object)","Entry");
		}
		if ((inlistener == null))
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("notification listener must not be null"),
												 "Exception occured trying to add a notification listener to a RequiredModelMBean");
		}
		if (generalBroadcaster == null)
		{
			generalBroadcaster = new NotificationBroadcasterSupport();
		}
		generalBroadcaster.addNotificationListener(inlistener, infilter, inhandback);
		if (tracing())
		{
			trace("addNotificationListener(NotificationListener, NotificationFilter, Object)","NotificationListener added");
			trace("addNotificationListener(NotificationListener, NotificationFilter, Object)","Exit");
		}
	}

	/**
	 * Removes a listener for Notifications from the RequiredModelMBean.
	 *
	 * @param listener The listener name which was handling notifications emitted by the registered MBean.
	 * This method will remove all information related to this listener.
	 *
	 * @exception ListenerNotFoundException The couple (listener,handback) is not registered in the MBean.
	 * The exception message contains either "listener", "handback" or the object name  depending on which object cannot be found.
	 *
	 *
	 */


	public void removeNotificationListener(NotificationListener inlistener)
	throws ListenerNotFoundException
	{
		if (tracing())
		{
			trace("removeNotificationListener(NotificationListener)","Entry");
		}
		if (inlistener == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Notification listener must not be null"),
												 "Exception occured trying to remove notifications of a RequiredModelMBean");
		}
		if (generalBroadcaster == null)
		{
			throw new ListenerNotFoundException("No notification listeners registered");
		} else
		{
			generalBroadcaster.removeNotificationListener(inlistener);
		}
		if (tracing())
		{
			trace("removeNotificationListener(NotificationListener)","Exit");
		}
	}


	/**
	 * Sends a Notification which is passed in to the registered Notification listeners on the RequiredModelMBean.
	 *
	 * @param notification The notification which is to be passed to the 'handleNotification' method
	 *   of the listener object.
	 *
	 * @return void
	 *
	 * @exception MBeanException for implementation exceptions
	 * @exception RuntimeOperationsException to wrapper IllegalArugementExceptions
	 *
	 */
	public void sendNotification(Notification ntfyObj)
	throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("sendNotification(Notification)","Entry");
		}
		if (ntfyObj == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("notification object must not be null"),
												 "Exception occured trying to send a notification from a RequiredModelMBean");
		}
		if (generalBroadcaster == null)
		{
			throw new MBeanException(new ListenerNotFoundException("No notification listeners registered"),
									 "Exception ocured trying to send notification from a RequiredModelMBean");
		} else
		{
			// log notification if specified in descriptor

			Descriptor ntfyDesc = modelMBeanInfo.getDescriptor(ntfyObj.getType(),"notification");
			Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();

			if (ntfyDesc != null)
			{
				String logging = (String) ntfyDesc.getFieldValue("log");
				if (logging == null)
				{
					if (mmbDesc != null) logging = (String) mmbDesc.getFieldValue("log");
				}
				if ((logging != null) && ( logging.equalsIgnoreCase("t") || logging.equalsIgnoreCase("true")))
				{
					String logfile = (String) ntfyDesc.getFieldValue("logfile");
					if (logfile == null)
					{
						if (mmbDesc != null)  mmbDesc.getFieldValue("logfile");
					}
					if (logfile != null)
					{
						writeToLog(logfile,"LogMsg: " + ntfyObj.getTimeStamp() + " " +
								   ntfyObj.getType() + " " +
								   ntfyObj.getMessage());
					}
				}
			}
			generalBroadcaster.sendNotification(ntfyObj);
			if (tracing()) trace("sendNotification(Notification)","sendNotification sent provided notification object");
		}
		if (tracing())
		{
			trace("sendNotification(Notification)","Exit");
		}
	}

	/**
	 * Sends a Notification which contains the text string that is passed in
	 * to the registered Notification listeners on the ModelMBean.
	 *
	 * @param ntfyText The text which is to be passed in the Notification to the 'handleNotification'
	 * method of the listener object.
	 * the constructed Notification will be:
	 *   type        "jmx.modelmbean.general"
	 *   source      this ModelMBean instance
	 *   sequence    1
	 *
	 * @exception MBeanException for implementation exceptions
	 * @exception RuntimeOperationsException to wrapper IllegalArugementExceptions
	 *
	 */
	public void sendNotification(String ntfyText)
	throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("sendNotification(String)","Entry");
		}
		if (ntfyText == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("notification message must not be null"),
												 "Exception occured trying to send a text notification from a ModelMBean");
		}

		Notification myNtfyObj = new Notification("jmx.modelmbean.generic",
												  this, 1,
												  ntfyText);
		sendNotification(myNtfyObj);
		if (tracing()) trace("sendNotification(string)", "Notification sent");

		if (tracing())
		{
			trace("sendNotification(String)","Exit");
		}
	}

	/**
	 * Returns the array of Notifications always generated by the RequiredModelMBean.
	 * This will include those specified by the application plus the jmx.modelmbean.generic and
	 * jmx.attribute.change notifications.
	 *
	 * @return MBeanNotificationInfo[]
	 *     
	 */
	public MBeanNotificationInfo[] getNotificationInfo()
	{
		//    throws MBeanException, RuntimeOperationsException                                                {
		if (tracing())
		{
			trace("getNotificationInfo()","Entry");
		}
		ModelMBeanNotificationInfo[] respInfo = null;
		MBeanNotificationInfo[] currInfo = modelMBeanInfo.getNotifications();


		if ((currInfo == null)  || (currInfo.length == 0))
		{
			respInfo = new ModelMBeanNotificationInfo[2];
		} else
		{
			respInfo = new ModelMBeanNotificationInfo[currInfo.length + 2];
		}
		Descriptor genericDescriptor = new DescriptorSupport( new String[] {"name=GENERIC", 
																	   "descriptorType=notification",
																	   "log=T",
																	   "severity=5",
																	   "displayName=jmx.modelmbean.generic"} );

		respInfo[0] = new ModelMBeanNotificationInfo(new String[] {"jmx.modelmbean.generic"},
													 "GENERIC",
													 "A text notification has been issued by the managed resource",
													 genericDescriptor);

		Descriptor attributeDescriptor = new DescriptorSupport(new String[] {"name=ATTRIBUTE_CHANGE", 
																		"descriptorType=notification",
																		"log=T",
																		"severity=5",
																		"displayName=jmx.attribute.change"});
		respInfo[1] = new ModelMBeanNotificationInfo(new String[] {"jmx.attribute.change"},
													 "ATTRIBUTE_CHANGE",
													 "Signifies that an observed MBean attribute value has changed",
													 attributeDescriptor );


		int i=2, j=0;
		for (j=0; j < currInfo.length; j++)
		{
			respInfo[i] = (ModelMBeanNotificationInfo)(currInfo[j]);
		}
		if (tracing())
		{
			trace("getNotificationInfo()","Exit");
		}
		return respInfo;
	}

	/**
	 * Registers an object which implements the NotificationListener interface as a listener.  This
	 * object's 'handleNotification()' method will be invoked when any attributeChangeNotification is issued through
	 * or by the ModelMBean.  This does not include other Notifications.  They must be registered
	 * for independently. An AttributeChangeNotification will be generated for this attributeName.
	 *
	 * @param listener The listener object which will handles notifications emitted by the registered MBean.
	 * @param attributeName The name of the ModelMBean attribute for which to receive change notifications.
	 *      If null, then all attribute changes will cause an attributeChangeNotification to be issued.
	 * @param handback The context to be sent to the listener with the notification when a notification is emitted.
	 *
	 * @exception IllegalArgumentException The MBean name doesn't correspond to a registered MBean.
	 *
	 */
	public void addAttributeChangeNotificationListener(NotificationListener inlistener,
													   String inAttributeName,
													   Object inhandback)
	throws MBeanException, RuntimeOperationsException, IllegalArgumentException
	{
		if (tracing())
		{
			trace("addAttributeChangeNotificationListener(NotificationListener, String, Object)","Entry");
		}

		if (inlistener == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Listener to be registered must not be null"),
												 "Exception occured trying to register for attribute change notification of a ModelMBean");
		}
		if (attributeBroadcaster == null)
		{
			attributeBroadcaster = new NotificationBroadcasterSupport();
		}
		try
		{
			AttributeChangeNotificationFilter currFilter = new AttributeChangeNotificationFilter();
			currFilter.enableAttribute(inAttributeName);
			if (tracing()) trace("addAttributeChangeNotificationListener", "Set attribute change filter to " + ((currFilter.getEnabledAttributes()).firstElement()).toString());
			attributeBroadcaster.addNotificationListener(inlistener, currFilter, inhandback);
			if (tracing()) trace("addAttributeChangeNotificationListener","added for " + inAttributeName);
		} catch (RuntimeException e)
		{
			throw new RuntimeOperationsException(e,"addAttributeChangeNotificationListener threw " + e.getMessage());
		}
		if (tracing())
		{
			trace("addAttributeChangeNotificationListener(NotificationListener, String, Object)","Exit");
		}

	}

	/**
	 * Removes a listener for attributeChangeNotifications from the RequiredModelMBean.
	 *
	 * @param listener The listener name which was handling notifications emitted by the registered MBean.
	 * This method will remove all information related to this listener.
	 * @param attributeName The attribute for which the listener no longer wants to receive attributeChangeNotifications.
	 *
	 * @exception ListenerNotFoundException The couple (listener,handback) is not registered in the MBean.
	 * The exception message contains either "listener", "handback" or the object name  depending on which object cannot be found.
	 *
	 */
	public void removeAttributeChangeNotificationListener(NotificationListener inlistener,
														  String inAttributeName)
	throws MBeanException, RuntimeOperationsException, ListenerNotFoundException
	{
		if (tracing())
		{
			trace("removeAttributeChangeNotificationListener(NotificationListener, String)","Entry");
		}
		if ((inlistener == null) || (inAttributeName == null))
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Notification Listener and attribute name must not be null"),
												 "Exception occured trying to remove an attribute change notification listener of a ModelMBean");
		}
		if (attributeBroadcaster == null)
		{
			throw new MBeanException(new ListenerNotFoundException("No attribute change notification listeners registered"),
									 "Exception ocured trying to remove an attribute change notification of a ModelMBean");
		} else
		{
			try
			{
				/* note: */
				/* this may be a problem if the same listener is registered for multiple
				   attributes with multiple filters and/or handback objects.  It may remove
				   all of them */

				attributeBroadcaster.removeNotificationListener((NotificationListener) inlistener);
			} catch (ListenerNotFoundException e)
			{
				throw new MBeanException(e,"removeAttributeNotificationListener threw " + e.getMessage());
			}

		}
		if (tracing())
		{
			trace("removeAttributeChangeNotificationListener(NotificationListener, String)","Exit");
		}
	}

	/**
	 * Sends an attributeChangeNotification which is passed in to the registered
	 * attributeChangeNotification listeners on the ModelMBean.
	 *
	 * @param notification The notification which is to be passed to the 'handleNotification' method
	 *   of the listener object.
	 *
	 * @exception MBeanException
	 * @exception RuntimeOperationsException
	 *
	 */
	public void sendAttributeChangeNotification(AttributeChangeNotification ntfyObj)
	throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("sendAttributeChangeNotification(AttributeChangeNotification)","Entry");
		}
		if (ntfyObj == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("attribute change notification object must not be null"),
												 "Exception occured trying to send attribute change notification of a ModelMBean");
		}
		if (attributeBroadcaster == null)
		{
			throw new MBeanException(new ListenerNotFoundException("No attribute change notification listeners registered"),
									 "Exception ocured trying to send attribute change notification of a ModelMBean");
		} else
		{
			if (tracing())
			{
				trace("sendAttributeChangeNotification(AttributeChangeNotification)",
					  ("Sending AttributeChangeNotification for " + attributeBroadcaster +
					   " with " + ntfyObj.getAttributeName() +
					   ntfyObj.getAttributeType() +
					   ntfyObj.getNewValue() +
					   ntfyObj.getOldValue()));
			}
			// log notification if specified in descriptor

			Descriptor ntfyDesc = modelMBeanInfo.getDescriptor(ntfyObj.getType(),"notification");
			Descriptor mmbDesc = modelMBeanInfo.getMBeanDescriptor();

			String logging, logfile;
			if (ntfyDesc != null)
			{
				logging =(String)  ntfyDesc.getFieldValue("log");
				if (logging == null)
				{
					if (mmbDesc != null) logging = (String) mmbDesc.getFieldValue("log");
				}
				if ((logging != null) && ( logging.equalsIgnoreCase("t") || logging.equalsIgnoreCase("true")))
				{
					logfile = (String) ntfyDesc.getFieldValue("logfile");
					if (logfile == null)
					{
						if (mmbDesc != null) mmbDesc.getFieldValue("logfile");
					}
					if (logfile != null)
					{
						writeToLog(logfile,"LogMsg: " + ((new Date(ntfyObj.getTimeStamp())).toString()) + " " +
								   ntfyObj.getType() + " " +
								   ntfyObj.getMessage());
					}
				}
			} else if (mmbDesc != null)
			{
				logging = (String) mmbDesc.getFieldValue("log");
				if ((logging != null) && ( logging.equalsIgnoreCase("t") || logging.equalsIgnoreCase("true") ))
				{
					logfile = (String) mmbDesc.getFieldValue("logfile");
					if (logfile != null)
					{
						writeToLog(logfile,"LogMsg: " + ((new Date(ntfyObj.getTimeStamp())).toString()) + " " +
								   ntfyObj.getType() + " " +
								   ntfyObj.getMessage());
					}

				}
			}
			attributeBroadcaster.sendNotification(ntfyObj);
			if (tracing()) trace("sendAttributeChangeNotification(AttributeChangeNotification)","sent notification");
		}
		if (tracing())
		{
			trace("sendAttributeChangeNotification(AttributeChangeNotification)","Exit");
		}

	}

	/**
	 * Sends an attributeChangeNotification which contains the old value and new value for the
	 * attribute to the registered AttributeChangeNotification listeners on the RequiredModelMBean.
	 * <P>
	 * @param oldValue The origional value for the Attribute
	 * @param newValue The current value for the Attribute
	 *<P>
	 * <PRE>
	 * The constructed attributeChangeNotification will be:
	 *   type        "jmx.attribute.change"
	 *   source      this RequiredModelMBean instance
	 *   sequence    1
	 *   attributeName oldValue.getName()
	 *   attributeType oldValue's class
	 *   attributeOldValue oldValue.getValue()
	 *   attributeNewValue newValue.getValue()
	 * </PRE>
	 *
	 * @exception MBeanException to wrapper implementation exceptions
	 * @exception RuntimeOperationsException to wrapper IllegalArgumentExceptions.
	 *
	 */
	public void sendAttributeChangeNotification(Attribute inOldVal, Attribute inNewVal)
	throws MBeanException, RuntimeOperationsException
	{
		if (tracing())
		{
			trace("sendAttributeChangeNotification(Attribute, Attribute)","Entry");
		}
		// do we really want to do this?
		if ((inOldVal == null) || (inNewVal == null))
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("attribute change notification object must not be null"),
												 "Exception occured trying to send attribute change notification of a ModelMBean");
		}
		if (tracing())
		{
			trace("sendAttributeChangeNotification(Attribute,Attribute)",
				  ("Creating notification object with " +
				   inOldVal.getName() +
				   inOldVal.getValue() +
				   inNewVal.getName() +
				   inNewVal.getValue() +
				   ((inNewVal.getValue()).getClass()).toString()));
		}
		AttributeChangeNotification myNtfyObj =
			 new AttributeChangeNotification(this, 
											 1,
											 ((new Date()).getTime()),
											 "AttributeChangeDetected",
											 inOldVal.getName(),
											 (((inNewVal.getValue()).getClass()).toString()),
											 inOldVal.getValue(),
											 inNewVal.getValue());

		if (tracing())
		{
			trace("sendAttributeChangeNotification(Attribute,Attribute)",
				  ("Sending AttributeChangeNotification for " + attributeBroadcaster +
				   " with " + myNtfyObj.getAttributeName() +
				   myNtfyObj.getAttributeType() +
				   myNtfyObj.getNewValue() +
				   myNtfyObj.getOldValue() ));
		}

		sendAttributeChangeNotification(myNtfyObj);

		if (tracing())
		{
			trace("sendAttributeChangeNotification(Attribute, Attribute)","Exit");
		}


	}

	// SUN Trace and debug functions
	private boolean tracing()
	{
		
	    //return true;
	    return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN);
	}

	private void trace(String inClass, String inMethod, String inText)
	{
	    //System.out.println("TRACE: " + inClass + ":" + inMethod + ": " + inText);
	    Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MODELMBEAN, inClass, inMethod, Integer.toHexString(this.hashCode()) + " " + inText);
	}

	private void trace(String inMethod, String inText)
	{
		trace(currClass, inMethod, inText);
	}

	private boolean debugging()
	{
		return Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_MODELMBEAN);
	}

	private void debug(String inClass, String inMethod, String inText)
	{
		Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_MODELMBEAN, inClass, inMethod, Integer.toHexString(this.hashCode()) + " " + inText);
	}

	private void debug(String inMethod, String inText)
	{
		debug(currClass, inMethod, inText);
	}

}





























