/*
 * @(#)file      CounterMonitor.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.43
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.monitor; 



// java imports
//
import java.util.Date;

// jmx imports
//
import javax.management.MBeanNotificationInfo;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.ReflectionException;

/**
 * Defines a monitor MBean designed to observe the values of a counter attribute.
 * <P>
 * A counter monitor sends a {@link MonitorNotification#THRESHOLD_VALUE_EXCEEDED threshold notification}
 * when the value of the counter reaches or exceeds a threshold known as the comparison level. 
 * The notify flag must be set to <CODE>true</CODE>.
 * <BR>In addition, an offset mechanism enables particular counting intervals to be detected.
 * <P>
 * If the offset value is not zero, whenever the threshold is triggered by the counter value 
 * reaching a comparison level, that comparison level is incremented by the offset value. 
 * This is regarded as taking place instantaneously, that is before the count is incremented.
 * Thus, for each level, the threshold triggers an event notification every time the count increases 
 * by an interval equal to the offset value.
 * <P>
 * If the counter can wrap around its maximum value, the modulus needs to be specified.
 * The modulus is the value at which the counter is reset to zero.
 * <P>
 * If the counter difference mode is used, the value of the derived gauge is calculated 
 * as the difference between the observed counter values for two successive observations. 
 * If this difference is negative, the value of the derived gauge is incremented by the value of the modulus.
 * <BR>The derived gauge value (V[t]) is calculated using the following method:
 * <UL>
 * <LI>if (counter[t] - counter[t-GP]) is positive then V[t] = counter[t] - counter[t-GP]
 * <LI>if (counter[t] - counter[t-GP]) is negative then V[t] = counter[t] - counter[t-GP] + MODULUS
 * </UL>
 * This implementation of the counter monitor requires the observed attribute to be of the type integer
 * (<CODE>Byte</CODE>, <CODE>Integer</CODE>, <CODE>Short</CODE>, <CODE>Long</CODE>).
 *
 * @version     1.43     03/29/00
 * @author      Sun Microsystems, Inc
 */

public class CounterMonitor extends Monitor implements CounterMonitorMBean { 

 
    /*
     * ------------------------------------------
     *  PRIVATE VARIABLES
     * ------------------------------------------
     */
        
    /**
     * Counter threshold.
     * <BR>The default value is a null Integer object.
     */
    private Number threshold = new Integer(0);

    /**
     * Counter modulus.
     * <BR>The default value is a null Integer object.
     */
    private Number modulus = new Integer(0);
    
    /**
     * Counter offset.
     * <BR>The default value is a null Integer object.
     */
    private Number offset = new Integer(0);
    
    /**
     * Flag indicating if the counter monitor notifies when exceeding the threshold.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean notify = false;

    /**
     * Flag indicating if the counter difference mode is used.
     * If the counter difference mode is used, the derived gauge is the difference between two consecutive observed values.
     * Otherwise, the derived gauge is directly the value of the observed attribute.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean differenceMode = false;
    
    /**
     * Initial counter threshold.
     * This value is used to reset the threshold to its initial value each time the counter resets.
     */
    private transient Number initThreshold = new Integer(0);
    
    /**
     * Derived gauge.
     */
    private transient Number derivedGauge = new Integer(0);

    /**
     * Derived gauge timestamp.
     */
    private transient long derivedGaugeTimestamp = (new Date()).getTime();
    
    /**
     * Scan counter value captured by the previous observation.
     */
    private transient Number previousScanCounter = null;
    
    /**
     * Flag indicating if the modulus has been exceeded by the threshold.
     * This flag is used to reset the threshold once we are sure that the counter has been resetted.
     */
    private transient boolean modulusExceeded = false;

    /**
     * Derived gauge captured when the modulus has been exceeded by the threshold.
     * This value is used to check if the counter has been resetted (in order to reset the threshold).
     */
    private transient Number derivedGaugeExceeded = null;
    
    /**
     * This flag is used to notify only once between two granularity periods.
     */
    private transient boolean eventAlreadyNotified = false;
        
    /**
     * This attribute is used to keep the derived gauge type.
     */
    private transient int type = INTEGER;
    
    // Flags needed to keep trace of the derived gauge type.
    // Integer types only are allowed.
    //
    private static final int INTEGER    = 0;
    private static final int BYTE       = 1;
    private static final int SHORT      = 2;
    private static final int LONG       = 3;
    
    // New flags defining possible counter monitor errors.
    // Flag denoting that a notification has occured after changing the threshold.
    // This flag is used to check that the threshold, offset and modulus types are the same as the counter.
    //
    private static final int THRESHOLD_ERROR_NOTIFIED   = 16;
    
    /**
     * AlarmClock MBean.
     */
    private transient CounterAlarmClock alarmClock = null;

    
    // TRACES & DEBUG
    //---------------
    
    String makeDebugTag() {
        return "CounterMonitor";
    }
    

    /*
     * ------------------------------------------
     *  CONSTRUCTORS
     * ------------------------------------------
     */
    
    /**
     * Default constructor.
     */
    public CounterMonitor() {
        dbgTag = makeDebugTag();
    }
    
    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */
    
    /**
     * Allows the counter monitor MBean to perform any operations it needs before being de-registered
     * by the MBean server.
     * <P>
     * Resets the threshold values.
     *
     * @exception java.lang.Exception
     */
    public void preDeregister() throws java.lang.Exception {
        
        // Stop the CounterMonitor.
        //
        super.preDeregister();
        
        if (isTraceOn()) {
            trace("preDeregister", "reset the threshold values");
        }
        
        // Reset values for serialization.
        //
        threshold = initThreshold;
    }
    
    /**
     * Starts the counter monitor.
     */
    public synchronized void start() {
        
        if (isTraceOn()) {
            trace("start", "start the counter monitor");
        }
        
        if (!isActive()) {
            
            // Start the AlarmClock.
            //
            alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
            alarmClock.start();
       
            isActive = true;
            
            // Reset values.
            //
            threshold = initThreshold;
            modulusExceeded = false;
            eventAlreadyNotified = false;
            previousScanCounter = null;
        } else {
            if (isTraceOn()) {
                trace("start", "the counter monitor is already activated");
            }
        }
    }
    
    /**
     * Stops the counter monitor.
     */
    public synchronized void stop() {

        if (isTraceOn()) {
            trace("stop", "stop the counter monitor");
        }
        
        if (isActive()) {
            
            // Stop the AlarmClock.
            //
            if (alarmClock != null) {
                alarmClock.interrupt();
                try {
                    // Wait until the thread die.
                    //
                    alarmClock.join();
                } catch (InterruptedException e) {
                    // Ignore...
                }
                // Remove the reference on the AlarmClock.
                //
                alarmClock = null;
            }

            isActive = false;
        } else {
            if (isTraceOn()) {
                trace("stop", "the counter monitor is already deactivated");
            }
        }
    }
    
    // GETTERS AND SETTERS
    //--------------------
    
    /**
     * Gets the derived gauge.
     *
     * @return The derived gauge.
     */
    public Number getDerivedGauge() { 
        return derivedGauge;
    } 

    /**
     * Gets the derived gauge timestamp.
     *
     * @return The derived gauge timestamp.
     */
    public long getDerivedGaugeTimeStamp() { 
        return derivedGaugeTimestamp ;
    } 
    
    /**
     * Gets the threshold value.
     *
     * @return The threshold value.
     */
    public Number getThreshold() { 
        return threshold;
    } 

    /**
     * Sets the threshold value.
     *
     * @param value The threshold value.
     * @exception java.lang.IllegalArgumentException The specified threshold is null or the threshold value is less than zero.
     */
    public void setThreshold(Number value) throws java.lang.IllegalArgumentException {
        
        if (value == null) {
            throw new java.lang.IllegalArgumentException("The threshold cannot be null.");
        }
        if (value.longValue() < 0L) {
            throw new java.lang.IllegalArgumentException("The threshold must be greater than or equal to zero.");
        }
        threshold = value;
        initThreshold = value;
        alreadyNotified &= ~THRESHOLD_ERROR_NOTIFIED;
        
        // Reset values.
        //
        modulusExceeded = false;
        eventAlreadyNotified = false;
    } 

    /**
     * Gets the offset value.
     *
     * @return The offset value.
     */
    public Number getOffset() { 
        return offset;
    } 

    /**
     * Sets the offset value.
     *
     * @param value The offset value.
     * @exception java.lang.IllegalArgumentException The specified offset is null or the offset value is less than zero.
     */
    public void setOffset(Number value) throws java.lang.IllegalArgumentException {
        
        if (value == null) {
            throw new java.lang.IllegalArgumentException("The offset cannot be null.");
        }
        if (value.longValue() < 0L) {
            throw new java.lang.IllegalArgumentException("The offset must be greater than or equal to zero.");
        }
        offset = value;
        alreadyNotified &= ~THRESHOLD_ERROR_NOTIFIED;
    } 

    /**
     * Gets the modulus value.
     *
     * @return The modulus value.
     */
    public Number getModulus() { 
        return modulus;
    } 

    /**
     * Sets the modulus value.
     *
     * @param value The modulus value.
     * @exception java.lang.IllegalArgumentException The specified modulus is null or the modulus value is less than zero.
     */
    public void setModulus(Number value) throws java.lang.IllegalArgumentException { 
        
        if (value == null) {
            throw new java.lang.IllegalArgumentException("The modulus cannot be null.");
        }
        if (value.longValue() < 0L) {
            throw new java.lang.IllegalArgumentException("The modulus must be greater than or equal to zero.");
        }
        modulus = value;
        alreadyNotified &= ~THRESHOLD_ERROR_NOTIFIED;
        
        // Reset values.
        //
        modulusExceeded = false;
    } 
    
    /**
     * Gets the notification's on/off switch value.
     *
     * @return <CODE>true</CODE> if the counter monitor notifies when exceeding the threshold, <CODE>false</CODE> otherwise.
     */
    public boolean getNotify() { 
        return notify;
    } 

    /**
     * Sets the notification's on/off switch value.
     *
     * @param value The notification's on/off switch value.
     */
    public void setNotify(boolean value) {
        notify = value;
    }

    /**
     * Gets the difference mode flag value.
     *
     * @return <CODE>true</CODE> if the difference mode is used, <CODE>false</CODE> otherwise.
     */
    public boolean getDifferenceMode() { 
        return differenceMode;
    } 

    /**
     * Sets the difference mode flag value.
     *
     * @param value The difference mode flag value.
     */
    public void setDifferenceMode(boolean value) { 
        
        differenceMode = value;
        
        // Reset values.
        //
        threshold = initThreshold;
        modulusExceeded = false;
        eventAlreadyNotified = false;
        previousScanCounter = null;
    }

        
    /**    
     * Returns a <CODE>NotificationInfo</CODE> object containing the name of the Java class of the notification
     * and the notification types sent by the counter monitor.  
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types  = { MonitorNotification.RUNTIME_ERROR, 
                            MonitorNotification.OBSERVED_OBJECT_ERROR, 
                            MonitorNotification.OBSERVED_ATTRIBUTE_ERROR,
                            MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR,
                            MonitorNotification.THRESHOLD_ERROR,
                            MonitorNotification.THRESHOLD_VALUE_EXCEEDED};
        MBeanNotificationInfo[] notifsInfo = { new MBeanNotificationInfo(types, "javax.management.monitor.MonitorNotification", 
                                                                         "Notifications sent by the CounterMonitor MBean")};
        return notifsInfo;
    }  

    /*
     * ------------------------------------------
     *  PRIVATE METHODS
     * ------------------------------------------
     */
    
    /**
     * Updates the derived gauge and the derived gauge timestamp attributes.
     *
     * @param scanCounter The value of the observed attribute.
     * @return <CODE>true</CODE> if the derived gauge value is valid, <CODE>false</CODE> otherwise.
     * The derived gauge value is invalid when the differenceMode flag is set to <CODE>true</CODE> and it is
     * the first notification (so we haven't 2 consecutive values to update the derived gauge).
     */
    private boolean updateDerivedGauge(Object scanCounter) {
        
        boolean is_derived_gauge_valid;
        
        derivedGaugeTimestamp = (new Date()).getTime();
        
        // The counter difference mode is used.
        //
        if (differenceMode == true) {
            
            // The previous scan counter has been initialized.
            //
            if (previousScanCounter != null) {
                setDerivedGaugeWithDifference((Number)scanCounter, null);
                
                // If derived gauge is negative it means that the counter has wrapped around and
                // the value of the threshold needs to be reset to its initial value.
                //
                if (derivedGauge.longValue() < 0L) {
                    if (modulus.longValue() > 0L) {
                        setDerivedGaugeWithDifference((Number)scanCounter, (Number)modulus);
                    }
                    threshold = initThreshold;
                    eventAlreadyNotified = false;
                }
                is_derived_gauge_valid = true;
            }
            // The previous scan counter has not been initialized.
            // We cannot update the derived gauge...
            //
            else {
                is_derived_gauge_valid = false;
            }
            previousScanCounter = (Number)scanCounter;
        }
        // The counter difference mode is not used.
        //
        else {
            derivedGauge = (Number)scanCounter;
            is_derived_gauge_valid = true;
        }
        return is_derived_gauge_valid;
    }

    /**
     * Updates the notification attributes and notifies the listeners 
     * only once if the notify flag is set to <CODE>true</CODE>.
     */
    private void updateNotifications() {
        
        MonitorNotification notif = null;
                        
        // Send notification if notify is true.
        //
        if (eventAlreadyNotified == false) {
            if (derivedGauge.longValue() >=  threshold.longValue()) {
                if (notify == true) {
                    sendNotification(MonitorNotification.THRESHOLD_VALUE_EXCEEDED,
                                     derivedGaugeTimestamp, "", derivedGauge, threshold);
                }
                eventAlreadyNotified = true;
            }
        } else {
            if (isTraceOn()) {
                trace("updateNotifications", "the notification:" +
                      "\n\tNotification observed object = " + getObservedObject() +
                      "\n\tNotification observed attribute = " + getObservedAttribute() +
                      "\n\tNotification derived gauge = " + derivedGauge +
                      "has already been sent");
            }
        }
    }

    /**
     * Updates the threshold attributes.
     */
    private void updateThreshold() {
                
        // Calculate the new threshold value if the threshold has been exceeded and
        // if the offset value is greater than zero.
        //
        if (derivedGauge.longValue() >=  threshold.longValue()) {
            
            if (offset.longValue() > 0L) {
            
                // Increment the threshold until its value is greater
                // than the one for the current derived gauge.
                //
                long threshold_value = threshold.longValue();
                while (derivedGauge.longValue() >= threshold_value) {
                    threshold_value = threshold_value + offset.longValue();
                }
            
                // Set threshold attribute.
                //
                switch(type) {
                case INTEGER:
                    threshold = new Integer((int)threshold_value);
                    break;
                case BYTE:
                    threshold = new Byte((byte)threshold_value);
                    break;
                case SHORT:
                    threshold = new Short((short)threshold_value);
                    break;
                case LONG:
                    threshold = new Long((long)threshold_value);
                    break;            
                default:
                    // Should never occured...
                    if (isDebugOn()) {
                        debug("updateThreshold", "the threshold type is invalid");
                    }
                    break;
                }
            
                // If the counter can wrap around when it reaches its maximum
                // and we are not dealing with counter differences then we need
                // to reset the threshold to its initial value too.
                //
                if (differenceMode == false) {
                    if (modulus.longValue() > 0L) {
                        if (threshold.longValue() > modulus.longValue()) {
                            modulusExceeded = true;
                            derivedGaugeExceeded = derivedGauge;
                        }
                    }
                }	 
            
                // Threshold value has been modified so we can notify again.
                //
                eventAlreadyNotified = false;
            }
            else {
                modulusExceeded = true;
                derivedGaugeExceeded = derivedGauge;
            }
        }
    }
    
    /*    
     * ------------------------------------------
     *  PACKAGE METHODS
     * ------------------------------------------
     */
    
    /**
     * This method is called by the counter monitor each time
     * the granularity period has been exceeded.
     */
    void notifyAlarmClock() {

        Object  scan_counter = null;
        String  notif_type = null;
        
        try {
            if (isActive()) {
                
                // Check if the observed object and observed attribute are valid.
                //

                // Check that neither the observed object nor the observed attribute are null.
                // If the observed object or observed attribute is null, this means that
                // the monitor started before a complete initialization and nothing is done.
                //
                if ((getObservedObject() == null) || (getObservedAttribute() == null)) {
                    // Start the AlarmClock with a timeout equal to the granularity period value.
                    //
                    alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                    alarmClock.start();
                    return;
                }
                
                // Check that the observed object is registered in the MBean server and
                // that the observed attribute belongs to the observed object.
                //
                try {
                    scan_counter = server.getAttribute(getObservedObject(), getObservedAttribute());
                } catch (NullPointerException np_ex) {
                    if ((alreadyNotified & RUNTIME_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.RUNTIME_ERROR;
                        alreadyNotified |= RUNTIME_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The counter monitor must be registered in the MBean server.");
                    }
                } catch (InstanceNotFoundException inf_ex) {
                    if ((alreadyNotified & OBSERVED_OBJECT_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_OBJECT_ERROR;
                        alreadyNotified |= OBSERVED_OBJECT_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed object must be registered in the MBean server.");
                    }
                } catch (AttributeNotFoundException anf_ex) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed attribute must be accessible in the observed object.");
                    }
                } catch (MBeanException mb_ex) {
                    if ((alreadyNotified & RUNTIME_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.RUNTIME_ERROR;
                        alreadyNotified |= RUNTIME_ERROR_NOTIFIED;
                        throw new MonitorSettingException(mb_ex.getMessage());
                    }
                } catch (ReflectionException ref_ex) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
                        throw new MonitorSettingException(ref_ex.getMessage());
                    }
                }

                // Check that the observed attribute is of type "Integer".
                //
                if (scan_counter instanceof Integer) {
                    type = INTEGER;
                } else if (scan_counter instanceof Byte) {
                    type = BYTE;
                } else if (scan_counter instanceof Short) {
                    type = SHORT;
                } else if (scan_counter instanceof Long) {
                    type = LONG;
                } else {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed attribute type must be an integer type.");
                    }
                }

                // Check that threshold, offset and modulus are of the same type as the counter.
                //
                if (!isThresholdTypeValid()) {
                    if ((alreadyNotified & THRESHOLD_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.THRESHOLD_ERROR;
                        alreadyNotified |= THRESHOLD_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The threshold, offset and modulus must be of the same type as the counter.");
                    }
                }

                // Clear all already notified flags.
                //
                alreadyNotified = RESET_FLAGS_ALREADY_NOTIFIED;
                                
                // Check if counter has wrapped around.
                //
                if (modulusExceeded) {
                    if (derivedGauge.longValue() < derivedGaugeExceeded.longValue()) {
                        threshold = initThreshold;
                        modulusExceeded = false;
                        eventAlreadyNotified = false;
                    }
                }
                
                // Update the derived gauge attributes and check the validity of the new value.
                // The derived gauge value is invalid when the differenceMode flag is set to true and it is
                // the first notification (so we haven't 2 consecutive values to update the derived gauge).
                //
                boolean is_derived_gauge_valid = updateDerivedGauge(scan_counter);
                
                // Notify the listeners and update the threshold if the updated derived gauge value is valid.
                // 
                if (is_derived_gauge_valid) {
                    updateNotifications();
                    updateThreshold();
                }
                                
                // Start the AlarmClock with a timeout equal to the granularity period value.
                //
                alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
                alarmClock.start();
            }
        } 
        catch (MonitorSettingException ms_ex) {
            
            // Create and send the monitor error notification.
            //
            sendNotification(notif_type, derivedGaugeTimestamp,  ms_ex.getMessage(), derivedGauge, null);
            
            // Reset values.
            //
            modulusExceeded = false;
            eventAlreadyNotified = false;
            previousScanCounter = null;
            
            // Start the AlarmClock with a timeout equal to the granularity period value.
            //
            alarmClock = new CounterAlarmClock(this, getGranularityPeriod());
            alarmClock.start();
        } 
    }
        
    /**
     * Tests if the threshold, offset and modulus are of the same type as the counter.
     * Integer types only are allowed.
     *
     * Note:
     *  If the optionnal offset or modulus have not been initialized, 
     *  their default value is an Integer object with a value equal to zero.
     *
     * @return <CODE>true</CODE> if type is the same, <CODE>false</CODE> otherwise.
     */
    boolean isThresholdTypeValid() {
        
        boolean is_same_type = false;
        
        switch(type) {
        case INTEGER:
            if ((threshold instanceof Integer) && 
                ((offset.equals(new Integer(0))) || (offset instanceof Integer)) && 
                ((modulus.equals(new Integer(0))) || (modulus instanceof Integer)))
                is_same_type = true;
            break;
        case BYTE:
            if ((threshold instanceof Byte) && 
                ((offset.equals(new Integer(0))) || (offset instanceof Byte)) && 
                ((modulus.equals(new Integer(0))) || (modulus instanceof Byte)))
                is_same_type = true;
            break;
        case SHORT:
            if ((threshold instanceof Short) && 
                ((offset.equals(new Integer(0))) || (offset instanceof Short)) && 
                ((modulus.equals(new Integer(0))) || (modulus instanceof Short)))
                is_same_type = true;
            break;
        case LONG:
            if ((threshold instanceof Long) && 
                ((offset.equals(new Integer(0))) || (offset instanceof Long)) && 
                ((modulus.equals(new Integer(0))) || (modulus instanceof Long)))
                is_same_type = true;
            break;            
        default:
            // Should never occured...
            if (isDebugOn()) {
                debug("isThresholdTypeValid", "the threshold type is invalid");
            }
            break;
        }
        return is_same_type;
    }  
        
    /**
     * Sets the derived gauge when the differenceMode flag is set to <CODE>true</CODE>.
     * Integer types only are allowed.
     *
     * @param scanCounter The value of the observed attribute.
     * @param mod The counter modulus value.
     */
    void setDerivedGaugeWithDifference(Number scanCounter, Number mod) { 
        
        switch (type) {
        case INTEGER:
            if (mod == null)
                derivedGauge = new Integer(((Integer)scanCounter).intValue() - ((Integer)previousScanCounter).intValue());
            else
                derivedGauge = new Integer(((Integer)scanCounter).intValue() - ((Integer)previousScanCounter).intValue() +
                                           ((Integer)modulus).intValue());
            break;
        case BYTE:
            if (mod == null)
                derivedGauge = new Byte((byte)(((Byte)scanCounter).byteValue() - ((Byte)previousScanCounter).byteValue()));
            else
                derivedGauge = new Byte((byte)(((Byte)scanCounter).byteValue() - ((Byte)previousScanCounter).byteValue() +
                                               ((Byte)modulus).byteValue()));
            break;
        case SHORT:
            if (mod == null)
                derivedGauge = new Short((short)(((Short)scanCounter).shortValue() - ((Short)previousScanCounter).shortValue()));
            else
                derivedGauge = new Short((short)(((Short)scanCounter).shortValue() - ((Short)previousScanCounter).shortValue() +
                                                 ((Short)modulus).shortValue()));
            break;
        case LONG:
            if (mod == null)
                derivedGauge = new Long(((Long)scanCounter).longValue() - ((Long)previousScanCounter).longValue());
            else
                derivedGauge = new Long(((Long)scanCounter).longValue() - ((Long)previousScanCounter).longValue() +
                                        ((Long)modulus).longValue());
            break;
        default:
            // Should never occured...
            if (isDebugOn()) {
                debug("setDerivedGaugeWithDifference", "the threshold type is invalid");
            }
            break;
        }
    } 
    
}


/**
 * CounterAlarmClock inner class:
 * This class provides a simple implementation of an alarm clock MBean.
 * The aim of this MBean is to set up an alarm which wakes up the counter monitor every granularity period.
 */

class CounterAlarmClock extends Thread { 

    CounterMonitor listener = null;
    long granularityPeriod = 10000;
    
    /*
     * ------------------------------------------
     *  CONSTRUCTORS
     * ------------------------------------------
     */
    
    public CounterAlarmClock(CounterMonitor listener, long granularityPeriod) {
        this.listener = listener;
        this.granularityPeriod = granularityPeriod;
    }

    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */

    /**
     * This method is called by the CounterAlarmClock thread when it is started.
     */
    public void run() {
        
        try {
            this.sleep(granularityPeriod);
            listener.notifyAlarmClock();
        } catch (InterruptedException e) {
            // Ignore...
        }
    }
}
