/*
 * @(#)file      GaugeMonitor.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.42
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.monitor; 


// java imports
//
import java.util.Date;

// jmx imports
//
import javax.management.MBeanNotificationInfo;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.ReflectionException;

/**
 * Defines a monitor MBean designed to observe the values of a gauge attribute.
 * <P>
 * A gauge monitor observes an attribute that is continuously variable with time.
 * A gauge monitor sends notifications as follows:
 * <UL>
 * <LI> if the attribute value is increasing and becomes equal to or greater than the high threshold value,
 *      a {@link MonitorNotification#THRESHOLD_HIGH_VALUE_EXCEEDED threshold high notification} is sent. 
 *      The notify high flag must be set to <CODE>true</CODE>
 *      <BR>Subsequent crossings of the high threshold value do not cause further notifications unless 
 *      the attribute value becomes equal to or less than the low threshold value.
 * <LI> if the attribute value is decreasing and becomes equal to or less than the low threshold value,
 *      a {@link MonitorNotification#THRESHOLD_LOW_VALUE_EXCEEDED threshold low notification} is sent.
 *      The notify low flag must be set to <CODE>true</CODE>.
 *      <BR>Subsequent crossings of the low threshold value do not cause further notifications unless 
 *      the attribute value becomes equal to or greater than the high threshold value.
 * </UL>
 * This provides a hysteresis mechanism to avoid repeated triggering of notifications when 
 * the attribute value makes small oscillations around the high or low threshold value.
 * <P>
 * If the gauge difference mode is used, the value of the derived gauge is calculated
 * as the difference between the observed gauge values for two successive observations.
 * <BR>The derived gauge value (V[t]) is calculated using the following method:
 * <UL>
 * <LI>V[t] = gauge[t] - gauge[t-GP]
 * </UL>
 * This implementation of the gauge monitor requires the observed attribute to be of the type integer or floating-point
 * (<CODE>Byte</CODE>, <CODE>Integer</CODE>, <CODE>Short</CODE>, <CODE>Long</CODE>, <CODE>Float</CODE>, <CODE>Double</CODE>).
 *
 * @version     1.42     03/29/00
 * @author      Sun Microsystems, Inc
 */

public class GaugeMonitor extends Monitor implements GaugeMonitorMBean { 

    
    /*
     * ------------------------------------------
     *  PRIVATE VARIABLES
     * ------------------------------------------
     */
    
    /**
     * Gauge high threshold.
     * <BR>The default value is a null Integer object.
     */
    private Number highThreshold = new Integer(0);
    
    /**
     * Gauge low threshold.
     * <BR>The default value is a null Integer object.
     */
    private Number lowThreshold = new Integer(0);
    
    /**
     * Flag indicating if the gauge monitor notifies when exceeding the high threshold.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean notifyHigh = false;
        
    /**
     * Flag indicating if the gauge monitor notifies when exceeding the low threshold.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean notifyLow = false;

    /**
     * Flag indicating if the gauge difference mode is used.
     * If the gauge difference mode is used, the derived gauge is the difference between two consecutive observed values.
     * Otherwise, the derived gauge is directly the value of the observed attribute.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean differenceMode = false;
    
    /**
     * Derived gauge.
     */
    private transient Number derivedGauge = new Integer(0);

    /**
     * Derived gauge timestamp.
     */
    private transient long derivedGaugeTimestamp = (new Date()).getTime();
        
    /**
     * Scan gauge value captured by the previous observation.
     */
    private transient Number previousScanGauge = null;

    /**
     * This attribute is used to handle the hysteresis mechanism.
     */
    private transient int status = RISING_OR_FALLING;

    /**
     * This attribute is used to keep the derived gauge type.
     */
    private transient int type = INTEGER;
    
    // Flags needed to implement the hysteresis mechanism.
    //
    private static final int RISING             = 0;
    private static final int FALLING            = 1;
    private static final int RISING_OR_FALLING  = 2;

    // Flags needed to keep trace of the derived gauge type.
    // Integer + floating-point types are allowed.
    //
    private static final int INTEGER    = 0;
    private static final int BYTE       = 1;
    private static final int SHORT      = 2;
    private static final int LONG       = 3;
    private static final int FLOAT      = 4;
    private static final int DOUBLE     = 5;

    // New flags defining possible gauge monitor errors.
    // Flag denoting that a notification has occured after changing the high/low threshold.
    // This flag is used to check that the high/low threshold type is the same as the gauge and that the 
    // threshold high value is greater than the threshold low value at the first notification time.
    //
    private static final int THRESHOLD_ERROR_NOTIFIED   = 16;

    /**
     * AlarmClock MBean.
     */
    private transient GaugeAlarmClock alarmClock = null;



    // TRACES & DEBUG
    //---------------
    
    String makeDebugTag() {
        return "GaugeMonitor";
    }

    /*
     * ------------------------------------------
     *  CONSTRUCTORS
     * ------------------------------------------
     */
    
    /**
     * Default constructor.
     */
    public GaugeMonitor() {
        dbgTag = makeDebugTag();
    }
    
    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */
            
    /**
     * Starts the gauge monitor.
     */
    public synchronized void start() {
        
        if (isTraceOn()) {
            trace("start", "start the gauge monitor");
        }
        
        if (!isActive()) {
                            
            // Start the AlarmClock.
            //
            alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
            alarmClock.start();
       
            isActive = true;
            
            // Reset values.
            //
            status = RISING_OR_FALLING;
            previousScanGauge = null;
        } else {
            if (isTraceOn()) {
                trace("start", "the gauge monitor is already activated");
            }
        }
    }
  
    /**
     * Stops the gauge monitor.
     */
    public synchronized void stop() {

        if (isTraceOn()) {
            trace("stop", "stop the gauge monitor");
        }
        
        if (isActive()) {
            
            // Stop the AlarmClock.
            //
            if (alarmClock != null) {
                alarmClock.interrupt();
                try {
                    // Wait until the thread die.
                    //
                    alarmClock.join();
                } catch (InterruptedException e) {
                    // Ignore...
                }
                // Remove the reference on the AlarmClock.
                //
                alarmClock = null;
            }

            isActive = false;
        } else {
            if (isTraceOn()) {
                trace("stop", "the counter monitor is already deactivated");
            }
        }
    }
    
    // GETTERS AND SETTERS
    //--------------------
    
    /**
     * Gets the derived gauge.
     *
     * @return The derived gauge.
     */
    public Number getDerivedGauge() { 
        return derivedGauge;
    } 

    /**
     * Gets the derived gauge timestamp.
     *
     * @return The derived gauge timestamp.
     */
    public long getDerivedGaugeTimeStamp() { 
        return derivedGaugeTimestamp;
    } 
    
    /**
     * Gets the high threshold value.
     *
     * @return The high threshold value.
     */
    public Number getHighThreshold() { 
        return highThreshold;
    } 

    /**
     * Gets the low threshold value.
     *
     * @return The low threshold value.
     */
    public Number getLowThreshold() { 
        return lowThreshold;
    } 

    /**
     * Sets the high and the low threshold values.
     *
     * @param highValue The high threshold value.
     * @param lowValue The low threshold value.
     * @exception java.lang.IllegalArgumentException The specified high/low threshold is null
     * or the low threshold is greater than the high threshold
     * or the high threshold and the low threshold are not of the same type.
     */
    public void setThresholds(Number highValue, Number lowValue) throws java.lang.IllegalArgumentException {
        
        if ((highValue == null) || (lowValue == null)) {
            throw new java.lang.IllegalArgumentException("The threshold values cannot be null.");
        }
        if (highValue.getClass() != lowValue.getClass()) {
            throw new java.lang.IllegalArgumentException("The high and the low thresholds must be of the same type.");
        }
        if (isFirstStrictlyGreaterThanLast(lowValue, highValue)) {
            throw new java.lang.IllegalArgumentException("The threshold high value must be greater than or equal to threshold low value.");
        }
        highThreshold = highValue;
        lowThreshold = lowValue;
        alreadyNotified &= ~THRESHOLD_ERROR_NOTIFIED;
    
        // Reset values.
        //
        status = RISING_OR_FALLING;
    }
    
    /**
     * Gets the high notification's on/off switch value.
     *
     * @return <CODE>true</CODE> if the gauge monitor notifies when exceeding the high threshold, <CODE>false</CODE> otherwise.
     */
    public boolean getNotifyHigh() { 
        return notifyHigh;
    } 

    /**
     * Sets the high notification's on/off switch value.
     *
     * @param value The high notification's on/off switch value.
     */
    public void setNotifyHigh(boolean value) {
        notifyHigh = value;
    }

    /**
     * Gets the low notification's on/off switch value.
     *
     * @return <CODE>true</CODE> if the gauge monitor notifies when exceeding the low threshold, <CODE>false</CODE> otherwise.
     */
    public boolean getNotifyLow() { 
        return notifyLow;
    } 

    /**
     * Sets the low notification's on/off switch value.
     *
     * @param value The low notification's on/off switch value.
     */
    public void setNotifyLow(boolean value) {
        notifyLow = value;
    }

    /**
     * Gets the difference mode flag value.
     *
     * @return <CODE>true</CODE> if the difference mode is used, <CODE>false</CODE> otherwise.
     */
    public boolean getDifferenceMode() { 
        return differenceMode;
    } 

    /**
     * Sets the difference mode flag value.
     *
     * @param value The difference mode flag value.
     */
    public void setDifferenceMode(boolean value) {
        
        differenceMode = value;
        
        // Reset values.
        //
        status = RISING_OR_FALLING;
        previousScanGauge = null;
    }

   /**    
     * Returns a <CODE>NotificationInfo</CODE> object containing the name of the Java class of the notification
     * and the notification types sent by the gauge monitor.  
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types  = { MonitorNotification.RUNTIME_ERROR, 
                            MonitorNotification.OBSERVED_OBJECT_ERROR, 
                            MonitorNotification.OBSERVED_ATTRIBUTE_ERROR,
                            MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR,
                            MonitorNotification.THRESHOLD_ERROR,
                            MonitorNotification.THRESHOLD_HIGH_VALUE_EXCEEDED,
                            MonitorNotification.THRESHOLD_LOW_VALUE_EXCEEDED};
        MBeanNotificationInfo[] notifsInfo = { new MBeanNotificationInfo(types, "javax.management.monitor.MonitorNotification", 
                                                                         "Notifications sent by the GaugeMonitor MBean")};
        return notifsInfo;
    }  

    /*
     * ------------------------------------------
     *  PRIVATE METHODS
     * ------------------------------------------
     */
    
    /**
     * Updates the derived gauge and the derived gauge timestamp attributes.
     *
     * @param scanGauge The value of the observed attribute.
     * @return <CODE>true</CODE> if the derived gauge value is valid, <CODE>false</CODE> otherwise.
     * The derived gauge value is invalid when the differenceMode flag is set to <CODE>true</CODE> and it is
     * the first notification (so we haven't 2 consecutive values to update the derived gauge).
     */
    private boolean updateDerivedGauge(Object scanGauge) {
        
        boolean is_derived_gauge_valid;
        
        derivedGaugeTimestamp = (new Date()).getTime();
        
        // The gauge difference mode is used.
        //
        if (differenceMode == true) {
            
            // The previous scan gauge has been initialized.
            //
            if (previousScanGauge != null) {
                setDerivedGaugeWithDifference((Number)scanGauge);
                is_derived_gauge_valid = true;
            } 
            // The previous scan gauge has not been initialized.
            // We cannot update the derived gauge...
            //
            else {
                is_derived_gauge_valid = false;
            }
            previousScanGauge = (Number)scanGauge;
        } 
        // The gauge difference mode is not used.
        //
        else {
            derivedGauge = (Number)scanGauge;
            is_derived_gauge_valid = true;
        }
        return is_derived_gauge_valid;
    }

    /**
     * Updates the notification attributes and notifies the listeners 
     * only once if the notifyHigh/notifyLow flag is set to <CODE>true</CODE>.
     */
    private void updateNotifications() {
        
        MonitorNotification notif = null;
                        
        // Send high notification if notifyHigh is true.
        // Send low notification if notifyLow is true.
        //
        if (status == RISING_OR_FALLING) {
            if (isFirstGreaterThanLast(derivedGauge, highThreshold)) {
                if (notifyHigh == true) {
                    sendNotification(MonitorNotification.THRESHOLD_HIGH_VALUE_EXCEEDED,
                                     derivedGaugeTimestamp, "", derivedGauge, highThreshold);
                }
                status = FALLING;
            }
            if (isFirstGreaterThanLast(lowThreshold, derivedGauge)) {
                if (notifyLow == true) {
                    sendNotification(MonitorNotification.THRESHOLD_LOW_VALUE_EXCEEDED,
                                     derivedGaugeTimestamp, "", derivedGauge, lowThreshold);
                }
                status = RISING;
            }
        } 
        else {
            if (status == RISING) {
                if (isFirstGreaterThanLast(derivedGauge, highThreshold)) {
                    if (notifyHigh == true) {
                        sendNotification(MonitorNotification.THRESHOLD_HIGH_VALUE_EXCEEDED,
                                         derivedGaugeTimestamp, "", derivedGauge, highThreshold);
                    }
                    status = FALLING;
                }
            } 
            else if (status == FALLING) {
                if (isFirstGreaterThanLast(lowThreshold, derivedGauge)) {
                    if (notifyLow == true) {
                        sendNotification(MonitorNotification.THRESHOLD_LOW_VALUE_EXCEEDED,
                                         derivedGaugeTimestamp, "", derivedGauge, lowThreshold);
                    }
                    status = RISING;
                }
            }
        }
    }
        
    /*    
     * ------------------------------------------
     *  PACKAGE METHODS
     * ------------------------------------------
     */
    
    /**
     * This method is called by the gauge monitor each time
     * the granularity period has been exceeded.
     */
    void notifyAlarmClock() {

        Object  scan_gauge = null;
        String  notif_type = null;
        
        try {
            if (isActive()) {
                
                // Check if the observed object and observed attribute are valid.
                //
                
                // Check that neither the observed object nor the observed attribute are null.
                // If the observed object or observed attribute is null, this means that
                // the monitor started before a complete initialization and nothing is done.
                //
                if ((getObservedObject() == null) || (getObservedAttribute() == null)) {
                    // Start the AlarmClock with a timeout equal to the granularity period value.
                    //
                    alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                    alarmClock.start();
                    return;
                }
                
                // Check that the observed object is registered in the MBean server and
                // that the observed attribute belongs to the observed object.
                //
                try {
                    scan_gauge = server.getAttribute(getObservedObject(), getObservedAttribute());
                } catch (NullPointerException np_ex) {
                    if ((alreadyNotified & RUNTIME_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.RUNTIME_ERROR;
                        alreadyNotified |= RUNTIME_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The gauge monitor must be registered in the MBean server.");
                    }
                } catch (InstanceNotFoundException inf_ex) {
                    if ((alreadyNotified & OBSERVED_OBJECT_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_OBJECT_ERROR;
                        alreadyNotified |= OBSERVED_OBJECT_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed object must be registered in the MBean server.");
                    }
                } catch (AttributeNotFoundException anf_ex) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed attribute must be accessible in the observed object.");
                    }
                } catch (MBeanException mb_ex) {
                    if ((alreadyNotified & RUNTIME_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.RUNTIME_ERROR;
                        alreadyNotified |= RUNTIME_ERROR_NOTIFIED;
                        throw new MonitorSettingException(mb_ex.getMessage());
                    }
                } catch (ReflectionException ref_ex) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
                        throw new MonitorSettingException(ref_ex.getMessage());
                    }
                }

                // Check that the observed attribute is either of type "Integer" or "Float".
                //
                if (scan_gauge instanceof Integer) {
                    type = INTEGER;
                } else if (scan_gauge instanceof Byte) {
                    type = BYTE;
                } else if (scan_gauge instanceof Short) {
                    type = SHORT;
                } else if (scan_gauge instanceof Long) {
                    type = LONG;
                } else if (scan_gauge instanceof Float) {
                    type = FLOAT;
                } else if (scan_gauge instanceof Double) {
                    type = DOUBLE;
                } else {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed attribute type must be an integer type or a floating-point type.");
                    }
                }

                // Check that threshold high and threshold low are of the same type as the gauge.
                //
                if (!isThresholdTypeValid()) {
                    if ((alreadyNotified & THRESHOLD_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.THRESHOLD_ERROR;
                        alreadyNotified |= THRESHOLD_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The threshold high and threshold low must be of the same type as the gauge.");
                    }
                }

                // Clear all already notified flags.
                //
                alreadyNotified = RESET_FLAGS_ALREADY_NOTIFIED;
                
                // Update the derived gauge attributes and check the validity of the new value.
                // The derived gauge value is invalid when the differenceMode flag is set to true and it is
                // the first notification (so we haven't 2 consecutive values to update the derived gauge).
                //
                boolean is_derived_gauge_valid = updateDerivedGauge(scan_gauge);
                
                // Notify the listeners if the updated derived gauge value is valid.
                //
                if (is_derived_gauge_valid) {
                    updateNotifications();
                }

                // Start the AlarmClock with a timeout equal to the granularity period value.
                //
                alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
                alarmClock.start();
            }
        }
        catch (MonitorSettingException ms_ex) {
            
            // Create and send the monitor error notification.
            //
            sendNotification(notif_type, derivedGaugeTimestamp, ms_ex.getMessage(), derivedGauge, null);
            
            // Reset values.
            //
            status = RISING_OR_FALLING;
            previousScanGauge = null;
            
            // Start the AlarmClock with a timeout equal to the granularity period value.
            //
            alarmClock = new GaugeAlarmClock(this, getGranularityPeriod());
            alarmClock.start();
        }
    }
    
    /**
     * Tests if the first specified Number is greater than or equal to the last.
     * Both integer and floating-point types are allowed.
     *
     * @param greater The first Number to compare with the second.
     * @param less The second Number to compare with the first.
     * @return <CODE>true</CODE> if the first specified Number is greater than or equal to the last, <CODE>false</CODE> otherwise.
     */
    boolean isFirstGreaterThanLast(Number greater, Number less) {
        
        boolean is_greater = false;
                
        switch(type) {
        case INTEGER:
        case BYTE:
        case SHORT:
        case LONG:
            if (greater.longValue() >= less.longValue())
                is_greater = true;
            break;
        case FLOAT:
        case DOUBLE:
            if (greater.doubleValue() >= less.doubleValue())
                is_greater = true;
            break;
        default:
            // Should never occured...
            if (isDebugOn()) {
                debug("isFirstGreaterThanLast", "the threshold type is invalid");
            }
            break;
        }
        return is_greater;
    }  
    
    /**
     * Tests if the first specified Number is strictly greater than the last.
     * Both integer and floating-point types are allowed.
     *
     * @param greater The first Number to compare with the second.
     * @param less The second Number to compare with the first.
     * @return <CODE>true</CODE> if the first specified Number is strictly greater than the last, <CODE>false</CODE> otherwise.
     */
    boolean isFirstStrictlyGreaterThanLast(Number greater, Number less) {
        
        boolean is_greater = false;
                
        switch(type) {
        case INTEGER:
        case BYTE:
        case SHORT:
        case LONG:
            if (greater.longValue() > less.longValue())
                is_greater = true;
            break;
        case FLOAT:
        case DOUBLE:
            if (greater.doubleValue() > less.doubleValue())
                is_greater = true;
            break;
        default:
            // Should never occured...
            if (isDebugOn()) {
                debug("isFirstStrictlyGreaterThanLast", "the threshold type is invalid");
            }
            break;
        }
        return is_greater;
    }  
    
    /**
     * Tests if the threshold high and threshold low are both of the same type as the gauge.
     * Both integer and floating-point types are allowed.
     *
     * @return <CODE>true</CODE> if type is the same, <CODE>false</CODE> otherwise.
     */
    boolean isThresholdTypeValid() {
        
        boolean is_same_type = false;
                
        switch(type) {
        case INTEGER:
            if ((highThreshold instanceof Integer) && (lowThreshold instanceof Integer))
                is_same_type = true;
            break;
        case BYTE:
            if ((highThreshold instanceof Byte) && (lowThreshold instanceof Byte))
                is_same_type = true;
            break;
        case SHORT:
            if ((highThreshold instanceof Short) && (lowThreshold instanceof Short))
                is_same_type = true;
            break;
        case LONG:
            if ((highThreshold instanceof Long) && (lowThreshold instanceof Long))
                is_same_type = true;
            break;
        case FLOAT:
            if ((highThreshold instanceof Float) && (lowThreshold instanceof Float))
                is_same_type = true;
            break;
        case DOUBLE:
            if ((highThreshold instanceof Double) && (lowThreshold instanceof Double))
                is_same_type = true;
            break;
        default:
            // Should never occured...
            if (isDebugOn()) {
                debug("isThresholdTypeValid", "the threshold type is invalid");
            }
            break;
        }
        return is_same_type;
    }  
        
    /**
     * Sets the derived gauge when the differenceMode flag is set to <CODE>true</CODE>.
     * Both integer and floating-point types are allowed.
     *
     * @param scanGauge The value of the observed attribute.
     */
    void setDerivedGaugeWithDifference(Number scanGauge) { 
        
        switch (type) {
        case INTEGER:
            derivedGauge = new Integer(((Integer)scanGauge).intValue() - ((Integer)previousScanGauge).intValue());
            break;
        case BYTE:
            derivedGauge = new Byte((byte)(((Byte)scanGauge).byteValue() - ((Byte)previousScanGauge).byteValue()));
            break;
        case SHORT:
            derivedGauge = new Short((short)(((Short)scanGauge).shortValue() - ((Short)previousScanGauge).shortValue()));
            break;
        case LONG:
            derivedGauge = new Long(((Long)scanGauge).longValue() - ((Long)previousScanGauge).longValue());
            break;
        case FLOAT:
            derivedGauge = new Float(((Float)scanGauge).floatValue() - ((Float)previousScanGauge).floatValue());
            break;
        case DOUBLE:
            derivedGauge = new Double(((Double)scanGauge).doubleValue() - ((Double)previousScanGauge).doubleValue());
        default:
            // Should never occured...
            if (isDebugOn()) {
                debug("setDerivedGaugeWithDifference", "the threshold type is invalid");
            }
            break;
        }
    } 
        
}


/**
 * GaugeAlarmClock inner class:
 * This class provides a simple implementation of an alarm clock MBean.
 * The aim of this MBean is to set up an alarm which wakes up the gauge monitor every granularity period.
 */

class GaugeAlarmClock extends Thread { 

    GaugeMonitor listener = null;
    long granularityPeriod = 10000;
    
    /*
     * ------------------------------------------
     *  CONSTRUCTORS
     * ------------------------------------------
     */
    
    public GaugeAlarmClock(GaugeMonitor listener, long granularityPeriod) {
        this.listener = listener;
        this.granularityPeriod = granularityPeriod;
    }

    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */

    /**
     * This method is called by the GaugeAlarmClock thread when it is started.
     */
    public void run() {
        
        try {
            this.sleep(granularityPeriod);
            listener.notifyAlarmClock();
        } catch (InterruptedException e) {
            // Ignore...
        }
    }
}
