/*
 * @(#)file      Monitor.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   4.15
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.monitor; 



// java imports
//
import java.util.*;

// jmx imports
//
import javax.management.MBeanServer;
import javax.management.MBeanRegistration;
import javax.management.ObjectName;

// RI imports
//
import com.sun.management.jmx.Trace;

/**
 * Defines the common part to all monitor MBeans.
 * A monitor MBean monitors values of an attribute in an observed MBean. The
 * observed attribute is monitored at intervals specified by the granularity
 * period. A gauge value (derived gauge) is derived from the values of
 * the observed attribute.
 *
 * @version     4.15     03/28/00
 * @author      Sun Microsystems, Inc
 */

public abstract class Monitor extends javax.management.NotificationBroadcasterSupport implements MonitorMBean, javax.management.MBeanRegistration, java.io.Serializable { 

    /*
     * ------------------------------------------
     *  PRIVATE VARIABLES
     * ------------------------------------------
     */
    
    /**
     * Object to which the attribute to observe belongs to.
     * <BR>The default value is set to null.
     */
    private ObjectName observedObject = null;

    /**
     * Attribute to observe.
     * <BR>The default value is set to null.
     */
    private String observedAttribute = null;

    /**
     * Monitor granularity period (in milliseconds).
     * <BR>The default value is set to 10 seconds.
     */
    private long granularityPeriod = 10000;
     

    /*
     * ------------------------------------------
     *  PROTECTED VARIABLES
     * ------------------------------------------
     */
    
    /**
     * Reference on the MBean server.
     * This reference is null when the monitor MBean is not registered in an MBean server.
     * This reference is initialized before the monitor MBean is registered in the MBean server.
     * @see #preRegister(MBeanServer server, ObjectName name) 
     */
    protected MBeanServer server = null;
            
    /**
     * Selected monitor errors that have already been notified.
     */
    protected int alreadyNotified = RESET_FLAGS_ALREADY_NOTIFIED;
            
    // Flags defining possible monitor errors.
    //
    
    /**
     * This flag is used to reset the {@link #alreadyNotified alreadyNotified} monitor attribute.
     */
    protected static final int RESET_FLAGS_ALREADY_NOTIFIED             = 0;
        
    /**
     * Flag denoting that a notification has occurred after changing the observed object.
     * This flag is used to check that the new observed object is registered in the MBean server
     * at the time of the first notification.
     */
    protected static final int OBSERVED_OBJECT_ERROR_NOTIFIED           = 1;
    
    /**
     * Flag denoting that a notification has occurred after changing the observed attribute.
     * This flag is used to check that the new observed attribute belongs to the observed object
     * at the time of the first notification.
     */
    protected static final int OBSERVED_ATTRIBUTE_ERROR_NOTIFIED        = 2;
    
    /**
     * Flag denoting that a notification has occurred after changing the observed object or the observed attribute.
     * This flag is used to check that the observed attribute type is correct (depending on the monitor in use)
     * at the time of the first notification.
     */
    protected static final int OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED   = 4;

    /**
     * Flag denoting that a notification has occurred after changing the observed object or the observed attribute.
     * This flag is used to notify any exception (except the cases described above) when trying to get the value
     * of the observed attribute at the time of the first notification.
     */
    protected static final int RUNTIME_ERROR_NOTIFIED                   = 8;
    
    protected String dbgTag = "Monitor";
        
    
    /*    
     * ------------------------------------------
     *  PACKAGE VARIABLES
     * ------------------------------------------
     */
    
    /**
     * Monitor state.
     * The default value is set to <CODE>false</CODE>.
     */
    transient boolean isActive = false;
    
    /**
     * Monitor sequence number.
     * The default value is set to 0.
     */
    transient long sequenceNumber = 0;

    // TRACES & DEBUG
    //---------------
    
    boolean isTraceOn() {
        return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_MONITOR);
    }

    void trace(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_MONITOR, clz, func, info);
    }

    void trace(String func, String info) {
        trace(dbgTag, func, info);
    }
    
    boolean isDebugOn() {
        return Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_MONITOR);
    }

    void debug(String clz, String func, String info) {
        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_MONITOR, clz, func, info);
    }

    void debug(String func, String info) {
        debug(dbgTag, func, info);
    }
        

    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */

    /**
     * Allows the monitor MBean to perform any operations it needs before being registered
     * in the MBean server.
     * <P>
     * Initializes the reference to the MBean server.
     *
     * @param server The MBean server in which the monitor MBean will be registered.
     * @param name The object name of the monitor MBean.
     *
     * @return The name of the monitor MBean registered.
     *
     * @exception java.lang.Exception
     */
    public ObjectName preRegister(MBeanServer server, ObjectName name) 
        throws java.lang.Exception {
        
        if (isTraceOn()) {
            trace("preRegister", "initialize the reference on the MBean server");
        }
        
        this.server = server;
        return name;
    } 
            
    /**
     * Allows the monitor MBean to perform any operations needed after having been
     * registered in the MBean server or after the registration has failed.
     * <P>
     * Not used in this context.
     */
    public void postRegister (Boolean registrationDone) {
    } 

    /**
     * Allows the monitor MBean to perform any operations it needs before being de-registered
     * by the MBean server.
     * <P>
     * Stops the monitor.
     *
     * @exception java.lang.Exception
     */
    public void preDeregister() throws java.lang.Exception {
        
        if (isTraceOn()) {
            trace("preDeregister", "stop the monitor");
        }
        
        // Stop the Monitor.
        //
        stop();
    }

    /**
     * Allows the monitor MBean to perform any operations needed after having been
     * de-registered by the MBean server.
     * <P>
     * Not used in this context.
     */
    public void postDeregister() {
    }
    
    /**
     * Starts the monitor.
     */
    public abstract void start();
  
    /**
     * Stops the monitor.
     */
    public abstract void stop();
  
    // GETTERS AND SETTERS
    //--------------------    
    
    /**
     * Gets the object name of the object being observed.
     * <BR>The observed object is not initialized by default (set to null).
     *
     * @return The object being observed.
     */
    public ObjectName getObservedObject() { 
        return observedObject;
    } 

    /**
     * Sets the object to observe identified by its object name.
     * <BR>The observed object is not initialized by default (set to null).
     *
     * @param object The object to observe.
     * @exception java.lang.IllegalArgumentException The specified object is null.
     */
    public void setObservedObject(ObjectName object) throws java.lang.IllegalArgumentException { 
        
        if (object == null) {
            throw new java.lang.IllegalArgumentException("The object to observe cannot be null.");
        }
        observedObject = object;
        alreadyNotified &= ~OBSERVED_OBJECT_ERROR_NOTIFIED;
        alreadyNotified &= ~OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
        alreadyNotified &= ~OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED;
    } 

    /**
     * Gets the attribute being observed.
     * <BR>The observed attribute is not initialized by default (set to null).
     *
     * @return The attribute being observed.
     */
    public String getObservedAttribute() { 
        return observedAttribute;
    } 

    /**
     * Sets the attribute to observe.
     * <BR>The observed attribute is not initialized by default (set to null).
     *
     * @param attribute The attribute to observe.
     * @exception java.lang.IllegalArgumentException The specified attribute is null.
     */
    public void setObservedAttribute(String attribute) throws java.lang.IllegalArgumentException { 
        
        if (attribute == null) {
            throw new java.lang.IllegalArgumentException("The attribute to observe cannot be null.");
        }
        observedAttribute = attribute;
        alreadyNotified &= ~OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
        alreadyNotified &= ~OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED;
    }

    /**
     * Gets the granularity period (in milliseconds).
     * <BR>The default value of the granularity period is 10 seconds.
     *
     * @return The granularity period value.
     */
    public long getGranularityPeriod() { 
        return granularityPeriod;
    } 
    
    /**
     * Sets the granularity period (in milliseconds).
     * <BR>The default value of the granularity period is 10 seconds.
     *
     * @param period The granularity period value.
     * @exception java.lang.IllegalArgumentException The granularity period is less than or equal to zero.
     */
    public void setGranularityPeriod(long period) throws java.lang.IllegalArgumentException { 
        
        if (period <= 0) {
            throw new java.lang.IllegalArgumentException("The granularity period must be greater than zero.");
        }
        granularityPeriod = period;
    }
    
    /**
     * Tests whether the monitor MBean is active.
     * A monitor MBean is marked active when the {@link #start start} method is called.
     * It becomes inactive when the {@link #stop stop} method is called.
     *
     * @return <CODE>true</CODE> if the monitor MBean is active, <CODE>false</CODE> otherwise.
     */
    public boolean isActive() {
        return isActive;
    }  
    
    /*    
     * ------------------------------------------
     *  PACKAGE METHODS
     * ------------------------------------------
     */
    
    /**
     * This method is used by the monitor MBean create and send a monitor
     * notification to all the listeners registered for this kind of notification.
     *
     * @param type The notification type.
     * @param timeStamp The notification emission date.
     * @param msg The notification message.
     * @param derGauge The derived gauge.
     * @param trigger The threshold/string (depending on the monitor type) that triggered off the notification.
     */
    void sendNotification(String type, long timeStamp, String msg, Object derGauge, Object trigger) {
        
        synchronized(this) {
            sequenceNumber = sequenceNumber + 1;
        }

        if (isTraceOn()) {
            trace("sendNotification", "send notification:" + 
                  "\n\tNotification observed object = " + observedObject +
                  "\n\tNotification observed attribute = " + observedAttribute +
                  "\n\tNotification derived gauge = " + derGauge);
        }
        MonitorNotification notif = new MonitorNotification(type, this, sequenceNumber, timeStamp, msg, 
                                                            observedObject, observedAttribute, derGauge, trigger);
            
        this.sendNotification(notif) ;
    }
}
