/*
 * @(#)file      StringMonitor.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   4.14
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.monitor; 


// java imports
//
import java.util.Date;

// jmx imports
//
import javax.management.MBeanNotificationInfo;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.ReflectionException;

/**
 * Defines a monitor MBean designed to observe the values of a string attribute.
 * <P>
 * A string monitor sends notifications as follows:
 * <UL>
 * <LI> if the attribute value matches the string to compare value, 
 *      a {@link MonitorNotification#STRING_TO_COMPARE_VALUE_MATCHED match notification} is sent. 
 *      The notify match flag must be set to <CODE>true</CODE>.
 *      <BR>Subsequent matchings of the string to compare values do not cause further notifications unless 
 *      the attribute value differs from the string to compare value.
 * <LI> if the attribute value differs from the string to compare value, 
 *      a {@link MonitorNotification#STRING_TO_COMPARE_VALUE_DIFFERED differ notification} is sent. 
 *      The notify differ flag must be set to <CODE>true</CODE>.
 *      <BR>Subsequent differences from the string to compare value do not cause further notifications unless 
 *      the attribute value matches the string to compare value.
 * </UL>
 *
 * @version     4.14     03/29/00
 * @author      Sun Microsystems, Inc
 */

public class StringMonitor extends Monitor implements StringMonitorMBean { 


    // TRACES & DEBUG
    //---------------
    
    String makeDebugTag() {
        return "StringMonitor";
    }
    
    /*
     * ------------------------------------------
     *  PRIVATE VARIABLES
     * ------------------------------------------
     */
    
    /**
     * String to compare with the observed attribute.
     * <BR>The default value is an empty character sequence.
     */
    private String stringToCompare = new String();
        
    /**
     * Flag indicating if the string monitor notifies when matching the string to compare.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean notifyMatch = false;
        
    /**
     * Flag indicating if the string monitor notifies when differing from the string to compare.
     * <BR>The default value is set to <CODE>false</CODE>.
     */
    private boolean notifyDiffer = false;
    
    /**
     * Derived gauge.
     */
    private transient String derivedGauge = new String();

    /**
     * Derived gauge timestamp.
     */
    private transient long derivedGaugeTimestamp = (new Date()).getTime();
    
    /**
     * This attribute is used to handle the matching/differing mechanism.
     */
    private transient int status = MATCHING_OR_DIFFERING;
    
    // Flags needed to implement the matching/differing mechanism.
    //
    private static final int MATCHING                   = 0;
    private static final int DIFFERING                  = 1;
    private static final int MATCHING_OR_DIFFERING      = 2;
    
    /**
     * AlarmClock MBean.
     */
    private transient StringAlarmClock alarmClock = null;


    /*
     * ------------------------------------------
     *  CONSTRUCTORS
     * ------------------------------------------
     */
    
    /**
     * Default constructor.
     */
    public StringMonitor() {
        dbgTag = makeDebugTag();
    }
    
    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */
        
    /**
     * Starts the string monitor.
     */
    public synchronized void start() {
        
        if (isTraceOn()) {
            trace("start", "start the string monitor");
        }
        
        if (!isActive()) {
                        
            // Start the AlarmClock.
            //
            alarmClock = new StringAlarmClock(this, getGranularityPeriod());
            alarmClock.start();
       
            isActive = true;
            
            // Reset values.
            //
            status = MATCHING_OR_DIFFERING;
        } else {
            if (isTraceOn()) {
                trace("start", "the string monitor is already activated");
            }
        }
    }
    
    /**
     * Stops the string monitor.
     */
    public synchronized void stop() {

        if (isTraceOn()) {
            trace("stop", "stop the string monitor");
        }
        
        if (isActive()) {
            
            // Stop the AlarmClock.
            //
            if (alarmClock != null) {
                alarmClock.interrupt();
                try {
                    // Wait until the thread die.
                    //
                    alarmClock.join();
                } catch (InterruptedException e) {
                    // Ignore...
                }
                // Remove the reference on the AlarmClock.
                //
                alarmClock = null;
            }

            isActive = false;
        } else {
            if (isTraceOn()) {
                trace("stop", "the string monitor is already deactivated");
            }
        }
    }
    
    // GETTERS AND SETTERS
    //--------------------
    
    /**
     * Gets the derived gauge.
     *
     * @return The derived gauge.
     */
    public String getDerivedGauge() { 
        return derivedGauge;
    } 

    /**
     * Gets the derived gauge timestamp.
     *
     * @return The derived gauge timestamp.
     */
    public long getDerivedGaugeTimeStamp() { 
        return derivedGaugeTimestamp;
    } 
    
    /**
     * Gets the string to compare with the observed attribute.
     *
     * @return The string value.
     */
    public String getStringToCompare() {
        return stringToCompare;
    }

    /**
     * Sets the string to compare with the observed attribute.
     *
     * @param value The string value.
     * @exception java.lang.IllegalArgumentException The specified string to compare is null.
     */
    public void setStringToCompare(String value) throws java.lang.IllegalArgumentException {
        
        if (value == null) {
            throw new java.lang.IllegalArgumentException("The string to compare cannot be null.");
        }
        stringToCompare = value;
        
        // Reset values.
        //
        status = MATCHING_OR_DIFFERING;
    }
    
    /**
     * Gets the matching notification's on/off switch value.
     *
     * @return <CODE>true</CODE> if the string monitor notifies when matching the string to compare, <CODE>false</CODE> otherwise.
     */
    public boolean getNotifyMatch() {
        return notifyMatch;
    }

    /**
     * Sets the matching notification's on/off switch value.
     *
     * @param value The matching notification's on/off switch value.
     */
    public void setNotifyMatch(boolean value) {
        notifyMatch = value;
    }

    /**
     * Gets the differing notification's on/off switch value.
     *
     * @return <CODE>true</CODE> if the string monitor notifies when differing from the string to compare, <CODE>false</CODE> otherwise.
     */
    public boolean getNotifyDiffer() {
        return notifyDiffer;
    }

    /**
     * Sets the differing notification's on/off switch value.
     *
     * @param value The differing notification's on/off switch value.
     */
    public void setNotifyDiffer(boolean value) {
        notifyDiffer = value;
    }
                
    /**    
     * Returns a <CODE>NotificationInfo</CODE> object containing the name of the Java class of the notification
     * and the notification types sent by the string monitor.  
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types  = { MonitorNotification.RUNTIME_ERROR, 
                            MonitorNotification.OBSERVED_OBJECT_ERROR, 
                            MonitorNotification.OBSERVED_ATTRIBUTE_ERROR,
                            MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR,
                            MonitorNotification.STRING_TO_COMPARE_VALUE_MATCHED,
                            MonitorNotification.STRING_TO_COMPARE_VALUE_DIFFERED};
        MBeanNotificationInfo[] notifsInfo = { new MBeanNotificationInfo(types, "javax.management.monitor.MonitorNotification", 
                                                                         "Notifications sent by the StringMonitor MBean")};
        return notifsInfo;
    }  
    
    /*
     * ------------------------------------------
     *  PRIVATE METHODS
     * ------------------------------------------
     */
    
    /**
     * Updates the derived gauge and the derived gauge timestamp attributes.
     *
     * @param scanString The value of the observed attribute.
     */
    private void updateDerivedGauge(Object scanString) {
        
        derivedGaugeTimestamp = (new Date()).getTime();
        derivedGauge = (String)scanString;
    }

    /**
     * Updates the notification attributes and notifies the listeners 
     * only once if the notifyMatch/notifyDiffer flag is set to <CODE>true</CODE>.
     */
    private void updateNotifications() {
        
        MonitorNotification notif = null;
                
        // Send matching notification if notifyMatch is true.
        // Send differing notification if notifyDiffer is true.
        //
        if (status == MATCHING_OR_DIFFERING) {
            if (derivedGauge.equals(stringToCompare)) {	      
                if (notifyMatch == true) {
                    sendNotification(MonitorNotification.STRING_TO_COMPARE_VALUE_MATCHED,
                                     derivedGaugeTimestamp, "", derivedGauge, stringToCompare);
                }
                status = DIFFERING;
            }
            else {	      
                if (notifyDiffer == true) {
                    sendNotification(MonitorNotification.STRING_TO_COMPARE_VALUE_DIFFERED,
                                     derivedGaugeTimestamp, "", derivedGauge, stringToCompare);
                }
                status = MATCHING;
            }
        }
        else {
            if (status == MATCHING) {
                if (derivedGauge.equals(stringToCompare)) {	      
                    if (notifyMatch == true) {
                        sendNotification(MonitorNotification.STRING_TO_COMPARE_VALUE_MATCHED,
                                         derivedGaugeTimestamp, "", derivedGauge, stringToCompare);
                    }
                    status = DIFFERING;
                }
            }
            else if (status == DIFFERING) {
                if (!derivedGauge.equals(stringToCompare)) {	      
                    if (notifyDiffer == true) {
                        sendNotification(MonitorNotification.STRING_TO_COMPARE_VALUE_DIFFERED,
                                         derivedGaugeTimestamp, "", derivedGauge, stringToCompare);
                    }
                    status = MATCHING;
                }
            }
        }
    }

    /*    
     * ------------------------------------------
     *  PACKAGE METHODS
     * ------------------------------------------
     */
    
    /**
     * This method is called by the string monitor each time
     * the granularity period has been exceeded.
     */
    void notifyAlarmClock() {

        Object  scan_string = null;
        String  notif_type = null;

        try {
            if (isActive()) {
                
                // Check if the observed object and observed attribute are valid.
                //

                // Check that neither the observed object nor the observed attribute are null.
                // If the observed object or observed attribute is null, this means that
                // the monitor started before a complete initialization and nothing is done.
                //
                if ((getObservedObject() == null) || (getObservedAttribute() == null)) {
                    // Start the AlarmClock with a timeout equal to the granularity period value.
                    //
                    alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                    alarmClock.start();
                    return;
                }

                // Check that the observed object is registered in the MBean server and
                // that the observed attribute belongs to the observed object.
                //
                try {
                    scan_string = server.getAttribute(getObservedObject(), getObservedAttribute());
                } catch (NullPointerException np_ex) {
                    if ((alreadyNotified & RUNTIME_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.RUNTIME_ERROR;
                        alreadyNotified |= RUNTIME_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The string monitor must be registered in the MBean server.");
                    }
                } catch (InstanceNotFoundException inf_ex) {
                    if ((alreadyNotified & OBSERVED_OBJECT_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_OBJECT_ERROR;
                        alreadyNotified |= OBSERVED_OBJECT_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed object must be registered in the MBean server.");
                    }
                } catch (AttributeNotFoundException anf_ex) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed attribute must be accessible in the observed object.");
                    }
                } catch (MBeanException mb_ex) {
                    if ((alreadyNotified & RUNTIME_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.RUNTIME_ERROR;
                        alreadyNotified |= RUNTIME_ERROR_NOTIFIED;
                        throw new MonitorSettingException(mb_ex.getMessage());
                    }
                } catch (ReflectionException ref_ex) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_ERROR_NOTIFIED;
                        throw new MonitorSettingException(ref_ex.getMessage());
                    }
                }

                // Check that the observed attribute is of type "String".
                //
                if (!(scan_string instanceof String)) {
                    if ((alreadyNotified & OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED) != RESET_FLAGS_ALREADY_NOTIFIED) {
                        alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                        alarmClock.start();
                        return;
                    } else {
                        notif_type = MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR;
                        alreadyNotified |= OBSERVED_ATTRIBUTE_TYPE_ERROR_NOTIFIED;
                        throw new MonitorSettingException("The observed attribute type must be a string type.");
                    }
                }

                // Clear all already notified flags.
                //
                alreadyNotified = RESET_FLAGS_ALREADY_NOTIFIED;

                // Update the derived gauge attributes.
                //
                updateDerivedGauge(scan_string);
                
                // Notify the listeners.
                //
                updateNotifications();
                
                // Start the AlarmClock with a timeout equal to the granularity period value.
                //
                alarmClock = new StringAlarmClock(this, getGranularityPeriod());
                alarmClock.start();
            }
        } 
        catch (MonitorSettingException ms_ex) {
            
            // Create and send the monitor error notification.
            //
            sendNotification(notif_type, derivedGaugeTimestamp, ms_ex.getMessage(), derivedGauge, null);
            
            // Reset values.
            //
            status = MATCHING_OR_DIFFERING;
            
            // Start the AlarmClock with a timeout equal to the granularity period value.
            //
            alarmClock = new StringAlarmClock(this, getGranularityPeriod());
            alarmClock.start();
        } 
    }
}


/**
 * StringAlarmClock inner class:
 * This class provides a simple implementation of an alarm clock MBean.
 * The aim of this MBean is to set up an alarm which wakes up the string monitor every granularity period.
 */

class StringAlarmClock extends Thread { 

    StringMonitor listener = null;
    long granularityPeriod = 10000;
    
    /*
     * ------------------------------------------
     *  CONSTRUCTORS
     * ------------------------------------------
     */
    
    public StringAlarmClock(StringMonitor listener, long granularityPeriod) {
        this.listener = listener;
        this.granularityPeriod = granularityPeriod;
    }

    /*
     * ------------------------------------------
     *  PUBLIC METHODS
     * ------------------------------------------
     */

    /**
     * This method is called by the StringAlarmClock thread when it is started.
     */
    public void run() {
        
        try {
            this.sleep(granularityPeriod);
            listener.notifyAlarmClock();
        } catch (InterruptedException e) {
            // Ignore...
        }
    }
}
