/*
 * @(#)file      MBeanServerNotificationFilter.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.12
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.relation;

import javax.management.Notification;
import javax.management.NotificationFilterSupport;
import javax.management.MBeanServerNotification;
import javax.management.ObjectName;

import java.util.Vector;

import com.sun.management.jmx.Trace;

/**
 * This filter allows to filter MBeanServerNotification notifications by
 * selecting the ObjectNames of interest and the operations (registration,
 * unregistration, both) of interest (corresponding to notification
 * types).
 */
public class MBeanServerNotificationFilter extends NotificationFilterSupport {

    //
    // Private members
    //

    // List of ObjectNames of interest
    // 'null' means all ObjectNames are implictly selected (check for explicit
    // deselections)
    // empty is no ObjectNames are explicitly selected
    private Vector mySelectObjNameList = new Vector();

    // List of ObjectNames NOT of interest
    // 'null' means all ObjectNames are implicitly deselected (check for
    // explicit selections)
    // empty is no ObjectNames are explicitly deselected
    private Vector myDeselectObjNameList = null;

    //
    // Constructor
    //

    /**
     * Creates a filter selecting all MBeanServerNotification notifications for
     * all ObjectNames
     */
    public MBeanServerNotificationFilter() {

	super();

        if (isTraceOn())
            trace("Constructor: entering", null);

	enableType(MBeanServerNotification.REGISTRATION_NOTIFICATION);
	enableType(MBeanServerNotification.UNREGISTRATION_NOTIFICATION);

        if (isTraceOn())
            trace("Constructor: exiting", null);
	return;
    }

    //
    // Accessors
    //

    /**
     * Disables any MBeanServerNotification (i.e. all ObjectNames
     * deselected)
     */
    public synchronized void disableAllObjectNames() {

        if (isTraceOn())
            trace("disableAllObjectNames: entering", null);

	mySelectObjNameList = new Vector();
	myDeselectObjNameList = null;

        if (isTraceOn())
            trace("disableAllObjectNames: exiting", null);
	return;
    }

    /**
     * Disables MBeanServerNotifications concerning given ObjectName.
     *
     * @param theObjName  ObjectName no longer of interest
     *
     * @exception IllegalArgumentException  if the given ObjectName is null
     */
    public synchronized void disableObjectName(ObjectName theObjName)
	throws IllegalArgumentException {

	if (theObjName == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	if (isTraceOn())
            trace("disableObjectName: entering", theObjName.toString());

	// Removes from selected ObjectNames, if present
	if (mySelectObjNameList != null) {
	    if (mySelectObjNameList.size() != 0) {
		mySelectObjNameList.removeElement(theObjName);
	    }
	}

	// Adds it in deselected ObjectNames
	if (myDeselectObjNameList != null) {
	    // If all are deselected, no need to do anything :)
	    if (!(myDeselectObjNameList.contains(theObjName))) {
		// ObjectName was not already deselected
		myDeselectObjNameList.addElement(theObjName);
	    }
	}

        if (isTraceOn())
            trace("disableObjectName: exiting", null);
	return;
    }

    /**
     * Enables all MBeanServerNotifications (i.e. all ObjectNames selected)
     */
    public synchronized void enableAllObjectNames() {

	if (isTraceOn())
            trace("enableAllObjectNames: entering", null);

    	mySelectObjNameList = null;
	myDeselectObjNameList = new Vector();

        if (isTraceOn())
            trace("enableAllObjectNames: exiting", null);
	return;
    }

    /**
     * Enables MBeanServerNotifications concerning given ObjectName.
     *
     * @param theObjName  ObjectName of interest
     *
     * @exception IllegalArgumentException  if the given ObjectName is null
     */
    public synchronized void enableObjectName(ObjectName theObjName)
	throws IllegalArgumentException {

	if (theObjName == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	if (isTraceOn())
            trace("enableObjectName: entering", theObjName.toString());

	// Removes from deselected ObjectNames, if present
	if (myDeselectObjNameList != null) {
	    if (myDeselectObjNameList.size() != 0) {
		myDeselectObjNameList.removeElement(theObjName);
	    }
	}

	// Adds it in selected ObjectNames
	if (mySelectObjNameList != null) {
	    // If all are selected, no need to do anything :)
	    if (!(mySelectObjNameList.contains(theObjName))) {
		// ObjectName was not already selected
		mySelectObjNameList.addElement(theObjName);
	    }
	}

        if (isTraceOn())
            trace("enableObjectName: exiting", null);
	return;
    }
		
    /**
     * Gets all the ObjectNames enabled.
     *
     * @return Vector of ObjectNames:
     * <P>- null means all ObjectNames are implicitly selected, except the
     * ObjectNames explicitly deselected
     * <P>- empty means all ObjectNames are deselected, i.e. no ObjectName
     * selected.
     */
    public synchronized Vector getEnabledObjectNames() {
	return (Vector)(mySelectObjNameList.clone());
    }

    /**
     * Gets all the ObjectNames disabled.
     *
     * @return Vector of ObjectNames:
     * <P>- null means all ObjectNames are implicitly deselected, except the
     * ObjectNames explicitly selected
     * <P>- empty means all ObjectNames are selected, i.e. no ObjectName
     * deselected.
     */
    public synchronized Vector getDisabledObjectNames() {
	return (Vector)(myDeselectObjNameList.clone());
    }

    //
    // NotificationFilter interface
    //

    /**
     * Invoked before sending the specified notification to the listener.
     * <P>If:
     * <P>- the ObjectName of the concerned MBean is selected (explicitly OR
     * (implicitly and not explicitly deselected))
     * <P>AND
     * <P>- the type of the operation (registration or unregistration) is
     * selected
     * <P>then the notification is sent to the listener.
     *
     * @param theNtf  The notification to be sent.
     *
     * @return true if the notification has to be sent to the listener, false
     * otherwise.
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public synchronized boolean isNotificationEnabled(Notification theNtf)
	throws IllegalArgumentException {

	if (theNtf == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	if (isTraceOn())
            trace("isNotificationEnabled: entering", theNtf.toString());

	// Checks the type first
	String ntfType = theNtf.getType();
	Vector enabledTypes = getEnabledTypes();
	if (!(enabledTypes.contains(ntfType))) {

	    if (isTraceOn())
		trace("isNotificationEnabled: type not selected, exiting", null);
	    return false;
	}

	// We have a MBeanServerNotification: downcasts it
	MBeanServerNotification mbsNtf = (MBeanServerNotification)theNtf;

	// Checks the ObjectName
	ObjectName objName = mbsNtf.getMBeanName();
	// Is it selected?
	boolean isSelectedFlg = false;
	if (mySelectObjNameList != null) {
	    // Not all are implicitly selected:
	    // checks for explicit selection
	    if (mySelectObjNameList.size() == 0) {
		// All are explicitly not selected
		if (isTraceOn())
		    trace("isNotificationEnabled: no ObjectNames selected, exiting", null);
		return false;
	    }

	    isSelectedFlg = mySelectObjNameList.contains(objName);
	    if (!isSelectedFlg) {
		// Not in the explicit selected list
		if (isTraceOn())
		    trace("isNotificationEnabled: ObjectName not in selected list, exiting", null);
		return false;
	    }
	}

	if (!isSelectedFlg) {
	    // Not explicitly selected: is it deselected?

	    if (myDeselectObjNameList == null) {
		// All are implicitly deselected and it is not explicitly
		// selected
		if (isTraceOn())
		    trace("isNotificationEnabled: ObjectName not selected and all deselectedm, exiting", null);
		return false;

	    } else if (myDeselectObjNameList.contains(objName)) {
		// Explicitly deselected
		if (isTraceOn())
		    trace("isNotificationEnabled: ObjectName explicitly not selected, exiting", null);
		return false;
	    }
	}

	if (isTraceOn())
	    trace("isNotificationEnabled: ObjectName selected, exiting", null);
	return true;
    }

    // stuff for Tracing

    private static String localClassName = "MBeanServerNotificationFilter";

    // trace level
    private boolean isTraceOn() {
        return Trace.isSelected(Trace.LEVEL_TRACE, Trace.INFO_RELATION);
    }

//    private void trace(String className, String methodName, String info) {
//        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_RELATION, className, methodName, info);
//    }

    private void trace(String methodName, String info) {
        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_RELATION, localClassName, methodName, info);
	Trace.send(Trace.LEVEL_TRACE, Trace.INFO_RELATION, "", "", "\n");
    }

//    private void trace(String className, String methodName, Exception e) {
//        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_RELATION, className, methodName, e);
//    }

//    private void trace(String methodName, Exception e) {
//        Trace.send(Trace.LEVEL_TRACE, Trace.INFO_RELATION, localClassName, methodName, e);
//    }

    // debug level
    private boolean isDebugOn() {
        return Trace.isSelected(Trace.LEVEL_DEBUG, Trace.INFO_RELATION);
    }

//    private void debug(String className, String methodName, String info) {
//        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_RELATION, className, methodName, info);
//    }

    private void debug(String methodName, String info) {
        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_RELATION, localClassName, methodName, info);
	Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_RELATION, "", "", "\n");
    }

//    private void debug(String className, String methodName, Exception e) {
//        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_RELATION, className, methodName, e);
//    }

//    private void debug(String methodName, Exception e) {
//        Trace.send(Trace.LEVEL_DEBUG, Trace.INFO_RELATION, localClassName, methodName, e);
//    }
}
