/*
 * @(#)file      RelationNotification.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.11
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.relation;

import javax.management.Notification;
import javax.management.ObjectName;

import java.util.ArrayList;
import java.util.List;

/**
 * A RelationNotification notification is sent when a relation is created via
 * the Relation Service, or a MBean is added as a relation in the Relation
 * Service, or a role is updated in a relation, or a relation is removed from
 * the Relation Service
 */
public class RelationNotification extends Notification {

    //
    // Notification types
    //

    /**
     * Type for the creation of an internal relation
     */
    public static final String RELATION_BASIC_CREATION = "jmx.relation.creation.basic";
    /**
     * Type for the relation MBean added into the Relation Service
     */
    public static final String RELATION_MBEAN_CREATION = "jmx.relation.creation.mbean";
    /**
     * Type for an update of an internal relation
     */
    public static final String RELATION_BASIC_UPDATE = "jmx.relation.update.basic";
    /**
     * Type for the update of a relation MBean
     */
    public static final String RELATION_MBEAN_UPDATE = "jmx.relation.update.mbean";
    /**
     * Type for the removal from the Relation Service of an internal relation
     */
    public static final String RELATION_BASIC_REMOVAL = "jmx.relation.removal.basic";
    /**
     * Type for the removal from the Relation Service of a relation MBean
     */
    public static final String RELATION_MBEAN_REMOVAL = "jmx.relation.removal.mbean";

    //
    // Private members
    //

    // Relation identifier of created/removed/updated relation
    private String myRelId = null;

    // Relation type name of created/removed/updated relation
    private String myRelTypeName = null;

    // ObjectName of the relation MBean of created/removed/updated relation
    // (only if the relation is represented by a MBean)
    private ObjectName myRelObjName = null;

    // List of ObjectNames of referenced MBeans to be unregistered due to
    // relation removal
    private ArrayList myUnregMBeanList = null;

    // Name of updated role (only for role update)
    private String myRoleName = null;

    // Old role value (ArrayList of ObjectNames) (only for role update)
    private ArrayList myOldRoleValue = null;

    // New role value (ArrayList of ObjectNames) (only for role update)
    private ArrayList myNewRoleValue = null;

    //
    // Constructors
    //

    /**
     * Creates a notification for either a relation creation (RelationSupport
     * object created internally in the Relation Service, or a MBean added as a
     * relation) or for a relation removal from the Relation Service
     *
     * @param theNtfType  type of the notification; either:
     * <P>- RELATION_BASIC_CREATION
     * <P>- RELATION_MBEAN_CREATION
     * <P>- RELATION_BASIC_REMOVAL
     * <P>- RELATION_MBEAN_REMOVAL
     * @param theSrcObj  source object, sending the notification. Will always
     * be a RelationService object.
     * @param TheSeqNbr  sequence number to identify the notification
     * @param theTimeStamp  time stamp
     * @param theMsg  human-readable message describing the notification
     * @param theRelId  relation id identifying the relation in the Relation
     * Service
     * @param theRelTypeName  name of the relation type
     * @param theRelObjName  ObjectName of the relation object if it is a MBean
     * (null for relations internally handled by the Relation Service)
     * @param theUnregMBeanList  list of ObjectNames of referenced MBeans
     * expected to be unregistered due to relation removal (only for removal,
     * due to CIM qualifiers, can be null)
     *
     * @exception IllegalArgumentException  if:
     * <P>- no value for the notification type
     * <P>- the notification type is not RELATION_BASIC_CREATION,
     * RELATION_MBEAN_CREATION, RELATION_BASIC_REMOVAL or
     * RELATION_MBEAN_REMOVAL
     * <P>- no source object
     * <P>- the source object is not a Relation Service
     * <P>- no relation id
     * <P>- no relation type name
     */
    public RelationNotification(String theNtfType,
				Object theSrcObj,
				long TheSeqNbr,
				long theTimeStamp,
				String theMsg,
				String theRelId,
				String theRelTypeName,
				ObjectName theRelObjName,
				List theUnregMBeanList)
	throws IllegalArgumentException {

	super(theNtfType, theSrcObj, TheSeqNbr, theTimeStamp, theMsg);

	// Can throw IllegalArgumentException
	initMembers(1,
		    theNtfType,
		    theSrcObj,
		    TheSeqNbr,
		    theTimeStamp,
		    theMsg,
		    theRelId,
		    theRelTypeName,
		    theRelObjName,
		    theUnregMBeanList,
		    null,
		    null,
		    null);
	return;
    }

    /**
     * Creates a notification for a role update in a relation
     *
     * @param theNtfType  type of the notification; either:
     * <P>- RELATION_BASIC_UPDATE
     * <P>- RELATION_MBEAN_UPDATE
     * @param theSrcObj  source object, sending the notification. Will always
     * be a RelationService object.
     * @param TheSeqNbr  sequence number to identify the notification
     * @param theTimeStamp  time stamp
     * @param theMsg  human-readable message describing the notification
     * @param theRelId  relation id identifying the relation in the Relation
     * Service
     * @param theRelTypeName  name of the relation type
     * @param theRelObjName  ObjectName of the relation object if it is a MBean
     * (null for relations internally handled by the Relation Service)
     * @param theRoleName  name of the updated role
     * @param theNewRoleValue  new value (ArrayList of ObjectName objects)
     * @param theOldRoleValue  old value (ArrayList of ObjectName objects)
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public RelationNotification(String theNtfType,
				Object theSrcObj,
				long TheSeqNbr,
				long theTimeStamp,
				String theMsg,
				String theRelId,
				String theRelTypeName,
				ObjectName theRelObjName,
				String theRoleName,
				List theNewRoleValue,
				List theOldRoleValue
				)
	throws IllegalArgumentException {

	super(theNtfType, theSrcObj, TheSeqNbr, theTimeStamp, theMsg);

	// Can throw IllegalArgumentException
	initMembers(2,
		    theNtfType,
		    theSrcObj,
		    TheSeqNbr,
		    theTimeStamp,
		    theMsg,
		    theRelId,
		    theRelTypeName,
		    theRelObjName,
		    null,
		    theRoleName,
		    theNewRoleValue,
		    theOldRoleValue);
	return;
    }

    //
    // Accessors
    //

    /**
     * Returns the relation identifier of created/removed/updated relation
     */
    public String getRelationId() {
	return new String(myRelId);
    }

    /**
     * Returns the relation type name of created/removed/updated relation
     */
    public String getRelationTypeName() {
	return  new String(myRelTypeName);
    }

    /**
     * Returns the ObjectName (if the relation is a MBean, else null) of the
     * created/removed/updated relation
     */
    public ObjectName getObjectName() {
	return myRelObjName;
    }

    /**
     * Returns the list of ObjectNames of MBeans expected to be unregistered
     * due to a relation removal (only for relation removal)
     */
    public List getMBeansToUnregister() {
	ArrayList result = null;
	if (myUnregMBeanList != null) {
	    result = (ArrayList)(myUnregMBeanList.clone());
	} else {
	    result = new ArrayList();
	}
	return result;
    }

    /**
     * Returns name of updated role of updated relation (only for role update)
     */
    public String getRoleName() {
	String result = null;
	if (myRoleName != null) {
	    result = new String(myRoleName);
	}
	return result;
    }

    /**
     * Returns old value of updated role (only for role update)
     */
    public List getOldRoleValue() {
	ArrayList result = null;
	if (myOldRoleValue != null) {
	    result = (ArrayList)(myOldRoleValue.clone());
	} else {
	    result = new ArrayList();
	}
	return result;
    }

    /**
     * Returns new value of updated role (only for role update)
     */
    public List getNewRoleValue() {
	ArrayList result = null;
	if (myNewRoleValue != null) {
	    result = (ArrayList)(myNewRoleValue.clone());
	} else {
	    result = new ArrayList();
	}
	return result;
    }

    //
    // Misc
    //

    // Initialises members
    //
    // -param theNtfKind  1 for creation/removal, 2 for update
    // -param theNtfType  type of the notification; either:
    //  - RELATION_BASIC_UPDATE
    //  - RELATION_MBEAN_UPDATE
    //  for an update, or:
    //  - RELATION_BASIC_CREATION
    //  - RELATION_MBEAN_CREATION
    //  - RELATION_BASIC_REMOVAL
    //  - RELATION_MBEAN_REMOVAL
    //  for a creation or removal
    // -param theSrcObj  source object, sending the notification. Will always
    //  be a RelationService object.
    // -param TheSeqNbr  sequence number to identify the notification
    // -param theTimeStamp  time stamp
    // -param theMsg  human-readable message describing the notification
    // -param theRelId  relation id identifying the relation in the Relation
    //  Service
    // -param theRelTypeName  name of the relation type
    // -param theRelObjName  ObjectName of the relation object if it is a MBean
    //  (null for relations internally handled by the Relation Service)
    // -param theUnregMBeanList  list of ObjectNames of MBeans expected to be
    //  removed due to relation removal
    // -param theRoleName  name of the updated role
    // -param theNewRoleValue  new value (ArrayList of ObjectName objects)
    // -param theOldRoleValue  old value (ArrayList of ObjectName objects)
    //
    // -exception IllegalArgumentException  if:
    //  - no value for the notification type
    //  - incorrect notification type
    //  - no source object
    //  - the source object is not a Relation Service
    //  - no relation id
    //  - no relation type name
    //  - no role name (for role update)
    //  - no role old value (for role update)
    //  - no role new value (for role update)
    private void initMembers(int theNtfKind,
			     String theNtfType,
			     Object theSrcObj,
			     long TheSeqNbr,
			     long theTimeStamp,
			     String theMsg,
			     String theRelId,
			     String theRelTypeName,
			     ObjectName theRelObjName,
			     List theUnregMBeanList,
			     String theRoleName,
			     List theNewRoleValue,
			     List theOldRoleValue)
	throws IllegalArgumentException {

      boolean badInitFlg = false;

      if (theNtfType == null ||	  
	  theSrcObj == null ||
	  (!(theSrcObj instanceof RelationService)) ||
	  theRelId == null ||
	  theRelTypeName == null) {

	  badInitFlg = true;
      }

      if (theNtfKind == 1) {

       if ((!(theNtfType.equals(RelationNotification.RELATION_BASIC_CREATION)))
	   &&
	   (!(theNtfType.equals(RelationNotification.RELATION_MBEAN_CREATION)))
	   &&
	   (!(theNtfType.equals(RelationNotification.RELATION_BASIC_REMOVAL)))
	   &&
	   (!(theNtfType.equals(RelationNotification.RELATION_MBEAN_REMOVAL)))
	   ) {

	      // Creation/removal
	      badInitFlg = true;
         }

       } else if (theNtfKind == 2) {

       if (((!(theNtfType.equals(RelationNotification.RELATION_BASIC_UPDATE)))
	    &&
	    (!(theNtfType.equals(RelationNotification.RELATION_MBEAN_UPDATE))))
	   || theRoleName == null ||
	   theOldRoleValue == null ||
	   theNewRoleValue == null) {

	   // Role update
	   badInitFlg = true;
       }
      }

    if (badInitFlg) {
	// Revisit [cebro] Localize message
	String excMsg = "Invalid parameter.";
	throw new IllegalArgumentException(excMsg);
    }

    myRelId = new String(theRelId);
    myRelTypeName = new String(theRelTypeName);
    myRelObjName = theRelObjName;
    if (theUnregMBeanList != null) {
	myUnregMBeanList = (ArrayList)
	    (((ArrayList)theUnregMBeanList).clone());
    }
    if (theRoleName != null) {
	myRoleName = new String(theRoleName);
    }
    if (theOldRoleValue != null) {
	myOldRoleValue = (ArrayList)
	    (((ArrayList)theOldRoleValue).clone());
    }
    if (theNewRoleValue != null) {
	myNewRoleValue = (ArrayList)
	    (((ArrayList)theNewRoleValue).clone());
    }
    return;
    }
}
