/*
 * @(#)file      Role.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.9
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.relation;

import javax.management.ObjectName;

import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

import java.io.Serializable;

/**
 * Represents a role: includes a role name and referenced MBeans (via their
 * ObjectNames). The role value is always represented as an ArrayList
 * collection (of ObjectNames) to homogeneise the access.
 */
public class Role implements Serializable {

    //
    // Private members
    //

    // Role name
    private String myName = null;

    // List of ObjectNames of referenced MBeans
    private ArrayList myObjNameList = new ArrayList();

    //
    // Constructors
    //

    /**
     * Constructor
     * <P>No verification about the provided ObjectNames, no need for them
     * to be registered at this level. This checking will be made when trying
     * to set a relation role.
     *
     * @param theRoleName  role name
     * @param theRoleValue  role value (ArrayList of ObjectName objects)
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public Role(String theRoleName,
		List theRoleValue)
	throws IllegalArgumentException {

	if (theRoleName == null || theRoleValue == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	setRoleName(theRoleName);
	setRoleValue(theRoleValue);

	return;
    }

    //
    // Accessors
    //

    /**
     * Retrieves role name
     */
    public String getRoleName() {
	return myName;
    }

    /**
     * Retrieves role value
     *
     * @return ArrayList of ObjectName objects for referenced MBeans.
     */
    public List getRoleValue() {
	return myObjNameList;
    }

    /**
     * Sets role name
     *
     * @param theRoleName  role name
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public void setRoleName(String theRoleName)
	throws IllegalArgumentException {

	if (theRoleName == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	myName = new String(theRoleName);
	return;
    }

    /**
     * Sets role value
     *
     * @param theRoleValue  ArrayList of ObjectName objects for referenced
     * MBeans.
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public void setRoleValue(List theRoleValue)
	throws IllegalArgumentException {

	if (theRoleValue == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	myObjNameList = (ArrayList)
	    (((ArrayList)theRoleValue).clone());
	return;
    }

    /**
     * Prints a string describing the role
     */
    public String toString() {
	StringBuffer result = new StringBuffer();
	result.append("role name: " + myName + "; role value: ");
	for (Iterator objNameIter = myObjNameList.iterator();
	     objNameIter.hasNext();) {
	    ObjectName currObjName = (ObjectName)(objNameIter.next());
	    result.append(currObjName.toString());
	    if (objNameIter.hasNext()) {
		result.append(", ");
	    }
	}
	return result.toString();
    }

    //
    // Misc
    //

    /**
     * Cloning
     *
     * @return a Role being an independent copy of the current Role object.
     */
    public Object clone() {

	try {
	    return new Role(myName, myObjNameList);
	} catch (IllegalArgumentException exc) {
	    return null; // :)
	}
    }

    /**
     * Returns a string for the given role value
     *
     * @param theRoleValue  ArrayList of ObjectName objects
     *
     * @return String being the ObjectNames separated by commas.
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public static String roleValueToString(List theRoleValue)
	throws IllegalArgumentException {

	if (theRoleValue == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	StringBuffer result = new StringBuffer();
	for (Iterator objNameIter = theRoleValue.iterator();
	     objNameIter.hasNext();) {
	    ObjectName currObjName = (ObjectName)(objNameIter.next());
	    result.append(currObjName.toString());
	    if (objNameIter.hasNext()) {
		result.append("\n");
	    }
	}
	return result.toString();
    }
}
