/*
 * @(#)file      RoleInfo.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.13
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.relation;

import javax.management.NotCompliantMBeanException;
import javax.management.MBeanInfo;

import com.sun.management.jmx.Introspector;

import java.io.Serializable;

/**
 * A RoleInfo object represents a role information in a relation type.
 */
public class RoleInfo implements Serializable {

    //
    // Public constants
    //

    /**
     * To specify an unlimited cardinality
     */
    public static int ROLE_CARDINALITY_INFINITY = -1;

    //
    // Private members
    //

    // Name of the role
    private String myName = null;

    // Read access mode: true if role is readable
    private boolean myIsReadableFlg;

    // Write access mode: true if role is writable
    private boolean myIsWritableFlg;

    // Description of role
    private String myDescription = null;

    // Minimum degree (i.e. minimum number of referenced MBeans in
    // corresponding role)
    private int myMinDegree;

    // Maximum degree (i.e. maximum number of referenced MBeans in
    // corresponding role)
    private int myMaxDegree;

    // Name of class of MBean(s) expected to be referenced in corresponding
    // role
    private String myRefMBeanClassName = null;

    //
    // Constructors
    //

    /**
     * Constructor
     *
     * @param theName  name of the role
     * @param theRefMBeanClassName  name of the class of MBean(s) expected to
     * be referenced in corresponding role
     * @param theIsReadable  flag to indicate if the corresponding role
     * can be read
     * @param theIsWritable  flag to indicate if the corresponding role
     * can be set
     * @param theMinDegree  minimum degree for role, i.e. minimum number of
     * MBeans to provide in corresponding role
     * Must be less or equal than theMaxDegree.
     * (ROLE_CARDINALITY_INFINITY for unlimited)
     * @param theMaxDegree  maximum degree for role, i.e. maximum number of
     * MBeans to provide in corresponding role
     * Must be greater or equal than theMinDegree
     * (ROLE_CARDINALITY_INFINITY for unlimited)
     * @param theDescription  description of the role (can be null)
     *
     * @exception IllegalArgumentException  if null parameter
     * @exception InvalidRoleInfoException  if the minimum degree is
     * greater than the maximum degree.
     * @exception ClassNotFoundException  if the class theRefMBeanClassName
     * does not exist.
     * @exception NotCompliantMBeanException  if the class theRefMBeanClassName
     * is not a MBean class.
     */
    public RoleInfo(String theName,
		    String theRefMBeanClassName,
		    boolean theIsReadable,
		    boolean theIsWritable,
		    int theMinDegree,
		    int theMaxDegree,
		    String theDescription)
    throws IllegalArgumentException,
	   InvalidRoleInfoException,
           ClassNotFoundException,
           NotCompliantMBeanException {

	init(theName,
	     theRefMBeanClassName,
	     theIsReadable,
	     theIsWritable,
	     theMinDegree,
	     theMaxDegree,
	     theDescription);
	return;
    }

    /**
     * Constructor
     *
     * @param theName  name of the role
     * @param theRefMBeanClassName  name of the class of MBean(s) expected to
     * be referenced in corresponding role
     * @param theIsReadable  flag to indicate if the corresponding role
     * can be read
     * @param theIsWritable  flag to indicate if the corresponding role
     * can be set
     *
     * <P>Minimum and maximum degrees defaulted to 1.
     * <P>Description of role defaulted to null.
     *
     * @exception IllegalArgumentException  if null parameter
     * @exception ClassNotFoundException  if the class theMBeanClassName does
     * not exist.
     * @exception NotCompliantMBeanException  if the class theRefMBeanClassName
     * is not a MBean class.
     */
    public RoleInfo(String theName,
		    String theRefMBeanClassName,
		    boolean theIsReadable,
		    boolean theIsWritable)
    throws IllegalArgumentException,
	   ClassNotFoundException,
           NotCompliantMBeanException {

	try {
	    init(theName,
		 theRefMBeanClassName,
		 theIsReadable,
		 theIsWritable,
		 1,
		 1,
		 null);
	} catch (InvalidRoleInfoException exc) {
	}

	return;
    }

    /**
     * Constructor
     *
     * @param theName  name of the role
     * @param theRefMBeanClassName  name of the class of MBean(s) expected to
     * be referenced in corresponding role
     *
     * <P>IsReadable and IsWritable defaulted to true.
     * <P>Minimum and maximum degrees defaulted to 1.
     * <P>Description of role defaulted to null.
     *
     * @exception IllegalArgumentException  if null parameter
     * @exception ClassNotFoundException  if the class theMBeanClassName does
     * not exist.
     * @exception NotCompliantMBeanException  if the class theRefMBeanClassName
     * is not a MBean class.
      */
    public RoleInfo(String theName,
		    String theRefMBeanClassName)
    throws IllegalArgumentException,
	   ClassNotFoundException,
           NotCompliantMBeanException {

	try {
	    init(theName,
		 theRefMBeanClassName,
		 true,
		 true,
		 1,
		 1,
		 null);
	} catch (InvalidRoleInfoException exc) {
	}

	return;
    }

    /**
     * Copy constructor
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public RoleInfo(RoleInfo theRoleInfo)
	throws IllegalArgumentException {

	if (theRoleInfo == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	try {
	    init(theRoleInfo.getName(),
		 theRoleInfo.getRefMBeanClassName(),
		 theRoleInfo.isReadable(),
		 theRoleInfo.isWritable(),
		 theRoleInfo.getMinDegree(),
		 theRoleInfo.getMaxDegree(),
		 theRoleInfo.getDescription());
	} catch (NotCompliantMBeanException exc1) {
	} catch (ClassNotFoundException exc2) {
	} catch (InvalidRoleInfoException exc3) {
	}
    }

    //
    // Accessors
    //

    /**
     * Returns the name of the role
     */
    public String getName() {
	return myName;
    }

    /**
     * Returns read access mode for the role (true if it is readable)
     */
    public boolean isReadable() {
	return myIsReadableFlg;
    }

    /**
     * Returns write access mode for the role (true if it is writable)
     */
    public boolean isWritable() {
	return myIsWritableFlg;
    }

    /**
     * Returns description text for the role
     */
    public String getDescription() {
	return myDescription;
    }

    /**
     * Returns minimum degree for corresponding role reference
     */
    public int getMinDegree() {
	return myMinDegree;
    }

    /**
     * Returns maximum degree for corresponding role reference
     */
    public int getMaxDegree() {
	return myMaxDegree;
    }

    /**
     * Returns name of class of MBean expected to be referenced in
     * corresponding role
     */
    public String getRefMBeanClassName() {
	return myRefMBeanClassName;
    }

    /**
     * Returns a boolean to specify if given value is greater or equal than
     * expected minimum degree (true if yes)
     *
     * @param theValue  value
     *
     * @return true if greater or equal than minimum degree, false else
     */
    public boolean checkMinDegree(int theValue) {
	if (theValue >= ROLE_CARDINALITY_INFINITY &&
	    (myMinDegree == ROLE_CARDINALITY_INFINITY
	     || theValue >= myMinDegree)) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Returns a boolean to specify if given value is less or equal than
     * expected maximum degree (true if yes)
     *
     * @param theValue  value
     *
     * @return true if less or equal than maximum degree, false else
     */
    public boolean checkMaxDegree(int theValue) {
	if (theValue >= ROLE_CARDINALITY_INFINITY &&
	    (myMaxDegree == ROLE_CARDINALITY_INFINITY ||
	     (theValue != ROLE_CARDINALITY_INFINITY &&
	      theValue <= myMaxDegree))) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Prints a string describing the role info
     */
    public String toString() {
	StringBuffer result = new StringBuffer();
	result.append("role info name: " + myName);
	result.append("; isReadable: " + myIsReadableFlg);
	result.append("; isWritable: " + myIsWritableFlg);
	result.append("; description: " + myDescription);
	result.append("; minimum degree: " + myMinDegree);
	result.append("; maximum degree: " + myMaxDegree);
	result.append("; MBean class: " + myRefMBeanClassName);
	return result.toString();
    }

    //
    // Misc
    //

    // Initialisation
    private void init(String theName,
		      String theRefMBeanClassName,
		      boolean theIsReadable,
		      boolean theIsWritable,
		      int theMinDegree,
		      int theMaxDegree,
		      String theDescription)
	throws IllegalArgumentException,
	       InvalidRoleInfoException,
               ClassNotFoundException,
               NotCompliantMBeanException {

	if (theName == null ||
	    theRefMBeanClassName == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	myName = new String(theName);
	myIsReadableFlg = theIsReadable;
	myIsWritableFlg = theIsWritable;
	if (theDescription != null) {
	    myDescription = new String(theDescription);
	}

	boolean invalidRoleInfoFlg = false;
	StringBuffer excMsgStrB = new StringBuffer();
	if (theMaxDegree != ROLE_CARDINALITY_INFINITY &&
	    (theMinDegree == ROLE_CARDINALITY_INFINITY ||
	     theMinDegree > theMaxDegree)) {
	    // Revisit [cebro] Localize message
	    excMsgStrB.append("Minimum degree ");
	    excMsgStrB.append(theMinDegree);
	    excMsgStrB.append(" is greater than maximum degree ");
	    excMsgStrB.append(theMaxDegree);
	    invalidRoleInfoFlg = true;

	} else if (theMinDegree < ROLE_CARDINALITY_INFINITY ||
		   theMaxDegree < ROLE_CARDINALITY_INFINITY) {
	    // Revisit [cebro] Localize message
	    excMsgStrB.append("Minimum or maximum degree has an illegal value, must be [0, ROLE_CARDINALITY_INFINITY].");
	    invalidRoleInfoFlg = true;
	}
	if (invalidRoleInfoFlg) {
	    throw new InvalidRoleInfoException(excMsgStrB.toString());
	}
	myMinDegree = theMinDegree;
	myMaxDegree = theMaxDegree;

	// Checks if class name provided for expected MBeans corresponds to
	// a class which exists
	// Can throw a ClassNotFoundException
	Class refMBeanClass = Class.forName(theRefMBeanClassName);

	// Checks if it is a MBean class
	// Can throw NotCompliantMBeanException
	MBeanInfo mbeanInfo = Introspector.testCompliance(refMBeanClass);

	myRefMBeanClassName = new String(theRefMBeanClassName);

	return;
    }    
}
