/*
 * @(#)file      RoleUnresolved.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.9
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.relation;

import javax.management.ObjectName;

import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

import java.io.Serializable;

/**
 * Represents an unresolved role: a role not retrieved from a relation due
 * to a problem. It provides the role name, value (if problem when trying to
 * set the role) and an integer defining the problem (constants defined in
 * RoleStatus).
 */
public class RoleUnresolved implements Serializable {

    //
    // Private members
    //

    // Role name
    private String myRoleName = null;

    // Role value (ArrayList of ObjectName objects)
    private ArrayList myRoleValue = null;

    // Problem type
    private int myPbType;

    //
    // Constructor
    //

    /**
     * Constructor
     *
     * @param theRoleName  name of the role
     * @param theRoleValue  value of the role (if problem when setting the
     * role)
     * @param thePbType  type of problem (according to known problem types,
     * listed as static final members).
     *
     * @exception IllegalArgumentException  if null parameter or incorrect
     * problem type
     */
    public RoleUnresolved(String theRoleName,
			  List theRoleValue,
			  int thePbType)
	throws IllegalArgumentException {

	if (theRoleName == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	setRoleName(theRoleName);
	setRoleValue(theRoleValue);
	// Can throw IllegalArgumentException
	setProblemType(thePbType);
	return;
    }

    //
    // Accessors
    //

    /**
     * Retrieves role name
     */
    public String getRoleName() {
	return myRoleName;
    }

    /**
     * Retrieves role value
     *
     * @return an ArrayList of ObjectName objects, the one provided to be set
     * in given role. Null if the unresolved role is returned for a read
     * access.
     */
    public List getRoleValue() {
	return myRoleValue;
    }

    /**
     * Retrieves problem type
     *
     * @return an integer corresponding to a problem, those being described as
     * static final members of current class.
     */
    public int getProblemType() {
	return myPbType;
    }

    /**
     * Sets role name
     *
     * @exception IllegalArgumentException  if null parameter
     */
    public void setRoleName(String theRoleName)
	throws IllegalArgumentException {

	if (theRoleName == null) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Invalid parameter.";
	    throw new IllegalArgumentException(excMsg);
	}

	myRoleName = new String(theRoleName);
	return;
    }

    /**
     * Sets role value
     *
     * @param theRoleValue  ArrayList of ObjectName objects for referenced
     * MBeans not set in role.
     */
    public void setRoleValue(List theRoleValue) {

	if (theRoleValue != null) {
	    myRoleValue = new ArrayList();
	    for (Iterator objNameIter = theRoleValue.iterator();
		 objNameIter.hasNext();) {
		ObjectName currObjName = (ObjectName)(objNameIter.next());
		myRoleValue.add(currObjName);
	    }
	} else {
	    myRoleValue = null;
	}
	return;
    }

    /**
     * Sets problem type
     *
     * @param thePbType  integer corresponding to a problem. Must be one of
     * those described as static final members of current class.
     *
     * @exception IllegalArgumentException  if incorrect problem type
     */
    public void setProblemType(int thePbType)
	throws IllegalArgumentException {

	if (!(RoleStatus.isRoleStatus(thePbType))) {
	    // Revisit [cebro] Localize message
	    String excMsg = "Incorrect problem type.";
	    throw new IllegalArgumentException(excMsg);
	}
	myPbType = thePbType;
	return;
    }

    /**
     * Clone
     *
     * @return an independent clone
     */
    public Object clone() {
	try {
	    return new RoleUnresolved(myRoleName, myRoleValue, myPbType);
	} catch (IllegalArgumentException exc) {
	    return null; // :)
	}
    }

    /**
     * Prints a string describing the role
     */
    public String toString() {
	StringBuffer result = new StringBuffer();
	result.append("role name: " + myRoleName);
	if (myRoleValue != null) {
	    result.append("; value: ");
	    for (Iterator objNameIter = myRoleValue.iterator();
		 objNameIter.hasNext();) {
		ObjectName currObjName = (ObjectName)(objNameIter.next());
		result.append(currObjName.toString());
		if (objNameIter.hasNext()) {
		    result.append(", ");
		}
	    }
	}
	result.append("; problem type: " + myPbType);
	return result.toString();
    }
}
