/*
 * @(#)file      TimerMBean.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   4.13
 * @(#)date      00/04/11
 *
 * Copyright 2000 Sun Microsystems, Inc. All rights reserved.
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 * 
 * Copyright 2000 Sun Microsystems, Inc. Tous droits rservs.
 * Ce logiciel est propriet de Sun Microsystems, Inc.
 * Distribu par des licences qui en restreignent l'utilisation. 
 */

package javax.management.timer; 



// java imports
//
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;

// jmx imports
//
import javax.management.InstanceNotFoundException;

/**
 * Exposes the management interface of the timer MBean.
 *
 * @version     4.13     02/04/00
 * @author      Sun Microsystems, Inc
 */

public interface TimerMBean { 

    /**
     * Starts the timer.
     * <P>
     * If there is one or more timer notifications before the time in the list of notifications, the notification
     * is sent according to the <CODE>sendPastNotifications</CODE> flag and then, updated
     * according to its period and remaining number of occurences.
     * If the timer notification date remains earlier than the current date, this notification is just removed 
     * from the list of notifications.
     */
    public void start();
    
    /**
     * Stops the timer.
     */
    public void stop();
    
    /**
     * Creates a new timer notification with the specified <CODE>type</CODE>, <CODE>message</CODE> 
     * and <CODE>userData</CODE> and inserts it into the list of notifications with a given date, 
     * period and number of occurences.
     * <P>
     * If the timer notification to be inserted has a date that is before the current date,
     * the method attempts to update the notification using the defined period.
     * If the defined period is null, the method generates an exception.
     * <P>
     * If the timer notification has a non null period, the date of the notification is updated
     * by adding the period until the notification date is later than the current date.
     * The timer notification is then inserted into the list of notifications.
     * <P>
     * If the specified number of occurences is non null, the previous mechanism 
     * is employed. The number of times that the period is added is limited by
     * the specified number of occurences. If the notification date remains earlier than
     * the current date, an exception is generated.
     * <P>
     * Note that once the timer notification has been added into the list of notifications,
     * its associated date, period and number of occurences cannot be updated.
     * 
     * @param type The timer notification type.
     * @param message The timer notification detailed message.
     * @param userData The timer notification user data object.
     * @param date The date when the notification occurs.
     * @param period The period of the timer notification (in milliseconds).
     * @param nbOccurences The total number the timer notification will be emitted.
     *
     * @return The identifier of the new created timer notification.
     *
     * @exception java.lang.IllegalArgumentException The period or the number of occurences is negative or
     * the date notification is before the current date.
     */
    public Integer addNotification(String type, String message, Object userData, 
                                   Date date, long period, long nbOccurences) 
        throws java.lang.IllegalArgumentException;
    
    /**
     * Creates a new timer notification with the specified <CODE>type</CODE>, <CODE>message</CODE> 
     * and <CODE>userData</CODE> and inserts it into the list of notifications with a given date
     * and period and a null number of occurences.
     * <P>
     * The timer notification will repeat continuously using the timer period.
     * 
     * @param type The timer notification type.
     * @param message The timer notification detailed message.
     * @param userData The timer notification user data object.
     * @param date The date when the notification occurs.
     * @param period The period of the timer notification (in milliseconds).
     *
     * @return The identifier of the new created timer notification.
     *
     * @exception java.lang.IllegalArgumentException The period is negative or
     * the date notification is before the current date.
     */
    public Integer addNotification(String type, String message, Object userData, 
                                   Date date, long period) 
        throws java.lang.IllegalArgumentException;
    
    /**
     * Creates a new timer notification with the specified <CODE>type</CODE>, <CODE>message</CODE> 
     * and <CODE>userData</CODE> and inserts it into the list of notifications with a given date
     * and a null period and number of occurences.
     * <P>
     * The timer notification will be handled once at the specified date.
     * 
     * @param type The timer notification type.
     * @param message The timer notification detailed message.
     * @param userData The timer notification user data object.
     * @param date The date when the notification occurs.
     *
     * @return The identifier of the new created timer notification.
     *
     * @exception java.lang.IllegalArgumentException The date notification is before the current date.
     */
    public Integer addNotification(String type, String message, Object userData, Date date) 
        throws java.lang.IllegalArgumentException;
    
    /**
     * Removes the timer notification corresponding to the specified identifier from the list of notifications.
     * 
     * @param id The timer notification identifier.
     *
     * @exception InstanceNotFoundException The specified identifier does not correspond to any timer notification
     * in the list of notifications of this timer MBean.
     */
    public void removeNotification(Integer id) throws InstanceNotFoundException;
      
    /**
     * Removes all the timer notifications corresponding to the specified type from the list of notifications.
     * 
     * @param type The timer notification type.
     *
     * @exception InstanceNotFoundException The specified type does not correspond to any timer notification
     * in the list of notifications of this timer MBean.
     */
    public void removeNotifications(String type) throws InstanceNotFoundException;
    
    /**
     * Removes all the timer notifications from the list of notifications
     * and resets the counter used to update the timer notification identifiers.
     */
    public void removeAllNotifications();
    
    // GETTERS AND SETTERS
    //--------------------    
        
    /**
     * Gets the number of timer notifications registered into the list of notifications.
     *
     * @return The number of timer notifications.
     */
    public int getNbNotifications();
    
    /**
     * Gets all timer notification identifiers registered into the list of notifications.
     *
     * @return A vector of <CODE>Integer</CODE> objects containing all the timer notification identifiers.
     * <BR>The vector is empty if there is no timer notification registered for this timer MBean.
     */
    public Vector getAllNotificationIDs();
    
    /**
     * Gets all the identifiers of timer notifications corresponding to the specified type.
     *
     * @param type The timer notification type.
     *
     * @return A vector of <CODE>Integer</CODE> objects containing all the identifiers of 
     * timer notifications with the specified <CODE>type</CODE>.
     * <BR>The vector is empty if there is no timer notifications registered for this timer MBean 
     * with the specified <CODE>type</CODE>.
     */
    public Vector getNotificationIDs(String type);
    
    /**
     * Gets the timer notification type corresponding to the specified identifier.
     *
     * @param id The timer notification identifier.
     *
     * @return The timer notification type or null if the identifier is not mapped to any 
     * timer notification registered for this timer MBean.
     */
    public String getNotificationType(Integer id);
    
    /**
     * Gets the timer notification detailed message corresponding to the specified identifier.
     *
     * @param id The timer notification identifier.
     *
     * @return The timer notification detailed message or null if the identifier is not mapped to any 
     * timer notification registered for this timer MBean.
     */
    public String getNotificationMessage(Integer id);
    
    /**
     * Gets the timer notification user data object corresponding to the specified identifier.
     *
     * @param id The timer notification identifier.
     *
     * @return The timer notification user data object or null if the identifier is not mapped to any 
     * timer notification registered for this timer MBean.
     */
    public Object getNotificationUserData(Integer id);
    
    /**
     * Gets a copy of the date associated to a timer notification.
     *
     * @param id The timer notification identifier.
     *
     * @return A copy of the date or null if the identifier is not mapped to any 
     * timer notification registered for this timer MBean.
     */
    public Date getDate(Integer id);
    
    /**
     * Gets a copy of the period (in milliseconds) associated to a timer notification.
     *
     * @param id The timer notification identifier.
     *
     * @return A copy of the period or null if the identifier is not mapped to any 
     * timer notification registered for this timer MBean.
     */
    public Long getPeriod(Integer id);
    
    /**
     * Gets a copy of the remaining number of occurences associated to a timer notification.
     *
     * @param id The timer notification identifier.
     *
     * @return A copy of the remaining number of occurences or null if the identifier is not mapped to any 
     * timer notification registered for this timer MBean.
     */
    public Long getNbOccurences(Integer id);
    
    /**
     * Gets the flag indicating whether or not the timer sends past notifications.
     *
     * @return The past notifications sending on/off flag value.
     */
    public boolean getSendPastNotifications();
    
    /**
     * Sets the flag indicating whether the timer sends past notifications or not.
     *
     * @param value The past notifications sending on/off flag value.
     */
    public void setSendPastNotifications(boolean value);
    
    /**
     * Tests whether the timer MBean is active.
     * A timer MBean is marked active when the {@link #start start} method is called.
     * It becomes inactive when the {@link #stop stop} method is called.
     *
     * @return <CODE>true</CODE> if the timer MBean is active, <CODE>false</CODE> otherwise.
     */
    public boolean isActive();

    /**
     * Tests whether the list of timer notifications is empty.
     *
     * @return <CODE>true</CODE> if the list of timer notifications is empty, <CODE>false</CODE> otherwise.
     */
    public boolean isEmpty();
}
